/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.66.2.8 $
|                                              $Date: 2007/05/10 09:54:15 $
\*============================================================================*/
 
/*******************************************************************************
*                           INTERNAL DPWS RUNTIME HEADER                       *
*******************************************************************************/
#ifndef _DPWS_H
# define _DPWS_H

#include "dpws.h"


#ifdef __cplusplus
extern "C" {
#endif

/* Macros */

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

/* ONVIF_LOG */
#ifdef USE_ONVIF /* Added by HG.W */
enum{
    ONVIF_LOG_NONE  = 0,
    ONVIF_LOG_ERROR,
    ONVIF_LOG_INFO,
};

#define ONVIF_LOG(level, msg) \
{\
    if(level <= ENABLE_DEBUG_LOG)  \
        fprintf(stderr, msg);     \
}
#define ONVIF_LOG1(level, msg, p1) \
{\
    if(level <= ENABLE_DEBUG_LOG)  \
        fprintf(stderr, msg, p1); \
}
#define ONVIF_LOG2(level, msg, p1, p2) \
{\
    if(level <= ENABLE_DEBUG_LOG)      \
        fprintf(stderr, msg, p1, p2); \
}
#define ONVIF_LOG3(level, msg, p1, p2, p3) \
{\
    if(level <= ENABLE_DEBUG_LOG)          \
        fprintf(stderr, msg, p1, p2, p3); \
}
#define ONVIF_LOG4(level, msg, p1, p2, p3, p4) \
{\
    if(level <= ENABLE_DEBUG_LOG)               \
        fprintf(stderr, msg, p1, p2, p3, p4);   \
}
#define ONVIF_LOG5(level, msg, p1, p2, p3, p4, p5) \
{\
    if(level <= ENABLE_DEBUG_LOG)                  \
        fprintf(stderr, msg, p1, p2, p3, p4, p5);  \
}
#endif

/* #define DPWS_DEBUG */

#ifdef DPWS_DEBUG
#  define DPWSLOG(MSG) \
{       FILE *fdebug = fopen("DPWS.log", "a");\
        fprintf(fdebug, MSG);\
        fclose(fdebug);\
}
#  define DPWSLOG1(FMT, P1) \
{       FILE *fdebug = fopen("DPWS.log", "a");\
        fprintf(fdebug, FMT, P1);\
        fclose(fdebug);\
}
#  define DPWSLOG2(FMT, P1, P2) \
{       FILE *fdebug = fopen("DPWS.log", "a");\
        fprintf(fdebug, FMT, P1, P2);\
        fclose(fdebug);\
}
#  define DPWSLOG3(FMT, P1, P2, P3) \
{       FILE *fdebug = fopen("DPWS.log", "a");\
        fprintf(fdebug, FMT, P1, P2, P3);\
        fclose(fdebug);\
}
#else
#  define DPWSLOG(MSG)
#  define DPWSLOG1(FMT, P1)
#  define DPWSLOG2(FMT, P1, P2)
#  define DPWSLOG3(FMT, P1, P2, P3)
#endif

/* DPWS Constants */
#ifdef USE_ONVIF /* Added by HG.W */
#define APP_MAX_DELAY                           500 
#else
#define APP_MAX_DELAY                           5000 // (ms) WS-Discovery (superseded by DPWS)
#endif /* Added by HG.W */

#define WSD_MATCH_TIMEOUT                       (APP_MAX_DELAY + 100) // (ms) WS-Discovery
#define MATCH_TIMEOUT                           10000 // (ms) DPWS
#define DISCOVERY_PORT                          3702 // WS-Discovery
#define DISCOVERY_PORT_STR                      "3702"
#ifdef WITH_IPV6
#define DISCOVERY_MULTICAST_ADDR_V6             "ff02::c" // WS-Discovery
#define DISCOVERY_ENDPOINT_V6                   "[ff02::c]:3702"// WS-Discovery
#endif
#define DISCOVERY_MULTICAST_ADDR                "239.255.255.250" // WS-Discovery
#define DISCOVERY_ENDPOINT                      "239.255.255.250:3702" // WS-Discovery

#define MAX_ENVELOPE_SIZE                       32767 // (octets) DPWS
#define MAX_FIELD_SIZE                          256 // (Unicode characters) DPWS
#define MAX_URI_SIZE                            2048 // (octets) DPWS
#define SOAP_UDP_SCHEME                         "soap.udp" // SOAP-over-UDP
#ifdef USE_ONVIF /* Added by HG.W */
#define MULTICAST_UDP_REPEAT            4
#else
#define MULTICAST_UDP_REPEAT            2 // SOAP-over-UDP (superseded by DPWS)
#endif
#define UDP_MAX_DELAY                           250 // (ms) SOAP-over-UDP
#define UDP_MIN_DELAY                           50 // (ms) SOAP-over-UDP
#define UDP_UPPER_DELAY                         450 // (ms) SOAP-over-UDP (superseded by DPWS)
#define UNICAST_UDP_REPEAT                      2 // SOAP-over-UDP
#define WAKE_UP_DATAGRAM                        "<WAKE/>" // start with '<' not to be considered as an HTTP header (blocks)
#define SHUTDOWN_DELAY                          200 // (ms) time to send the last bye.
#define SHUTDOWN_DELAY_MAX_REPEAT       10 // Number of times to wait until servers have sent scheduled messages.

/**
 * WSA error codes
 */
#define MIN_WSA_ERROR                                                   1000
#define WSA_ERR_INVALID_HEADER_INFO_MESSAGE_ID  1001
#define WSA_ERR_INVALID_HEADER_INFO_RELATES_TO  1002
#define WSA_ERR_INVALID_HEADER_INFO_TO                  1003
#define WSA_ERR_INVALID_HEADER_INFO_ACTION              1004
#define WSA_ERR_INVALID_HEADER_INFO_FROM                1005
#define WSA_ERR_INVALID_HEADER_INFO_REPLY_TO    1006
#define WSA_ERR_INVALID_HEADER_INFO_FAULT_TO    1007
#define WSA_ERR_REQUIRED_HEADER_INFO                    1011
#define WSA_ERR_DESTINATION_UNREACHABLE                 1012
#define WSA_ERR_ACTION_NOT_SUPPORTED                    1013
#define WSA_ERR_ENDPOINT_UNAVAILABLE                    1014

/**
 *  WSE error codes
 */
#define MIN_WSE_ERROR                                                           1100
#define WSE_ERR_DELIVERY_MODE_REQUESTED_UNAVAILABLE     1101
#define WSE_ERR_INVALID_EXPIRATION_TIME                         1102
#define WSE_ERR_UNSUPPORTED_EXPIRATION_TYPE                     1103
#define WSE_ERR_FILTERING_NOT_SUPPORTED                         1104
#define WSE_ERR_FILTERING_REQUESTED_UNAVAILABLE         1105
#define WSE_ERR_EVENT_SOURCE_UNABLE_TO_PROCESS          1106
#define WSE_ERR_UNABLE_TO_RENEW                                         1107
#define WSE_ERR_INVALID_MESSAGE                                         1108

#define WSE_ERR_INVALID_SUBSCRIPTION                            1109    // not standard: re-routed to INVALID message

#define MDM_ERR_MISSING_SCOPES                                          1150


#ifdef USE_ONVIF /* Added by HG.W (for discovery fault) */
#define WSD_ERR_MATCHING_RULE_NOT_SUPPORTED                1201
#define WSD_ERR_UNKNOWN_UDP_DESTINATION_ADDRESS            1202
#define WSD_ERR_NON_XML_DATA                               1203
#endif

/* Status codes */
// bit 0 exchange step: resquest = 0, response = 1
// bit 1 communication way: send = 0, receive = 1
// bit 2 transport: tcp = 0, udp = 1
// bit 3 mode: unicast = 0, multicast = 1
// bit 4-9 message type. None = 'user'
// bit 4: hello
// bit 5: bye
// bit 6: probe
// bit 7: probe match
// bit 8: resolve
// bit 9: resolve match

#define DPWS_MSG_TYPES                  (0x03F0)
#define DPWS_MSG_TYPES_MASK             (0xFC0F)

#define DPWS_MSG_HELLO                  (0x0010) 
#define DPWS_MSG_BYE                    (0x0020) 
#define DPWS_MSG_PROBE                  (0x0040) 
#define DPWS_MSG_PROBE_MATCH    (0x0080) 
#define DPWS_MSG_RESOLVE                (0x0100) 
#define DPWS_MSG_RESOLVE_MATCH  (0x0200) 

#define DPWS_GET_MSG_TYPE(s) (s & 0x03F0) 
#define DPWS_IS_DISCOVERY(s) (DPWS_GET_MSG_TYPE(s)) 

#define DPWS_SET_REQUEST(s) (s &= 0xFFFE)
#define DPWS_SET_RESPONSE(s) (s |= 0x0001)
#define DPWS_IS_RESPONSE(s) (s & 0x0001)

#define DPWS_SET_SEND(s) (s &= 0xFFFD)
#define DPWS_SET_RCV(s) (s |= 0x0002)
#define DPWS_IS_RCV(s) (s & 0x0002)

#define DPWS_SET_TCP(s) (s &= 0xFFFB)
#define DPWS_SET_UDP(s) (s |= 0x0004)
#define DPWS_IS_UDP(s) (s & 0x0004)

#define DPWS_SET_UNICAST(s) (s &= 0xFFF7)
#define DPWS_SET_MULTICAST(s) (s |= 0x0008)
#define DPWS_IS_MULTICAST(s) (s & 0x0008)

#define DPWS_SET_USER(s) (s &= DPWS_MSG_TYPES_MASK)
#define DPWS_IS_USER(s) (DPWS_GET_MSG_TYPE(s) == 0x0000)

#define DPWS_SET_HELLO(s) (s = DPWS_MSG_HELLO | (s & DPWS_MSG_TYPES_MASK))
#define DPWS_IS_HELLO(s) (DPWS_GET_MSG_TYPE(s) == DPWS_MSG_HELLO)

#define DPWS_SET_BYE(s) (s = DPWS_MSG_BYE | (s & DPWS_MSG_TYPES_MASK))
#define DPWS_IS_BYE(s) ((DPWS_GET_MSG_TYPE(s) == DPWS_MSG_BYE)

#define DPWS_SET_PROBE(s) (s = DPWS_MSG_PROBE | (s & DPWS_MSG_TYPES_MASK))
#define DPWS_IS_PROBE(s) (DPWS_GET_MSG_TYPE(s) == DPWS_MSG_PROBE)

#define DPWS_SET_PROBE_MATCH(s) (s = DPWS_MSG_PROBE_MATCH | (s & DPWS_MSG_TYPES_MASK))
#define DPWS_IS_PROBE_MATCH(s) (DPWS_GET_MSG_TYPE(s) == DPWS_MSG_PROBE_MATCH)

#define DPWS_SET_RESOLVE(s) (s = DPWS_MSG_RESOLVE | (s & DPWS_MSG_TYPES_MASK))
#define DPWS_IS_RESOLVE(s) (DPWS_GET_MSG_TYPE(s) == DPWS_MSG_RESOLVE)

#define DPWS_SET_RESOLVE_MATCH(s) (s = DPWS_MSG_RESOLVE_MATCH | (s & DPWS_MSG_TYPES_MASK))
#define DPWS_IS_RESOLVE_MATCH(s) (DPWS_GET_MSG_TYPE(s) == DPWS_MSG_RESOLVE_MATCH)

#define DPWS_INIT_STATUS(s) (s = 0x0)   //request,send,tcp,unicast,user

#define DPWS_SET_UDP_MULTICAST(s) (s |= 0x000C)
#define DPWS_IS_UDP_RCV_MULTICAST(s) ((s & 0x000E) == 0x000E)
#define DPWS_POLICY_MSG(s) (s & 0x0290)
#define DPWS_APP_SEQ_MSG(s) (s & 0x02B0)
#define DPWS_APP_DELAY_MSG(s) (s & 0x0090)
#define DPWS_RESET_MSG_TYPE(s) (s &= DPWS_MSG_TYPES_MASK) 

#define DPWS_ERR_UDP_TIMEOUT                                    1300    // internal


#define QNAME_PREFIX_GEN_PREFIX "ns"
#define QNAME_PREFIX_GEN_LEN    (sizeof(QNAME_PREFIX_GEN_PREFIX) + sizeof(MAX_SERVICES_PER_DEVICE))

#define WSA_ANONYMOUS_ENDPOINT_REF_URI "http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"
#define WSA_SOAP_FAULT_URI "http://schemas.xmlsoap.org/ws/2004/08/addressing/fault"
#define WSA_URI "http://schemas.xmlsoap.org/ws/2004/08/addressing"

#define WDP_URI "http://schemas.xmlsoap.org/ws/2006/02/devprof"
#define WSDL_URI "http://schemas.xmlsoap.org/wsdl"
#define WDP_MODEL_URI "http://schemas.xmlsoap.org/ws/2006/02/devprof/ThisModel"
#define WDP_DEVICE_URI "http://schemas.xmlsoap.org/ws/2006/02/devprof/ThisDevice"
#define WDP_HOSTING_URI "http://schemas.xmlsoap.org/ws/2006/02/devprof/Relationship"
#define WDP_RELATIONSHIP_HOST_TYPE_URI "http://schemas.xmlsoap.org/ws/2006/02/devprof/host"

/* Added by HG.W (when onvif only support 3986 matching rule) */
#ifdef USE_ONVIF
#define WSD_MATCH_BY_RFC3986_URI    "http://schemas.xmlsoap.org/ws/2005/04/discovery/rfc3986"
#define WSD_SOAP_FAULT_URI "http://schemas.xmlsoap.org/ws/2005/04/discovery/fault"
#else
#define WSD_MATCH_BY_RFC2396_URI        "http://schemas.xmlsoap.org/ws/2005/04/discovery/rfc2396"
#define WSD_MATCH_BY_UUID_URI           "http://schemas.xmlsoap.org/ws/2005/04/discovery/uuid"
#define WSD_MATCH_BY_LDAP_URI           "http://schemas.xmlsoap.org/ws/2005/04/discovery/ldap"
#define WSD_MATCH_BY_STRCMP_URI         "http://schemas.xmlsoap.org/ws/2005/04/discovery/strcmp0"
#endif

#define WSD_URN "urn:schemas-xmlsoap-org:ws:2005:04:discovery"
#define WSD_URI "http://schemas.xmlsoap.org/ws/2005/04/discovery"
#define WSP_URI "http://schemas.xmlsoap.org/ws/2004/09/policy"
#define WSM_URI "http://schemas.xmlsoap.org/ws/2004/09/mex"
#define WSE_URI "http://schemas.xmlsoap.org/ws/2004/08/eventing"
#define MDM_URI "http://www.sirena-itea.org/dpws/2005/01/mdm"

#define DISC_PROXY_PORT_TYPE "DiscoveryProxy"
#define TGS_PORT_TYPE "TargetService"
#define WSM_PORT_TYPE "MetadataExchange"
#define MDM_PORT_TYPE "ManageDeviceMetadata"
#define WDP_DEVICE_TYPE "Device"

#define WSE_MODE_PUSH "http://schemas.xmlsoap.org/ws/2004/08/eventing/DeliveryModes/Push"
#define DPWS_WSE_FILTERING_DIALECT "http://schemas.xmlsoap.org/ws/2006/02/devprof/Action"

#define DPWS_EVENT_SUBSCRIBE_ACTION                             "http://schemas.xmlsoap.org/ws/2004/08/eventing/Subscribe"

#define DPWS_EVENT_SUBSCRIPTION_END_REASON              "Commanded device shutdown"

#define dpws_endpoint_ref_is_null(endpoint) (!((endpoint).address))

/*----------------------------------------------------------------------------*\
 *                                Type Definitions                            *
\*----------------------------------------------------------------------------*/
typedef unsigned long   unsigned32;
typedef unsigned short  unsigned16;
typedef unsigned char   unsigned8;
typedef unsigned char   byte;

typedef struct _unsigned64_t {
        unsigned32          lo;
        unsigned32          hi;
} unsigned64_t;

#define MAKE_32b(hi, lo, dest) {dest = hi&0x7FFF; dest = dest<<16; dest += lo&0x7FFF;}

/*
**  Add two unsigned 64-bit long integers.
*/
#define ADD_64b_2_64b(A, B, sum) \
      { \
          if (!(((A)->lo & 0x80000000UL) ^ ((B)->lo & 0x80000000UL))) { \
              if (((A)->lo&0x80000000UL)) { \
                  (sum)->lo = (A)->lo + (B)->lo; \
                  (sum)->hi = (A)->hi + (B)->hi + 1; \
              } \
              else { \
                  (sum)->lo  = (A)->lo + (B)->lo; \
                  (sum)->hi = (A)->hi + (B)->hi; \
              } \
          } \
          else { \
              (sum)->lo = (A)->lo + (B)->lo; \
              (sum)->hi = (A)->hi + (B)->hi; \
              if (!((sum)->lo&0x80000000UL)) (sum)->hi++; \
          } \
      }

/*
**  Add a 16-bit unsigned integer to a 64-bit unsigned integer.
*/
#define ADD_16b_2_64b(A, B, sum) \
      { \
          (sum)->hi = (B)->hi; \
          if ((B)->lo & 0x80000000UL) { \
              (sum)->lo = (*A) + (B)->lo; \
              if (!((sum)->lo & 0x80000000UL)) (sum)->hi++; \
          } \
          else \
              (sum)->lo = (*A) + (B)->lo; \
      }
      
#define IS_TIME_INF(t1, t2) (t1.hi < t2.hi || (t1.hi == t2.hi && t1.lo < t2.lo))
#define IS_TIME_SUP(t1, t2) (t1.hi > t2.hi || (t1.hi == t2.hi && t1.lo > t2.lo))
#define IS_TIME_EQUAL(t1, t2) (t1.hi == t2.hi && t1.lo == t2.lo)

typedef struct schedule_item {
        unsigned64_t expDate;
        int (*handler) (struct dpws*, void *);
        void *param;
        struct schedule_item *next;
} schedule_t;

typedef struct discovery_msg {
        void *param;    // not freed by this struct user
        char uuid[SCHEME_UUID_STRING_SIZE];
        char req_uuid[SCHEME_UUID_STRING_SIZE];
        unsigned short msg_nb;
} discovery_msg_t;

/*----------------------------------------------------------------------------*\
 *                   RUNTIME (used by the generated code)                     *
\*----------------------------------------------------------------------------*/
#ifndef WITH_NONAMESPACES
extern struct Namespace snd_namespaces[];
#endif

SOAP_FMAC1 void SOAP_FMAC2 dpws_wsa_headers_out(struct dpws*);
SOAP_FMAC1 int SOAP_FMAC2 dpws_wsa_header_in(struct dpws*);
SOAP_FMAC1 void SOAP_FMAC2 dpws_init_wsa_headers(struct dpws* dpws);
SOAP_FMAC1 int SOAP_FMAC2 dpws_check_wsa_headers(struct dpws *);
SOAP_FMAC1 void SOAP_FMAC2 dpws_set_call_wsa_headers(struct dpws*, struct wsa_endpoint_ref *, struct wsa_endpoint_ref *, char *);
SOAP_FMAC1 void SOAP_FMAC2 dpws_set_oneway_wsa_headers(struct dpws*, struct wsa_endpoint_ref *, char *);
SOAP_FMAC1 void SOAP_FMAC2 dpws_set_reply_wsa_headers(struct dpws*, char *);
SOAP_FMAC1 int SOAP_FMAC2 dpws_is_asynchronous(struct dpws*);
SOAP_FMAC1 int SOAP_FMAC2 dpws_send_fault(struct dpws*);
SOAP_FMAC1 int SOAP_FMAC2 dpws_response(struct dpws *dpws, int status);
SOAP_FMAC1 void SOAP_FMAC2 dpws_set_namespaces(struct dpws *, struct Namespace *);
SOAP_FMAC1 int SOAP_FMAC2 dpws_connect(struct dpws* dpws, const char* address, const char* action);
SOAP_FMAC1 struct wsa_endpoint_ref** SOAP_FMAC2 dpws_get_subscriptors(struct dpws* dpws, short, char* action);
SOAP_FMAC1 int SOAP_FMAC2 dpws_serve_subscribe(struct dpws* dpws);
SOAP_FMAC1 void SOAP_FMAC2 dpws_invalidate(struct dpws* dpws, struct wsa_endpoint_ref* device);
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_build_local_endpoint_ref(struct wsa_endpoint_ref * dest, short service_port_href);


/*******************************************************************************
*                                    SHARED                                    *
*******************************************************************************/
SOAP_FMAC1 int SOAP_FMAC2 dpws_send_empty_response(struct dpws*);

#ifdef __cplusplus
}
#endif

#endif
