/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.206.2.9 $
|                                              $Date: 2007/06/28 17:10:17 $
\*============================================================================*/
 
/*! \file dpws.h
    DPWS Toolkit API Reference.
*/
/* \mainpage DPWS Toolkit API Reference

Version 2.0M11.2
  
This manual is divided in the following sections:
- \subpage const 
- \subpage struct
- \subpage func
*/

/*******************************************************************************
*                                 PUBLIC API                                   *
*******************************************************************************/
#ifndef DPWS_H
# define DPWS_H

#include "stdsoap2.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Sizing Constants */
#define DISCOVERY_ALLOC_INCREMENT       4
#define EVENT_ALLOC_INCREMENT           5
#define SCHEME_UUID_STRING_SIZE         46
#define UUID_SCHEME_SIZE                9
#define UUID_SCHEME                     "urn:uuid:"
#define UUID_STRING_SIZE                37
#define UUID_SIZE                       36
//#define UUID_SCHEME_SIZE              9
#define UUID_CACHE_SIZE                 10
#define DURATION_MAX_SIZE               32
#define MAX_SERVER_OBJECTS              1024
#define MAX_CACHE_OBJECTS               4096
#define MAX_SERVICES_PER_DEVICE         512
#define DISCOVERY_HANDLE_POOL_SIZE      10
#define DISCOVERY_MAX_SCHEDULED_MSG     50  // TODO: link to the max number of device

/* Default configuration */
#define DPWS_INT_HTTP_PORT_DEFAULT              80
#define DPWS_INT_HTTP_BACKLOG_DEFAULT           10
#define DPWS_STR_MAX_SUBSC_DURATION_DEFAULT     "P1D"


/* \page const Constants
*/

/**
 *  DPWS API error codes
 */
#define DPWS_OK                                     SOAP_OK

// TODO: DOC
    
// runtime errors (can use dpws_get_error() on dpws rt structure and may able to get extended text message)
// NOTE: -1 reserved for SOAP_EOF
#define DPWS_ERR_COULD_NOT_SET_REUSE_ADDR_OPTION    -2
#define DPWS_ERR_COULD_NOT_BIND_SOCKET              -3
#define DPWS_ERR_COULD_NOT_JOIN_MULTICAST_GROUP     -4
#define DPWS_ERR_COULD_NOT_SEND_DATA                -5
#define DPWS_ERR_INVALID_PARAMETER                  -6  // no message
#define DPWS_ERR_NO_AN_MDM_ENDPOINT                 -7  // no message
#define DPWS_ERR_COULD_NOT_FIND_CACHE_ENTRY         -8  // no message
#define DPWS_ERR_SERVER_STOPPED                     -9  // no message
#define DPWS_ERR_COULD_NOT_CREATE_SOCKET            -10 // Could not create IP socket.
#define DPWS_ERR_UNSUPPORTED_METADATA_SECTION       -11 // no message

// local error (mainly configuration/initialization: no message)
#define MIN_RT_API_ERROR                            -100
#define DPWS_ERR_OUT_OF_MEMORY                      -100
#define DPWS_ERR_CREATING_MUTEX                     -101
#define DPWS_ERR_DELETING_MUTEX                     -102
#define DPWS_ERR_WINSOCK_INIT                       -103
#define DPWS_ERR_RETRIEVING_NETWORK_INTERFACE_INFO  -104
#define DPWS_ERR_RETRIEVING_MAC_ADDRESS             -105
#define DPWS_ERR_RETRIEVING_IP_ADDRESS              -106
#define DPWS_ERR_MISSING_METADATA_VERSION           -107
#define DPWS_ERR_MISSING_MANUFACTURER               -108
#define DPWS_ERR_MISSING_MODEL_NAME                 -109
#define DPWS_ERR_MISSING_FRIENDLY_NAME              -110
#define DPWS_ERR_MISSING_BOOT_SEQUENCE              -111
#define DPWS_ERR_PREFIX_ALREADY_USED                -112
#define DPWS_ERR_NO_HANDLE_FOUND                    -113
#define DPWS_ERR_NO_SUCH_ATT_ON_OBJECT              -114
#define DPWS_ERR_MONOVALUED_ATT_TYPE                -115
#define DPWS_ERR_CONFLICTING_OPERATION              -116
#define DPWS_ERR_IMMUTABLE_USED_OBJECT              -117
#define DPWS_ERR_COULD_NOT_FIND_ENDPOINT            -118
#define DPWS_ERR_MAX_HANDLE_REACHED                 -119
#define DPWS_ERR_UNCOMPLETE_SERVICE                 -120
#define DPWS_ERR_INVALID_PROXY                      -121
#define DPWS_ERR_IS_A_DEVICE                        -122
#define DPWS_ERR_IS_A_HOSTED_SERVICE                -123
#ifdef USE_ONVIF /* Added by HG.W */
#define DPWS_ERR_MISSING_WAKEUP_SOCKS_PATH          -201
#define DPWS_ERR_TOO_LONG_WAKEUP_SOCKS_PATHNAME     -202
#endif


/* \page struct Structures
*/

/** WS-Adressing endpoint reference
 * Runtime structure that contains WS-addressing endpoint reference data.
 * Used for the source, reply to and fault to WSA header fields.
 */
struct wsa_endpoint_ref {
    char* address;  /**< WSA endpoint Address element */
    char* subscription_id;  /**< WS-Eventing subscription identifier */
};

// dynamic array: used internally
typedef struct dynamic_array {
    void * tab;
    int nb;
    int size;
    size_t f_size;
    int inc;
} dyn_array;

#ifdef USE_ONVIF /* Added by HG.W */
typedef enum ipproto_type{
    DPWS_IPPROTO_UNSPEC = 0,
    DPWS_IPPROTO_IPV4,
#ifdef WITH_IPV6
    DPWS_IPPROTO_IPV6,
#endif
}dpws_ipproto_type;

#ifdef WITH_IPV6
typedef enum xaddr_ipv6_scope{
    DPWS_XADDR_IPV6_SCOPE_LINKLOCAL = 0,
    DPWS_XADDR_IPV6_SCOPE_GLOBAL,
}dpws_xaddr_ipv6_scope;
#endif
#endif

/** DPWS toolkit runtime structure
 * Main DPWS API runtime data structure.
 * The fields in the dpws structure provide support for the WS-Addressing
 * specification. Most of those fields are set automatically or through API
 * calls by the DPWS toolkit, and should therefore be accessed only for
 * reading in user-defined client or server code. The notable exception is
 * the fault_to field, which should be set by the user before calling a proxy
 * function, if required. A future version of the toolkit may add parameters
 * to the proxy function signatures to pass this optional information to the
 * toolkit runtime.
 * As mentioned previously, it is always possible to retrieve the soap
 * structure embedded within the dpws structure, and use it to call gSOAP
 * runtime functions. It is also possible to retrieve the dpws structure
 * corresponding to a soap structure. The dpws_dpws2soap(dpws) and
 * dpws_soap2dpws(soap) macros are available to cast from one structure to
 * the other.
 */
struct dpws {
    struct soap soap;   /**< The gSOAP runtime data structure */
    char* to;   /**< WSA header To element */
    struct wsa_endpoint_ref *source;    /**< This fields stores the source endpoint reference information set through the API and that is used to fill the WSA header source tag (from field) before every sending. */
    char* action;   /**< WSA header Action element */
    char* message_id;   /**< WSA header MessageID element */
    char* relates_to;   /**< WSA header RelatesTo element */
    struct wsa_endpoint_ref from;   /**< WSA header From element */
    struct wsa_endpoint_ref reply_to;   /**< WSA header ReplyTo element */
    struct wsa_endpoint_ref fault_to;   /**< WSA header FaultTo element */
    unsigned long instance_id;  /**< WS-Discovery application sequence instance id */
    unsigned short msg_nb;  /**< WS-Discovery application sequence message number */
    char* subscription_id;  /**< WS-Eventing subscription identifier */
    unsigned short status;  /**< internal DPWS processing status */
#ifdef USE_ONVIF /* Added by HG.W */
    dpws_ipproto_type ipproto_type; /**< server type */
#ifdef WITH_IPV6
    dpws_xaddr_ipv6_scope xaddr_ipv6_scope;
#endif
    int wsd_ws;  /**< Wake up socket */
#endif
    int wsd_ls; /**< WS-Discovery listening socket (receiving hello, bye, probe, resolve) */
    int err;    /**< API error code */
    void * discovery_msg;   /**< Stores the parsed discovery message */
    int tmp_buf_count;  /**< internal use */
    int (*backup_send_f)(struct soap *, const char *, size_t);  /**< internal use */
    int err_detail_type;    /**< internal use */
    void * err_detail_data; /**< internal use */
    char * err_detail_data_elt; /**< internal use */
    short href_endpoint;    /**< internal use */
    struct schedule_item *schedule; /**< internal use */
    struct Namespace *namespaces;   /**< internal use */
    dyn_array proxyHandles;
    char* out_ip_addr;  /**< ip used to select the interface used to send UDP datagrams */
};

// WS-Discovery structures

/**
 * An XML qualified name used for instance for WS-Discovery port types.
 */
struct qname
{
    char * ns;  /**< Namespace. */
    char * lname;   /**< Local name. */
};

/**
 * An XML qualified name with a prefix specification used for instance
 * for WS-Discovery port types.
 */
struct prefixed_qname
{
    struct qname qname; /**< Qualified name. */
    char *prefix;   /**< Prefix to use for the qname namespace. */
};

// WS-MetadataExchange structures

/**
 * Container for WS-MedataExchange APIs results. This particular one
 * relates to the http://schemas.xmlsoap.org/ws/2005/05/devprof/ThisModel
 * dialect.
 */
struct model_info
{
    char * manufacturer;    /**< See [DPWS]. Only one is provided. If the metadata message returns several, the one with default language or in english is assumed. */
    char * manufacturer_url;    /**< Cf. [DPWS]. */
    char * model_name;  /**< See [DPWS]. Only one is provided. If the metadata message returns several, the one with default language or in english is assumed. */
    char * model_number;    /**< See [DPWS]. */
    char * model_url;   /**< See [DPWS]. */
    char * presentation_url;    /**< See [DPWS]. */
};

/**
 * Container for WS-MedataExchange APIs results. This particular one
 * relates to the http://schemas.xmlsoap.org/ws/2005/05/devprof/ThisDevice
 * dialect.
 */
struct device_info
{
    char * friendly_name;   /**< See [DPWS]. Only one is provided. If the metadata message returns several, the one with default language or in english is assumed. */
    char * firmware_version;    /**< See [DPWS]. */
    char * serial_number;   /**< See [DPWS]. */
};

/**
 * Container for WS-MedataExchange APIs results. This particular one relates
 * to the http://schemas.xmlsoap.org/wsdl dialect.
 */
struct wsdl_info
{
    char * target_ns;   /**< The target namespace for the web service endpoint. */
    char * location;    /**< An URL that enables WDSL location for the service. Note that the current version of the toolkit does not support the inline retrieval of WSDL. */
};

/**
 * Container for WS-MedataExchange APIs results. This particular one relates
 * to the http://schemas.xmlsoap.org/ws/2005/05/devprof/Relationship dialect.
 */
struct service_info
{
    char * service_id;  /**< The service identifier (URI). */
    struct wsa_endpoint_ref** endpoints;    /**< A list of endpoint references for the service. */
};

/**
 * Container for WS-MedataExchange APIs results. This particular one relates
 * to all dialects supported by the toolkit.
 */
struct device_metadata {
    struct wsa_endpoint_ref** host_endpoints;   /**< A list of endpoint references for the device. */
    struct model_info model_info;   /**< The device model DPWS information. */
    struct device_info device_info; /**< The device own information. */
    struct service_info** services; /**< A list of the services hosted by the device. */
    struct wsdl_info** wsdls;   /**< List of WSDL files for the device. */
};

/**
 * Container for WS-MedataExchange APIs results. This particular one relates
 * to all dialects supported by the toolkit.
 */
struct service_metadata {
    struct wsa_endpoint_ref** host_endpoints;   /**< A list of endpoint references for the device. */
    struct service_info* service_info;  /**< The service information. */
    struct wsdl_info** wsdls;   /**< List of WSDL files for the device. */
};

/**
 * Container for WS-Eventing subscription end event. This event is sent when a
 * subscription manager terminates abnormally a subscription.
 */
struct event_end_info
{
    struct wsa_endpoint_ref * subscription_manager; /**< The endpoint reference of the subscription manager that cancels the subscription. */
    char * status;  /**< See [WS-Eventing]. */
    char * reason;  /**< See [WS-Eventing]. */
};

// Metadata Manager structures

/**
 * Used for API parameters that are localized using the xml:lang attribute.
 */
struct localized_string
{
    char * lang;    /**< The values used for this field are defined in the IETF RFC 3066. */
    char * s;   /**< A string in the language specified by the lang attribute. */
};

/*
 * Utilities
 */

/**
 * Macro for transforming a DPWS runtime structure into a gSOAP runtime
 * structure.
 * @param dpws a DPWS runtime structure
 */
#define dpws_dpws2soap(dpws) (&dpws->soap)

/**
 * Macro for transforming a gSOAP runtime structure into a DPWS runtime
 * structure (provided the gSOAP structure is in fact embedded in a DPWS
 * structure).
 * @param soap a gSOAP runtime structure
 */
#define dpws_soap2dpws(soap) ((struct dpws*)(soap))


/*----------------------------------------------------------------------------*\
 *                                   API                                      *
\*----------------------------------------------------------------------------*/
/*
 * Configuration macro Setters
 */
#define DPWS_SET_PTR_ATT(h,p,v) dpws_set_ptr_att(h,p,v)
#define DPWS_SET_STR_ATT(h,p,v) dpws_set_ptr_att(h,p,v)
#define DPWS_SET_INT_ATT(h,p,v) dpws_set_int_att(h,p,v)
#define DPWS_SET_BOOL_ATT(h,p,v) dpws_set_int_att(h,p,v)
#define DPWS_ADD_PTR_ATT(h,p,v) dpws_add_ptr_att(h,p,v)
#define DPWS_ADD_STR_ATT(h,p,v) dpws_add_ptr_att(h,p,v)
#define DPWS_ADD_INT_ATT(h,p,v) dpws_add_int_att(h,p,v)
#define DPWS_ADD_BOOL_ATT(h,p,v) dpws_add_int_att(h,p,v)

/* Configuration parameters
 * PTR are supposed to point at static data
 * STR are string that are copied
 * INT are integer parameters
 */
// TODO: OXYDOC !
/* toolkit configuration parameters */
#define DPWS_STR_MAC_ADDR                       1   // 6 bytes given as hexadecimal strings separated by '-' or ':'. Optional (generated if missing)
#define DPWS_STR_IP_ADDR                        2   // Optional (generated if missing).
#define DPWS_INT_BOOT_SEQ                       3   // Mandatory.
#define DPWS_BOOL_TCP_LISTEN                    4   // Optional (TRUE if missing).
#define DPWS_STR_HTTP_ADDR                      5   // Optional (generated if missing).
#define DPWS_INT_HTTP_PORT                      6   // Optional. 80
#define DPWS_INT_HTTP_BACKLOG                   7   // Optional. 10
#define DPWS_PTR_CALLBACK_HELLO                 8   // Optional. NULL. void(*)(struct dpws *, short)    // NOTE: DPWS rt structure is passed to be able to use user data
#define DPWS_PTR_CALLBACK_BYE                   9   // Optional. NULL. void(*)(struct dpws *, short)    // NOTE: DPWS rt structure is passed to be able to use user data
#define DPWS_INT_DEFAULT_DEVICE                 10  // Optional. href (short). Default -1.

/* toolkit configuration parameters */
#define DPWS_STR_PREFERRED_LANG                 11  // Optional.

/* toolkit configuration, service parameters */
#define DPWS_STR_MAX_SUBSC_DURATION             12  // Optional.

/* toolkit configuration, device class parameters */
#define DPWS_BOOL_MANAGEMENT_SERVICE            13  // Optional. TRUE or FALSE. Default FALSE.

/* device class, service class parameters */
#define DPWS_PTR_TYPE                           14  // Mandatory. qname struct.
#define DPWS_PTR_PREFIXED_TYPE                  15  // Mandatory. prefixed_qname struct.
#define DPWS_PTR_WSDL                           16  // struct wsdl_info
#define DPWS_STR_ID                             17  // Optional. default: string form of the object handle reference

/* device class, service class and endpoint parameters */
#define DPWS_PTR_FAULT_FUNCTION                 18  // Optional. int (*)(struct dpws*)
#define DPWS_PTR_CALLBACK_EVENT_END             19  // Optional. void (*) (struct event_end_info *)
#define DPWS_PTR_HANDLING_FUNCTION              20  // Mandatory. int (*)(struct dpws*)

/* device class parameters */
#define DPWS_STR_MODEL_NAME                     21  // Mandatory (@lang not supported yet).
#define DPWS_STR_MODEL_NUMBER                   22  // Optional.
#define DPWS_STR_MODEL_URL                      23  // Optional.
#define DPWS_STR_PRESENTATION_URL               24  // Optional.
#define DPWS_STR_MANUFACTURER                   25  // Mandatory (@lang not supported yet).
#define DPWS_STR_MANUFACTURER_URL               26  // Optional.

/* device class and instance parameters */
#define DPWS_INT_METADATA_VERSION               27  // Mandatory for class. Initialized to 0 for instance.

/* service port parameters */
#define DPWS_STR_ADDRESS                        28  // Optional. Relative or absolute. Default: URL made of device IP + generated UUID

/* service instance parameters */
#define DPWS_STR_SERVICE_ID                     29  // Optional. Default: service class ID

/* device instance parameters */
#define DPWS_STR_DEVICE_ID                      30  // Optional (generated if missing).
#define DPWS_STR_SCOPE                          31  // Optional. "http://schemas.xmlsoap.org/ws/2004/02/discovery/adhoc".
#define DPWS_STR_FRIENDLY_NAME                  32  // Mandatory (@lang not supported yet).
#define DPWS_STR_FIRMWARE_VERSION               33  // Optional.
#define DPWS_STR_SERIAL_NUMBER                  34  // Optional.

/* device, service instance and class parameters */
#define DPWS_PTR_USER_DATA                      35  // Optional.

/* device and service instance parameters */
#define DPWS_INT_CLASS_HREF                     36  // Optional.


/* Added by HG.W */
#ifdef USE_ONVIF
#define DPWS_STR_WAKEUP_SOCKS_DIR               40
#define DPWS_STR_EXTERNAL_PATH                  41
#endif

/*
 * Constants
 */
#define TOOLKIT_HANDLE      -1
#define DEVICE_SERVICE_ID   NULL
#define WSD_DEFAULT_SCOPE   "http://schemas.xmlsoap.org/ws/2004/02/discovery/adhoc"

#define WSD_MATCH_BY_UNSPECIFIED    -1

/* Added by HG.W (when onvif only support 3986 matching rule) */
#ifdef USE_ONVIF
#define WSD_MATCH_BY_RFC3986        4
#else
#define WSD_MATCH_BY_RFC2396        0
#define WSD_MATCH_BY_UUID           1
#define WSD_MATCH_BY_LDAP           2
#define WSD_MATCH_BY_STRCMP         3
#endif

#define WSD_LOOKUP_MODE_CACHE_FIRST     0
#define WSD_LOOKUP_MODE_CACHE_ONLY      1
#define WSD_LOOKUP_MODE_FORCE_NETWORK   2


#define DPWS_EXPIRATION_DURATION    0
#define DPWS_EXPIRATION_ABS_TIME    1


/* Custom C types */
#ifndef WIN32
# ifndef BOOLEAN
#  ifdef VXWORKS
typedef int BOOLEAN;
# else
typedef enum {FALSE = 0, TRUE = 1} BOOLEAN;
#  endif
# endif
#endif

typedef enum {UNKNOWN_PROXY_TYPE = 0, DEVICE_PROXY_TYPE = 1, SERVICE_PROXY_TYPE = 2} proxy_type;

/*
 * Initialization APIs
 */


/** DPWS tookit initialization function.
 * Must be called before any other API call.
 * @return a potential API error code.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_init();

/** Shutdowns the DPWS toolkit.
 * After this call, all DPWS toolkit resources are freed and a call to
 * dpws_init() must be performed to be able to operate again.
 * This api is called in the Windows DLL version delivery when a process
 * stops to use the DLL.
 * @return a potential API error code.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_shutdown();

/**
 * Creates an empty device class object.
 * @return A handle reference, if no error occurred. -1 otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_device_class();

/**
 * Creates an empty service class object.
 * @return A handle reference, if no error occurred. -1 otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_service_class();

/** Associates a service class to a device class.
 * Devices that will be created from the device class will host one service
 * by service class registered to the device class. If the registration is not
 * performed, hosted services will have to be created manually using
 * dpws_create_hosted service. Note that the 65535 value is reserved.
 * @param device_class_href The device class handle reference.
 * @param service_class_href The service class handle reference.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_register_service_class(short device_class_href, short service_class_href);

/** Creates a device instance object with all associated hosted services in 
 * one API call.
 * This function creates an operational device object that will host a set of
 * services corresponding to the service classes registered to the class of
 * the device. Each service created will also have its own HTTP service port
 * to which is will be bound.
 * @param local_id A numeric identifier for the device instance that, with the
 * listening port and the MAC address, will be used to generate the device
 * logical address (i.e., the uuid). Note that the 65535 value is reserved.
 * @param class_href The device class handle reference.
 * @return A handle reference, if no error occurred. a negative error code
 * otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_device(unsigned short local_id, short class_href);

/** Creates a device instance object without hosted services.
 * This function creates an operational device object that host no services.
 * The hosted services can be added later with the dpws_create_hosted_service
 * API.
 * @param local_id A numeric identifier for the device instance that, with the
 * listening port and the MAC address, will be used to generate the device
 * logical address (i.e., the uuid).
 * @param class_href The device class handle reference.
 * @return A handle reference, if no error occurred. a negative error code
 * otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_custom_device(unsigned short local_id, short class_href);

/** Creates a hosted service (and its associated HTTP service port) on a device
 * instance.
 * The hosted service can only be added if that service class is registered to
 * the class of the device. The service is created with the service class
 * identifier as identifier. The service id attribute is an attribute that can
 * be overwritten later.
 * @param device_href The device instance handle reference.
 * @param service_class_href The service class handle reference.
 * @return A handle reference, if no error occurred. a negative error code
 * otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_hosted_service(short device_href, short service_class_href);

/** Creates a new service port on the default HTTP listener. A default, HTTP
 * address is created for the port with the toolkit IP address and port plus
 * a generated UUID which can be superseded by any absolute or relative URL
 * by setting the DPWS_STR_ADDRESS parameter on the server object.
 * @return A handle reference, if no error occurred. a negative error code
 * otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_service_port();

/** Publishes a service endpoint on a specific address (service port).
 * @param hrefService The service endpoint handle reference.
 * @param hrefServicePort The service port handle reference.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_bind_service(short hrefService, short hrefServicePort);

/** Creates an endpoint and a bound service port. Services (and then devices)
 * are endpoints, from the DPWS specification point of vue. An endpoint can be
 * considered as a service which is not hosted, i.e., which is not advertized
 * in device's metadata.
 * The endpoint is a kind of service which can be configured like a service
 * class since its main use is to register handlers on it in order to receive
 * events or asynchronous responses.
 * An endpoint reference can be retrieved using dpws_get_local_endpoint_ref
 * @return A handle reference, if no error occurred. a negative error code
 * otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_create_endpoint();

/** Retrieves a device class handle by its id attribute.
 * @param id The device class id attribute.
 * @return A handle reference, if found. -1 otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_device_class_handle(char * id);

/** Retrieves a service class handle by its id attribute.
 * @param id The service class id attribute.
 * @return A handle reference, if found. -1 otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_service_class_handle(char * id);

/** Retrieves a device handle by its uuid.
 * This method may be used when a discovery event is received using the user
 * callback and device data has to be retrieved.
 * @param uuid the device uuid.
 * @return the handle reference for the device if local. -1 else.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_device_handle(char * uuid);

/** Retrieves a hosted service handle by its service identifier.
 * This function may be used to retrieve a hosted service when the
 * dpws_create_device opaque function was used to create a device.
 * @param device_href The device instance handle reference.
 * @param service_id the service identifier used as a key.
 * @return A handle reference, if found. -1 otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_service_handle(short device_href, char * service_id);

/** Retrieves a hosted service handle by its service class.
 * This function may be used to retrieve a hosted service when the
 * dpws_create_device opaque function was used to create a device.
 * @param device_href The device instance handle reference.
 * @param service_class_href The service class handle reference used as a key.
 * @return A handle reference, if found. -1 otherwise.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_service_handle_by_class(short device_href, short service_class_href);

/** Retrieves service handles attached to a device instance.
 * This function may be used to retrieve a hosted services when the
 * dpws_create_device opaque function was used to create a device.
 * @param device_href The device instance handle reference.
 * @param[in,out] hrefs A short array provided by the caller to receive results.
 * @param[in,out] len Must contain initially the length of the array hrefs.
 * When the function returns it contains the number of results actually retrieved.
 * @return DPWS_OK if the array supplied was large enough to receive results. 1,
 * if extra-results are available.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_get_service_handles(short device_href, short* hrefs, int* len);

/** Retrieves service port handle references bound to a local device or service
 * endpoint. This function may be used to retrieve service ports when the
 * dpws_create_device opaque function was used to create a device.
 * @param endpoint_href The (service) endpoint handle reference.
 * @param[in,out] hrefs A short array provided by the caller to receive results.
 * @param[in,out] len Must contain initially the length of the array hrefs.
 * When the function returns it contains the number of results actually retrieved.
 * @return DPWS_OK if the array supplied was large enough to receive results. 1,
 * if extra-results are available.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_get_service_port_handles(short endpoint_href, short* hrefs, int* len);

/** Get the default service port handle reference bound to a device or service
 * endpoint. Practically the first HTTP service port.
 * @param endpoint_href The (service) endpoint handle reference.
 * @return The handle reference of the default service port.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_default_service_port(short endpoint_href);

/** Allocates and initialize a wsa_endpoint_ref structure for a service port
 * attached to a device, an hosted service or any local endpoint.
 * Must be called after the call to dpws_server_init() or after the
 * DPWS_STR_HTTP_ADDR attribute has been set on the toolkit.
 * @param service_port_href A service port handle reference.
 * @return A heap allocated endpoint reference. The caller is responsible for
 * liberation when no more needed (using dpws_endpoint_ref_free for instance).
 */
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_get_local_endpoint_ref(short service_port_href);

/** Builds the physical address for a local service port.
 * Must be called after the call to dpws_server_init().
 * @param href_service_port The handle for a local service port.
 * @return A heap allocated URL. The caller is responsible for liberation
 * when not needed anymore.
 */
SOAP_FMAC1 char * SOAP_FMAC2 dpws_get_transport_address(short href_service_port);

/** Deletes a hosted service.
 * Undeploys a hosted service.
 * WARNING ! The current version DPWS does not really support "hot"
 * configuration. This API was added for DPWS Interop Workshop (October 2005)
 * purpose.
 * @param service_href The service handle reference.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_delete_hosted_service(short service_href);

/** DPWS metadata change commit function.
 * This function must be called after a live update of a device metadata
 * (increments metadata and send a hello).
 * WARNING ! The current version DPWS does not really support "hot"
 * configuration. This API was added for DPWS Interop Workshop (October 2005)
 * purpose.
 * @param device_href The device handle reference.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_update_metadata(short device_href);


/** Make a device available on the network.
 * This API schedules a Hello message that is sent if the server is running.
 * WARNING ! The current version DPWS does not really support "hot"
 * configuration.
 * @param device_href The device handle reference.
 * @return DPWS_OK, if no error occurred. A DPWS error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_enable_device(short device_href);

/** Removes a device from the network.
 * This API schedules a Bye message that is sent if the server is running.
 * WARNING ! The current version DPWS does not really support "hot"
 * configuration.
 * @param device_href The device handle reference.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_disable_device(short device_href);

/**
 * Used to set a server object attribute value for pointer parameter types
 * (strings included). If the attribute is multivalued, replace the existing
 * attributes.
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @param value The parameter value. Must be of the appropriate type, depending
 * on the attribute.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_set_ptr_att(short href, int att, const void * value);

/**
 * Used to set a server object attribute value for integer types (boolean
 * included). If the attribute is multivalued, replace the existing attributes.
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @param value The integer parameter value.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_set_int_att(short href, int att, unsigned long value);

/**
 * Used to add a server object attribute value for pointer parameter types
 * (strings included) when this attribute is multivalued. If the attribute is
 * monovalued, replace the existing attribute.
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @param value The parameter value. Must be of the appropriate type, depending
 * on the attribute.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_add_ptr_att(short href, int att, const void * value);

/**
 * Used to add a server object attribute value for integer types (boolean
 * included) when this attribute is multivalued. If the attribute is
 * monovalued, replace the existing attribute.
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @param value The integer parameter value.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_add_int_att(short href, int att, unsigned long value);

/** Returns server object attributes for pointer types (string or not) attributes.
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @return A pointer to the asked attribute value. Must not be freed since no
 * copy is performed. Multivalued attributes will be returned as NULL
 * terminated arrays.
 */
SOAP_FMAC1 void * SOAP_FMAC2 dpws_get_ptr_att(short href, int att);

/** Returns a specific instance of a multivalued server object attributes for
 * pointer types (string or not) attributes.
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @param index The index of the multivalued attribute.
 * @return A pointer to the asked attribute value. Must not be freed since no
 * copy is performed. Multivalued attributes will be returned as NULL
 * terminated arrays.
 */
SOAP_FMAC1 void * SOAP_FMAC2 dpws_get_ptr_att_item(short href, int att, int index);

/** Returns server object attributes of integer type (or boolean).
 * @param href The server object handle reference.
 * @param att The attribute identifier.
 * @return The asked attribute value.
 */
SOAP_FMAC1 unsigned long SOAP_FMAC2 dpws_get_int_att(short href, int att);

/** Releases a toolkit configuration object handle when not needed anymore by
 * the user. This is good practice to release object handles in order to avoid
 * memory exhaustion.
 * @param href The handle reference on the object to release.
 * @return DPWS_OK, if no error occurred. A DPWS error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_release_handle(short href);


/*
 * Server Runtime APIs
 */
// TODO: make endpoint_ref disappear of the API ? hrefs ? Or simple strings ?

/** Initialize the dpws structure and perform server-side initialization.
 * The initialization includes:
 * - the creation of the listening sockets,
 * - the scheduling of hello messages for all enabled devices,
 * - the mandatory configuration parameters check (according to the server mode),
 * - default parameters setting. Must be called before starting the main server
 *   loop.
 * This function sets automatically the server mode according to the
 * configuration calls performed before:
 *  - Level 1: WS-Discovery listener (no endpoint have been created but at least
 *    one discovery handler has been registered),
 *  - Level 2: WS-Eventing listener (one endpoint at least has been created but
 *    no device),
 *  - Level 3: Device (one device at least has been created).
 * @param[in,out] dpws A pointer on the main DPWS API runtime data structure.
 * @param source This structure will be copied in every SOAP message as the WSA
 * source endpoint. The structure passed by the user will be copied and can be
 * released.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise. For instance
 * if a needed parameter is missing and no default value is available.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_server_init(struct dpws* dpws, struct wsa_endpoint_ref* source);
#ifdef WITH_IPV6 /* Added by HG.W */
SOAP_FMAC1 int SOAP_FMAC2 dpws_server_init_v6(struct dpws* dpws, struct wsa_endpoint_ref* source);
#endif

/** Blocks and wait for an incoming SOAP message.
 * To be used in a server or handler context.
 * @param dpws A pointer on the main DPWS API runtime data structure. When returning
 * this structure contains both the listening and the accepted sockets.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_accept(struct dpws* dpws);

/** Blocks and wait for an incoming SOAP message.
 * To be used in a server or handler context. This alternate version of
 * dpws_accept is particularly suited to multi-thread implementations since it
 * make unnecessary to perform the dpws_copy_contex call.
 * @param dpws A pointer on the main DPWS API runtime data structure. When returning
 * this structure is not altered.
 * @param dpws_thr A pointer on the service thread target DPWS API runtime data
 * structure. When returning this structure contains the accepted socket.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_accept_thr(struct dpws* dpws, struct dpws* dpws_thr);

/** Handles a request message.
 * This function starts reading the SOAP envelope, the SOAP headers and the
 * SOAP body start tag before delegating further processing to the message
 * handling function passed by a previous call to the server configuration
 * functions.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_serve(struct dpws * dpws);

/** Stops smoothly the locally running server for multi-thread servers.
 * This function schedules the Eventing end notifications and Bye WS-Discovery
 * messages for all the active devices, unblocks the main loop blocking on dpws_accept(
 * and closes all open listening sockets. This functions blocks first until the scheduled
 * messages are all being processed by server threads and lets an additional amount
 * of time to let the server threads to send the messages. The dpws_stop_server_no_wait()
 * function should be preferred if the developper can supply thread synchronization to
 * avoid server threads to be stopped before having performed their tasks or in mono-thread
 * servers where the call is performed in the server thread.
 * A new call to dpws_server_init() must be then performed to restart the server.
 * @param delay_ms the number of milliseconds to wait after the last "bye" task has been
 * passed to a server thread and that must allow the thread to send pending messages. Must
 * be tuned, especially if the number of threads is high. This is why one should use
 * dpws_stop_server_no_wait() whenever possible.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_stop_server(long delay_ms);

#ifdef WITH_IPV6
SOAP_FMAC1 int SOAP_FMAC2 dpws_stop_server_v6(long delay_ms);
#endif

/** Stops smoothly the locally running server.
 * This function schedules the Eventing end notifications and Bye WS-Discovery
 * messages for all the active devices, unblocks the main loop blocking on dpws_accept(
 * and closes all open listening sockets. This functions returns immediately so either
 * the developper can supply thread synchronization to avoid server threads to be
 * stopped before having performed their tasks or use it in mono-thread servers.
 * A new call to dpws_server_init() must be then performed to restart the server.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_stop_server_no_wait();

/** Retrieves the user data associated to the endpoint that received the
 * message.
 * @param dpws A pointer on the main DPWS API runtime data structure.
 * @return A pointer on the data set on the device by the user.
 */
SOAP_FMAC1 void * SOAP_FMAC2 dpws_get_endpoint_user_data(struct dpws* dpws);

/** Retrieves the user data associated to the device instance that received the
 * message.
 * @param dpws A pointer on the main DPWS API runtime data structure.
 * @return A pointer on the data set on the device by the user.
 */
SOAP_FMAC1 void * SOAP_FMAC2 dpws_get_device_user_data(struct dpws* dpws);

/** Clears structures and free dynamically allocated memory used for request
 * processing.
 * Must be called after processing a message on a client, a server or a
 * handler.
 * @param dpws A pointer on a dpws data structure that was used to process the
 * request.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_end(struct dpws* dpws);


/*
 * Client Runtime APIs
 */
 
// WS-Discovery APIs

/** Initialize the dpws structure and perform one-time client-side
 * initialisation.
 * Must be called once before using a dpws structure.
 * @param[in,out] dpws A pointer on a DPWS API runtime data structure.
 * @param source This structure will be copied in every SOAP message as the WSA
 * source endpoint. The structure passed by the user will be copied and can be
 * released.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_client_init(struct dpws* dpws, struct wsa_endpoint_ref* source);
#ifdef WITH_IPV6 /* Added by HG.W */
SOAP_FMAC1 int SOAP_FMAC2 dpws_client_init_v6(struct dpws* dpws, struct wsa_endpoint_ref* source);
#endif

/** Retrieves one or several device proxies basing on a type and a scope.
 * This function will first look for proxies into the local cache, and then
 * generate a WS-Discovery Probe message if the number of requested proxies
 * is not available in the cache. This function implements the semantics of the
 * WS-Discovery Probe message, for a single device type. Note that this API
 * uses a timeout which value is the MATCH_TIMEOUT constant defined in the
 * WS-Discovery specifications and is 600 ms. To specify a custom value, use
 * the dpws_lookup_ex API.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param ns The target namespace of the device type. May be NULL only if the
 * next parameter (type) is also NULL.
 * @param type The local name of the device type. May be null, in which case no
 * filtering on type is performed.
 * @param scope An URI specifying the scope of the search. May be NULL, in
 * which case the scope is unconstrained. The matching rule used for scope is
 * the default one (RFC2396).
 * @param[in,out] res_size The maximum number of proxies to retrieve. When
 * the function returns, it contains the actual number of retrieved device
 * proxies. WARNING : if 0 or less is supplied, the number of devices to
 * be retrieved is not constrained.
 * @return An array of device proxy handle references that will be released
 * during the next call to dpws_end in order to make associated resources
 * available for liberation. To avoid the proxy to be freed and recycled, one
 * should use the dpws_pin_proxy api. Note that between the moment it is
 * retrieved and the one it is used, the proxy may become invalid.
 * The size of the returned array can be retrieved using the output value of
 * the res_size parameter. Note: the returned array is dynamically allocated
 * on the gSOAP heap. This data is available until dpws_end is called to free
 * memory. If an error occurs or no device is found, NULL is returned. To
 * distinguish between the two, use the res_size fields that is 0 is no
 * device were retrieved and -1 in case of error.
 */
SOAP_FMAC1 short * SOAP_FMAC2 dpws_lookup(struct dpws* dpws, char * ns, char * type, char * scope, int *res_size);

/** Retrieves one or several device endpoints basing on types and scopes.
 * This function implements the semantics of the WS-Discovery Probe message,
 * for a single device type. A matching device must match all the types and
 * scopes provided to the API. This is the complete version of the lookup
 * functions which enables fine-grained control of the API behavior.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param types A null-terminated qname struct array of port types used for
 * filtering. The last element must be a qname struct filled with zeros.
 * @param scopes A null-terminated URI string array of scopes used for
 * filtering.
 * @param match_rule The matching rule to be used for scope comparison. Should
 * be one of:
 *  - WSD_MATCH_BY_RFC2396,
 *  - WSD_MATCH_BY_UUID,
 *  - WSD_MATCH_BY_LDAP (not supported yet),
 *  - WSD_MATCH_BY_STRCMP,
 *  - WSD_MATCH_BY_UNSPECIFIED.
 * In the latter case, the default matching rule is used.
 * @param[in,out] res_size The maximum number of endpoints to retrieve. When
 * the function returns, it contains the actual number of retrieved device
 * endpoints. WARNING : if 0 or less is supplied, the number of devices to
 * be retrieved is not constrained which could lead to memory exhaustion.
 * @param time_out A timeout in milliseconds after which the search should
 * return, when the maximum number of requested endpoints has not been reached.
 * This must be a positive value. Note that is parameter is upbounded by the
 * DPWS constant MATCH_TIMEOUT which is 10 seconds.
 * @param mode A specification on how the API should use the local cache.
 * Specify a value among:
 *  - WSD_LOOKUP_MODE_CACHE_FIRST: The defaut mode where the cache will be
 *    checked first before sending a probe on the network if retrieved
 *    endpoints are not as many as asked.
 *  - WSD_LOOKUP_MODE_CACHE_ONLY: Only look in the local cache whatever be the
 *    number of endpoints retrieved.
 *  - WSD_LOOKUP_MODE_FORCE_NETWORK: Forces a cache update with a network probe.
 *    Note that invalid endpoint lying in the cache (because no Bye message was
 *    received) are not discarded an may be retrieved.
 * @return An array of device proxy handle references that will be released
 * during the next call to dpws_end in order to make associated resources
 * available for liberation. To avoid the proxy to be freed and recycled, one
 * should use the dpws_pin_proxy api. Note that between the moment it is
 * retrieved and the one it is used, the proxy may become invalid.
 * The size of the returned array can be retrieved using the output value of
 * the res_size parameter. Note: the returned array is dynamically allocated
 * on the gSOAP heap. This data is available until dpws_end is called to free
 * memory. If an error occurs or no device is found, NULL is returned. To
 * distinguish between the two, use the res_size fields that is 0 is no
 * device were retrieved and -1 in case of error.
 */
SOAP_FMAC1 short * SOAP_FMAC2 dpws_lookup_ex(struct dpws* dpws, struct qname *types, char ** scopes, int match_rule, int * res_size, long time_out, int mode);

/** Checks a DPWS device is available on the local network and retrieve its
 * proxy.
 * This function will first look for a device proxy into the local cache, and
 * then generate a WS-Discovery Resolve message if the requested endpoint is
 * not available in the cache. This function implements the semantics of the
 * WS-Discovery Resolve message. Note that this API uses a timeout which value
 * is the MATCH_TIMEOUT constant defined in the WS-Discovery specifications
 * and is 600 ms.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param uuid The device logical ID.
 * @return The device proxy handle reference if found. -1, if not. The handle
 * will be released during the next call to dpws_end in order to make associated
 * resources available for liberation. To avoid the proxy to be freed and
 * recycled, one should use the dpws_pin_proxy api. Note that between the
 * moment it is retrieved and the one it is used, the proxy may become invalid.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_lookup_by_id(struct dpws* dpws, char * uuid);

/** Retrieves a device proxy checking a physical address matches a type and
 * a scope.
 * This function issues a directed WS-Discovery Probe message to the supplied
 * physical address and implements the semantics of this message.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param address The physical address to which the directed probe is sent.
 * @param ns The target namespace of the type, as defined in the WSDL document.
 * May be NULL only if the next parameter (type) is also NULL.
 * @param type The local name of the device type. May be null, in which case no
 * filtering on type is performed.
 * @param scope An URI specifying the scope of the search. May be NULL, in
 * which case the scope is unconstrained. The matching rule used for scope is
 * the default one (RFC2396).
 * @return The device proxy handle reference if the device matched the
 * requirements at the supplied address. -1, if not. The handle
 * will be released during the next call to dpws_end in order to make associated
 * resources available for liberation. To avoid the proxy to be freed and
 * recycled, one should use the dpws_pin_proxy api. Note that between the
 * moment it is retrieved and the one it is used, the proxy may become invalid.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_probe_address(struct dpws* dpws, char* address, char * ns, char * type, char * scope);

/** Retrieves a device endpoint checking a physical address matches a set of
 * types and scopes.
 * This function issues a directed WS-Discovery Probe message to the supplied
 * physical address and implements the semantics of this message.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param address The physical address to which the directed probe is sent.
 * @param types A null-terminated qname struct array of port types used for
 * filtering. The last element must be a qname struct filled with zeros.
 * @param scopes A null-terminated URI string array of scopes used for
 * filtering.
 * @param match_rule The matching rule to be used for scope comparison. Should
 * be one of:
 *  - WSD_MATCH_BY_RFC2396,
 *  - WSD_MATCH_BY_UUID,
 *  - WSD_MATCH_BY_LDAP (not supported yet),
 *  - WSD_MATCH_BY_STRCMP,
 *  - WSD_MATCH_BY_UNSPECIFIED.
 * In the latter case, the default matching rule is used.
 * @return The device proxy handle reference if the device matched the
 * requirements at the supplied address. -1, if not. The handle
 * will be released during the next call to dpws_end in order to make associated
 * resources available for liberation. To avoid the proxy to be freed and
 * recycled, one should use the dpws_pin_proxy api. Note that between the
 * moment it is retrieved and the one it is used, the proxy may become invalid.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_probe_address_ex(struct dpws* dpws, char* address, struct qname *types, char ** scopes, int match_rule);

/** Retrieve one or several service proxies basing on a device endpoint and a
 * type.
 * This function will look for endpoints into the local cache or generate a
 * WS-MetadataExchange GetMetadata message if necessary.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param href_device A valid device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @param ns The target namespace of the port type, as defined in the WSDL
 * document.
 * @param port_type The name of the port type. May be null, in which case no
 * filtering on type is performed.
 * @param[out] res_size The number of service proxies retrieved.
 * @return An array of svice proxy handle references that will be released
 * during the next call to dpws_end in order to make associated resources
 * available for liberation. To avoid the proxy to be freed and recycled, one
 * should use the dpws_pin_proxy api. Note that between the moment it is
 * retrieved and the one it is used, the proxy may become invalid.
  * The size of the returned array can be retrieved using the output value of
 * the res_size parameter. Note: the returned array is dynamically allocated
 * on the gSOAP heap. This data is available until dpws_end is called to free
 * memory. If an error occurs or no service is found, NULL is returned. To
 * distinguish between the two, use the res_size fields that is 0 is no
 * device were retrieved and -1 in case of error.
 */
SOAP_FMAC1 short * SOAP_FMAC2 dpws_get_services(struct dpws* dpws, short href_device, char * ns, char * port_type, int *res_size);

#define dpws_get_service dpws_get_services  // for compatibility

/** Return the type of proxy returned by the dpws_get_services. This
 * API is useful to set if dpws_get_device_metadata or dpws_get_service_metadata
 * must be used to retrieve proxy metadata.
 * @param href_proxy The service proxy handle reference (a device is also a
 * service).
 * @return An enum value representing the proxy type.
 */
SOAP_FMAC1 proxy_type SOAP_FMAC2 dpws_get_proxy_type(short href_proxy);

/** Retrieve a service proxy based on its service ID.
 * This function will first look for the service into the local cache, and
 * then generate a WS-MetadataExchange GetMetadata message if the requested
 * service is not available in the cache. This function implements the
 * semantics of the WS-MetadataExchange GetMetadata message.
 * @param dpws A pointer on a dpws runtime data structure used to process the
 * request.
 * @param href_device A valid device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @param id The service ID.
 * @return The retrieved service proxy handle reference, if available, -1
 * otherwise. Note: the returned endpoint is dynamically allocated on the
 * gSOAP heap. This data is available until dpws_end is called to free memory.
 * @return The service proxy handle reference if found. -1, if not. The handle
 * will be released during the next call to dpws_end in order to make associated
 * resources available for liberation. To avoid the proxy to be freed and
 * recycled, one should use the dpws_pin_proxy api. Note that between the
 * moment it is retrieved and the one it is used, the proxy may become invalid.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_service_by_id(struct dpws* dpws, short href_device, char * id);

/** Retrieves the scopes of a device in the cache.
 * @param dpws A pointer on a dpws data structure used to process the API call.
 * @param href_device A valid device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @return a null-terminated array of strings featuring the scopes. Data is
 * allocated on the dpws structure and will be freed during the next call to
 * dpws_end. An empty result may result from an API error.
 */
SOAP_FMAC1 char ** SOAP_FMAC2 dpws_cache_get_scopes(struct dpws* dpws, short href_device);

/** Retrieves the uuid of a device in the cache.
 * @param dpws A pointer on a dpws data structure used to process the API call.
 * @param href_device A valid device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @return a string featuring the device uuid. Data is allocated on the dpws
 * structure and will be freed during the next call to dpws_end. An empty result
 * may result from an API error.
 */
SOAP_FMAC1 char * SOAP_FMAC2 dpws_cache_get_uuid(struct dpws* dpws, short href_device);

/** Checks a device is of a given type.
 * This function may be used when a discovery callback has been triggered.
 * @param dpws A pointer on a dpws runtime data structure.
 * @param ns The target namespace of the type, as defined in the WSDL document.
 * May be NULL only if the next parameter (type) is also NULL.
 * @param type The local name of the device type. May be null, in which case no
 * filtering on type is performed.
 * @param match_rule The matching rule to be used for scope comparison. Should
 * be one of:
 *  - WSD_MATCH_BY_RFC2396,
 *  - WSD_MATCH_BY_UUID,
 *  - WSD_MATCH_BY_LDAP (not supported yet),
 *  - WSD_MATCH_BY_STRCMP,
 *  - WSD_MATCH_BY_UNSPECIFIED.
 * In the latter case, the default matching rule is used.
 * @param href_device The device proxy handle reference.
 * @return 1 if the device is of the specified type. 0, if not. A negative API
 * code if an error occured.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_check_device_type(struct dpws* dpws, char * ns, char * type, int match_rule, short href_device);

/** Retrieves the port types for a device or service proxy lying in the cache.
 * @param dpws A pointer on a dpws data structure used to process the API call.
 * @param href_proxy The device or hosted service proxy.
 * @return a null-terminated array of pointers on qname structures featuring
 * the port types. Data is allocated on the dpws structure and will be freed
 * during the next call to dpws_end. An empty result may result from an API
 * error.
 */
SOAP_FMAC1 struct qname ** SOAP_FMAC2 dpws_cache_get_types(struct dpws* dpws, short href_proxy);

/** Retrieves a cached proxy handle reference basing on an id. For device
 * proxies, the id considered is the device UUID. For service proxies, the
 * service ID is defined by the DPWS specifications.
 * @param id the proxy id to be searched.
 * @return the handle reference if retrieved. -1 else.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_proxy_by_id(char * id);

/** Retrieves a cached proxy handle reference basing on a physical address.
 * @param address the proxy physical address to be searched.
 * @return the handle reference if retrieved. -1 else.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_proxy_by_address(char * address);

/** Check if a cached device or service is still valid from the cache point of
 * view.
 * @param href_proxy The proxy handle reference to be checked.
 * @return TRUE if the handle exist and is valid.
 */
SOAP_FMAC1 BOOLEAN SOAP_FMAC2 dpws_check_proxy(short href_proxy);

/** Increase the use counter of a cached device or service so that is cannot
 * be freed by the toolkit even if invalid from the cache point of view. When the
 * proxy is not used anymore, the user should use dpws_release_proxy to let it be
 * freed.
 * @param href_proxy The proxy handle reference to be pin.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_pin_proxy(short href_proxy);

/** Decrease the use counter of a cached device or service so that it can be
 * freed by the toolkit if invalid from the cache point of view.
 * @param href_proxy The proxy handle reference to be released.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_release_proxy(short href_proxy);

/** Gets the default endpoint reference for a device or service proxy.
 * In this version, the first endpoint is considered. If a specific endpoint
 * reference is required, use the dpws_get_endpoint_refs.
 * @param dpws A pointer on a dpws data structure used to allocate the result.
 * If NULL, the result is allocated on the process heap.
 * @param href_proxy The device or service proxy which endpoint reference must
 * be retrieved for.
 * @return An endpoint reference structure. Must be freed by the user using
 * dpws_end if a dpws structure was provided, dpws_endpoint_ref_free else.
 */
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_get_default_endpoint_ref(struct dpws* dpws, short href_proxy);

#define dpws_get_default_endpoint_reference dpws_get_default_endpoint_ref   // for compatibility

/** Gets all the endpoint references retrieved for a device or service proxy.
 * @param dpws A pointer on a dpws data structure used to allocate the result.
 * If NULL, the result is allocated on the process heap.
 * @param href_proxy The device or service proxy which endpoint references must
 * be retrieved for.
 * @param[out] res_size The number of endpoints retrieved.
 * @return An array of endpoint references.
 * The returned array and endpoint structures are dynamically allocated
 * on the gSOAP or process heap. This data must be freed using dpws_end or 
 * the  dpws_endpoint_ref_array_free utility according to the fact that a dpws 
 * structure has been provided or not. If an error occurs or no proxy is found,
 * NULL is returned. To distinguish between the two, use the res_size
 * fields that is 0 is no endpoint were retrieved and -1 in case of error.
 */
SOAP_FMAC1 struct wsa_endpoint_ref ** SOAP_FMAC2 dpws_get_endpoint_refs(struct dpws* dpws, short href_proxy, int * res_size);

#define dpws_get_endpoint_references dpws_get_endpoint_refs // for compatibility

/** Retrieves the metadata version of a cached device proxy.
 * @param href_device The device proxy.
 * @return The metadata version or a DPWS error code.
 */
SOAP_FMAC1 unsigned long SOAP_FMAC2 dpws_cache_get_device_metadata_version(short href_device);

/** Retrieves in the cache the URLs for a device.
 * @param dpws A pointer on a dpws data structure used to allocate the result.
 * @param href_device The device proxy handle reference.
 * @param[out] res_size The number of URLs retrieved.
 * @return An array of strings containing the transport addresses of the device.
 */
SOAP_FMAC1 char ** SOAP_FMAC2 dpws_cache_get_device_transport_addresses(struct dpws* dpws, short href_device, int * res_size);

/** Remove a device or service proxy from the cache.
 * This function should be used when a call to a service using a given endpoint
 * is unsuccessful because the endpoint cannot be reached.
 * @param href The device or hosted service proxy href.
 * @return DPWS_OK, if no error occurred. A DPWS code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_invalidate_proxy(short href);

// WS-Metadata Exchange APIs

/** Retrieves the whole DPWS metadata for a device.
 * This API uses the WS-Transfer Get request and the WS-MetadataExchange message
 * format. Like, does not cache the results of network
 * queries performed so that a query is performed every time.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param href_device The device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @return A structure containing metadata allocated on the dpws structure (
 * freed during the next call to dpws_end).
 */
SOAP_FMAC1 struct device_metadata * SOAP_FMAC2 dpws_get_device_metadata(struct dpws* dpws, short href_device);

/** Retrieves the service DPWS metadata for a hosted service.
 * This function uses no dialect in the GetMetadata request. This API, like
 * all the functions WS-Metadata Exchange APIs, does not cache the results of
 * network queries performed so that a query is performed every time.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param href_service A hosted service proxy handle reference.
 * @return A structure containing metadata allocated on the dpws structure (
 * freed during the next call to dpws_end).
 */
SOAP_FMAC1 struct service_metadata * SOAP_FMAC2 dpws_get_service_metadata(struct dpws* dpws, short href_service);

/** Retrieves in the cache the hosting device proxy of a hosted service.
 * This function only uses the cache.
 * @param href_service A hosted service proxy handle reference.
 * @return The host device proxy handle reference. -1 if not found.
 */
SOAP_FMAC1 short SOAP_FMAC2 dpws_get_host_device_proxy(short href_service);


// WS-Eventing APIs

/** Subscribes to a WS-Eventing event source.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param event_source A valid endpoint (that may have been retrieved using
 * dpws_lookup or dpws_get_services) of an WS-Eventing event source.
 * @param notify_to A valid device endpoint (that may have been retrieved using
 * dpws_lookup) that will receive the events.
 * @param end_to A valid device endpoint (that may have been retrieved using
 * dpws_lookup) that will receive abnormal subscription termination. Optional
 * (if NULL no termination event is not sent).
 * @param filter A NULL-terminated string array of DPWS actions to limit events
 * to receive from the event source.
 * @param[in,out] expiration An XML Schema dateTime or duration, to specify the
 * subscription duration requested. This parameter will be modified with the
 * value granted by the subscription manager.
 * @return The subscription manager endpoint that shall be used as a ticket to
 * manage the subscription. NULL if an error occurred.
 */
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_event_subscribe(struct dpws* dpws, struct wsa_endpoint_ref* event_source, struct wsa_endpoint_ref* notify_to, struct wsa_endpoint_ref* end_to, char ** filter, char ** expiration);

/** Renews a subscription to a WS-Eventing event source.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param subscription_manager The subscription ticket gained during initial
 * call to dpws_event_subscribe.
 * @param expiration XML Schema dateTime or duration, to specify the new
 * subscription duration requested.
 * @return The subscription duration (XML Schema dateTime or duration) granted
 * by the manager. NULL if an error occurred.
 */
SOAP_FMAC1 char * SOAP_FMAC2 dpws_event_renew(struct dpws* dpws, struct wsa_endpoint_ref* subscription_manager, char* expiration);

/** Gets the remaining duration for a subscription to a WS-Eventing event source.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param subscription_manager The subscription ticket gained during initial
 * call to dpws_event_subscribe.
 * @return The subscription duration (XML Schema dateTime or duration) still
 * running. NULL if an error occurred.
 */
SOAP_FMAC1 char * SOAP_FMAC2 dpws_event_status(struct dpws* dpws, struct wsa_endpoint_ref* subscription_manager);

/** Cancel a running subscription to a WS-Eventing event source.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param subscription_manager The subscription ticket gained during initial
 * call to dpws_event_subscribe.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_event_unsubscribe(struct dpws* dpws, struct wsa_endpoint_ref* subscription_manager);

// Metadata Manager APIs

/** Changes a remote device scope that exposes the MDM service.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param href_device The device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @param scopes A NULL-terminated URI array that specify the new scopes for
 * the device.
 * @return The scopes actually set for the device (a NULL-terminated URI
 * array). If null, may result from an API error or a remote fault.
 */
SOAP_FMAC1 char ** SOAP_FMAC2 dpws_set_device_scopes(struct dpws* dpws, short href_device, char ** scopes);

/** Changes a remote device friendly name that exposes the MDM service.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @param href_device The device proxy handle reference that may have been
 * retrieved using dpws_lookup.
 * @param names A NULL-terminated localized string array that specify the new
 * friendly names for the device. NOTE: the current version of the toolkit only
 * supports one friendly name so that the remote device may choose one name
 * among the one supplied (basing on the preferred language).
 * @return The friendly names actually set for the device (a NULL-terminated
 * array). If null, may result from an API error or a remote fault.
 */
SOAP_FMAC1 struct localized_string ** SOAP_FMAC2 dpws_set_device_friendly_names(struct dpws* dpws, short href_device, struct localized_string ** names);

// Other

/** Receives one way or asynchronous empty response.
 * May be used only if a client wants to block waiting for and empty response.
 * @param dpws A pointer on a dpws runtime data structure.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_recv_empty_response(struct dpws* dpws);

/*
 * Utilities
 */
 
/** Returns the last DPWS or gSOAP runtime error.
 * @param dpws A pointer on a dpws data structure used when the error occured.
 * @return A DPWS or gSOAP api error code.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_get_error(struct dpws* dpws);

/** Gets a message for a SOAP fault or a DPWS/gSOAP error that has
 * previously occurred.
 * @param dpws A pointer on a dpws runtime data structure used when the error
 * occured.
 * @return A readonly error message.
 */
SOAP_FMAC1 const char * SOAP_FMAC2 dpws_get_error_msg(struct dpws* dpws);

/** Initialize a wsa_endpoint_ref structure that the user may use for a
 * subsequent API call.
 * @param endpoint_ref his structure will be reset.
 */
SOAP_FMAC1 void SOAP_FMAC2 dpws_default_wsa_endpoint_ref(struct wsa_endpoint_ref *endpoint_ref);

/** Allocates and initialize a wsa_endpoint_ref structure for a device.
 * No verification is performed for device existence (this could be done thanks
 * to the dpws_lookup_by_id API).
 * @param uuid A device uuid.
 * @return A DPWS library heap allocated endpoint reference. The caller is
 * responsible for liberation when no more needed (using dpws_endpoint_ref_free).
 */
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_create_device_endpoint_ref(char * uuid);

/** Creates a deep copy of a wsa_endpoint_ref structure on DPWS library heap.
 * May be used for instance to keep a WS-Eventing subscription endpoint after
 * dpws_end has been called.
 * @param endpoint The structure to copy.
 * @return A heap allocated endpoint reference. The caller is responsible for
 * liberation when no more needed (using dpws_endpoint_ref_free).
 */
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_endpoint_ref_dup(struct wsa_endpoint_ref * endpoint);

/** Copy of a wsa_endpoint_ref structure to a provided destination.
 * This function uses the standard malloc function to allocate the structure
 * contents.
 * @param dest he destination structure.
 * @param src The structure to copy.
 * @return The dest param. The API user will have the responsibility to free
 * the allocated contents of the structure using dpws_dpws_endpoint_ref_content_free.
 */
SOAP_FMAC1 struct wsa_endpoint_ref * SOAP_FMAC2 dpws_endpoint_ref_copy(struct wsa_endpoint_ref * dest, struct wsa_endpoint_ref * src);

/** Frees a string or any buffer allocated on the DPWS library heap.
 * @param buffer The memory block to be freed.
 */
SOAP_FMAC1 void SOAP_FMAC2 dpws_free(void * buffer);

/** Frees the contents of an endpoint reference.
 * Liberates the contents of a wsa_endpoint_ref structure allocated on the
 * DPWS library heap. This function can be used along with dpws_endpoint_ref_copy.
 * @param endpoint The structure which contents must be freed.
 */
SOAP_FMAC1 void SOAP_FMAC2 dpws_endpoint_ref_content_free(struct wsa_endpoint_ref * endpoint);

/** Frees an endpoint reference and its contents.
 * This function can be used along with dpws_endpoint_ref_dup.
 * @param endpoint The structure which must be freed.
 */
SOAP_FMAC1 void SOAP_FMAC2 dpws_endpoint_ref_free(struct wsa_endpoint_ref * endpoint);

/** Frees an endpoint reference array and its contents.
 * This function can be used along with dpws_endpoint_ref_dup that allows for
 * instance the user to keep the selected endpoint reference before array
 * deletion.
 * @param endpoints A array of endpoint reference pointers.
 * @param nb_endpoints The size of the array to free.
 */
SOAP_FMAC1 void SOAP_FMAC2 dpws_endpoint_ref_array_free(struct wsa_endpoint_ref ** endpoints, int nb_endpoints);

/** Retrieves the system time in an UNIX-like date format.
 * @return An absolute date featuring the seconds since 00:00:00 GMT,January 1,1970.
 */
SOAP_FMAC1 time_t SOAP_FMAC2 get_system_time();

/** Converts an XML Schema duration value into an UNIX-like date.
 * @param xml_duration A valid XML Schema duration.
 * @return An absolute date calculated from the system time and the duration
 * specified. The time features the seconds since 00:00:00 GMT,January 1,1970.
 */
SOAP_FMAC1 time_t SOAP_FMAC2 xmlduration2time(char *xml_duration);

/** Converts an UNIX-like date into an XML Schema duration value.
 * @param dest The buffer to receive the XML duration. Note that the 
 * DURATION_MAX_SIZE macro can be used to size the buffer.
 * @param time An absolute datetime to convert. The time features the seconds
 * since 00:00:00 GMT,January 1,1970.
 * @return An XML Schema duration calculated basing on the system time.
 */
SOAP_FMAC1 char * SOAP_FMAC2 time2xmlduration(char * dest, time_t time);

/** A portable version of a case-insensitive strcmp.
 * May be used to compare uuid, for instance.
 * @param s1 The first string to compare.
 * @param s2 The second string to compare.
 * @return 0 if strings are equal. <0 if s1 is alphabetically before s2. >0 else.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_strcasecmp(const char *s1, const char *s2);

#ifdef __cplusplus
}
#endif

#endif
