/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.15.4.2 $
|                                              $Date: 2007/03/23 16:41:34 $
\*============================================================================*/
 
/*! \file dpwsX.h
    DPWS Toolkit eXtended API Reference.
*/

/*******************************************************************************
*                                 PUBLIC API                                   *
*******************************************************************************/
#ifndef DPWSX_H
# define DPWSX_H


#include "dpws.h"

#ifdef __cplusplus
extern "C" {
#endif

// Generic msg handling types

typedef int (* wscbk) (struct dpws *, void *, char *);			//  dpws, cbk_data, msgBody

// Generic message functions

/** Sends a custom SOAP message to a service endpoind using a one-way message
 * pattern.
 * This version uses single-byte characters for the XML body.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @action action The WS-Addressing action.
 * @param to A destination service endpoint.
 * @param XML_msg_body A well-formed SOAP message body.
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_send_message(struct dpws *dpws, char * action, struct wsa_endpoint_ref *to, char *XML_msg_body);

/** Invokes a service using custom SOAP messages.
 * This version uses single-byte characters for the XML body. Use UTF-8 encoding
 * for Unicode.
 * @param dpws A pointer on a dpws data structure used to process the request.
 * @action action The WS-Addressing action.
 * @param to A destination service endpoint.
 * @param reply_to The endpoint to which the service must send the response. If
 * this parameter is NULL, the response will be sent to the calling endpoint
 * and retrieved in the XML_response_body parameter. Else the function returns
 * without receiving the response message (asynchronous call).
 * @param XML_request_body A well-formed SOAP message body used for the request.
 * @param[in,out] response The well-formed SOAP message body of the
 * response in no reply_to was specified. The message ID of the request in case
 * of asynchronous invokation (reply_to NULL).
 * @return DPWS_OK, if no error occurred. A DPWS or gSOAP error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_call_service(struct dpws *dpws, char * action, struct wsa_endpoint_ref *to, struct wsa_endpoint_ref *reply_to, char *XML_request_body, char **response);

/** Registers a generic handler on a local endpoint.
 * @param href_endpoint The handle reference of a local endpoint that will
 * receive events or asynchronous responses. The endpoint may have been created
 * using dpws_create_endpoint or be an hosted service.
 * @param callback The generic handler.
 * @param cbk_data Callback private data enabling pseudo-polymorphism.
 * @return DPWS_OK, if no error occurred. A DPWS API error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_register_generic_handler(short href_endpoint, wscbk callback, void *cbk_data);	// TODO: check handle_msg is registered once on an endpoint

/** Helper function that frees the callback entry of the endpoint.
 * The callback entry is created on the endpoint by the dpws_register_generic_handler
 * and  dpws_event_subscribe_generic and should be freed when shuting down the server.
 * @param href_endpoint The handle reference of the local endpoint that will
 * be cleaned.
 * @param clean_cbk called when the callback entry is freed with the callback data as
 * a 1st parameter and cbk_param as a second parameter.
 * @param cbk_param an additional parameter passed to the clean callback.
 * @return DPWS_OK, if no error occurred. A DPWS API error code otherwise.
 */
SOAP_FMAC1 int SOAP_FMAC2 dpws_clean_endpoint(short href_endpoint, void (* clean_cbk) (void *, void *), void *cbk_param);

#ifdef __cplusplus
}
#endif

#endif
