/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.4.2.4 $
|                                              $Date: 2007/06/05 16:17:35 $
\*============================================================================*/
#include "dpws.h"

#include "litStub.h"
#include "lit.nsmap"

#define LIGHTING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Lighting"
#define SWITCH_POWER_TYPE "SwitchPower"
#define DIMMING_TYPE "Dimming"

static char command[32], param1[32];
static int status, nbParams;
	

static short getServiceEndpoint(struct dpws *dpws, short deviceProxy, char * type)
{
	int nbServices;
	short * serviceProxies = dpws_get_services(dpws, deviceProxy, LIGHTING_NS, type, &nbServices);
	if (nbServices != 1) {
		if (nbServices <= 0)
			printf("No service endpoint found\n");
		else
			printf("Too many service endpoints found\n");
		dpws_end(dpws);
		return -1;
	}
	dpws_pin_proxy(serviceProxies[0]);
	return serviceProxies[0];
}

void simpleLightClient(struct dpws *dpws, short deviceProxy)
{
	short serviceProxy;
	struct wsa_endpoint_ref * servEndPt;

	printf ("\nSimpleLight device type commands (Q to quit this menu):\n\n");
	printf ("Switch (ON|OFF)\nGetStatus\n");

	// command loop
	for (;;) {
		
		printf ("\nEnter command: ");
		nbParams = scanf("%32s", command);
		if (nbParams == 0) {
			printf("Type mismatch !\n");
			continue;
		}
		
		// finding service endpoint
		serviceProxy = getServiceEndpoint(dpws, deviceProxy, SWITCH_POWER_TYPE);
		if (serviceProxy < 0)
			break;
		servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

		// Command parsing
		if (!stricmp(command, "Switch")) {
			nbParams = scanf("%32s", param1);
			if (nbParams != 1)
				printf("Bad parameter count !\n");
			else {
				enum lit__PowerState state = (stricmp(param1, "ON") ? lit__PowerState__OFF : lit__PowerState__ON);
				status = dpws_send___lit__Switch(dpws, servEndPt, state);
			}
		}
		else if (!stricmp(command,"GetStatus")) {
			enum lit__PowerState state;
			status = dpws_call___lit__GetStatus(dpws, servEndPt, NULL, &state);
			if (!status)
				printf("\n<- The light is %s\n", state == lit__PowerState__ON ? "ON" : "OFF");
		}
		else if ( stricmp(command,"Q")==0 ){
			return;
		}
		else
			printf("Command Unknown\n");
		if (status)
			fprintf(stderr, dpws_get_error_msg(dpws));

		dpws_release_proxy(serviceProxy);
		dpws_end(dpws);
	}
}


void dimmerLightClient(struct dpws *dpws, short deviceProxy, struct wsa_endpoint_ref * event_endpoint)
{
	// dimmer event subscription
	short serviceProxy;
	struct wsa_endpoint_ref * subscManager = NULL, * servEndPt;
	char *duration = "PT1H";

	// finding service endpoint
	serviceProxy = getServiceEndpoint(dpws, deviceProxy, DIMMING_TYPE);
	if (serviceProxy < 0) {
		printf ("Could not find service for subscription.\n");
		return;
	}
	servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

	subscManager = dpws_endpoint_ref_dup(dpws_event_subscribe(dpws, servEndPt, event_endpoint, NULL, NULL, &duration));
	if (!subscManager) {
		printf ("Could not subscribe.\n");
		fprintf(stderr, dpws_get_error_msg(dpws));
		return;
	}
	printf ("Subscribed to Dimming events for %s\n", duration);
	dpws_end(dpws);
	
	printf ("\nDimmerLight device type commands (Q to quit this menu):\n\n");
	printf ("Switch (ON|OFF)\nGetStatus\n");
	printf ("SetLevel <%% of power> <duration (s)>\nGetLevel\nGetTarget\n");

	// command loop
	for (;;) {
		
		printf ("\nEnter command:\n");
		nbParams = scanf("%32s", command);
		if (nbParams == 0) {
			printf("Type mismatch !\n");
			continue;
		}
		
		// Command parsing
		if (!stricmp(command, "Switch")) {
			nbParams = scanf("%32s", param1);
			if (nbParams != 1)
				printf("Bad parameter count !\n");
			else {
				enum lit__PowerState state = (stricmp(param1, "ON") ? lit__PowerState__OFF : lit__PowerState__ON);

				// finding service endpoint (again, since dpws_end has been called)
				serviceProxy = getServiceEndpoint(dpws, deviceProxy, SWITCH_POWER_TYPE);
				if (serviceProxy < 0) {
					printf ("Could not find service.\n");
					break;
				}
				servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

				status = dpws_send___lit__Switch(dpws, servEndPt, state);
			}
		}
		else if (!stricmp(command,"GetStatus")) {
			enum lit__PowerState state;

			// finding service endpoint (again, since dpws_end has been called)
			serviceProxy = getServiceEndpoint(dpws, deviceProxy, SWITCH_POWER_TYPE);
			if (serviceProxy < 0) {
				printf ("Could not find service.\n");
				break;
			}
			servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

			status = dpws_call___lit__GetStatus(dpws, servEndPt, NULL, &state);
			if (!status)
				printf("\n<- The light is %s\n", state == lit__PowerState__ON ? "ON" : "OFF");
		}
		else if (!stricmp(command, "SetLevel")) {
			int level, duration;
			nbParams = scanf("%d %d", &level, &duration);
			if (nbParams < 1 || nbParams > 2)
				printf("Bad parameter count !\n");
			else {
				struct _lit__LightLevelTarget levelTarget = {level, NULL};
				if (duration != 0)
					levelTarget.TransitionDuration = (short *)&duration;

				// finding service endpoint
				serviceProxy = getServiceEndpoint(dpws, deviceProxy, DIMMING_TYPE);
				if (serviceProxy < 0) {
					printf ("Could not find service.\n");
					break;
				}
				servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

				status = dpws_send___lit__SetLevel(dpws, servEndPt, &levelTarget);
			}
		}
		else if (!stricmp(command,"GetLevel")) {
			short level;

			// finding service endpoint
			serviceProxy = getServiceEndpoint(dpws, deviceProxy, DIMMING_TYPE);
			if (serviceProxy < 0) {
				printf ("Could not find service.\n");
				break;
			}
			servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

			status = dpws_call___lit__GetLevel(dpws, servEndPt, NULL, &level);
			if (!status)
				printf("\n<- The light level is %d%%\n", level);
		}
		else if (!stricmp(command,"GetTarget")) {
			short target;

			// finding service endpoint
			serviceProxy = getServiceEndpoint(dpws, deviceProxy, DIMMING_TYPE);
			if (serviceProxy < 0) {
				printf ("Could not find service.\n");
				break;
			}
			servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);

			status = dpws_call___lit__GetLevelTarget(dpws, servEndPt, NULL, &target);
			if (!status)
				printf("\n<- The light target level is %d%%\n", target);
		}
		else if ( stricmp(command,"Q")==0 ){
			return;
		}
		else
			printf("Command Unknown\n");
		if (status)
			fprintf(stderr, dpws_get_error_msg(dpws));

		dpws_release_proxy(serviceProxy);
		dpws_end(dpws);
	}
	
	// unsubscribe
	dpws_event_unsubscribe(dpws, subscManager);
	dpws_endpoint_ref_free(subscManager);
}

/*
 *  HANDLER Implementations
 */
int __lit__TargetReached(struct dpws* dpws, short lit__LightLevel)
{
	printf("<- Event 'TargetReached' received ! Level = %d\n", lit__LightLevel);
	return SOAP_OK;
}

// NOTE: the following handlers must be implemented even if not used
int __lit__GetStatus_handler(struct dpws* dpws, enum lit__PowerState *a)
{
	return SOAP_OK;
}

int __lit__GetLevelTarget_handler(struct dpws* dpws, short *a)
{
	return SOAP_OK;
}

int __lit__GetLevel_handler(struct dpws* dpws, short *a)
{
	return SOAP_OK;
}

