/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.14.2.4 $
|                                              $Date: 2007/03/09 15:48:27 $
\*============================================================================*/
#include <stdio.h>

#include "litStub.h"
#include "wshStub.h"
#include "prjStub.h"

#include "dpws.h"
#include "serverDPWS.h"
#include "LightingClient.h"
#include "WasherClient.h"
#include "ProjectorClient.h"

#define DEMONSTRATOR_SCOPE "http://www.schneider-electric.com/HCDemo/"

#define TRAINING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training"
#define SIMPLE_LIGHT_TYPE "SimpleLight"
#define DIMMER_LIGHT_TYPE "DimmerLight"
#define WASHER_TYPE "WashingMachine"
#define PROJECTOR_TYPE "Projector"

static void device_joining(struct dpws *dpws, short href_device)
{
	printf("\n<- New device on the LAN: %s\n", dpws_cache_get_uuid(dpws, href_device));
}

static void device_leaving(struct dpws *dpws, short href_device)
{
	printf("\n<- Device %s leaving the LAN.\n", dpws_cache_get_uuid(dpws, href_device));
}

static struct dpws listen_dpws;

/*
 *  Utilities
 */
static char scope_buffer[255];

static char * getDiversifiedScope(char *suffix)
{
	strcpy(scope_buffer, DEMONSTRATOR_SCOPE);
	if (suffix)
		strcat(scope_buffer, suffix);
	return scope_buffer;
}

/*
 *  Main function performing server intialization.
 *	Usage: client [<scope suffix>] [<port>]
 */
int main(int argc, char* argv[]) { 

	int nbDevEndPt = 10, indexDev = 0, port = 8889, ret = DPWS_OK, i;
	char *scope_suffix = NULL;
	
	short hTkEndpoint;
	short deviceProxy = -1, * deviceProxies;
	struct wsa_endpoint_ref * tkEndpoint = NULL;
	struct dpws dpws;
	struct device_metadata *devMetadata;
	char ** scopes;

	if (argc > 1)
		scope_suffix = argv[1];
	
	if (argc > 2)
		port = atoi(argv[2]);

	if (port == 0) {
		fprintf(stderr, "Incorrect TCP port.\n");
		exit(1);
	}

	// DPWS initialization
	dpws_init();
	dpws_client_init(&dpws, NULL);

	// initialize DPWS server for event reception
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_BOOT_SEQ, 0);
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_HTTP_PORT, port);
	DPWS_SET_PTR_ATT(TOOLKIT_HANDLE, DPWS_PTR_CALLBACK_HELLO, device_joining);
	DPWS_SET_PTR_ATT(TOOLKIT_HANDLE, DPWS_PTR_CALLBACK_BYE, device_leaving);

	hTkEndpoint = dpws_create_endpoint(TOOLKIT_HANDLE);
	DPWS_ADD_PTR_ATT(hTkEndpoint, DPWS_PTR_HANDLING_FUNCTION, lit_handle_event);
	DPWS_ADD_PTR_ATT(hTkEndpoint, DPWS_PTR_HANDLING_FUNCTION, wsh_handle_event);

	if (initServer(&listen_dpws)) {
		fprintf(stderr, "Could not initialize DPWS server...\n");
		exit(1);
	}
	tkEndpoint = dpws_get_local_endpoint_ref(dpws_get_default_service_port(hTkEndpoint));

	if (bootServer(&listen_dpws)) {
		fprintf(stderr, "Could not boot server...\n");
		exit(1);
	}


	// Device menu loop;
	for (;;) {

		// lookup devices
		int nbDevices = 10;
		deviceProxies = dpws_lookup(&dpws, NULL, NULL, getDiversifiedScope(scope_suffix), &nbDevices);
		if (nbDevices <= 0) {
			printf("\nNo Device found.\n");
			goto exit;
		}
	
		// print available devices
		printf("\n### Main Menu ###\n\n%d devices found.\n\n", nbDevices);
		for ( i = 0; i< nbDevices; i++ ) {
			devMetadata = dpws_get_device_metadata(&dpws, deviceProxies[i]);
			if (devMetadata) {
				printf("\n-> Index %d - %s\n   Model: %s\n", i, 
				devMetadata->device_info.friendly_name, devMetadata->model_info.model_name);
				scopes = dpws_cache_get_scopes(&dpws, deviceProxies[i]);
				for(;scopes && *scopes; scopes++)
					printf("   Scope: %s\n", *scopes);
			}
			else {
				printf("-> Device %s not reachable.\n", dpws_cache_get_uuid(&dpws, deviceProxies[i]));
				fprintf(stderr, dpws_get_error_msg(&dpws));
			}
		}

		// select device
		for (;;) {
			printf("\nEnter device index (or Q to exit): ");
			if (scanf("%d", &indexDev) == 0)
				goto exit;
			printf("\n");
			if (indexDev >= 0 || indexDev < nbDevEndPt)
				break;
			printf("Incorrect device number\n");
		}

		// Once the device chosen, release the device proxies except the chosen one
		deviceProxy = deviceProxies[indexDev];
		dpws_pin_proxy(deviceProxy);
		dpws_end(&dpws);

		// Route processing basing on type
		if (dpws_check_device_type(&dpws, TRAINING_NS, SIMPLE_LIGHT_TYPE, WSD_MATCH_BY_STRCMP, deviceProxy))
			simpleLightClient(&dpws, deviceProxy);
		else if (dpws_check_device_type(&dpws, TRAINING_NS, DIMMER_LIGHT_TYPE, WSD_MATCH_BY_STRCMP, deviceProxy))
			dimmerLightClient(&dpws, deviceProxy, tkEndpoint);
		else if (dpws_check_device_type(&dpws, TRAINING_NS, WASHER_TYPE, WSD_MATCH_BY_STRCMP, deviceProxy))
			washerClient(&dpws, deviceProxy, tkEndpoint);
		else if (dpws_check_device_type(&dpws, TRAINING_NS, PROJECTOR_TYPE, WSD_MATCH_BY_STRCMP, deviceProxy))
			projectorClient(&dpws, deviceProxy);
		else
			printf("Unknown device type.");
		
	}

exit:
	dpws_endpoint_ref_free(tkEndpoint);
	dpws_release_proxy(deviceProxy);
	stopServer();
	exit(0);
}


