/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.21.2.4 $
|                                              $Date: 2007/06/05 16:17:35 $
\*============================================================================*/
#include <stdio.h>

#include "litStub.h"
#include "wshStub.h"
#include "prjStub.h"
#include "lit.nsmap"
#include "wsh.nsmap"
#include "prj.nsmap"

#include "dpws.h"
#include "fileio.h"
#include "serverDPWS.h"

/*
 * IMPORTANT NOTE: This multi-thread server is supplied as an example for studying purpose
 * and does not meet the usual standards of industrial software. For instance, device data
 * should be protected against concurrent access from server threads.
 */

/*
 *	Configuration static data
 */
 
#define DEMONSTRATOR_SCOPE "http://www.schneider-electric.com/HCDemo/"
#define KITCHEN_SUFFIX_SCOPE "/Kitchen"
#define BEDROOM_SUFFIX_SCOPE "/Bedroom"
#define LIVING_SUFFIX_SCOPE "/LivingRoom"
#define LAUNDRY_SUFFIX_SCOPE "/LaundryRoom"

#define TRAINING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training"
#define LIGHTING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Lighting"
#define LAUNDRY_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Laundry"
#define HOMECINEMA_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/HomeCinema"

#define BOOT_SEQ_FILE "bootseq.log"

struct qname SimpleLightType = {TRAINING_NS, "SimpleLight"};
struct qname DimmerLightType = {TRAINING_NS, "DimmerLight"};
struct qname WashingMachineType = {TRAINING_NS, "WashingMachine"};
struct qname ProjectorType = {TRAINING_NS, "Projector"};

struct prefixed_qname SwitchPowerPortType = {{LIGHTING_NS, "SwitchPower"}, "lit"};
struct prefixed_qname DimmingPortType = {{LIGHTING_NS, "Dimming"}, "lit"};
struct prefixed_qname WashPortType = {{LAUNDRY_NS, "Wash"}, "wsh"};
struct prefixed_qname DisplayPortType = {{HOMECINEMA_NS, "Display"}, "prj"};

struct wsdl_info LightingWsdl = {LIGHTING_NS, "http://wsdl.schneider-electric.com/Lighting.wsdl"};
struct wsdl_info WashingMachineWsdl = {LAUNDRY_NS, "http://wsdl.schneider-electric.com/WashingMachine.wsdl"};
struct wsdl_info ProjectorWsdl = {HOMECINEMA_NS, "http://wsdl.schneider-electric.com/Projector.wsdl"};

/*
 *	device instance runtime data
 */
enum lit__PowerState simpleLight1 = lit__PowerState__OFF, simpleLight2 = lit__PowerState__OFF;

struct dimmerData {
	enum lit__PowerState state;	// must be first to be able to cast to enum lit__PowerState
	short level;
	short target;
	struct dpws dpws;	// for event notification
	short endpointRef;	// for event notification
} dimmer;

struct washerData {
	int cycle;	// -1 means no cycle
	time_t scheduledEnd;
	struct dpws dpws;	// for event notification & client invokation
	short endpointRef;	// for event notification
} washer;

static struct dpws  master_dpws;

/*
 *  Utilities
 */
static char scope_buffer[255];
static char * StudentID = NULL;
static portableSleep(int seconds)
{
	#ifdef WIN32
			Sleep(seconds * 1000);
	#else
			sleep(seconds);
	#endif
}

static char * getDiversifiedScope(char *diversifier, char *suffix)
{
	strcpy(scope_buffer, DEMONSTRATOR_SCOPE);
	strcat(scope_buffer, diversifier);
	strcat(scope_buffer, suffix);
	return scope_buffer;
}

static void printDeviceInfo(short hrefDevice) {
	int len = 10, i; 
	short hrefs[10];
	char * transportAddr = NULL;
	transportAddr = dpws_get_transport_address(dpws_get_default_service_port(hrefDevice));
	printf("\nDevice started\n       ID: %s\n  Address: %s\n", dpws_get_ptr_att(hrefDevice, DPWS_STR_DEVICE_ID), transportAddr);
	dpws_free(transportAddr);
	if (!dpws_get_service_handles(hrefDevice, hrefs, &len)) {
		printf("\nHosted services:");
		for (i = 0; i < len; i++) {
			transportAddr = dpws_get_transport_address(dpws_get_default_service_port(hrefs[i]));
			printf("\n  * Service ID: %s\n       Address: %s\n", dpws_get_ptr_att(hrefs[i], DPWS_STR_SERVICE_ID), transportAddr);
			dpws_free(transportAddr);
		}
	}
}

/*
 *  Main function performing server intialization.
 *	Usage: server [<tcp port>]
 */
#ifdef VXWORKS
void startHome(char * studentId, char * port) {
	char* params[] =  {"", studentId, port};
	homeMain(3, params);	
}

int homeMain (int argc, char **argv)
#else
int main (int argc, char **argv)
#endif
{
	int port = 9876, c, bootSeq;
	short hDevClass, hServClass, hKitchenLight, hBedroomLight, hDimmer, hWasher, hProjector,
			hDimmerService, hWasherService;

	if (argc > 2)
		port = atoi(argv[2]);

	if (port == 0 || argc < 2) {
		fprintf(stderr, "Syntax is server StudentID [<tcp port>]\n");
		exit(1);
	}
	StudentID = argv[1];

	printf("Configuring DPWS devices...\n\n");


	dpws_init();

	bootSeq = readintfromfile(BOOT_SEQ_FILE);
	writeinttofile(bootSeq + 1, BOOT_SEQ_FILE);

	// DPWS server configuration
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_HTTP_PORT, port);
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_BOOT_SEQ, bootSeq);	// NOTE: should store the value in persistent memory

	printf("Initializing DPWS light devices...\n");

	// create simple light device class carrying device metadata
	hDevClass = dpws_create_device_class();
	DPWS_ADD_PTR_ATT(hDevClass, DPWS_PTR_TYPE, &SimpleLightType);
	DPWS_SET_INT_ATT(hDevClass, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should be updated everytime device class metadata changes
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NAME, "BrightBulb");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NUMBER, "1.0");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_URL, "http://www.schneider-electric.com/BrightBulb.html");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_PRESENTATION_URL, "index.html");	// Any relative URI will use the default presentation server
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER, "Schneider Electric SA");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER_URL, "http://www.schneider-electric.com");
	
	// create a service class carrying the switch power port type
	hServClass = dpws_create_service_class();
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &SwitchPowerPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_WSDL, &LightingWsdl);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_HANDLING_FUNCTION, &lit_serve_request);
	DPWS_SET_STR_ATT(hServClass, DPWS_STR_ID, "http://www.schneider-electric.com/DPWS/2006/03/Training/Light1");

	// Register the service class to the device class
	dpws_register_service_class(hDevClass, hServClass);
	
	// create 2 light instances of the same class
	hKitchenLight = dpws_create_device(0, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hKitchenLight, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hKitchenLight, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, KITCHEN_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hKitchenLight, DPWS_STR_FRIENDLY_NAME, "Kitchen light");
	DPWS_SET_STR_ATT(hKitchenLight, DPWS_STR_FIRMWARE_VERSION, "1.0");
	DPWS_SET_STR_ATT(hKitchenLight, DPWS_STR_SERIAL_NUMBER, "56080001");

	DPWS_SET_PTR_ATT(hKitchenLight, DPWS_PTR_USER_DATA, &simpleLight1);	// instance device specific information

	dpws_enable_device(hKitchenLight);	// prepares hello and activates the device (prefiguration of hot-deployment)

	hBedroomLight = dpws_create_device(1, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hBedroomLight, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hBedroomLight, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, BEDROOM_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hBedroomLight, DPWS_STR_FRIENDLY_NAME, "Bedroom light");
	DPWS_SET_STR_ATT(hBedroomLight, DPWS_STR_FIRMWARE_VERSION, "1.0");
	DPWS_SET_STR_ATT(hBedroomLight, DPWS_STR_SERIAL_NUMBER, "05609540");

	DPWS_SET_PTR_ATT(hBedroomLight, DPWS_PTR_USER_DATA, &simpleLight2);	// instance device specific information

	dpws_enable_device(hBedroomLight);	// prepares hello and activates the device (prefiguration of hot-deployment)



	// create a dimmer light device class
	hDevClass = dpws_create_device_class();
	DPWS_ADD_PTR_ATT(hDevClass, DPWS_PTR_TYPE, &DimmerLightType);
	DPWS_SET_INT_ATT(hDevClass, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should be updated everytime device class metadata changes
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NAME, "SmoothDimmer");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NUMBER, "2.0");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_URL, "http://www.schneider-electric.com/SmoothDimmer.html");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_PRESENTATION_URL, "index.html");	// Any relative URI will use the default presentation server
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER, "Schneider Electric SA");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER_URL, "http://www.schneider-electric.com");

	// create a new service class carrying the switch power and dimming port type
	hServClass = dpws_create_service_class();
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &SwitchPowerPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &DimmingPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_WSDL, &LightingWsdl);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_HANDLING_FUNCTION, &lit_serve_request);
	DPWS_SET_STR_ATT(hServClass, DPWS_STR_ID, "http://www.schneider-electric.com/DPWS/2006/03/Training/Light1");

	// Register the service class to the device class
	dpws_register_service_class(hDevClass, hServClass);
	
	// create a dimming light
	hDimmer = dpws_create_device(2, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hDimmer, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hDimmer, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, LIVING_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hDimmer, DPWS_STR_FRIENDLY_NAME, "Living light");
	DPWS_SET_STR_ATT(hDimmer, DPWS_STR_FIRMWARE_VERSION, "1.0");
	DPWS_SET_STR_ATT(hDimmer, DPWS_STR_SERIAL_NUMBER, "8609701");

	dimmer.state = lit__PowerState__OFF;
	dimmer.level = 100;
	dimmer.target = 100;
	dpws_client_init(&dimmer.dpws, NULL);
	DPWS_SET_PTR_ATT(hDimmer, DPWS_PTR_USER_DATA, &dimmer);	// instance device specific information
	hDimmerService = dpws_get_service_handle_by_class(hDimmer, hServClass);

	dpws_enable_device(hDimmer);	// prepares hello and activates the device (prefiguration of hot-deployment)



	printf("Initializing DPWS washing machine device...\n");

	// create a Washing machine device class
	hDevClass = dpws_create_device_class();
	DPWS_ADD_PTR_ATT(hDevClass, DPWS_PTR_TYPE, &WashingMachineType);
	DPWS_SET_INT_ATT(hDevClass, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should be updated everytime device class metadata changes
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NAME, "Salsa");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NUMBER, "2.0");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_URL, "http://www.whurlpool.com/Salsa.html");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_PRESENTATION_URL, "index.html");	// Any relative URI will use the default presentation server
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER, "Whurlpool");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER_URL, "http://www.whurlpool.com");

	// create a new service class carrying the wash port type
	hServClass = dpws_create_service_class();
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &WashPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_WSDL, &WashingMachineWsdl);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_HANDLING_FUNCTION, &wsh_serve_request);
	DPWS_SET_STR_ATT(hServClass, DPWS_STR_ID, "http://www.schneider-electric.com/DPWS/2006/03/Training/Washer1");

	// Register the service class to the device class
	dpws_register_service_class(hDevClass, hServClass);
	
	// create a washer device
	hWasher = dpws_create_device(3, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hWasher, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hWasher, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, LAUNDRY_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hWasher, DPWS_STR_FRIENDLY_NAME, "Washing machine");
	DPWS_SET_STR_ATT(hWasher, DPWS_STR_FIRMWARE_VERSION, "1.0");
	DPWS_SET_STR_ATT(hWasher, DPWS_STR_SERIAL_NUMBER, "85374012");

	washer.cycle = -1;	// -1 means no cycle
	washer.scheduledEnd = 0;
	dpws_client_init(&washer.dpws, NULL);
	DPWS_SET_PTR_ATT(hWasher, DPWS_PTR_USER_DATA, &washer);	// instance device specific information
	hWasherService = dpws_get_service_handle_by_class(hWasher, hServClass);

	dpws_enable_device(hWasher);	// prepares hello and activates the device (prefiguration of hot-deployment)



	printf("Initializing DPWS home cinema device...\n");

	// create a projector device class
	hDevClass = dpws_create_device_class();
	DPWS_ADD_PTR_ATT(hDevClass, DPWS_PTR_TYPE, &ProjectorType);
	DPWS_SET_INT_ATT(hDevClass, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should be updated everytime device class metadata changes
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NAME, "UltraVision");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NUMBER, "1.1.0");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_URL, "http://www.canyon.com/UltraVision.html");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_PRESENTATION_URL, "index.html");	// Any relative URI will use the default presentation server
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER, "Canyon");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER_URL, "http://www.canyon.com");

	// create a new service class carrying the projector port type
	hServClass = dpws_create_service_class();
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &DisplayPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_WSDL, &ProjectorWsdl);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_HANDLING_FUNCTION, &prj_serve_request);
	DPWS_SET_STR_ATT(hServClass, DPWS_STR_ID, "http://www.schneider-electric.com/DPWS/2006/03/Training/Projector1");

	// Register the service class to the device class
	dpws_register_service_class(hDevClass, hServClass);
	
	// create a projector device
	hProjector = dpws_create_device(4, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hProjector, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hProjector, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, LIVING_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hProjector, DPWS_STR_FRIENDLY_NAME, "Home Cinema Projector");
	DPWS_SET_STR_ATT(hProjector, DPWS_STR_FIRMWARE_VERSION, "2.1");
	DPWS_SET_STR_ATT(hProjector, DPWS_STR_SERIAL_NUMBER, "54272855997");

	dpws_enable_device(hProjector);	// prepares hello and activates the device (prefiguration of hot-deployment)

	// initialize DPWS server
	if (initServer(&master_dpws)) {
		fprintf(stderr, "Could not initialize DPWS server...\n");
		exit(1);
	}
	soap_set_mode(&master_dpws.soap, SOAP_ENC_MTOM);	// To be able to process MTOM attachments

	// Create local endpoints for event notification
	// WARNING ! Must be performed after the dpws_server_init() call (when local url are initialized)
	dimmer.endpointRef = dpws_get_default_service_port(hDimmerService);
	washer.endpointRef = dpws_get_default_service_port(hWasherService);	// NOTE: should free it when exiting

	if (bootServer(&master_dpws)) {
		fprintf(stderr, "Could not boot server...\n");
		exit(1);
	}

	printDeviceInfo(hBedroomLight);
	printDeviceInfo(hKitchenLight);
	printDeviceInfo(hDimmer);
	printDeviceInfo(hWasher);
	printDeviceInfo(hProjector);

	printf("\n\nDPWS server ready. Type 'q' or 'Q' to stop it smoothly.\n\n");
	do {
		c = getchar();
	} 
	while (c !='q' && c != 'Q');

	// Stop DPWS server smoothly
	stopServer();

	// NOTE: should release server object handles...
	exit(0);
}

/*
 *                   Service operations
 */

int __lit__Switch(struct dpws* dpws, enum lit__PowerState lit__Power)
{
	enum lit__PowerState * pState = (enum lit__PowerState *)dpws_get_device_user_data(dpws);
	*pState = lit__Power;
	printf("-> Setting light %s...\n", lit__Power == lit__PowerState__ON ? "ON" : "OFF");
	return SOAP_OK;
}

int __lit__GetStatus(struct dpws* dpws, enum lit__PowerState *lit__Power)
{
	*lit__Power = *(enum lit__PowerState *)dpws_get_device_user_data(dpws);	// NOTE: should protect device data from concurrent access
	return SOAP_OK;
}

int __lit__SetLevel(struct dpws* dpws, struct _lit__LightLevelTarget *lit__LightLevelTarget)
{
	// NOTE: Should protect device data
	short inc;
	struct dimmerData * pState = (struct dimmerData *)dpws_get_device_user_data(dpws);

	if (lit__LightLevelTarget->__item < 0 || lit__LightLevelTarget->__item > 100)
		return soap_sender_fault(dpws_dpws2soap(dpws), "Level out of bounds.", NULL);

	pState->target = lit__LightLevelTarget->__item;
	if (lit__LightLevelTarget->TransitionDuration) {
		inc = (pState->target - pState->level)/(*lit__LightLevelTarget->TransitionDuration);
		
		printf("-> Light level:");
		if (inc > 0) {
			for (; (pState->level + inc) <= pState->target; pState->level += inc) {
				portableSleep(1);
				printf(" %d%%", pState->level);
			}
		}
		else {
			for (; (pState->level + inc) >= pState->target; pState->level += inc) {
				portableSleep(1);
				printf(" %d%%", pState->level);
			}
		}
	}
	pState->level = lit__LightLevelTarget->__item;
	printf("\n-> Target %d%% reached.\n", pState->level);
	dpws_notify___lit__TargetReached(&pState->dpws, pState->endpointRef, pState->level);
	dpws_end(&pState->dpws);
	return SOAP_OK;
}

int __lit__GetLevelTarget(struct dpws* dpws, short *lit__LightLevel)
{
	struct dimmerData * pState = (struct dimmerData *)dpws_get_device_user_data(dpws);
	*lit__LightLevel = pState->target;	// NOTE: should protect device data from concurrent access
	return SOAP_OK;
}

int __lit__GetLevel(struct dpws* dpws, short *lit__LightLevel)
{
	struct dimmerData * pState = (struct dimmerData *)dpws_get_device_user_data(dpws);
	*lit__LightLevel = pState->level;	// NOTE: should protect device data from concurrent access
	return SOAP_OK;
}

int __wsh__LaunchCycle(struct dpws* dpws, struct _wsh__LaunchCycle *wsh__LaunchCycle)
{
	struct washerData * pState = (struct washerData *)dpws_get_device_user_data(dpws);
	struct _wsh__CycleEnd cycleEnd;
	time_t cycleDuration = 0;
	int nbProxies = 1;
	short * deviceProxies;

	if (pState->cycle != -1)
		return soap_receiver_fault(dpws_dpws2soap(dpws), "Cycle running", NULL);
	printf("-> Starting cycle ") ;
	switch (wsh__LaunchCycle->Name)
	{
		case wsh__Cycle__Gentle:
			cycleDuration = 10;
			printf("GENTLE (duration 10s)\n");
			break;
		case wsh__Cycle__Regular:
			cycleDuration = 20;
			printf("REGULAR (duration 20s)\n");
			break;
		case wsh__Cycle__Heavy:
			cycleDuration = 30;
			printf("HEAVY (duration 30s)\n");
			break;
	}
	pState->scheduledEnd = get_system_time() + cycleDuration;	// API provided by the toolkit
	pState->cycle = wsh__LaunchCycle->Name;
	portableSleep(cycleDuration);	// NOTE: server must be multithread or won't be able to answer any request

	pState->cycle = -1;

	// End notification
	cycleEnd.CycleName = wsh__LaunchCycle->Name;
	dpws_notify___wsh__CycleEnded(&pState->dpws, pState->endpointRef, &cycleEnd);
	dpws_end(&pState->dpws);

	// Light on kitchen light
	printf("<- Cycle ended: switch on kitchen lights.\n") ;
	deviceProxies = dpws_lookup(&pState->dpws, SimpleLightType.ns, SimpleLightType.lname, getDiversifiedScope(StudentID, KITCHEN_SUFFIX_SCOPE), &nbProxies);
	if (nbProxies == 1) {	// No more : one asked
		short * serviceProxies;
		nbProxies = 1;
		serviceProxies = dpws_get_services(&pState->dpws, deviceProxies[0], SwitchPowerPortType.qname.ns, SwitchPowerPortType.qname.lname, &nbProxies);
		if (nbProxies > 0) {
			dpws_send___lit__Switch(&pState->dpws, dpws_get_default_endpoint_reference(&pState->dpws, serviceProxies[0]), lit__PowerState__ON);
			dpws_release_proxy(*serviceProxies);
		}
		dpws_release_proxy(*deviceProxies);
	}
	dpws_end(&pState->dpws);

	return SOAP_OK;
}

int __wsh__GetCycleStatus(struct dpws* dpws, struct _wsh__CycleStatus *wsh__CycleStatus)
{
	// NOTE: should protect device data from concurrent access
	struct washerData * pState = (struct washerData *)dpws_get_device_user_data(dpws);
	if (pState->cycle == -1)
		wsh__CycleStatus->TimeLeft = NULL;
	else {	// Look here for result allocation using soap_malloc
		char * duration = soap_malloc(dpws_dpws2soap(dpws), DURATION_MAX_SIZE + 1);	// Constant defined in the DPWS API
		wsh__CycleStatus->TimeLeft = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct struct_1));
		wsh__CycleStatus->TimeLeft->__item = time2xmlduration(duration, pState->scheduledEnd);
		wsh__CycleStatus->TimeLeft->CycleName = pState->cycle;
	}

	return SOAP_OK;
}

int __prj__DisplayImage(struct dpws* dpws, struct prj__AttachmentType *prj__Image)
{
	fprintf(stderr, "-> Attachment received and saved under MyImage.jpg (size: %d, type: %s)\n",
		prj__Image->Param.__size, prj__Image->Param.type);
	if (writefile("MyImage.jpg", prj__Image->Param.__ptr, prj__Image->Param.__size) < 0) {
  		printf("Cannot save attachment\n");
		return SOAP_OK;
	}
	return SOAP_OK;
}
