/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.10.2.4 $
|                                              $Date: 2007/06/05 16:17:35 $
\*============================================================================*/
#include <stdio.h>

#include "litStub.h"
#include "lit.nsmap"

#include "dpws.h"

#define DEMONSTRATOR_SCOPE "http://www.schneider-electric.com/HCDemo/"

#define TRAINING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training"
#define SIMPLE_LIGHT_TYPE "SimpleLight"
#define LIGHTING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Lighting"
#define SWITCH_POWER_TYPE "SwitchPower"

enum commands {ON, OFF, STATUS};

/*
 *  Utilities
 */
static char scope_buffer[255];

static char * getDiversifiedScope(char *suffix)
{
	strcpy(scope_buffer, DEMONSTRATOR_SCOPE);
	if (suffix)
		strcat(scope_buffer, suffix);
	return scope_buffer;
}

static void usage() {
	fprintf(stderr, "Usage: client [-s <scope suffix>] <ON|OFF|STATUS>");
	exit(1);
}
/*
 *  Main function performing server intialization.
 *	Usage: client [-s <scope suffix>] <ON|OFF|STATUS>
 */
int main(int argc, char* argv[]) { 

	int nbDevices = 1,  nbServices, status = DPWS_OK, i, c_index = -1;
	char *scope_suffix = NULL;
	
	short deviceProxy = -1, * deviceProxies, serviceProxy = -1, * serviceProxies;
	struct wsa_endpoint_ref * invokationEPR;
	struct dpws dpws;
	struct device_metadata *devMetadata;
	char ** scopes;
	enum commands command;

	// Command line parsing
	if (argc == 1)
		usage();
	
	for (i = 1; i < argc; i++) {
		if (!strncmp(argv[i], "-s", 2)) {
			if (strlen(argv[i]) == 2) {
				if (++i == argc)
					usage();
				scope_suffix = argv[i];
			}
			else
				scope_suffix = argv[i] + 2;
		}
		else
			c_index = i;
	}
	if (c_index < 0)
		usage();
		
	if (!strcmp(argv[c_index], "ON"))
		command = ON;
	else if (!strcmp(argv[c_index], "OFF"))
		command = OFF;
	else if (!strcmp(argv[c_index], "STATUS"))
		command = STATUS;
	else
		usage();
	
	// DPWS initialization
	dpws_init();
	dpws_client_init(&dpws, NULL);

	// lookup (WS-Discovery PROBE message) simple light device
	nbDevices = 1;
	deviceProxies = dpws_lookup(&dpws, TRAINING_NS, SIMPLE_LIGHT_TYPE, getDiversifiedScope(scope_suffix), &nbDevices);
	if (nbDevices <= 0) {
		printf("\nNo Device found.\n");
		goto exit;
	}

	// take the first (but could use a discriminant or consider multiple matches as an error)
	deviceProxy = deviceProxies[0];
	dpws_pin_proxy(deviceProxy);
	dpws_end(&dpws);

	// print its metadata
	devMetadata = dpws_get_device_metadata(&dpws, deviceProxy);
	if (devMetadata) {
		printf("\n-> Friendly name: %s\n   Model: %s\n", devMetadata->device_info.friendly_name, devMetadata->model_info.model_name);
		scopes = dpws_cache_get_scopes(&dpws, deviceProxy);
		for(;scopes && *scopes; scopes++)
			printf("   Scope: %s\n", *scopes);
	}
	else {
		printf("-> Device %s not reachable.\n", dpws_cache_get_uuid(&dpws, deviceProxy));
		fprintf(stderr, dpws_get_error_msg(&dpws));
	}

	// finding service endpoint
	serviceProxies = dpws_get_services(&dpws, deviceProxy, LIGHTING_NS, SWITCH_POWER_TYPE, &nbServices);
	if (nbServices != 1) {
		if (nbServices <= 0)
			printf("No service endpoint found\n");
		else
			printf("Too many service endpoints found\n");
		goto exit;
	}
	serviceProxy = serviceProxies[0];
	dpws_pin_proxy(serviceProxy);
	dpws_end(&dpws);

	// invokation
	invokationEPR = dpws_get_default_endpoint_reference(&dpws, serviceProxy);
	switch(command)
	{
		enum lit__PowerState state;

		case ON:
			status = dpws_send___lit__Switch(&dpws, invokationEPR, lit__PowerState__ON);
			break;
		case OFF:
			status = dpws_send___lit__Switch(&dpws, invokationEPR, lit__PowerState__OFF);
			break;
		case STATUS:
			status = dpws_call___lit__GetStatus(&dpws, invokationEPR, NULL, &state);
			if (!status)
				printf("\n<- The light is %s\n", state == lit__PowerState__ON ? "ON" : "OFF");
			break;
	}
	
	if (status)
		fprintf(stderr, dpws_get_error_msg(&dpws));

exit:
	dpws_end(&dpws);
	dpws_release_proxy(deviceProxy);
	dpws_release_proxy(serviceProxy);
	exit(0);
}


