/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.6.2.3 $
|                                              $Date: 2007/06/05 16:17:35 $
\*============================================================================*/
#include <stdio.h>

#include "litStub.h"
#include "lit.nsmap"

#include "dpws.h"
#include "fileio.h"
#include "serverDPWS.h"

/*
 *	Configuration static data
 */
 
#define DEMONSTRATOR_SCOPE "http://www.schneider-electric.com/HCDemo/"
#define KITCHEN_SUFFIX_SCOPE "/Kitchen"

#define TRAINING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training"
#define LIGHTING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Lighting"

#define BOOT_SEQ_FILE "bootseq.log"

struct qname SimpleLightType = {TRAINING_NS, "SimpleLight"};

struct prefixed_qname SwitchPowerPortType = {{LIGHTING_NS, "SwitchPower"}, "lit"};

struct wsdl_info LightingWsdl = {LIGHTING_NS, "http://wsdl.schneider-electric.com/Lighting.wsdl"};

/*
 *	device instance runtime data
 */
enum lit__PowerState simpleLight1 = lit__PowerState__OFF;

static struct dpws  master_dpws;

/*
 *  Utilities
 */
static char scope_buffer[255];
static char * StudentID = NULL;

static char * getDiversifiedScope(char *diversifier, char *suffix)
{
	strcpy(scope_buffer, DEMONSTRATOR_SCOPE);
	strcat(scope_buffer, diversifier);
	strcat(scope_buffer, suffix);
	return scope_buffer;
}

/*
 *  Main function performing server intialization.
 *	Usage: server [<tcp port>]
 */
#ifdef VXWORKS
void startLight(char * studentId, char * port) {
	char* params[] =  {"", studentId, port};
	lightMain(3, params);	
}

int lightMain (int argc, char **argv)
#else
int main (int argc, char **argv)
#endif
{
	int port = 9867, c, bootSeq;
	short hDevClass, hServClass, hKitchenLight;
	char * StudentID = NULL;

	if (argc > 2)
		port = atoi(argv[2]);

	if (port == 0 || argc < 2) {
		fprintf(stderr, "Syntax is server StudentID [<tcp port>]\n");
		exit(1);
	}
	StudentID = argv[1];

	printf("Configuring DPWS device...\n\n");

	dpws_init();
	
	bootSeq = readintfromfile(BOOT_SEQ_FILE);
	writeinttofile(bootSeq + 1, BOOT_SEQ_FILE);
	
	// DPWS server configuration
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_HTTP_PORT, port);
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_BOOT_SEQ, bootSeq);

	printf("Initializing DPWS light devices...\n");

	// create simple light device class carrying device metadata
	hDevClass = dpws_create_device_class();
	DPWS_ADD_PTR_ATT(hDevClass, DPWS_PTR_TYPE, &SimpleLightType);
	DPWS_SET_INT_ATT(hDevClass, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should be updated everytime device class metadata changes
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NAME, "BrightBulb");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NUMBER, "1.0");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_URL, "http://www.schneider-electric.com/BrightBulb.html");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_PRESENTATION_URL, "LightsOverview.html");	// Any relative URI will use the default presentation server
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER, "Schneider Electric SA");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER_URL, "http://www.schneider-electric.com");
	
	// create a service class carrying the switch power port type
	hServClass = dpws_create_service_class();
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &SwitchPowerPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_WSDL, &LightingWsdl);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_HANDLING_FUNCTION, &lit_serve_request);
	DPWS_SET_STR_ATT(hServClass, DPWS_STR_ID, "http://www.schneider-electric.com/DPWS/2006/03/Training/Light1");

	// Register the service class to the device class
	dpws_register_service_class(hDevClass, hServClass);
	
	// create a light instance of the same class
	hKitchenLight = dpws_create_device(0, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hKitchenLight, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hKitchenLight, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, KITCHEN_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hKitchenLight, DPWS_STR_FRIENDLY_NAME, "Kitchen light");
	DPWS_SET_STR_ATT(hKitchenLight, DPWS_STR_FIRMWARE_VERSION, "1.0");
	DPWS_SET_STR_ATT(hKitchenLight, DPWS_STR_SERIAL_NUMBER, "56080001");

	DPWS_SET_PTR_ATT(hKitchenLight, DPWS_PTR_USER_DATA, &simpleLight1);	// instance device specific information

	dpws_enable_device(hKitchenLight);	// prepares hello and activates the device (prefiguration of hot-deployment)

	// initialize DPWS server
	if (initServer(&master_dpws)) {
		fprintf(stderr, "Could not initialize DPWS server...\n");
		exit(1);
	}

	if (bootServer(&master_dpws)) {
		fprintf(stderr, "Could not boot server...\n");
		exit(1);
	}

	printf("\n\nDPWS server ready. Type 'q' or 'Q' to stop it smoothly.\n\n");
	do {
		c = getchar();
	} 
	while (c !='q' && c != 'Q');

	// Stop DPWS server smoothly
	stopServer();

	// NOTE: should release server object handles...
	exit(0);
}

/*
 *                   Service operations
 */

int __lit__Switch(struct dpws* dpws, enum lit__PowerState lit__Power)
{
	enum lit__PowerState * pState = (enum lit__PowerState *)dpws_get_device_user_data(dpws);
	*pState = lit__Power;
	printf("-> Setting light %s...\n", lit__Power == lit__PowerState__ON ? "ON" : "OFF");
	return SOAP_OK;
}

int __lit__GetStatus(struct dpws* dpws, enum lit__PowerState *lit__Power)
{
	*lit__Power = *(enum lit__PowerState *)dpws_get_device_user_data(dpws);	// NOTE: should protect device data from concurrent access
	return SOAP_OK;
}

// The following operations must be implemented since they were generated from the same WSDL
int __lit__SetLevel(struct dpws* dpws, struct _lit__LightLevelTarget *lit__LightLevelTarget)
{
	return SOAP_OK;
}

int __lit__GetLevelTarget(struct dpws* dpws, short *lit__LightLevel)
{
	return SOAP_OK;
}

int __lit__GetLevel(struct dpws* dpws, short *lit__LightLevel)
{
	return SOAP_OK;
}

