/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.10.2.4 $
|                                              $Date: 2007/06/05 16:17:35 $
\*============================================================================*/
#include "dpws.h"

#include "wshStub.h"
#include "wsh.nsmap"

#define LAUNDRY_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Laundry"
#define WASH_TYPE "Wash"

static char command[32], param1[32];
static int status, nbParams;
	

static short getServiceEndpoint(struct dpws *dpws, short deviceProxy)
{
	int nbServices;
	short * serviceProxies = dpws_get_services(dpws, deviceProxy, LAUNDRY_NS, WASH_TYPE, &nbServices);
	if (nbServices != 1) {
		if (nbServices <= 0)
			printf("No service endpoint found\n");
		else
			printf("Too many service endpoints found\n");
		dpws_end(dpws);
		return -1;
	}
	dpws_pin_proxy(serviceProxies[0]);
	return serviceProxies[0];
}

void washerClient(struct dpws *dpws, short deviceProxy, struct wsa_endpoint_ref * event_endpoint)
{
	short serviceProxy;

	// dimmer event subscription
	struct wsa_endpoint_ref * subscManager = NULL, * servEndPt;
	char *duration = "PT1H";

	// finding service endpoint
	serviceProxy = getServiceEndpoint(dpws, deviceProxy);
	if (serviceProxy < 0) {
		printf ("Could not find service for subscription.\n");
		return;
	}
	servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);
	
	subscManager = dpws_endpoint_ref_dup(dpws_event_subscribe(dpws, servEndPt, event_endpoint, NULL, NULL, &duration));
	if (!subscManager) {
		printf ("Could not subscribe.\n");
		fprintf(stderr, dpws_get_error_msg(dpws));
		return;
	}
	printf ("Subscribed to Wash events for %s\n", duration);
	dpws_end(dpws);
	
	printf ("\nWasher device type commands (Q to quit this menu):\n\n");
	printf ("Launch (GENTLE|REGULAR|HEAVY) <temp (0-90)> <spin speed (0-1500)>\nGetCycleStatus\n");

	// command loop
	for (;;) {
		
		printf ("\nEnter command:\n");
		nbParams = scanf("%32s", command);
		if (nbParams == 0) {
			printf("Type mismatch !\n");
			continue;
		}
		
		// finding service endpoint (again, since dpws_end has been called)
		serviceProxy = getServiceEndpoint(dpws, deviceProxy);
		if (serviceProxy < 0) {
			printf ("Could not find service.\n");
			break;
		}
		servEndPt = dpws_get_default_endpoint_reference(dpws, serviceProxy);
		
		// Command parsing
		if (!stricmp(command, "Launch")) {
			int temp, spin;
			nbParams = scanf("%32s %d %d", param1, &temp, &spin);
			if (nbParams != 3)
				printf("Bad parameter count !\n");
			else {
				struct _wsh__LaunchCycle cycle;
				cycle.Temperature = temp;
				cycle.SpinDryingSpeed = spin;
				if (!stricmp(param1, "GENTLE"))
					cycle.Name = wsh__Cycle__Gentle;
				else if (!stricmp(param1, "HEAVY"))
					cycle.Name = wsh__Cycle__Heavy;
				else
					cycle.Name = wsh__Cycle__Regular;

				status = dpws_send___wsh__LaunchCycle(dpws, servEndPt, &cycle);
			}
		}
		else if (!stricmp(command,"GetCycleStatus")) {
			struct _wsh__CycleStatus cycleStatus;

			status = dpws_call___wsh__GetCycleStatus(dpws, servEndPt, NULL, &cycleStatus);
			if (!status) {
				if (cycleStatus.TimeLeft) {
					printf("\n<- Cycle ");
					switch (cycleStatus.TimeLeft->CycleName) {
						case wsh__Cycle__Gentle:
							printf("GENTLE running for %s.\n", cycleStatus.TimeLeft->__item);
							break;
						case wsh__Cycle__Regular:
							printf("REGULAR running for %s.\n", cycleStatus.TimeLeft->__item);
							break;
						case wsh__Cycle__Heavy:
							printf("HEAVY running for %s.\n", cycleStatus.TimeLeft->__item);
							break;
					}
				}
				else
					printf("\n<- No cycle running\n");
			}
		}
		else if ( stricmp(command,"Q")==0 ){
			return;
		}
		else
			printf("Command Unknown\n");
		if (status)
			fprintf(stderr, dpws_get_error_msg(dpws));
		
		dpws_release_proxy(serviceProxy);
		dpws_end(dpws);
	}
	
	// unsubscribe
	dpws_event_unsubscribe(dpws, subscManager);
	dpws_endpoint_ref_free(subscManager);
}

/*
 *  HANDLER Implementations
 */
int __wsh__CycleEnded(struct dpws* dpws, struct _wsh__CycleEnd *wsh__CycleEnd)
{
	printf("<- Event 'CycleEnded' received ! Cycle: ");
	switch (wsh__CycleEnd->CycleName) {
		case wsh__Cycle__Gentle:
			printf("GENTLE\n");
			break;
		case wsh__Cycle__Regular:
			printf("REGULAR\n");
			break;
		case wsh__Cycle__Heavy:
			printf("HEAVY\n");
			break;
	}
	return SOAP_OK;
}

// NOTE: the following handlers must be implemented even if not used
int __wsh__GetCycleStatus_handler(struct dpws* dpws, struct _wsh__CycleStatus *a)
{
	return SOAP_OK;
}
