/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.9.2.2 $
|                                              $Date: 2007/03/05 15:50:29 $
\*============================================================================*/
#include <stdio.h>

#include "wshStub.h"

#include "dpws.h"
#include "serverDPWS.h"
#include "WasherClient.h"

#define DEMONSTRATOR_SCOPE "http://www.schneider-electric.com/HCDemo/"

#define TRAINING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training"
#define WASHER_TYPE "WashingMachine"

static void device_joining(struct dpws *dpws , short href_device)
{
	printf("\n<- New device on the LAN: %s\n", dpws_cache_get_uuid(dpws, href_device));
}

static void device_leaving(struct dpws *dpws , short href_device)
{
	printf("\n<- Device %s leaving the LAN.\n", dpws_cache_get_uuid(dpws, href_device));
}

static struct dpws listen_dpws;

/*
 *  Utilities
 */
static char scope_buffer[255];

static char * getDiversifiedScope(char *suffix)
{
	strcpy(scope_buffer, DEMONSTRATOR_SCOPE);
	if (suffix)
		strcat(scope_buffer, suffix);
	return scope_buffer;
}

/*
 *  Main function performing server intialization.
 *	Usage: client [<scope suffix>] [<port>]
 */
int main(int argc, char* argv[]) { 

	int nbDevices = 1, port = 8834, ret = DPWS_OK;
	char *scope_suffix = NULL;
	
	short hTkEndpoint;
	short deviceProxy, * deviceProxies;
	struct wsa_endpoint_ref * tkEndpoint = NULL;
	struct dpws dpws;
	struct device_metadata *devMetadata;
	char ** scopes;

	if (argc > 1)
		scope_suffix = argv[1];
	
	if (argc > 2)
		port = atoi(argv[2]);

	if (port == 0) {
		fprintf(stderr, "Incorrect TCP port.\n");
		exit(1);
	}

	// DPWS initialization
	dpws_init();
	dpws_client_init(&dpws, NULL);

	// initialize DPWS server for event reception
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_BOOT_SEQ, 0);
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_HTTP_PORT, port);
	DPWS_SET_PTR_ATT(TOOLKIT_HANDLE, DPWS_PTR_CALLBACK_HELLO, device_joining);
	DPWS_SET_PTR_ATT(TOOLKIT_HANDLE, DPWS_PTR_CALLBACK_BYE, device_leaving);

	hTkEndpoint = dpws_create_endpoint();
	DPWS_ADD_PTR_ATT(hTkEndpoint, DPWS_PTR_HANDLING_FUNCTION, wsh_handle_event);

	if (initServer(&listen_dpws)) {
		fprintf(stderr, "Could not initialize DPWS server...\n");
		exit(1);
	}
	tkEndpoint = dpws_get_local_endpoint_ref(dpws_get_default_service_port(hTkEndpoint));

	if (bootServer(&listen_dpws)) {
		fprintf(stderr, "Could not boot server...\n");
		exit(1);
	}

	// lookup devices
	nbDevices = 1;
	deviceProxies = dpws_lookup(&dpws, TRAINING_NS, WASHER_TYPE, getDiversifiedScope(scope_suffix), &nbDevices);
	if (nbDevices <= 0) {
		printf("\nNo Device found.\n");
		goto exit;
	}
	deviceProxy = deviceProxies[0];
	dpws_pin_proxy(deviceProxy);
	dpws_end(&dpws);

	// print available devices
	devMetadata = dpws_get_device_metadata(&dpws, deviceProxy);
	if (devMetadata) {
		printf("\n-> Friendly name: %s\n   Model: %s\n", devMetadata->device_info.friendly_name, devMetadata->model_info.model_name);
		scopes = dpws_cache_get_scopes(&dpws, deviceProxy);
		for(;scopes && *scopes; scopes++)
			printf("   Scope: %s\n", *scopes);
	}
	else {
		printf("-> Device %s not reachable.\n", dpws_cache_get_uuid(&dpws, deviceProxy));
		fprintf(stderr, dpws_get_error_msg(&dpws));
	}

	// Once the device chosen, clean the device endpoint array
	dpws_end(&dpws);

	// call command input procedure
	washerClient(&dpws, deviceProxy, tkEndpoint);
	
exit:
	dpws_endpoint_ref_free(tkEndpoint);
	dpws_release_proxy(deviceProxy);
	stopServer();
	exit(0);
}


