/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.4.4.1 $
|                                              $Date: 2007/02/13 10:03:09 $
\*============================================================================*/
#include <pthread.h>
#include <stdio.h>

#include "dpws.h"

/*
 * IMPORTANT NOTE: This multi-thread server is supplied as an example for studying purpose
 * and does not meet the usual standards of industrial software. For instance, threads
 * should not create threads or allocate a runtime structure for every service but implement
 * pools instead.
 */
pthread_t bootThread = NULL;

static void* serve(struct dpws * dpws)
{
	dpws_serve(dpws);
	dpws_end(dpws);
	free(dpws);
	return NULL;
}

static void* runServer(struct dpws * m_dpws)
{
	pthread_t hThread;
	struct dpws * s_dpws;
	int status = DPWS_OK;

    while (1) {
		/* Allocate a DPWS runtime structure for the next serve task */
		s_dpws = malloc(sizeof(struct dpws));
	
		status = dpws_accept_thr(m_dpws, s_dpws);
		if (status) {
			if (status != DPWS_ERR_SERVER_STOPPED)
				fprintf(stderr, dpws_get_error_msg(m_dpws));
			goto exit;
		}

		/* spawn a new thread */ // NOTE: implement a pool instead
		if ((status = pthread_create(&hThread, NULL, serve, s_dpws))) {
	        fprintf (stderr, "Thread Creation ERROR (%d) - runServer()\n", status);
	        goto exit;
		}
	}
	
exit:
	// Clean the DPWS registry
	dpws_shutdown(m_dpws);
	return NULL;
}

int initServer(struct dpws *dpws)
{
	int status = dpws_server_init(dpws, NULL);
	if (status)
		fprintf(stderr, dpws_get_error_msg(dpws));
	return status;
}

int bootServer(struct dpws *dpws)
{
	int status = 0;
	if ((status = pthread_create(&bootThread, NULL, runServer, dpws))) {
	    fprintf (stderr, "Thread Creation ERROR (%d) - bootServer()\n", status);
	    return status;
	}
	return 0;
}

void stopServer()
{
	dpws_stop_server(1000);
	// wait for the master thread to exit
	pthread_join(bootThread, NULL);
}

