/*============================================================================*\
|                                                                              |
|                         =S= DPWS toolkit Samples                             |
|                                                                              |
|           ->>  Copyright 2004-2006 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.6.2.3 $
|                                              $Date: 2007/06/05 16:17:35 $
\*============================================================================*/
#include <stdio.h>

#include "wshStub.h"
#include "wsh.nsmap"

#include "dpws.h"
#include "fileio.h"
#include "serverDPWS.h"

/*
 * IMPORTANT NOTE: This multi-thread server is supplied as an example for studying purpose
 * and does not meet the usual standards of industrial software. For instance, device data
 * should be protected against concurrent access from server threads.
 */

/*
 *	Configuration static data
 */
 
#define DEMONSTRATOR_SCOPE "http://www.schneider-electric.com/HCDemo/"
#define LAUNDRY_SUFFIX_SCOPE "/LaundryRoom"

#define TRAINING_NS "http://www.schneider-electric.com/DPWS/2006/03/Training"
#define LAUNDRY_NS "http://www.schneider-electric.com/DPWS/2006/03/Training/Laundry"

#define BOOT_SEQ_FILE "bootseq.log"

struct qname WashingMachineType = {TRAINING_NS, "WashingMachine"};

struct prefixed_qname WashPortType = {{LAUNDRY_NS, "Wash"}, "wsh"};

struct wsdl_info WashingMachineWsdl = {LAUNDRY_NS, "http://wsdl.schneider-electric.com/WashingMachine.wsdl"};

/*
 *	device instance runtime data
 */
struct washerData {
	int cycle;	// -1 means no cycle
	time_t scheduledEnd;
	struct dpws dpws;	// for event notification & client invokation
	short endpointRef;	// for event notification
} washer;

static struct dpws  master_dpws;

/*
 *  Utilities
 */
static char scope_buffer[255];
static char * StudentID = NULL;

static portableSleep(int seconds)
{
	#ifdef WIN32
			Sleep(seconds * 1000);
	#else
			sleep(seconds);
	#endif
}

static char * getDiversifiedScope(char *diversifier, char *suffix)
{
	strcpy(scope_buffer, DEMONSTRATOR_SCOPE);
	strcat(scope_buffer, diversifier);
	strcat(scope_buffer, suffix);
	return scope_buffer;
}

/*
 *  Main function performing server intialization.
 *	Usage: server [<tcp port>]
 */
#ifdef VXWORKS
void startWashingMachine(char * studentId, char * port) {
	char* params[] =  {"", studentId, port};
	washingMachineMain(3, params);	
}

int washingMachineMain (int argc, char **argv)
#else
int main (int argc, char **argv)
#endif
{
	int port = 9876, c, bootSeq;
	short hDevClass, hServClass, hWasher, hWasherService;
	char * StudentID = NULL;

	if (argc > 2)
		port = atoi(argv[2]);

	if (port == 0 || argc < 2) {
		fprintf(stderr, "Syntax is server StudentID [<tcp port>]\n");
		exit(1);
	}
	StudentID = argv[1];

	printf("Configuring DPWS device...\n\n");

	dpws_init();

	bootSeq = readintfromfile(BOOT_SEQ_FILE);
	writeinttofile(bootSeq + 1, BOOT_SEQ_FILE);

	// DPWS server configuration
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_HTTP_PORT, port);
	DPWS_SET_INT_ATT(TOOLKIT_HANDLE, DPWS_INT_BOOT_SEQ, bootSeq);

	printf("Initializing DPWS washing machine device...\n");

	// create a Washing machine device class
	hDevClass = dpws_create_device_class();
	DPWS_ADD_PTR_ATT(hDevClass, DPWS_PTR_TYPE, &WashingMachineType);
	DPWS_SET_INT_ATT(hDevClass, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should be updated everytime device class metadata changes
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NAME, "Salsa");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_NUMBER, "2.0");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MODEL_URL, "http://www.whurlpool.com/Salsa.html");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_PRESENTATION_URL, "WashersOverview.html");	// Any relative URI will use the default presentation server
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER, "Whurlpool");
	DPWS_SET_STR_ATT(hDevClass, DPWS_STR_MANUFACTURER_URL, "http://www.whurlpool.com");

	// create a new service class carrying the wash port type
	hServClass = dpws_create_service_class();
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_PREFIXED_TYPE, &WashPortType);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_WSDL, &WashingMachineWsdl);
	DPWS_ADD_PTR_ATT(hServClass, DPWS_PTR_HANDLING_FUNCTION, &wsh_serve_request);
	DPWS_SET_STR_ATT(hServClass, DPWS_STR_ID, "http://www.schneider-electric.com/DPWS/2006/03/Training/Washer1");

	// Register the service class to the device class
	dpws_register_service_class(hDevClass, hServClass);
	
	// create a washer device
	hWasher = dpws_create_device(3, hDevClass);	// the id must be unique in the physical local device
	DPWS_SET_INT_ATT(hWasher, DPWS_INT_METADATA_VERSION, 1);	// NOTE: should change everytime a change on the device impacts DPWS metadata
	DPWS_ADD_STR_ATT(hWasher, DPWS_STR_SCOPE, getDiversifiedScope(StudentID, LAUNDRY_SUFFIX_SCOPE));
	DPWS_SET_STR_ATT(hWasher, DPWS_STR_FRIENDLY_NAME, "Washing machine");
	DPWS_SET_STR_ATT(hWasher, DPWS_STR_FIRMWARE_VERSION, "1.0");
	DPWS_SET_STR_ATT(hWasher, DPWS_STR_SERIAL_NUMBER, "85374012");

	washer.cycle = -1;	// -1 means no cycle
	washer.scheduledEnd = 0;
	dpws_client_init(&washer.dpws, NULL);
	DPWS_SET_PTR_ATT(hWasher, DPWS_PTR_USER_DATA, &washer);	// instance device specific information
	hWasherService = dpws_get_service_handle_by_class(hWasher, hServClass);

	dpws_enable_device(hWasher);	// prepares hello and activates the device (prefiguration of hot-deployment)


	// initialize DPWS server
	if (initServer(&master_dpws)) {
		fprintf(stderr, "Could not initialize DPWS server...\n");
		exit(1);
	}

	// Create local endpoints for event notification
	// WARNING ! Must be performed after the dpws_server_init() call (when local url are initialized)
	washer.endpointRef = dpws_get_default_service_port(hWasherService);	// NOTE: should free it when exiting

	if (bootServer(&master_dpws)) {
		fprintf(stderr, "Could not boot server...\n");
		exit(1);
	}

	printf("\n\nDPWS server ready. Type 'q' or 'Q' to stop it smoothly.\n\n");
	do {
		c = getchar();
	} 
	while (c !='q' && c != 'Q');

	// Stop DPWS server smoothly
	stopServer();

	// NOTE: should release server object handles...
	exit(0);
}

/*
 *                   Service operations
 */

int __wsh__LaunchCycle(struct dpws* dpws, struct _wsh__LaunchCycle *wsh__LaunchCycle)
{
	struct washerData * pState = (struct washerData *)dpws_get_device_user_data(dpws);
	struct _wsh__CycleEnd cycleEnd;
	time_t cycleDuration = 0;
	int nbEndpt = 1;

	if (pState->cycle != -1)
		return soap_receiver_fault(dpws_dpws2soap(dpws), "Cycle running", NULL);
	printf("-> Starting cycle ") ;
	switch (wsh__LaunchCycle->Name)
	{
		case wsh__Cycle__Gentle:
			cycleDuration = 10;
			printf("GENTLE (duration 10s)\n");
			break;
		case wsh__Cycle__Regular:
			cycleDuration = 20;
			printf("REGULAR (duration 20s)\n");
			break;
		case wsh__Cycle__Heavy:
			cycleDuration = 30;
			printf("HEAVY (duration 30s)\n");
			break;
	}
	pState->scheduledEnd = get_system_time() + cycleDuration;	// API provided by the toolkit
	pState->cycle = wsh__LaunchCycle->Name;
	portableSleep(cycleDuration);	// NOTE: server must be multithread or won't be able to answer any request

	pState->cycle = -1;

	// End notification
	cycleEnd.CycleName = wsh__LaunchCycle->Name;
	dpws_notify___wsh__CycleEnded(&pState->dpws, pState->endpointRef, &cycleEnd);
	dpws_end(&pState->dpws);

	return SOAP_OK;
}

int __wsh__GetCycleStatus(struct dpws* dpws, struct _wsh__CycleStatus *wsh__CycleStatus)
{
	// NOTE: should protect device data from concurrent access
	struct washerData * pState = (struct washerData *)dpws_get_device_user_data(dpws);
	if (pState->cycle == -1)
		wsh__CycleStatus->TimeLeft = NULL;
	else {	// Look here for result allocation using soap_malloc
		char * duration = soap_malloc(dpws_dpws2soap(dpws), DURATION_MAX_SIZE + 1);	// Constant defined in the DPWS API
		wsh__CycleStatus->TimeLeft = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct struct_1));
		wsh__CycleStatus->TimeLeft->__item = time2xmlduration(duration, pState->scheduledEnd);
		wsh__CycleStatus->TimeLeft->CycleName = pState->cycle;
	}

	return SOAP_OK;
}
