/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.102.2.9 $
|                                              $Date: 2007/05/03 16:56:48 $
\*============================================================================*/
 
/******************************************************************************\
 *                              COMMON DYNAMIC DATA                           *
\******************************************************************************/
#include <string.h>
#include "cache.h"
#include "metadata.h"

// NOTE: PERF could use soap_unlink to avoid free + malloc when adding dynamic
// data to cache. But it would prevent to implement a more efficient memory
// allocation for gSOAP (global free)

short cache_lock_handle = -1;

handlePool_t cachePool;

/*----------------------------------------------------------------------------*\
 *                               PRIVATE STUFF                                *
\*----------------------------------------------------------------------------*/
static int service_free(struct service_proxy *service)
{
	free_buffer_dyn_array(&service->super.types);
	free(service->id);
	free_buffer_dyn_array(&service->endpoint_references);
	free(service);
	return DPWS_OK;
}

static void service_release(short * href)
{
	struct service_proxy * service = (struct service_proxy *)getObject(&cachePool, *href);
	service->host = NULL;
	releaseHandle(&cachePool, *href);
}

static void clean_hosted_services(struct device_proxy * device)
{
	free_dyn_array(&device->hosted_services, (void (*)(void *))&service_release);
	device->hosted_retrieved = FALSE;
}

static int device_content_free(struct device_proxy * device)
{
	free_buffer_dyn_array(&device->super.types);
	dpws_endpoint_ref_content_free(&device->superEPR.super);
	free_buffer_dyn_array(&device->scopes);
	free_buffer_dyn_array(&device->transport_addresses);
	clean_hosted_services(device);
	return DPWS_OK;
}

static int device_free(struct device_proxy * device)
{
	device_content_free(device);
	free(device);
	return DPWS_OK;
}

static void overwrite_entry(struct device_proxy * dest, struct dpws * dpws, struct wsd__HelloType *source)
{
	string_array_dup(NULL, &dest->super.types, source->Types);
	dest->superEPR.status = CACHED;
	dpws_endpoint_ref_copy(&dest->superEPR.super, source->wsa__EndpointReference);
	dest->metadata_version = source->MetadataVersion;
	dest->message_number = dpws->msg_nb;
	dest->instance_id = dpws->instance_id;
	
	if (!source->Scopes || source->Scopes->__item->nb == 0)	// TODO: see why a missing tag could lead to a NULL pointer (see what users expect)
	{
		init_dyn_array(&dest->scopes, sizeof(char *), 1);
		*(char **) new_entry(NULL, &dest->scopes) = dpws_strdup(WSD_DEFAULT_SCOPE);	// to be able to free it
	}
	else
		string_array_dup(NULL, &dest->scopes, source->Scopes->__item);
		
	if (!source->XAddrs || source->XAddrs->nb == 0)
		init_dyn_array(&dest->transport_addresses, sizeof(char *), 1);
	else
		string_array_dup(NULL, &dest->transport_addresses, source->XAddrs);

	dest->hosted_retrieved = FALSE;
	init_dyn_array(&dest->hosted_services, sizeof(struct service_proxy), 2);
}

static BOOLEAN qname_struct_equals(struct qname *type1, char **type2)
{
	return qname_equals(type1->ns, type1->lname, *type2);
}

static BOOLEAN device_match(struct qname * types, char ** scopes,
		int match_rule, dyn_array *dev_types, dyn_array *dev_scopes)
{
	int i;
	dyn_array typeArray = {types, 0, 0, sizeof (struct qname), 1};
	dyn_array scopeArray = {scopes, 0, 0, sizeof (char *), 1};
	
	// NOTE: size must absolute be found because of polymorphism that prevents the easy use of boundaries
	if (types) {
		for (i = 0; (types + i)->lname; i++);
		typeArray.nb = typeArray.size = i;
	}
	
	if (scopes) {
		for (i = 0; *(scopes + i); i++);
		scopeArray.nb = scopeArray.size = i;
	}

	return probe_match(&typeArray, &scopeArray, match_rule,	dev_types, dev_scopes, (type_compare_cbk)qname_struct_equals);
}

static int releaseHandleHook(handle_s *handleInfo)
{
	int ret = DPWS_OK;
	switch (handleInfo->type)
	{
		case DEVICE_PROXY_TYPE:
			ret = device_free((struct device_proxy *)handleInfo->pObject);
			break;
		case SERVICE_PROXY_TYPE:
			ret = service_free((struct service_proxy *)handleInfo->pObject);
			break;
		default:
			break;
	}
	return ret;
}

static BOOLEAN is_matching_proxy(dyn_array *types, char * ns, char * port_type)
{
	int j;
	char * type;

	for (j = 0; j < types->nb; j++)
	{
		type = *(char **)get_entry(types, j);
		if (!port_type ||qname_equals(ns, port_type, type))
			return TRUE;
	}
	return FALSE;
}

struct lookup_info {
	struct qname * types;
	char ** scopes;
	int match_rule;
	struct dpws * dpws;
	int max_result;
	int count;
	dyn_array result;
};

static BOOLEAN match_device_by_discovery(short type, struct device_proxy *device, struct lookup_info *info)
{
	if (device->superEPR.status == CACHED
		&& device_match(info->types, info->scopes, info->match_rule, &device->super.types, &device->scopes)) {
		// copy entry if needed because count_cache_entry() only needs a count...
		if (info->dpws)
			*( short *)new_entry(info->dpws, &info->result) = device->super.href;
		info->count++;
	}
	return info->count == info->max_result;
}

static short * _cache_lookup(struct dpws * dpws, struct qname * types, char ** scopes, int match_rule, int *res_size)
{
	struct lookup_info info = {NULL, NULL, 0, NULL, 0, 0, DYN_ARRAY_INITIALIZER(sizeof(short), 5)};
	info.types = types;
	info.scopes = scopes;
	info.match_rule = match_rule;
	info.dpws = dpws;
	info.max_result = *res_size;

	
	lock(cache_lock_handle);
	handlePoolIterate(&cachePool, DEVICE_PROXY_TYPE, (handle_cbk)match_device_by_discovery, (void *)&info);
	unlock(cache_lock_handle);
	*res_size = info.count;
	return info.result.tab;
}

static struct wsa_endpoint_ref * create_EPR(struct dpws * dpws, char * Xaddr)
{
	struct wsa_endpoint_ref * ret = NULL; 
	if (dpws)
		ret = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct wsa_endpoint_ref));
	else
		ret = malloc(sizeof(struct wsa_endpoint_ref));
		
	dpws_default_wsa_endpoint_ref(ret);
	if (dpws)
		ret->address = soap_strdup(dpws_dpws2soap(dpws), Xaddr);
	else		
		ret->address = dpws_strdup(Xaddr);
		
	return ret;
}

struct grab_service_info {
	struct dpws * dpws;
	dyn_array * grabbed;
	char * ns;
	char * port_type;
};

static BOOLEAN grab_service(short * href, struct grab_service_info * grabInfo)
{
	struct service_proxy * serviceProxy = (struct service_proxy *)getObject(&cachePool, *href);
	
	if (serviceProxy && is_matching_proxy(&serviceProxy->super.types, grabInfo->ns, grabInfo->port_type))
		*(short *)new_entry(grabInfo->dpws, grabInfo->grabbed) = *href;
	
	return FALSE;
}

static BOOLEAN find_service_hook(short * href, char * id)
{
	struct service_proxy * serviceProxy = (struct service_proxy *)getObject(&cachePool, *href);
	return serviceProxy && serviceProxy->id && !strcmp(id, serviceProxy->id);
}

static BOOLEAN compare_proxies_id(short type, void * p_object, char * id)
{
	char * proxy_id;
	switch (type) {
		case DEVICE_PROXY_TYPE:
			proxy_id = ((struct device_proxy *)p_object)->superEPR.super.address;
			break;
		case SERVICE_PROXY_TYPE:
			proxy_id = ((struct service_proxy *)p_object)->id;
			break;
		default:
			return FALSE;
	}
	return !strcmp(proxy_id, id);
}

static BOOLEAN string_equals(char * s1, char * s2)
{
	return !strcmp(s1, s2);
}

static BOOLEAN lookup_proxy_addresses(short type, void * p_object, char * id)
{
	dyn_array * addresses;
	switch (type) {
		case DEVICE_PROXY_TYPE:
			addresses = &((struct device_proxy *)p_object)->transport_addresses;
			break;
		case SERVICE_PROXY_TYPE:
			addresses = &((struct service_proxy *)p_object)->endpoint_references;
			break;
		default:
			return FALSE;
	}
	return browse_dyn_array(addresses, (browse_entry_cbk)string_equals, id) < addresses->nb;
}

/*----------------------------------------------------------------------------*\
 *                               Discovery Cache                              *
\*----------------------------------------------------------------------------*/

int initCache()
{
	createHandlePool(
					&cachePool,
					CACHE_POOL_NB_BUCKETS,
					MAX_SERVER_OBJECTS,
					-1,	// enables "outside" protection
					(handle_release_cbk)releaseHandleHook
					);
	if ((cache_lock_handle = lock_create()) < 0)	// NOTE: no need for registry
		return cache_lock_handle;
	return DPWS_OK;
}

void uninitCache()
{
	if (cache_lock_handle >= 0)
	{
		deleteHandlePool(&cachePool);
	  	lock_destroy(cache_lock_handle);
	  	cache_lock_handle = -1;
	}
}

void clean_cache()
{
	if (cache_lock_handle >= 0)
	{
		lock(cache_lock_handle);
		deleteHandles(&cachePool, ANY_TYPE);
		unlock(cache_lock_handle);
	}
}

static BOOLEAN match_device_by_id(short type, struct device_proxy *device, char *id)
{
	return device->superEPR.status == CACHED && !strcmp(id, device->superEPR.super.address);
}

struct device_proxy * find_device_proxy(char * logical_id)
{
	handle_s * h_device = handlePoolIterate(&cachePool, DEVICE_PROXY_TYPE, (handle_cbk)match_device_by_id, logical_id);
	return h_device ? (struct device_proxy *) h_device->pObject : NULL;
}

struct device_proxy * get_device_proxy(short href, BOOLEAN *isDevice)
{
	handle_s *handleInfo = getHandle(&cachePool, href);
	if (!handleInfo)
		return NULL;
		
	switch (handleInfo->type)
	{
		case DEVICE_PROXY_TYPE:
			*isDevice = TRUE;
			return (struct device_proxy *)handleInfo->pObject;
		case SERVICE_PROXY_TYPE:
			*isDevice = FALSE;
			return ((struct service_proxy *)handleInfo->pObject)->host;
	}
	return NULL;
}

struct wsa_endpoint_ref * get_default_EPR(struct dpws * dpws, short href)
{
	dyn_array * Xaddrs;
	struct wsa_endpoint_ref * ret = NULL;
	handle_s *handleInfo = getHandle(&cachePool, href);

	if (!handleInfo)
		return NULL;
		
	switch (handleInfo->type)
	{
		case DEVICE_PROXY_TYPE:
			if (dpws)
				ret = endpoint_ref_dup(dpws, &((struct device_proxy *)handleInfo->pObject)->superEPR.super);
			else
				ret = dpws_endpoint_ref_dup(&((struct device_proxy *)handleInfo->pObject)->superEPR.super);
			break;
		case SERVICE_PROXY_TYPE:
			Xaddrs = &((struct service_proxy *)handleInfo->pObject)->endpoint_references;
			if (Xaddrs && Xaddrs->nb > 0)
				ret = create_EPR(dpws, *(char **)get_entry(Xaddrs, 0));	// TODO: use a better creterion
			break;
	}
	return ret;
}

struct wsa_endpoint_ref ** get_EPRs(struct dpws * dpws, short href, int * res_size)
{
	struct wsa_endpoint_ref ** ret;
	int i, offset;
	dyn_array * Xaddrs;
	handle_s *handleInfo = getHandle(&cachePool, href);

	if (!handleInfo) {
		*res_size = -1;
		return NULL;
	}
		
	switch (handleInfo->type)
	{
		case DEVICE_PROXY_TYPE:
			Xaddrs = &((struct device_proxy *)handleInfo->pObject)->transport_addresses;
			offset = 1;
			break;
		case SERVICE_PROXY_TYPE:
			Xaddrs = &((struct service_proxy *)handleInfo->pObject)->endpoint_references;
			if (!Xaddrs || Xaddrs->nb == 0)
				return NULL;
			offset = 0;
			break;
	}
	*res_size = offset + Xaddrs->nb;
	
	if (dpws)
		ret = soap_malloc(dpws_dpws2soap(dpws), *res_size * sizeof(struct wsa_endpoint_ref *));
	else
		ret = malloc(*res_size * sizeof(struct wsa_endpoint_ref *));
		
	if (handleInfo->type == DEVICE_PROXY_TYPE)
		ret[0] = dpws_endpoint_ref_dup(&((struct device_proxy *)handleInfo->pObject)->superEPR.super);
		
	for (i = 0; i < Xaddrs->nb; i++) {
		ret[offset + i] = create_EPR(dpws, *(char **)get_entry(Xaddrs, i));
	}
		
	return ret;
}

int count_cache_entry(struct dpws * dpws, struct qname * types,	char ** scopes, int match_rule, int max_res)
{
	_cache_lookup(NULL, types, scopes, match_rule, &max_res);
	return max_res;
}

short * cache_lookup(struct dpws * dpws, struct qname * types, char ** scopes, int match_rule, int *res_size)
{
	return _cache_lookup(dpws, types, scopes, match_rule, res_size);
}

struct device_proxy * add_cache_entry(struct dpws * dpws, struct wsd__HelloType *entry)
{
	struct device_proxy * new_entry = malloc(sizeof(struct device_proxy));
	new_entry->super.href = createHandle(&cachePool, DEVICE_PROXY_TYPE, new_entry);
	overwrite_entry(new_entry, dpws, entry);
	// lock is performed by the caller
	return new_entry;
}

void replace_cache_entry(struct device_proxy * dest, struct dpws * dpws, struct wsd__HelloType *source)
{
	// lock(cache_lock_handle) performed outside
	device_content_free(dest);
	overwrite_entry(dest, dpws, source);
}

int is_sequenced_msg(struct dpws* dpws, struct device_proxy * device)
{
	return (dpws->instance_id >= device->instance_id && dpws->msg_nb > device->message_number);
}

short * filter_services_by_type(struct dpws * dpws,	struct device_proxy * device_proxy, char * ns, char * port_type, int *res_size)
{
	dyn_array result = DYN_ARRAY_INITIALIZER(sizeof(short), 5);	// TODO: param for increment ?
	struct grab_service_info grabInfo = {dpws, &result, ns, port_type};
	
	// loop over device endpoint types
	if (is_matching_proxy(&device_proxy->super.types, ns, port_type))
		*(short *)new_entry(dpws, &result) = device_proxy->super.href;

	// loop over services
	browse_dyn_array(&device_proxy->hosted_services, (browse_entry_cbk)grab_service, &grabInfo);
	*res_size = result.nb;

	return result.tab;
}

short filter_services_by_id(struct dpws * dpws, struct device_proxy * device_proxy, char * id)
{
	// loop over services
	int index = browse_dyn_array(&device_proxy->hosted_services, (browse_entry_cbk)find_service_hook, id);
	return index < device_proxy->hosted_services.nb ? *(short *)get_entry(&device_proxy->hosted_services, index) : -1;
}

struct service_proxy * add_service_proxy(struct device_proxy * device, struct wsa_endpoint_ref **EPRs, int nbEPRs, dyn_array* types, char* service_id)
{
	int i;

	struct service_proxy * service = malloc(sizeof(struct service_proxy));
	service->super.href = createHandle(&cachePool, SERVICE_PROXY_TYPE, service);
	*(short *)new_entry(NULL, &device->hosted_services) = service->super.href;
	string_array_dup(NULL, &service->super.types, types);
	service->id = dpws_strdup(service_id);

	init_dyn_array(&service->endpoint_references, sizeof(char *), nbEPRs);
	for (i = 0; i < nbEPRs; i++) {
		*(char **)new_entry(NULL, &service->endpoint_references) = dpws_strdup(EPRs[i]->address);
	}
	
	service->host = device;
	return service;
}

int checkout_proxy(struct dpws * dpws, short href)
{
	int ret = DPWS_OK;
	* (short *) new_entry(dpws, &dpws->proxyHandles) = href;
	ret = checkoutHandle(&cachePool, href) ? DPWS_OK : DPWS_ERR_INVALID_PROXY;
	return ret;
}

int checkout_proxies(struct dpws * dpws, short * handles, int nb_handles)
{
	int i = 0;
	for (; i < nb_handles; i++)
		checkout_proxy(dpws, handles[i]);
	return DPWS_OK;
}

static BOOLEAN release_proxy(short * href, void * dummy)
{
	if (*href >= 0)
		releaseHandle(&cachePool, *href);
	return FALSE;
}

int release_proxies(struct dpws * dpws)
{
	lock(cache_lock_handle);
	browse_dyn_array(&dpws->proxyHandles, (browse_entry_cbk)release_proxy, NULL);
	dpws->proxyHandles.nb = dpws->proxyHandles.size = 0;
	dpws->proxyHandles.tab = NULL;	// tab contents will be freed by dpws_end
	unlock(cache_lock_handle);
	return DPWS_OK;
}

/*----------------------------------------------------------------------------*\
 *                                  USER API                                  *
\*----------------------------------------------------------------------------*/

struct qname ** dpws_cache_get_types(struct dpws* dpws, short href_proxy)
{
	struct qname ** ret = NULL;
	struct device_proxy * device = NULL;
	struct service_proxy * service = NULL;
	dyn_array type_infos = DYN_ARRAY_INITIALIZER(sizeof(struct qname *), 2), * types;
	char ** p_type;
	handle_s *hProxy;
	
	lock(cache_lock_handle);
	// find proxy types
	hProxy = getHandle(&cachePool, href_proxy);
	if (!hProxy) {
		dpws->err = DPWS_ERR_COULD_NOT_FIND_CACHE_ENTRY;
		goto exit;
	}
		
	switch (hProxy->type)
	{
		case DEVICE_PROXY_TYPE:
			types = &((struct device_proxy *)hProxy->pObject)->super.types;
			break;
		case SERVICE_PROXY_TYPE:
			types = &((struct service_proxy *)hProxy->pObject)->super.types;
			break;
		default:
			goto exit;
	}

	// duplicate & format results
	for (p_type = (char **)types->tab; *p_type; p_type++) {
		char *p1 = *p_type, *p2;
		struct qname * type_info = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct qname));
		if (*p1 == '"') {	// expanded form
			for(p2 = ++p1; *p2 != 0 && *p2 != '"'; p2++);
			if (*p2 == '"') {
				int ns_size = (int)(p2 - p1);
				type_info->ns = soap_malloc(dpws_dpws2soap(dpws), ns_size + 1);
				strncpy(type_info->ns, p1, ns_size);
				type_info->ns[ns_size] = 0;
				if (*++p2 == ':')
					p2++;
			}
		}
		else {
			type_info->ns = NULL;
			p2 = p1;
		}
		type_info->lname = soap_strdup(dpws_dpws2soap(dpws), p2);
		*(struct qname **)new_entry(dpws, &type_infos) = type_info;
	}
	ret = type_infos.tab;
exit:
	unlock(cache_lock_handle);
	return ret;
}

int dpws_check_device_type(struct dpws* dpws, char * ns, char * type, int match_rule, short href_device)
{
	struct device_proxy * w_device = NULL;
	struct qname qn_type[] = {{ns, type}, {NULL, NULL}};
	int ret;
	
	lock(cache_lock_handle);
	w_device = getTypedObject(&cachePool, href_device, DEVICE_PROXY_TYPE);
	if (!w_device)
		return dpws->err = DPWS_ERR_COULD_NOT_FIND_CACHE_ENTRY;

	ret = device_match(qn_type, NULL, match_rule, &w_device->super.types, &w_device->scopes);
	unlock(cache_lock_handle);
	return ret;
}

int dpws_release_proxy(short handleRef)
{
	int ret = DPWS_OK;
	lock(cache_lock_handle);
	if (handleRef >= 0)
		ret = releaseHandle(&cachePool, handleRef);
	unlock(cache_lock_handle);
	return ret;
}

int dpws_pin_proxy(short href_proxy)
{
	int ret;
	lock(cache_lock_handle);
	ret = checkoutHandle(&cachePool, href_proxy) ? DPWS_OK : DPWS_ERR_INVALID_PROXY;
	unlock(cache_lock_handle);
	return ret;
}

int invalidate_device_proxy(struct device_proxy * device)
{
	device->superEPR.status = INVALID;
	return releaseHandle(&cachePool, device->super.href);
}

int dpws_invalidate_proxy(short href)
{
	struct device_proxy * device;
	handle_s *handleInfo;
		
	lock(cache_lock_handle);
	handleInfo = getHandle(&cachePool, href);
	if (!handleInfo)
		return DPWS_ERR_INVALID_PROXY;
	switch (handleInfo->type)
	{
		case DEVICE_PROXY_TYPE:
			device = (struct device_proxy *)handleInfo->pObject;
			invalidate_device_proxy(device);
			break;
		case SERVICE_PROXY_TYPE:
			device = ((struct service_proxy *)handleInfo->pObject)->host;
			clean_hosted_services(device);
			break;
	}
	unlock(cache_lock_handle);
	return DPWS_OK;
}

void dpws_invalidate(struct dpws* dpws, struct wsa_endpoint_ref* device)
{
	short href = dpws_get_proxy_by_address(device->address);
	if (href >= 0)
		dpws_invalidate_proxy(href);
}



short dpws_get_proxy_by_id(char * id)
{
	handle_s * h;
	short ret = -1;
	lock(cache_lock_handle);
	h = handlePoolIterate(&cachePool, ANY_TYPE, (handle_cbk)compare_proxies_id, id);
	if (h)
		ret = h->handleRef;
	unlock(cache_lock_handle);
	return ret;
}

short dpws_get_proxy_by_address(char * address)
{
	handle_s * h;
	short ret = -1;
	lock(cache_lock_handle);
	h = handlePoolIterate(&cachePool, ANY_TYPE, (handle_cbk)lookup_proxy_addresses, address);
	if (h)
		ret = h->handleRef;
	unlock(cache_lock_handle);
	return ret;
}

BOOLEAN dpws_check_proxy(short href_proxy)
{
	BOOLEAN ret = FALSE;
	handle_s *handleInfo;

	lock(cache_lock_handle);
	handleInfo = getHandle(&cachePool, href_proxy);
	if (!handleInfo)
		return FALSE;
		
	switch (handleInfo->type)
	{
		case DEVICE_PROXY_TYPE:
			ret = ((struct device_proxy *)handleInfo->pObject)->superEPR.status != INVALID;
			break;
		case SERVICE_PROXY_TYPE:
			ret = ((struct service_proxy *)handleInfo->pObject)->host != NULL;
			break;
	}
	unlock(cache_lock_handle);
	return ret;
}

proxy_type dpws_get_proxy_type(short href_proxy)
{
	proxy_type ret = UNKNOWN_PROXY_TYPE;
	handle_s *handleInfo;

	lock(cache_lock_handle);
	handleInfo = getHandle(&cachePool, href_proxy);
	if (handleInfo)
		ret = handleInfo->type;
	unlock(cache_lock_handle);
	return ret;
}

struct wsa_endpoint_ref * dpws_get_default_endpoint_ref(struct dpws* dpws, short href_proxy)
{
	struct wsa_endpoint_ref * ret = NULL;
	lock(cache_lock_handle);
	ret = get_default_EPR(dpws, href_proxy);
	unlock(cache_lock_handle);
	return ret;
}

struct wsa_endpoint_ref ** dpws_get_endpoint_refs(struct dpws* dpws, short href_proxy, int * res_size)
{
	struct wsa_endpoint_ref ** ret = NULL;
	lock(cache_lock_handle);
	ret = get_EPRs(dpws, href_proxy, res_size);
	unlock(cache_lock_handle);
	return ret;
}

char ** dpws_cache_get_scopes(struct dpws* dpws, short href_device)
{
	struct device_proxy * device_proxy;
	dyn_array scopes;
	
	lock(cache_lock_handle);
	device_proxy = getTypedObject(&cachePool, href_device, DEVICE_PROXY_TYPE);
	if (!device_proxy) {
		dpws->err = DPWS_ERR_NO_HANDLE_FOUND;
		goto exit;
	}
		
	string_array_dup(dpws, &scopes, &device_proxy->scopes);

exit:
	unlock(cache_lock_handle);
	return (char **)scopes.tab;
}

char * dpws_cache_get_uuid(struct dpws* dpws, short href_device)
{
	struct device_proxy * device_proxy;
	char * uuid;
	
	lock(cache_lock_handle);
	device_proxy = getTypedObject(&cachePool, href_device, DEVICE_PROXY_TYPE);
	if (!device_proxy) {
		dpws->err = DPWS_ERR_NO_HANDLE_FOUND;
		goto exit;
	}
		
	uuid = dpws ? soap_strdup(dpws_dpws2soap(dpws), device_proxy->superEPR.super.address)
				: dpws_strdup(device_proxy->superEPR.super.address);

exit:
	unlock(cache_lock_handle);
	return uuid;
}

unsigned long dpws_cache_get_device_metadata_version(short href_device)
{
	struct device_proxy * device;
	unsigned long ret = 0;
	
	lock(cache_lock_handle);
	// find device entry
	device = getObject(&cachePool, href_device);
	if (!device) {
		ret = DPWS_ERR_NO_HANDLE_FOUND;
		goto exit;
	}
	ret = device->metadata_version;
exit:
	unlock(cache_lock_handle);
	return ret;
}

char** dpws_cache_get_device_transport_addresses(struct dpws* dpws, short href_device, int * res_size)
{
	struct device_proxy * device;
	dyn_array ret;
	
	lock(cache_lock_handle);
	// find device entry
	device = getObject(&cachePool, href_device);
	if (!device) {
		dpws->err = DPWS_ERR_NO_HANDLE_FOUND;
		goto exit;
	}
	string_array_dup(dpws, &ret, &device->transport_addresses);
	*res_size = device->transport_addresses.nb;
exit:
	unlock(cache_lock_handle);
	return (char **)ret.tab;
}

short dpws_get_host_device_proxy(short href_service)
{
	struct service_proxy *service = NULL;
	service = getTypedObject(&cachePool, href_service, SERVICE_PROXY_TYPE);
	return service ? service->host->super.href : -1;
}
