/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.45.2.5 $
|                                              $Date: 2007/03/23 16:41:34 $
\*============================================================================*/
 
#ifndef CACHE_H
# define CACHE_H

# include "tools.h"
# include "_dpws.h"
# include "discovery.h"

/*----------------------------------------------------------------------------*\
 *                               Discovery Cache                              *
\*----------------------------------------------------------------------------*/
struct proxy
{
	short href;
	dyn_array types;	// NOTE: of the form "nsURI":localName
};

typedef enum {CACHED = 0, USER = 1, INVALID = 2} endpoint_status;

struct client_endpoint_ref
{
	struct wsa_endpoint_ref super;
	endpoint_status status;
};

struct device_proxy
{
	struct proxy super;
	struct client_endpoint_ref superEPR;
	unsigned long metadata_version;
	unsigned short message_number;
	unsigned long instance_id;
	dyn_array scopes;	// URIs (char *)
	dyn_array transport_addresses;	// URIs (char *)
	BOOLEAN hosted_retrieved;
	dyn_array hosted_services;	//  services handle refs
};

struct service_proxy
{
	struct proxy super;
	char *id;
	dyn_array endpoint_references;	// char * instead of full EPR
	struct device_proxy * host;	// The proxy is a "zombie" if null
};

int initCache();
void uninitCache();
void clean_cache();
struct device_proxy * find_device_proxy(char *id);
/* Get host if a hosted service proxy is supplied */
struct device_proxy * get_device_proxy(short href, BOOLEAN *isDevice);
int count_cache_entry(struct dpws * dpws, struct qname *types, char ** scopes, int match_rule, int max_res);
short * cache_lookup(struct dpws * dpws, struct qname *types, char ** scopes, int match_rule, int *res_size);
int invalidate_device_proxy(struct device_proxy * device);
struct device_proxy * add_cache_entry(struct dpws*, struct wsd__HelloType *);	// Hello, RM, PM
void replace_cache_entry(struct device_proxy *, struct dpws*, struct wsd__HelloType *);	// Hello, RM, PM
int is_sequenced_msg(struct dpws* dpws, struct device_proxy * device);
short * filter_services_by_type(struct dpws *, struct device_proxy *, char *, char *, int *);
struct service_proxy * add_service_proxy(struct device_proxy * device, struct wsa_endpoint_ref **EPRs, int nbEPRS, dyn_array* types, char* service_id);
short filter_services_by_id(struct dpws * dpws, struct device_proxy * cache_entry, char * id);
struct wsa_endpoint_ref * get_default_EPR(struct dpws * dpws, short href);
struct wsa_endpoint_ref ** get_EPRs(struct dpws * dpws, short href, int * res_size);
int checkout_proxy(struct dpws * dpws, short href);
int checkout_proxies(struct dpws * dpws, short * handles, int nb_handles);
int release_proxies(struct dpws * dpws);

#define CACHE_POOL_NB_BUCKETS	64

#endif
