/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.178.2.5 $
|                                              $Date: 2007/05/25 17:10:50 $
\*============================================================================*/
 
/******************************************************************************\
 *                                  SERVER DATA                               *
\******************************************************************************/
#include "_dpws.h"
#include "config.h"
#include "metadata.h"
#include "mdm.h"
#include "event.h"
#include <sys/un.h>

// NOTE: PERF could use soap_unlink to avoid free + malloc when adding dynamic
// data to registry. But it would prevent to implement a more efficient memory
// allocation for gSOAP (global free)

/*----------------------------------------------------------------------------*\
 *                               REGISTRY DATA                                *
\*----------------------------------------------------------------------------*/
#ifdef USE_ONVIF /* Added by HG.W */
struct dpws_config tk_config = 
{ 0xFFFFFFFF , NULL, NULL, FALSE, TRUE,
  /* for na_info */ 
  {
      NULL, NULL, /* mac_addr, ip_addr */
#ifdef WITH_IPV6
      NULL, NULL, NULL, /* ip_addr_v6_g, ip_add_v6_l, ip_addr_l_with_if */
#endif      
  },
  -1, NULL, FALSE, -1,
  FALSE, /* wakeup_socks_full_path_generated */
  NULL, NULL, /* wakeup_socks_dir, wakeup_sock_cli */
  NULL, /* wakeup_sock_svr_v4*/
#ifdef WITH_IPV6
  FALSE, NULL /* server_up_v6, wakeup_sock_svr_v6 */
#endif
};
#else /* !USE_ONVIF */
struct dpws_config tk_config = {0xFFFFFFFF , NULL, NULL, FALSE, TRUE, {NULL, NULL}, -1, NULL, FALSE, -1};
#endif

static char subsc_eventing[16];

// common namespace table for headers (copied into runtime tables for WS-Discovery and WS-Mex)
#define DEF_NS_SIZE 8
struct Namespace default_namespaces[DEF_NS_SIZE] =
{
  {"SOAP-ENV", "http://www.w3.org/2003/05/soap-envelope", "http://www.w3.org/*/soap-envelope", NULL},
  {"SOAP-ENC", "http://www.w3.org/2003/05/soap-encoding", "http://www.w3.org/*/soap-encoding", NULL},
  {"xsi", "http://www.w3.org/2001/XMLSchema-instance", "http://www.w3.org/*/XMLSchema-instance", NULL},
  {"wsa", WSA_URI, "http://schemas.xmlsoap.org/ws/*/addressing", NULL},
#ifndef USE_ONVIF /* by HG.W */
  {"wdp", WDP_URI, "http://schemas.xmlsoap.org/ws/*/devprof", NULL},
  {"wse", WSE_URI, "http://schemas.xmlsoap.org/ws/*/eventing", NULL},
#endif /* by HG.W */
  {"wsd", WSD_URI, "http://schemas.xmlsoap.org/ws/*/discovery", NULL},
  {NULL, NULL, NULL, NULL}
};

#define DISC_NS_SIZE 4  // TODO: use generated table ?
struct Namespace discovery_ns[] =
{
        {"SOAP-ENV", "http://www.w3.org/2003/05/soap-envelope", "http://www.w3.org/*/soap-envelope", NULL},
        {"wsa", WSA_URI, "http://schemas.xmlsoap.org/ws/*/addressing", NULL},
        {"wsd", WSD_URI, NULL, NULL},
        {NULL, NULL, NULL, NULL}
};

#define METADATA_NS_SIZE 5
extern struct Namespace metadata_snd_namespaces[];

struct device *registry = NULL;
short hidden_device_href = -1;

struct transport HTTP_TRANSPORT = {"http", "http:"};

#ifdef USE_ONVIF /* Added by HG.W */
struct listener http_listener = {
                &HTTP_TRANSPORT, NULL, TRUE, -1, -1,
                NULL,/* external_path = NULL */
#ifdef WITH_IPV6
                NULL, /* url_v6_g = NULL */
                NULL, /* url_v6_l = NULL */
#endif
                DYN_ARRAY_INITIALIZER(sizeof(short), 5), NULL
        };
#else /* !USE_ONVIF */
struct listener http_listener = {
                &HTTP_TRANSPORT, NULL, TRUE, -1, -1,
                DYN_ARRAY_INITIALIZER(sizeof(short), 5), NULL
        };      // NOTE: only one listener in the 2.0 version of the tk
#endif

struct listener * listeners = &http_listener;

handlePool_t registryPool;

short registry_lock_handle = -1;
static int prefix_counter = 0;

static int set_service_class_ptr_att(struct service_class * pServiceClass, int att, const void * value, BOOLEAN reset);

// TODO: perf use hashmap for uuids and service ids & store binary form (comparison would require parsing) ?
// TODO: hot plug & configuration ? Use href indirection and a copy
/*----------------------------------------------------------------------------*\
 *                               PRIVATE TOOLS                                *
\*----------------------------------------------------------------------------*/

#ifdef USE_ONVIF /* Modified by HG.W */

static int __build_url(char **url, const char *scheme, const char *host, const char *path)
{
    if(!url || !scheme || !host) return DPWS_ERR_INVALID_PARAMETER; 

    if(path){
        *url = malloc(14 + strlen(host) + strlen(path));
        if((path[0] == '/') || (path[0] == ':')){
            sprintf(*url, "%s//%s%s", scheme, host, path);
        }
        else{
            sprintf(*url, "%s//%s/%s", scheme, host, path);
        }
    }
    else{
        *url = malloc(14 + strlen(host));
        sprintf(*url, "%s//%s/", scheme, host);
    }

    return DPWS_OK;
}

#ifdef WITH_IPV6
static int default_url_v6(struct listener * listener) {
    int ret = DPWS_OK;

    free(listener->url_v6_g);
    listener->url_v6_g = NULL;
    free(listener->url_v6_l);
    listener->url_v6_l = NULL;

    if(!tk_config.network_adapter.ip_addr_v6_g
#ifndef USE_IPV6_GLOBAL_ADDR_IN_XADDRS_ALWAYS
       && !tk_config.network_adapter.ip_addr_v6_l
#endif
      ){
        ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: No available IPv6 address.\n");
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }

    if(tk_config.network_adapter.ip_addr_v6_g){    
        ret = __build_url(&listener->url_v6_g,
                          listener->transport->scheme,
                          tk_config.network_adapter.ip_addr_v6_g,
                          listener->external_path);
        if(ret == DPWS_OK){
            listener->url_generated = TRUE;
            ONVIF_LOG1(ONVIF_LOG_INFO,
                       "[INFO]: Built url with IPv6 global addr=[%s].\n",
                       listener->url_v6_g);
        }
    }

    if(tk_config.network_adapter.ip_addr_v6_l){
        ret = __build_url(&listener->url_v6_l,
                          listener->transport->scheme,
                          tk_config.network_adapter.ip_addr_v6_l,
                          listener->external_path);
        if(ret == DPWS_OK){
            listener->url_generated = TRUE;
            ONVIF_LOG1(ONVIF_LOG_INFO,
                       "[INFO]: Built url with IPv6 link-local addr=[%s].\n",
                       listener->url_v6_l);
        }
    }


    return ret;
}
#endif /* WITH_IPV6 */

static int default_url(struct listener * listener) {
    int ret = DPWS_OK;
    
    free(listener->url);
    listener->url = NULL;

    if(tk_config.network_adapter.ip_addr){
        ret = __build_url(&listener->url,
                          listener->transport->scheme,
                          tk_config.network_adapter.ip_addr,
                          listener->external_path);
        if(ret == DPWS_OK){
            listener->url_generated = TRUE;
            ONVIF_LOG1(ONVIF_LOG_INFO,
                       "[INFO]: Built url with IPv4 addr=[%s].\n",
                       listener->url);
        }
        return ret;
    }
    else{
        ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: No available IPv4 address.\n");
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }
}

#else /* !USE_ONVIF */
static int default_url(struct listener * listener) {
        free(listener->url);
        if (tk_config.network_adapter.ip_addr) {
                listener->url = malloc(14 + strlen(tk_config.network_adapter.ip_addr));
                sprintf(listener->url, "%s//%s:%d", listener->transport->scheme, tk_config.network_adapter.ip_addr, listener->port);
                return DPWS_OK;
        }
        else
                return DPWS_ERR_RETRIEVING_IP_ADDRESS;          
}
#endif

/* Namespace table management (gSOAP) format */
static int add_ns_entry(char * prefix, char * ns, char * ns_pattern, dyn_array *table)
{
        struct Namespace *nns = NULL, *wns = NULL;
        int i;

        // check that if prefix is used twice, the namespace is the same...
        for(i = 0; i < table->nb; i++) {
                wns = (struct Namespace *)get_entry(table, i);
                if (!strcmp(prefix, wns->id)) {
                        if (strcmp(ns, wns->ns))
                                return DPWS_ERR_PREFIX_ALREADY_USED;
                        else
                                return SOAP_OK; // ns is the same
                }
        }
        // add entry
        nns = (struct Namespace *) new_entry(NULL, table);
        nns->id = prefix;
        nns->ns = ns;
        nns->in = ns_pattern;
        nns->out = NULL;
        return SOAP_OK;
}

static void wsdlInfoFree(struct wsdl_info *pWsdl)
{
        free(pWsdl->target_ns);
        free(pWsdl->location);
}

static void prefixedQNameFree(struct prefixed_qname *pPQName)
{
        free(pPQName->prefix);
        free(pPQName->qname.ns);
        free(pPQName->qname.lname);
}

static int set_service_endpoint_ptr_att(struct service_endpoint * pService, int att, const void * value, BOOLEAN reset)
{
        time_t t;
        switch (att)
        {
                case DPWS_STR_MAX_SUBSC_DURATION:
                        if (*(char *)value != 'P')
                                return DPWS_ERR_INVALID_PARAMETER;
                        t = xmlduration2duration((char *)value);
                        if (t == -1)
                                return DPWS_ERR_INVALID_PARAMETER;
                        else
                                pService->max_subsc_duration = t;
                        break;
                case DPWS_STR_SERVICE_ID:
                        free(pService->service_id);
                        pService->service_id = dpws_strdup(value);
                        break;
                case DPWS_PTR_USER_DATA:
                        pService->user_data = value;
                        break;
                case DPWS_PTR_HANDLING_FUNCTION: case DPWS_PTR_FAULT_FUNCTION: case DPWS_PTR_CALLBACK_EVENT_END:        // NOTE: Not ideal, a service class may be shared if it is a Hosted Service endpoint...
                        set_service_class_ptr_att((struct service_class *)getObject(&registryPool, pService->hrefServiceClass), att, value, reset);
                        break;
                case DPWS_INT_CLASS_HREF:
                        return DPWS_ERR_IMMUTABLE_USED_OBJECT;
                default:
                        return DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
        }
        return DPWS_OK;
}

static int set_service_port_ptr_att(struct service_port * pServicePort, int att, const void * value, BOOLEAN reset)
{
        switch (att)
        {
                case DPWS_STR_ADDRESS:
                        free(pServicePort->context_path);
                        pServicePort->context_path = dpws_strdup(value);
                        break;
                default:
                        return DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
        }
        return DPWS_OK;
}

static int set_service_class_ptr_att(struct service_class * pServiceClass, int att, const void * value, BOOLEAN reset)
{
        struct prefixed_qname * pqName, * pqnParam, pqn;
        struct wsdl_info *wsdlInfo;
        char buffer[QNAME_PREFIX_GEN_LEN], **pStr;      // TODO: check overflow
        dyn_array *handlers;
        int (**pHandler)(struct dpws *) = NULL;

        switch (att)
        {
                case DPWS_STR_ID:
                        free(pServiceClass->id);
                        pServiceClass->id = dpws_strdup((char *)value);
                        break;
                case DPWS_PTR_TYPE:
                case DPWS_PTR_PREFIXED_TYPE:
                        if (reset)
                                free_dyn_array(&pServiceClass->types, (void (*)(void *))&prefixedQNameFree);
                        pqName = (struct prefixed_qname *)new_entry(NULL, &pServiceClass->types);
                        
                        if (att == DPWS_PTR_PREFIXED_TYPE)
                                pqnParam = (struct prefixed_qname *)value;
                        else {
                                pqnParam = memcpy(&pqn, value, sizeof(struct qname));
                                sprintf(buffer, "%s%d", QNAME_PREFIX_GEN_PREFIX, prefix_counter++);     // NOTE: counter is global since a prefix must be unique for a device but a service class can be shared
                                pqnParam->prefix = buffer;
                        }                       
                        pqName->prefix = dpws_strdup(pqnParam->prefix);
                        pqName->qname.ns = dpws_strdup(pqnParam->qname.ns);
                        pqName->qname.lname = dpws_strdup(pqnParam->qname.lname);
                        pStr = (char **)new_entry(NULL, &pServiceClass->all_types);
                        *pStr = malloc(strlen(pqName->prefix) + strlen(pqName->qname.lname) + 2);
                        strcpy(*pStr, pqName->prefix);
                        strcat(*pStr, ":");
                        strcat(*pStr, pqName->qname.lname);
                        break;
                case DPWS_PTR_WSDL:
                        if (reset)
                                free_dyn_array(&pServiceClass->wsdl_files, (void (*)(void *))&wsdlInfoFree);
                        wsdlInfo = (struct wsdl_info *)new_entry(NULL, &pServiceClass->wsdl_files);
                        wsdlInfo->target_ns     = dpws_strdup(((struct wsdl_info *)value)->target_ns);
                        wsdlInfo->location      = dpws_strdup(((struct wsdl_info *)value)->location);
                        break;
                case DPWS_PTR_HANDLING_FUNCTION:
                        handlers = &pServiceClass->funcs;
                        if (reset)
                                free_dyn_array(handlers, NULL);
                        /* Check for handler doubles */
                        if (handlers->tab)
                                for(pHandler = handlers->tab; *pHandler && *pHandler != value; pHandler++);
                        if (!pHandler || !*pHandler)
                                *(int (**)())new_entry(NULL, handlers) = (int (*)(struct dpws *))value;
                        break;
                case DPWS_PTR_FAULT_FUNCTION:
                        pServiceClass->fault_func = (int (*)(struct dpws *))value;
                        break;
                case DPWS_PTR_CALLBACK_EVENT_END:
                        pServiceClass->event_end_hook = (void (*) (struct event_end_info *))value;
                        set_service_class_ptr_att(pServiceClass, DPWS_PTR_HANDLING_FUNCTION, &eventing_handle_event, FALSE);
                        break;
                case DPWS_PTR_USER_DATA:
                        pServiceClass->user_data = value;
                        break;
                default:
                        return DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
        }
        return DPWS_OK;
}

static int set_device_class_ptr_att(struct device_class * pDeviceClass, int att, const void * value, BOOLEAN reset)
{
        struct service_class * pServiceClass;
        switch (att)
        {
                case DPWS_STR_ID:
                        free(pDeviceClass->id);
                        pDeviceClass->id = dpws_strdup((char *)value);
                        break;
                case DPWS_INT_METADATA_VERSION:
                        pDeviceClass->metadata_version = *(unsigned short *)value;
                        break;
                case DPWS_STR_MODEL_NAME:
                        free(pDeviceClass->model_name);
                        pDeviceClass->model_name = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_MODEL_NUMBER:
                        free(pDeviceClass->model_number);
                        pDeviceClass->model_number = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_MODEL_URL:
                        free(pDeviceClass->model_url);
                        pDeviceClass->model_url = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_PRESENTATION_URL:
                        free(pDeviceClass->presentation_url);
                        pDeviceClass->presentation_url = dpws_strdup((char *)value);
                        break;
                case DPWS_BOOL_MANAGEMENT_SERVICE:
                        pDeviceClass->mdm_enabled = *(BOOLEAN *)value;
                        break;
                case DPWS_STR_MANUFACTURER:
                        free(pDeviceClass->manufacturer);
                        pDeviceClass->manufacturer = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_MANUFACTURER_URL:
                        free(pDeviceClass->manufacturer_url);
                        pDeviceClass->manufacturer_url = dpws_strdup((char *)value);
                        break;
                case DPWS_PTR_USER_DATA:
                        pDeviceClass->user_data = value;
                        break;
                default:
                        pServiceClass = (struct service_class *)
                                getObject(&registryPool, pDeviceClass->hrefBuiltinServiceClass);
                        return set_service_class_ptr_att(pServiceClass, att, value, reset);
        }
        return DPWS_OK;
}

static int set_config_ptr_att(int att, const void * value, BOOLEAN reset)
{
        time_t t;
        switch (att)
        {
                case DPWS_STR_MAC_ADDR:
                        if (reset)
                        {
                                free(tk_config.network_adapter.mac_addr);
                                tk_config.network_adapter.mac_addr = parse_mac_address((char *)value);
                        }
                        else
                                return DPWS_ERR_MONOVALUED_ATT_TYPE;
                        break;
                case DPWS_STR_IP_ADDR:
                        tk_config.network_adapter.ip_addr = dpws_strdup((char *)value);
                        break;
                case DPWS_INT_BOOT_SEQ:
                        tk_config.boot_seq = *(unsigned long *)value;
                        break;
                case DPWS_BOOL_TCP_LISTEN:
                        tk_config.tcp_listen = *(BOOLEAN *)value;
                        break;
                case DPWS_PTR_CALLBACK_HELLO:
                        if (reset)
                                tk_config.hello_hook = (void (*) (struct dpws *, short))value;
                        else
                                return DPWS_ERR_MONOVALUED_ATT_TYPE;
                        break;
                case DPWS_PTR_CALLBACK_BYE:
                        if (reset)
                                tk_config.bye_hook = (void (*) (struct dpws *, short))value;
                        else
                                return DPWS_ERR_MONOVALUED_ATT_TYPE;
                        break;
                case DPWS_STR_HTTP_ADDR:
                        free(listeners->url);
                        listeners->url = dpws_strdup((char *)value);
                        listeners->url_generated = FALSE;
                        break;
                case DPWS_INT_HTTP_PORT:
                        listeners->port = *(unsigned long *)value;
                        break;
                case DPWS_INT_HTTP_BACKLOG:
                        listeners->tcp_backlog = *(unsigned long *)value;
                        break;
                case DPWS_STR_PREFERRED_LANG:
                        free(tk_config.preferred_lang);
                        tk_config.preferred_lang = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_MAX_SUBSC_DURATION:
                        if (*(char *)value != 'P')
                                return DPWS_ERR_INVALID_PARAMETER;
                        t = xmlduration2duration((char *)value);
                        if (t == -1)
                                return DPWS_ERR_INVALID_PARAMETER;
                        else
                                tk_config.max_subsc_duration = t;
                        break;
                case DPWS_BOOL_MANAGEMENT_SERVICE:
                        tk_config.mdm_enabled = *(BOOLEAN *)value;
                        break;
                case DPWS_INT_DEFAULT_DEVICE:
                        tk_config.default_device_handle = (short)*(int *)value;
                        break;
/* Added by honggnag */
#ifdef USE_ONVIF
        case DPWS_STR_EXTERNAL_PATH:
            free(listeners->external_path);
            listeners->external_path = dpws_strdup((char *)value);
            break;
        case DPWS_STR_WAKEUP_SOCKS_DIR:
            free(tk_config.wakeup_socks_dir);
            tk_config.wakeup_socks_dir = dpws_strdup((char *)value);
            break;
#endif
                default:
                        return DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
        }
        return DPWS_OK;
}

static void * get_config_ptr_att(int att, int index)
{
        switch (att)
        {
                case DPWS_STR_MAC_ADDR:
                        return tk_config.network_adapter.mac_addr;      // TODO: string hexa form
                case DPWS_STR_IP_ADDR:
                        return tk_config.network_adapter.ip_addr;
                case DPWS_INT_BOOT_SEQ:
                        return &tk_config.boot_seq;
                case DPWS_BOOL_TCP_LISTEN:
                        return &tk_config.tcp_listen;
                case DPWS_PTR_CALLBACK_HELLO:
                        return tk_config.hello_hook;
                case DPWS_PTR_CALLBACK_BYE:
                        return tk_config.bye_hook;
                case DPWS_STR_HTTP_ADDR:
                        return listeners->url;
                case DPWS_INT_HTTP_PORT:
                        return &listeners->port;
                case DPWS_INT_HTTP_BACKLOG:
                        return &listeners->tcp_backlog;
                case DPWS_STR_PREFERRED_LANG:
                        return tk_config.preferred_lang;
                case DPWS_STR_MAX_SUBSC_DURATION:
                        return dpws_duration2xmlduration(subsc_eventing, tk_config.max_subsc_duration);
                case DPWS_BOOL_MANAGEMENT_SERVICE:
                        return &tk_config.mdm_enabled;
                default:
                        return NULL;
        }
}

static void * get_service_class_ptr_att(struct service_class * pServiceClass, int att, int index)
{
        switch (att)
        {
                case DPWS_STR_ID:
                        return pServiceClass->id;
                case DPWS_PTR_TYPE:
                        return (struct prefixed_qname *)get_entry(&pServiceClass->types, index);
                case DPWS_PTR_WSDL:
                        return (struct wsdl_info *)get_entry(&pServiceClass->wsdl_files, index);
                case DPWS_PTR_HANDLING_FUNCTION:
                        return *(int (**)(struct dpws*))get_entry(&pServiceClass->funcs, index);
                case DPWS_PTR_FAULT_FUNCTION:
                        return pServiceClass->fault_func;
                case DPWS_PTR_CALLBACK_EVENT_END:
                        return pServiceClass->event_end_hook;
                case DPWS_PTR_USER_DATA:
                        return (void *)pServiceClass->user_data;
                default:
                        return NULL;
        }
}

static void * get_device_class_ptr_att(struct device_class * pDeviceClass, int att, int index)
{
        struct service_class * pServiceClass = (struct service_class *)
                getObject(&registryPool, pDeviceClass->hrefBuiltinServiceClass);
        switch (att)
        {
                case DPWS_STR_ID:
                        return pDeviceClass->id;
                case DPWS_INT_METADATA_VERSION:
                        return &pDeviceClass->metadata_version;
                case DPWS_STR_MODEL_NAME:
                        return pDeviceClass->model_name;
                case DPWS_STR_MODEL_NUMBER:
                        return pDeviceClass->model_number;
                case DPWS_STR_MODEL_URL:
                        return pDeviceClass->model_url;
                case DPWS_STR_PRESENTATION_URL:
                        return pDeviceClass->presentation_url;
                case DPWS_BOOL_MANAGEMENT_SERVICE:
                        return &pDeviceClass->mdm_enabled;
                case DPWS_STR_MANUFACTURER:
                        return pDeviceClass->manufacturer;
                case DPWS_STR_MANUFACTURER_URL:
                        return pDeviceClass->manufacturer_url;
                case DPWS_PTR_USER_DATA:
                        return (void *)pDeviceClass->user_data;
                default:
                        return get_service_class_ptr_att(pServiceClass, att, index);
        }
}

static void * get_service_endpoint_ptr_att(struct service_endpoint * pService, int att, int index)
{
        switch (att)
        {
                case DPWS_STR_MAX_SUBSC_DURATION:
                        return dpws_duration2xmlduration(subsc_eventing, pService->max_subsc_duration);
                case DPWS_STR_SERVICE_ID:
                        return pService->service_id;
                case DPWS_PTR_USER_DATA:
                        return (void *)pService->user_data;
                case DPWS_INT_CLASS_HREF:
                        return &pService->hrefServiceClass;
                case DPWS_PTR_HANDLING_FUNCTION: case DPWS_PTR_FAULT_FUNCTION: case DPWS_PTR_CALLBACK_EVENT_END:
                        return get_service_class_ptr_att((struct service_class *)getObject(&registryPool, pService->hrefServiceClass), att, index);
                default:
                        return NULL;
        }
}

static void * get_service_port_ptr_att(struct service_port * pServicePort, int att, int index)
{
        switch (att)
        {
                case DPWS_STR_ADDRESS:
                        return pServicePort->context_path;      // TODO: check unicity
                default:
                        return NULL;
        }
}

static void * get_device_ptr_att(struct device * pDevice, int att, int index)
{
        switch (att)
        {
                case DPWS_STR_DEVICE_ID:
                        return pDevice->uuid;
                case DPWS_STR_SCOPE:
                        return *(char **)get_entry(&pDevice->scopes, index);
                case DPWS_STR_FRIENDLY_NAME:
                        return pDevice->friendly_name;
                case DPWS_STR_FIRMWARE_VERSION:
                        return pDevice->firmware_version;
                case DPWS_STR_SERIAL_NUMBER:
                        return pDevice->serial_number;
                case DPWS_INT_METADATA_VERSION:
                        return &pDevice->metadata_version;
                case DPWS_PTR_USER_DATA:
                        return (void *)pDevice->user_data;
                case DPWS_INT_CLASS_HREF:
                        return &pDevice->hrefClass;
                default:
                        return get_service_endpoint_ptr_att(get_endpoint(pDevice->hrefEndpoint), att, index);
        }
}

static short create_hosted_service(struct device * device, short hrefClass)
{
        struct service_endpoint * service;
        struct device_class * devClass;
        struct service_class * servClass = (struct service_class *)checkoutObject(&registryPool, hrefClass);

        service = malloc(sizeof(struct service_endpoint));
        memset(service, 0, sizeof(struct service_endpoint));
        service->service_id = dpws_strdup(servClass->id);
        service->hrefServiceClass = hrefClass;
        devClass = (struct device_class *)getObject(&registryPool, device->hrefClass);
        service->max_subsc_duration = tk_config.max_subsc_duration;
        init_dyn_array(&service->service_ports, sizeof(short), 5);
        service->device = device;
        if (device->hosted_services)
                device->hosted_services->previous = service;
        service->next = device->hosted_services;
        device->hosted_services = service;
        
        service->href = createHandle(&registryPool, SERVICE_ENDPOINT_TYPE, service);
        if (service->href < 0)
                free(service);
        else
                device->service_count++;

        return service->href;
}

static BOOLEAN findBigger(short * hrefServPort, char * ref)
{
        char * current = ((struct service_port *)getObject(&registryPool, *hrefServPort))->context_path;
        return strcmp(current, ref) > 0;
}

static short create_service_port(char * context_path)
{
        short href = -1;
        struct service_port *serv_port = malloc(sizeof(struct service_port));
        if (!serv_port)
                return DPWS_ERR_OUT_OF_MEMORY;
                
        // init service port
        serv_port->context_path = context_path;
        serv_port->listener = listeners;
        serv_port->service = NULL;

        href = createHandle(&registryPool, SERVICE_PORT_TYPE, serv_port);
        if (href < 0)
                free(serv_port);
        
        // Add the service_port to the HTTP listener ordered array
        *(short *)insert_entry(NULL, &listeners->service_ports,
                browse_dyn_array(&listeners->service_ports, (browse_entry_cbk)findBigger, serv_port->context_path)) = href;

        return href;
}

static short create_random_service_port()
{
        char * context_path = malloc(UUID_STRING_SIZE);
        return create_service_port(uuid_create(context_path));
}

static int bind_service(short hrefService, short hrefServicePort)
{
        struct service_endpoint * serviceEndpoint;
        struct service_port * servicePort;
        
        serviceEndpoint = (struct service_endpoint *)getObject(&registryPool, hrefService);
        servicePort = (struct service_port *)checkoutObject(&registryPool, hrefServicePort);
        
        if (!serviceEndpoint || !servicePort)
                return DPWS_ERR_NO_HANDLE_FOUND;
                
        * (short *)new_entry(NULL, &serviceEndpoint->service_ports) = hrefServicePort;
        servicePort->service = serviceEndpoint;

        return DPWS_OK;
}

static int freeDevice(handle_s *handleInfo)
{
        struct device * device = (struct device *)handleInfo->pObject;
        int ret = 1;

        if (ret)
                ret = releaseHandle(&registryPool, device->hrefClass);
                
        while(ret && device->hosted_services)   // NOTE: only one handle for API user and device (device hold no handle on service, one device per service) however we liberate it for the user...
                ret = releaseHandle(&registryPool, device->hosted_services->href);
                
        if (ret)
        {
                lock_destroy(device->lock);
                free_buffer_dyn_array(&device->scopes);
                free(device->friendly_name);
                free(device->firmware_version);
                free(device->serial_number);
                if (device->previous) {
                        device->previous->next = device->next;
                        device->next->previous = device->previous;
                }
                else {
                        registry = device->next;
#ifdef USE_ONVIF /* Added by HG.W (fixed null ptr access) */
                        if(device->next) device->next->previous = NULL;
#else
                        device->next->previous = NULL;
#endif
                }
                free(device);
        }

        return ret;
}

static void hrefRelease(short * href)
{
        releaseHandle(&registryPool, *href);
}

static int freeService(handle_s *handleInfo)
{
        struct service_endpoint * service = (struct service_endpoint *)handleInfo->pObject;
        struct event_subscription * pwes;
        int ret = 1;

        ret = releaseHandle(&registryPool, service->hrefServiceClass);
        if (ret) {
                free(service->service_id);
                free_dyn_array(&service->service_ports, (free_entry_cbk)hrefRelease);
                
                service->device->service_count--;
                while (service->subscriptions) {
                        pwes = service->subscriptions;
                        service->subscriptions = service->subscriptions->next;
                        free_subscription(pwes);
                }

                if (service->previous) {
                        service->previous->next = service->next;
                        service->next->previous = service->previous;
                }
                else {

                        service->device->hosted_services = service->next;
#ifdef USE_ONVIF /* Added by HG.W (fixed null ptr access) */
                        if(service->next) service->next->previous = NULL;
#else
                        service->next->previous = NULL;
#endif
                }
                free(service);
        }
        return ret;
}

static int freeDeviceClass(handle_s *handleInfo)
{
        struct device_class * deviceClass = (struct device_class *)handleInfo->pObject;
        int ret = 1;
        
        ret = releaseHandle(&registryPool, deviceClass->hrefBuiltinServiceClass);

        free_dyn_array(&deviceClass->service_classes, (free_entry_cbk)hrefRelease);

        if (ret) {
#ifdef USE_ONVIF /* Added by HG.W (fixed memory leaks) */
            free(deviceClass->id);
#endif
                free(deviceClass->model_name);
                free(deviceClass->model_number);
                free(deviceClass->model_url);
                free(deviceClass->presentation_url);
                free_dyn_array(&deviceClass->discovery_namespaces, NULL);       // compilated data, allocated elsewhere
                free_dyn_array(&deviceClass->services_namespaces, NULL);        // compilated data, allocated elsewhere
                free(deviceClass->manufacturer);
                free(deviceClass->manufacturer_url);
                free(deviceClass);
        }
        return ret;
}

static int freeServiceClass(handle_s *handleInfo)
{
        struct service_class * servClass = (struct service_class *)handleInfo->pObject;

        free_dyn_array(&servClass->types, (void (*)(void *))&prefixedQNameFree);
        free_dyn_array(&servClass->wsdl_files, (void (*)(void *))&wsdlInfoFree);
        free_dyn_array(&servClass->funcs, NULL);
        free_buffer_dyn_array(&servClass->all_types);
        free(servClass);
        return DPWS_OK;
}

static BOOLEAN findServicePort(short * p_href, short * p_searched_href)
{
        return *p_href == *p_searched_href ? TRUE : FALSE;
}

static int freeServicePort(handle_s *handleInfo)
{
        struct service_port * servPort = (struct service_port *)handleInfo->pObject;
        dyn_array * servPorts = &servPort->listener->service_ports;

        // Remove port from listener since listener is not a public "holder" and cannot be removed
        int i = browse_dyn_array(servPorts, (browse_entry_cbk)findServicePort, &handleInfo->handleRef);
        if (i < servPorts->nb)
                remove_entry(servPorts, i, NULL);
        free(servPort->context_path);
        free(servPort);
        return DPWS_OK;
}

static BOOLEAN isServiceClassRegistered(struct device_class * devClass, short hrefServiceClass)
{
        short *href;
        int i;
        for (i = 0; i < devClass->service_classes.nb; i++){
                href = ((short *)devClass->service_classes.tab) + i;
                if (*href == hrefServiceClass)
                        return 1;
        }
        return 0;
}

static short create_custom_device(short localID, short hrefClass)
{
        struct device * device;
        struct device_class * devClass;
        short lock = -1;

        devClass = (struct device_class *)getObject(&registryPool, hrefClass);
        if (!devClass)
                return  DPWS_ERR_NO_HANDLE_FOUND;

        if ((lock = lock_create()) < 0)
                return lock;

        device = malloc(sizeof(struct device));
        if (!device)
                return DPWS_ERR_OUT_OF_MEMORY;

        memset(device, 0, sizeof(struct device));
        device->lock = lock;
        device->local_id = localID;
        device->hrefClass = hrefClass;
        device->metadata_version = 0;
        init_dyn_array(&device->scopes, sizeof(char *), DISCOVERY_ALLOC_INCREMENT);
        if (registry)
                registry->previous = device;
        device->next = registry;
        device->previous = NULL;
        registry = device;
        device->href = createHandle(&registryPool, DEVICE_INSTANCE_TYPE, device);
        if (device->href < 0)
                free(device);
        else
                device->hrefEndpoint = create_hosted_service(device, devClass->hrefBuiltinServiceClass);

        return device->href;
}

static int releaseHandleHook(handle_s *handleInfo)
{
        int ret;
        switch (handleInfo->type)
        {
                case DEVICE_INSTANCE_TYPE:
                        ret = freeDevice(handleInfo);
                        break;
                case SERVICE_ENDPOINT_TYPE:
                        ret = freeService(handleInfo);
                        break;
                case DEVICE_CLASS_TYPE:
                        ret = freeDeviceClass(handleInfo);
                        break;
                case SERVICE_CLASS_TYPE:
                        ret = freeServiceClass(handleInfo);
                        break;
                case SERVICE_PORT_TYPE:
                        ret = freeServicePort(handleInfo);
                        break;
                default:
                        break;
        }
        return ret;
}

static struct device * find_device(char * device_uuid)
{
        struct device * wDevice;
        
        // TODO: the next protection is used for dynamic registration, disabling/enabling. Is cost too heavy (AM opinion) ? Could multiply mutexes at least.
        lock(registry_lock_handle);
        for (wDevice = registry; wDevice; wDevice = wDevice->next) {
                if (wDevice->device_up && !dpws_strcasecmp(wDevice->uuid + UUID_SCHEME_SIZE, device_uuid))      // jump over scheme
                        break;
        }
        unlock(registry_lock_handle);
        return wDevice;
}

static BOOLEAN device_class_equals(short type, struct device_class * devClass, char * id)
{
        return !strcmp(devClass->id, id);
}

static BOOLEAN service_class_equals(short type, struct service_class * servClass, char * id)
{
        return !strcmp(servClass->id, id);
}

static BOOLEAN device_equals(short type, struct device * device, char * uuid)
{
        return !strcmp(device->uuid, uuid);
}

static dyn_array * find_service_endpoint_ports(short href)
{
        struct service_endpoint * endpoint;
        handle_s *handle = getHandle(&registryPool, href);
        if (!handle)
                return NULL;
        
        switch (handle->type)
        {
                case DEVICE_INSTANCE_TYPE:
                        endpoint = (struct service_endpoint *)getObject(&registryPool, ((struct device *)handle->pObject)->hrefEndpoint);
                        break;
                case SERVICE_ENDPOINT_TYPE:
                        endpoint = (struct service_endpoint *)handle->pObject;
                        break;
                default:
                        return NULL;
        }
        
        return &endpoint->service_ports;
}

static char* build_transport_address(struct dpws* dpws, struct service_port * port)
{
        char* result = NULL;
        
        if (port->context_path && is_physical_address(port->context_path))
            result = dpws ? port->context_path: dpws_strdup(port->context_path);
        else
        {
#ifdef USE_ONVIF /* Added by HG.W */
            if(!dpws) return NULL;

            char *tmp_url = NULL;
            if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
                tmp_url = listeners->url;
            }
#ifdef WITH_IPV6
            else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
                if(dpws->xaddr_ipv6_scope == DPWS_XADDR_IPV6_SCOPE_GLOBAL){
                    tmp_url = listeners->url_v6_g;
                }
                else{
#ifdef USE_IPV6_GLOBAL_ADDR_IN_XADDRS_ALWAYS                    
                    tmp_url = listeners->url_v6_g;
#else
                    tmp_url = listeners->url_v6_l;
#endif
                }
            }
#endif /* WITH_IPV6 */
            else{
                return NULL;
            }

            if(!tmp_url) return NULL;

            int len_addr = strlen(tmp_url) + 2;
            result = (char*)soap_malloc(dpws_dpws2soap(dpws), len_addr);
            strcpy(result, tmp_url);

#else /* !USE_ONVIF */
                int len_addr = strlen(listeners->url) + strlen(port->context_path) + 2;
                if (dpws)
                        result = (char*)soap_malloc(dpws_dpws2soap(dpws), len_addr);
                else 
                        result = (char*)malloc(len_addr);
                strcpy(result, listeners->url);
                strcat(result, "/");
                strcat(result, port->context_path);
#endif
        }

        return result;
}

/*----------------------------------------------------------------------------*\
 *               INTERNAL API: CONFIGURATION & INITIALIZATION                 *
\*----------------------------------------------------------------------------*/

// Used directly by MDM (mdm.c)
int set_device_ptr_att(struct device * pDevice, int att, const void * value, BOOLEAN reset)
{
        switch (att)
        {
                case DPWS_STR_DEVICE_ID:
                        strcpy(pDevice->uuid, (char *)value);
                        break;
                case DPWS_STR_SCOPE:
                        if (value && strlen((char *)value) > MAX_URI_SIZE)
                                return DPWS_ERR_INVALID_PARAMETER;
                        if (reset)
                                free_buffer_dyn_array(&pDevice->scopes);
                        if (value)
                                *(char **)new_entry(NULL, &pDevice->scopes) = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_FRIENDLY_NAME:
                        pDevice->friendly_name = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_FIRMWARE_VERSION:
                        pDevice->firmware_version = dpws_strdup((char *)value);
                        break;
                case DPWS_STR_SERIAL_NUMBER:
                        pDevice->serial_number = dpws_strdup((char *)value);
                        break;
                case DPWS_INT_METADATA_VERSION:
                        pDevice->metadata_version = *(unsigned short *)value;
                        break;
                case DPWS_PTR_USER_DATA:
                        pDevice->user_data = value;
                        break;
                case DPWS_INT_CLASS_HREF:
                        return DPWS_ERR_IMMUTABLE_USED_OBJECT;
                default:
                        set_service_endpoint_ptr_att(get_endpoint(pDevice->hrefEndpoint), att, value, reset);
        }
        return DPWS_OK;
}

int init_device_class(short hrefDeviceClass)
{
        struct device_class *deviceClass =
                (struct device_class *)getObject(&registryPool, hrefDeviceClass);
        
        if (!deviceClass)
                return DPWS_ERR_NO_HANDLE_FOUND;
        
        if (deviceClass->discovery_namespaces.nb == 0)
        {
                struct service_class *serviceClass = (struct service_class *)
                        getObject(&registryPool, deviceClass->hrefBuiltinServiceClass);
                struct prefixed_qname *pPQName, PQName;
                int i, j;

                // init default ns
                init_entries(&deviceClass->discovery_namespaces, discovery_ns, DISC_NS_SIZE - 1);       // NOTE: this table is also used by the client to process probe match
                init_entries(&deviceClass->services_namespaces, metadata_snd_namespaces, METADATA_NS_SIZE - 1);

                // set special types after to avoid prefix generation
                if (deviceClass->service_classes.nb > 0 || serviceClass->types.nb > 0)  // NOTE: no types if eventing fake device
                {
#ifdef USE_DEVPROF /* by HG.W (remove the ns of devprof) */
                        PQName.qname.ns = WDP_URI;
                        PQName.qname.lname = WDP_DEVICE_TYPE;
                        PQName.prefix = "wdp";
                        set_device_class_ptr_att(deviceClass, DPWS_PTR_PREFIXED_TYPE, &PQName, FALSE);
#endif /* by HG.W */
                        if (deviceClass->mdm_enabled) {
                                PQName.qname.ns = MDM_URI;
                                PQName.qname.lname = MDM_PORT_TYPE;
                                PQName.prefix = "mdm";
                                set_device_class_ptr_att(deviceClass, DPWS_PTR_PREFIXED_TYPE, &PQName, FALSE);
                                set_device_class_ptr_att(deviceClass, DPWS_PTR_HANDLING_FUNCTION, &mdm_serve_request, FALSE);
                        }
                }

                // set discovery ns table
                for(i = 0; i < serviceClass->types.nb; i++)
                {
                        pPQName = ((struct prefixed_qname *)serviceClass->types.tab) + i;
                        add_ns_entry(pPQName->prefix, pPQName->qname.ns, NULL, &deviceClass->discovery_namespaces);
                }
                        
                // set Metadata Exchange ns table
                for (i = 0; i < deviceClass->service_classes.nb; i++)
                {
                        serviceClass = (struct service_class *)getObject(&registryPool,
                                                        *(short *)get_entry(&deviceClass->service_classes, i));
                        for(j = 0; j < serviceClass->types.nb; j++)
                        {
                                pPQName = ((struct prefixed_qname *)serviceClass->types.tab) + j;
                                add_ns_entry(pPQName->prefix, pPQName->qname.ns, NULL, &deviceClass->services_namespaces);
                        }
                }
        }
        
        return DPWS_OK;
}

void checkout_device_classes(struct device *device)
{
        struct service_endpoint *wEndpoint;
        
        // checkout the device class first
        checkoutHandle(&registryPool, device->hrefClass);
        for (
                wEndpoint = device->hosted_services;
                wEndpoint;
                wEndpoint = wEndpoint->next
                )
                checkoutHandle(&registryPool, wEndpoint->hrefServiceClass);
}

void release_device_classes(struct device *device)
{
        struct service_endpoint *wEndpoint;
        
        // checkout the device class first
        releaseHandle(&registryPool, device->hrefClass);
        for (
                wEndpoint = device->hosted_services;
                wEndpoint;
                wEndpoint = wEndpoint->next
                )
                releaseHandle(&registryPool, wEndpoint->hrefServiceClass);
}

int get_toolkit_mode()
{
        if (registry)   // device or eventing endpoint
        {
                if (registry->next || get_device_mode(registry) == DPWS_DEVICE_USE)     // if two cannot be eventing only (one hidden device)
                        return DPWS_DEVICE_USE;
                else
                        return DPWS_EVENTING_USE;
        }
        else    // client or discovery monitor
        {
                if (tk_config.bye_hook || tk_config.hello_hook)
                        return DPWS_DISCOVERY_USE;
                else
                        return DPWS_CLIENT_USE;
        }
}

int get_device_mode(struct device* pDevice)
{
        /* Check if it is the hidden device */
        return pDevice->href == hidden_device_href ? DPWS_EVENTING_USE : DPWS_DEVICE_USE;
}

#ifdef USE_ONVIF /* Added by HG.W */
static int __generate_wakeup_socks_full_path(void)
{
    int path_len,  max_local_name_len, full_path_len;
    int ret = DPWS_OK;
    struct sockaddr_un tmp_sockaddr;
    
    path_len = strlen(tk_config.wakeup_socks_dir);
#ifdef WITH_IPV6
    max_local_name_len = MAX(strlen(WAKEUP_SOCK_FNAME_CLI),
                             MAX(strlen(WAKEUP_SOCK_FNAME_SVR_V4),
                                 strlen(WAKEUP_SOCK_FNAME_SVR_V6)));
#else
    max_local_name_len = MAX(strlen(WAKEUP_SOCK_FNAME_CLI),
                             strlen(WAKEUP_SOCK_FNAME_SVR_V4));
#endif
    full_path_len = path_len + max_local_name_len + 2;

    if(full_path_len > sizeof(tmp_sockaddr.sun_path))
        return DPWS_ERR_TOO_LONG_WAKEUP_SOCKS_PATHNAME;
       
    tk_config.wakeup_sock_cli = malloc(full_path_len);
    tk_config.wakeup_sock_svr_v4 = malloc(full_path_len);
#ifdef WITH_IPV6
    tk_config.wakeup_sock_svr_v6 = malloc(full_path_len);
#endif
    if(tk_config.wakeup_socks_dir[path_len-1] == '/'){
        sprintf(tk_config.wakeup_sock_cli, "%s%s",
                tk_config.wakeup_socks_dir, WAKEUP_SOCK_FNAME_CLI);
        sprintf(tk_config.wakeup_sock_svr_v4, "%s%s",
                tk_config.wakeup_socks_dir, WAKEUP_SOCK_FNAME_SVR_V4);
#ifdef WITH_IPV6
        sprintf(tk_config.wakeup_sock_svr_v6, "%s%s",
                tk_config.wakeup_socks_dir, WAKEUP_SOCK_FNAME_SVR_V6);
#endif
    }
    else{
        sprintf(tk_config.wakeup_sock_cli, "%s/%s",
                tk_config.wakeup_socks_dir, WAKEUP_SOCK_FNAME_CLI);
        sprintf(tk_config.wakeup_sock_svr_v4, "%s/%s",
                tk_config.wakeup_socks_dir, WAKEUP_SOCK_FNAME_SVR_V4);
#ifdef WITH_IPV6
        sprintf(tk_config.wakeup_sock_svr_v6, "%s/%s",
                tk_config.wakeup_socks_dir, WAKEUP_SOCK_FNAME_SVR_V6);
#endif
    }

    tk_config.wakeup_socks_full_path_generated = TRUE;

    ONVIF_LOG1(ONVIF_LOG_INFO, "[INFO]: wakeup_sock_cli=[%s].\n",
               tk_config.wakeup_sock_cli);
    ONVIF_LOG1(ONVIF_LOG_INFO, "[INFO]: wakeup_sock_svr_v4=[%s].\n",
               tk_config.wakeup_sock_svr_v4);
#ifdef WITH_IPV6
    ONVIF_LOG1(ONVIF_LOG_INFO, "[INFO]: wakeup_sock_svr_v6=[%s].\n",
               tk_config.wakeup_sock_svr_v6);
#endif
    return ret;
}
#endif


#ifdef USE_ONVIF /* Added by HG.W */
#ifdef WITH_IPV6
int check_configuration_v6()
{
        int ret = DPWS_OK;
        struct listener * http_listener = listeners;
        char *str_ptr = NULL;
                
        if (tk_config.boot_seq == 0xFFFFFFFF && get_toolkit_mode() == DPWS_DEVICE_USE)
                return DPWS_ERR_MISSING_BOOT_SEQUENCE;
        if (http_listener->url_generated){
            ret = default_url_v6(http_listener);
            if(ret != DPWS_OK){
                ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Failed to build url for web service on IPv6.\n");
                return ret;
            }
        }

        if (http_listener->tcp_backlog == -1)
                http_listener->tcp_backlog = DPWS_INT_HTTP_BACKLOG_DEFAULT;

        if(!tk_config.wakeup_socks_dir)
            return DPWS_ERR_MISSING_WAKEUP_SOCKS_PATH;

        if(!tk_config.wakeup_socks_full_path_generated)
            ret = __generate_wakeup_socks_full_path();

        return ret;
}
#endif /* WITH_IPV6 */
int check_configuration()
{
        int ret = DPWS_OK;
        struct listener * http_listener = listeners;
        char *str_ptr = NULL;
                
        if (tk_config.boot_seq == 0xFFFFFFFF && get_toolkit_mode() == DPWS_DEVICE_USE)
                return DPWS_ERR_MISSING_BOOT_SEQUENCE;
        if (http_listener->url_generated){
            ret = default_url(http_listener);
            if(ret != DPWS_OK){
                ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Failed to build url for web service on IPv4.\n");
                return ret;
            }
        }
        if (http_listener->tcp_backlog == -1)
                http_listener->tcp_backlog = DPWS_INT_HTTP_BACKLOG_DEFAULT;

        if(!tk_config.wakeup_socks_dir)
            return DPWS_ERR_MISSING_WAKEUP_SOCKS_PATH;

        if(!tk_config.wakeup_socks_full_path_generated)
            ret = __generate_wakeup_socks_full_path();

        return ret;
}
#else /* !USE_ONVIF */
int check_configuration()
{
        int ret = DPWS_OK;
        struct listener * http_listener = listeners;
        char *str_ptr = NULL;
                
        if (tk_config.boot_seq == 0xFFFFFFFF && get_toolkit_mode() == DPWS_DEVICE_USE)
                return DPWS_ERR_MISSING_BOOT_SEQUENCE;
        if (http_listener->url_generated) // TODO: should the consistency problem be considered and this param removed ?
                ret = default_url(http_listener);       // do it after ip and port are set
        if (http_listener->tcp_backlog == -1)
                http_listener->tcp_backlog = DPWS_INT_HTTP_BACKLOG_DEFAULT;

        return ret;
}
#endif

int check_device(struct device *device)
{
        int ret = SOAP_OK;
        struct device_class *deviceClass;
        struct service_endpoint *service;
        
        deviceClass = (struct device_class *)getObject(&registryPool, device->hrefClass);
        
        switch (get_device_mode(device))
        {
                case DPWS_DEVICE_USE:
                
                        if (deviceClass->metadata_version == 0xFFFF)
                                return DPWS_ERR_MISSING_METADATA_VERSION;
#ifndef USE_ONVIF /* by HG.W (Do not check for the madatory params and subscriptions for devprof) */
                        if (!deviceClass->manufacturer)
                                return DPWS_ERR_MISSING_MANUFACTURER;
                        if (!deviceClass->model_name)
                                return DPWS_ERR_MISSING_MODEL_NAME;
                        if (!device->friendly_name)
                                return DPWS_ERR_MISSING_FRIENDLY_NAME;

                        for(service = device->hosted_services; service; service = service->next)
                        {
                                if (service->max_subsc_duration == -1)
                                        service->max_subsc_duration = xmlduration2duration(DPWS_STR_MAX_SUBSC_DURATION_DEFAULT);
                        }
#endif /* by HG.W */

                        // no break
                        
                case DPWS_EVENTING_USE:
                
                        if (listeners->port == -1)
                                listeners->port = DPWS_INT_HTTP_PORT_DEFAULT;
                        if (strlen(device->uuid) == 0)
                        {
                                unsigned32 uuid;
                                MAKE_32b(device->local_id, (unsigned16)listeners->port, uuid)
                                stable_uuid_create(device->uuid, uuid);
                        }
                        // create device physical address
                        bind_service(device->hrefEndpoint, create_service_port(dpws_strdup(device->uuid + UUID_SCHEME_SIZE)));
                        if (device->scopes.nb == 0)
                                *(char **)new_entry(NULL, &device->scopes) = dpws_strdup(WSD_DEFAULT_SCOPE);    // to be able to free it
                        break;
                        
                default:
                        break;
        }
        return ret;
}

int initRegistry()
{
        createHandlePool(
                                        &registryPool,
                                        REGISTRY_POOL_NB_BUCKETS,
                                        MAX_SERVER_OBJECTS,
                                        -1,     // enables "outside" protection
                                        (handle_release_cbk)releaseHandleHook
                                        );
        if ((registry_lock_handle = lock_create()) < 0)
                return registry_lock_handle;
        return DPWS_OK;
}

void uninitRegistry()
{
        if (registry_lock_handle >= 0)
        {
                deleteHandlePool(&registryPool);
                lock_destroy(registry_lock_handle);
                registry_lock_handle = -1;
        }
}

/*----------------------------------------------------------------------------*\
 *                          INTERNAL API: RUNTIME                             *
\*----------------------------------------------------------------------------*/

struct service_endpoint * find_endpoint(char * address)
{
        struct device * wDevice;

        // Test if the address is a physical http address or a logical UUID
        if (is_physical_address(address))
        {       // NOTE: dwps_config.url/service_context_path
                // NOTE: do not compare the addresses (could have different forms : IP, NS...)
                char prefix[16];
                
                strcpy(prefix, listeners->transport->scheme);
                strcat(prefix, "//");
                if (!strncmp(address, prefix, 7))
                {
                        char * service_context_path = NULL;
                        
                        service_context_path = address + 7;
                        service_context_path = strchr(service_context_path, '/');       // Skip hostname/IP
                        
                        // default device handling
                        if (!service_context_path || strlen(service_context_path) == 1) {
                                struct device * device = NULL;
                                if (tk_config.default_device_handle != -1) 
                                        device = (struct device *)getObject(&registryPool, tk_config.default_device_handle);
                                return device && device->device_up ? (struct service_endpoint *)getObject(&registryPool, device->hrefEndpoint) : NULL;
                        }
                        else {
                                dyn_array servPorts = listeners->service_ports;
                                int i = 0, j = servPorts.nb - 1, k = j, comp;
                                short href;
                                struct service_port * serv_port;
                                
                                service_context_path++;
                                
                                // find the corresponding service port using dichotomy
                                // look first the upper bound because the integral division may prevent convergence on it
                                DPWSLOG1("\n-> Entering dichotomy algo, searching: %s\n", service_context_path);
#ifdef DPWS_DEBUG
                                {
                                        int n;
                                        DPWSLOG("\t+ Invocation table contents\n");
                                        for (n = 0; n < servPorts.nb; n++) {
                                                DPWSLOG2("\t[%d, %s]\n", n, ((struct service_port *)getObject(&registryPool, *(short *)get_entry(&servPorts, n)))->context_path);
                                        }
                                }
#endif
                                do {
                                        href = *(short *)get_entry(&servPorts, k);
                                        serv_port = (struct service_port *)getObject(&registryPool, href);
                                        DPWSLOG2("\tEvaluating %s at index %d\n", serv_port->context_path, k);
                                        comp = strcmp(service_context_path, serv_port->context_path);
                                        if (!comp) {
                                                DPWSLOG("<- Endpoint found.\n");
                                                return serv_port->service;
                                        }
                                        if (comp < 0)
                                                j = k;
                                        else 
                                                i = k;
                                        k = (j + i) / 2;
                                }
                                while (i < j);

                                DPWSLOG("<- #### Endpoint not found !!!!\n");
                                return NULL;    // no match
                        }
                }
                // else ignore
        }
        else if (!strncmp(address, "urn:uuid:", UUID_SCHEME_SIZE)) {
                wDevice = find_device(address + UUID_SCHEME_SIZE);
                return wDevice ? (struct service_endpoint *)getObject(&registryPool, wDevice->hrefEndpoint) : NULL;
        }
        // NOTE: not protected since only device can conflict with the device_up field (other fields are protected by this field)
        return NULL;
}

unsigned long getDeviceMetadataVersion(struct device * device)
{
        unsigned32 metaVersion;

/* Modified by HG.W */
#ifdef USE_ONVIF
        return ((struct device_class *)getObject(&registryPool, device->hrefClass))->metadata_version;
#else
        MAKE_32b(((struct device_class *)getObject(&registryPool,
                        device->hrefClass))->metadata_version, device->metadata_version, metaVersion)
#endif
        return  metaVersion;
}

struct service_endpoint * get_endpoint(short href)
{
        struct service_endpoint * endpoint = NULL;
        if (href >= 0) {
                handle_s * h = getHandle(&registryPool, href);
                switch (h->type)
                {
                        case SERVICE_ENDPOINT_TYPE:
                                endpoint = (struct service_endpoint *)h->pObject;
                                break;
                        case SERVICE_PORT_TYPE: // NOTE: added for compatibility reasons in notify. Could remove in v3
                                endpoint = ((struct service_port *)h->pObject)->service;
                                break;
                }
        }
        return endpoint;
}

struct event_subscription * find_subscription(struct service_endpoint * endpoint, char * id)
{
        struct event_subscription * w_sub = NULL, *prev = NULL;
        
        for (w_sub = endpoint->subscriptions; w_sub; w_sub = w_sub->next) {
                 if (!strcmp(id, w_sub->id)) {
                        if ((unsigned32)w_sub->expiration < system_time_s()) {
                                if (prev) 
                                        prev->next = w_sub->next;
                                else 
                                        endpoint->subscriptions = w_sub->next;
                                // check that subscription must not be discarded (expired, lazy update)
                                free_subscription(w_sub);
                                return NULL;
                        }
                        else
                                return w_sub;
                 }
                 else
                         prev = w_sub;
        }
        return NULL;
}

void remove_subscription(struct service_endpoint * endpoint, char * id)
{
        struct event_subscription * w_sub = NULL, *r_sub = NULL;
        
        if (endpoint->subscriptions) {
                if (!strcmp(id, endpoint->subscriptions->id)) {
                        r_sub = endpoint->subscriptions;
                        endpoint->subscriptions = endpoint->subscriptions->next;
                }
                else {
                        for (w_sub = endpoint->subscriptions; w_sub->next; w_sub = w_sub->next) {
                                if (!strcmp(id, w_sub->next->id)) {
                                        r_sub = w_sub->next;
                                        w_sub->next = w_sub->next->next;
                                        break;
                                }
                        }
                }
        }
        if (r_sub)
                free_subscription(r_sub);
}

dyn_array * get_device_types(struct device * device)
{
        return get_service_types(get_endpoint(device->hrefEndpoint));
}

dyn_array * get_service_types(struct service_endpoint * service)
{
        return &((struct service_class *)getObject(&registryPool, service->hrefServiceClass))->types;
}

dyn_array * get_string_device_types(struct device * device)
{
        return get_string_service_types(get_endpoint(device->hrefEndpoint));
}

dyn_array * get_string_service_types(struct service_endpoint * service)
{
        return &((struct service_class *)getObject(&registryPool, service->hrefServiceClass))->all_types;
}

int count_tns_wsdl(struct service_endpoint *service)
{
        struct service_class * servClass = (struct service_class *)getObject(&registryPool, service->hrefServiceClass);
        return servClass->wsdl_files.nb;        // TODO: Consider doubles
}

void free_configuration()
{
        if (registry_lock_handle >= 0)
        {
                lock(registry_lock_handle);
                
                tk_config.boot_seq = 0xFFFFFFFF;
                tk_config.hello_hook = NULL;
                tk_config.bye_hook = NULL;
                tk_config.server_up = FALSE;
                tk_config.tcp_listen = TRUE;
                free(tk_config.network_adapter.mac_addr);
                tk_config.network_adapter.mac_addr = NULL;
                if(tk_config.network_adapter.ip_addr){
                    free(tk_config.network_adapter.ip_addr);
                    tk_config.network_adapter.ip_addr = NULL;
                }

#ifdef USE_ONVIF /* Added by HG.W */
                free(tk_config.wakeup_socks_dir);
                tk_config.wakeup_socks_dir = NULL;
                if(tk_config.wakeup_sock_cli){
                    unlink(tk_config.wakeup_sock_cli);
                    free(tk_config.wakeup_sock_cli);
                    tk_config.wakeup_sock_cli = NULL;
                }
                if(tk_config.wakeup_sock_svr_v4){
                    unlink(tk_config.wakeup_sock_svr_v4);
                    free(tk_config.wakeup_sock_svr_v4);
                    tk_config.wakeup_sock_svr_v4 = NULL;
                }
#ifdef WITH_IPV6
                tk_config.server_up_v6 = FALSE;
                if(tk_config.wakeup_sock_svr_v6){
                    unlink(tk_config.wakeup_sock_svr_v6);
                    free(tk_config.wakeup_sock_svr_v6);
                    tk_config.wakeup_sock_svr_v6 = NULL;
                }
                free(tk_config.network_adapter.ip_addr_v6_g);
                tk_config.network_adapter.ip_addr_v6_g = NULL;
                free(tk_config.network_adapter.ip_addr_v6_l);
                tk_config.network_adapter.ip_addr_v6_l = NULL;
                free(tk_config.network_adapter.ip_addr_v6_l_with_if);
                tk_config.network_adapter.ip_addr_v6_l_with_if = NULL;
                free(listeners->url_v6_g);
                listeners->url_v6_g = NULL;
                free(listeners->url_v6_l);
                listeners->url_v6_l = NULL;
#endif /* WITH_IPV6 */
                free(listeners->external_path);
                listeners->external_path = NULL;
#endif /* USE_ONVIF */
                free(listeners->url);
                listeners->url = NULL;
                listeners->port = -1;
                listeners->tcp_backlog = -1;
                tk_config.max_subsc_duration = -1;
                free(tk_config.preferred_lang);
                tk_config.preferred_lang = NULL;
                tk_config.mdm_enabled = TRUE;
                
                // delete registry objects
                deleteHandles(&registryPool, ANY_TYPE);
        
                unlock(registry_lock_handle);
        }
}

void free_subscription(struct event_subscription * wes)
{
        dpws_endpoint_ref_free(wes->notify_to);
//      free(wes->notify_to);
        dpws_endpoint_ref_free(wes->end_to);
//      free(wes->end_to);
        free_buffer_dyn_array(&wes->actions);
        free(wes);
}

char* get_transport_address(struct dpws* dpws, short hrefServicePort)
{
        struct service_port * port;
        
        port = (struct service_port *)getObject(&registryPool, hrefServicePort);
        if (!port)
                return NULL;

        return build_transport_address(dpws, port);
}

int browse_service_ports(short href, BOOLEAN (*hook)(short *, void *), void * param)
{
        dyn_array * service_port_hrefs = find_service_endpoint_ports(href);
        if (!service_port_hrefs)
                return -1;
        return browse_dyn_array(service_port_hrefs, (browse_entry_cbk)hook, param);
}

struct wsa_endpoint_ref * dpws_build_local_endpoint_ref(struct wsa_endpoint_ref * dest, short service_port_href)
{
        struct service_port * service_port = (struct service_port *)getTypedObject(&registryPool, service_port_href, SERVICE_PORT_TYPE);
        if (!service_port)
                return NULL;
        dpws_default_wsa_endpoint_ref(dest);
        dest->address = build_transport_address(NULL, service_port);
        return dest;
}


/*----------------------------------------------------------------------------*\
 *                     USER API: REGISTRY CONFIGURATION                       *
\*----------------------------------------------------------------------------*/

short dpws_create_service_port()
{
        short href = -1;

        lock(registry_lock_handle);
        href = create_random_service_port();
        unlock(registry_lock_handle);
        
        return href;
}

int dpws_bind_service(short hrefService, short hrefServicePort)
{
        int ret = DPWS_OK;
        
        lock(registry_lock_handle);
        ret = bind_service(hrefService, hrefServicePort);
        unlock(registry_lock_handle);

        return ret;
}

int dpws_release_handle(short handleRef)
{
        int ret = DPWS_OK;
        lock(registry_lock_handle);
        ret = releaseHandle(&registryPool, handleRef);
        unlock(registry_lock_handle);
        return ret;
}

short dpws_create_device_class()
{
        short href = -1;
        struct device_class *class = malloc(sizeof(struct device_class));
        if (!class)
                return DPWS_ERR_OUT_OF_MEMORY;
                
        // init device class
        memset(class, 0, sizeof(struct device_class));
        init_dyn_array(&class->service_classes, sizeof(short), DISCOVERY_ALLOC_INCREMENT);
        init_dyn_array(&class->discovery_namespaces, sizeof(struct Namespace), DISCOVERY_ALLOC_INCREMENT);
        init_dyn_array(&class->services_namespaces, sizeof(struct Namespace), DISCOVERY_ALLOC_INCREMENT);
        class->mdm_enabled = tk_config.mdm_enabled;
        class->hrefBuiltinServiceClass = dpws_create_service_class();
        DPWS_SET_STR_ATT(class->hrefBuiltinServiceClass, DPWS_STR_ID, NULL); // Added to remove default id which breaks everything!
        class->metadata_version = 0xFFFF;
        class->id = malloc(6);  // max string size for short

        lock(registry_lock_handle);
        href = createHandle(&registryPool, DEVICE_CLASS_TYPE, class);
        if (href < 0)
                free(class);
        else
                sprintf(class->id, "%d", href);
        unlock(registry_lock_handle);
        return href;
}

short dpws_create_service_class()
{
        short href = -1;
        struct service_class *sClass = malloc(sizeof(struct service_class));

        if (!sClass)
                return DPWS_ERR_OUT_OF_MEMORY;

        sClass->fault_func = NULL;
        init_dyn_array(&sClass->types, sizeof(struct prefixed_qname), DISCOVERY_ALLOC_INCREMENT);
        init_dyn_array(&sClass->wsdl_files, sizeof(struct wsdl_info), DISCOVERY_ALLOC_INCREMENT);
        init_dyn_array(&sClass->funcs, sizeof(int (*)()), DISCOVERY_ALLOC_INCREMENT);
        init_dyn_array(&sClass->all_types, sizeof(char *), DISCOVERY_ALLOC_INCREMENT);
        sClass->id = malloc(6); // max string size for short
        
        lock(registry_lock_handle);
        href = createHandle(&registryPool, SERVICE_CLASS_TYPE, sClass);
        if (href < 0)
                free(sClass);
        else
                sprintf(sClass->id, "%d", href);
        unlock(registry_lock_handle);
        return href;
}

int dpws_register_service_class(short deviceClassHandleRef, short hrefServiceClass)
{
        struct service_class * servClass;
        struct device_class * devClass;
        int ret = DPWS_OK;
        short *phref;
        
        lock(registry_lock_handle);
        servClass = (struct service_class *)checkoutObject(&registryPool, hrefServiceClass);
        
        if (!servClass) {
                ret = DPWS_ERR_NO_HANDLE_FOUND;
                goto error;
        }
                
        // check that service class has types and handlers
        if (servClass->types.nb == 0 || servClass->funcs.nb == 0) {
                ret = DPWS_ERR_UNCOMPLETE_SERVICE;
                goto error;
        }
        
        devClass = (struct device_class *)getObject(&registryPool, deviceClassHandleRef);
        if (!devClass) {
                ret = DPWS_ERR_NO_HANDLE_FOUND;
                goto error;
        }
                
        // check for doubles
        if (isServiceClassRegistered(devClass, hrefServiceClass)) {
                ret = DPWS_ERR_CONFLICTING_OPERATION;
                goto error;
        }
        
        // create device class/service class association
        phref = (short *)new_entry(NULL, &devClass->service_classes);
        *phref = hrefServiceClass;

        goto exit;
        
error:
        releaseHandle(&registryPool, hrefServiceClass);
exit:
        unlock(registry_lock_handle);
        return ret;
}

short dpws_create_custom_device(unsigned short localID, short hrefClass)
{
        short hRef = -1;
        if (localID == RESERVED_DEVICE_LOCAL_ID)
                return DPWS_ERR_INVALID_PARAMETER;
        lock(registry_lock_handle);
        hRef = create_custom_device(localID, hrefClass);
        unlock(registry_lock_handle);
        return hRef;
}

short dpws_create_hosted_service(short deviceHandleRef, short hrefClass)
{
        short href = -1;
        handle_s *handle;
        struct device * device;
        
        lock(registry_lock_handle);
        handle = getHandle(&registryPool, hrefClass);
        if (!handle)
                goto exit;
        
        device = getObject(&registryPool, deviceHandleRef);
        if (!device)
                goto exit;
                
        handle = getHandle(&registryPool, device->hrefClass);
        
//      NOTE: this control had no real justification : if (isServiceClassRegistered((struct device_class *)handle->pObject, hrefClass))
        href = create_hosted_service(device, hrefClass);
exit:
        unlock(registry_lock_handle);
        return href;
}

short dpws_create_endpoint()
{
        short ret = -1, href = -1;

        // create the hidden device if necessary
        if (hidden_device_href < 0)
        {
                struct device* hiddenDev = NULL;
                href = dpws_create_device_class();              
                lock(registry_lock_handle);
                hidden_device_href = create_custom_device(RESERVED_DEVICE_LOCAL_ID, href);
                if (hidden_device_href < 0)
                        return hidden_device_href;
                hiddenDev = (struct device *)getObject(&registryPool, hidden_device_href);
                hiddenDev->device_up = TRUE; // must be able to receive soap messages
                unlock(registry_lock_handle);
        }
        href = dpws_create_service_class();             // NOTE: a little heavy but needed to carry handlers
        ret = dpws_create_hosted_service(hidden_device_href, href);
        href = dpws_create_service_port();
        dpws_bind_service(ret, href);
        return ret;
}

int dpws_delete_hosted_service(short serviceHandleRef)
{
        struct device * device = NULL;
        struct service_endpoint * service = NULL, *wServ;
        int ret = DPWS_OK;
        
        lock(registry_lock_handle);
        
        service = (struct service_endpoint *)getObject(&registryPool, serviceHandleRef);
        if (!service) {
                ret = DPWS_ERR_NO_HANDLE_FOUND;
                goto exit;
        }
        device = service->device;
        if (device->hosted_services == service)
                device->hosted_services = service->next;
        else {
                for (wServ = device->hosted_services; wServ->next != service; wServ = wServ->next);
                wServ->next = wServ->next->next; // should find
        }
        device->service_count--;
        ret = releaseHandle(&registryPool, serviceHandleRef);

exit:   
        unlock(registry_lock_handle);
        return ret;
}

short dpws_get_device_class_handle(char * id)
{
        handle_s * h;
        short ret = -1;
        lock(registry_lock_handle);
        h = handlePoolIterate(&registryPool, DEVICE_CLASS_TYPE, (handle_cbk)device_class_equals, id);
        if (h)
                ret = h->handleRef;
        unlock(registry_lock_handle);
        return ret;
}

short dpws_get_service_class_handle(char * id)
{
        handle_s * h;
        short ret = -1;
        lock(registry_lock_handle);
        h = handlePoolIterate(&registryPool, SERVICE_CLASS_TYPE, (handle_cbk)service_class_equals, id);
        if (h)
                ret = h->handleRef;
        unlock(registry_lock_handle);
        return ret;
}

short dpws_get_device_handle(char * uuid)
{
        handle_s * h;
        short ret = -1;
        lock(registry_lock_handle);
        h = handlePoolIterate(&registryPool, DEVICE_INSTANCE_TYPE, (handle_cbk)device_equals, uuid);
        if (h)
                ret = h->handleRef;
        unlock(registry_lock_handle);
        return ret;
}

short dpws_get_service_handle(short deviceHandleRef, char * service_id)
{
        short href = -1;
        struct device *device;
        struct service_endpoint *wEndpoint;
                
        lock(registry_lock_handle);
        device = getObject(&registryPool, deviceHandleRef);
        if (!device)
                goto exit;
                
        for (wEndpoint = device->hosted_services; wEndpoint; wEndpoint = wEndpoint->next){
                if (!strcmp(wEndpoint->service_id, service_id)) {
                        href = wEndpoint->href;
                        break;
                }
        }

exit:
        unlock(registry_lock_handle);
        return href;
}

short dpws_get_service_handle_by_class(short deviceHandleRef, short serviceClassHandleRef)
{
        short href = -1;
        struct device *device;
        struct service_endpoint *wEndpoint;
                
        lock(registry_lock_handle);
        device = getObject(&registryPool, deviceHandleRef);
        if (!device)
                goto exit;
                
        for (wEndpoint = device->hosted_services; wEndpoint; wEndpoint = wEndpoint->next){
                if (wEndpoint->hrefServiceClass == serviceClassHandleRef) {
                        href = wEndpoint->href;
                        break;
                }
        }

exit:
        unlock(registry_lock_handle);
        return href;
}

short dpws_create_device(unsigned short localID, short hrefClass)
{
        short hRef;
        
        lock(registry_lock_handle);

        hRef = create_custom_device(localID, hrefClass);
        if (hRef >= 0)
        {
                struct device * device;
                struct device_class * deviceClass;
                short * phrefServiceClass, hrefServiceEndpoint, hrefServicePort;
                int i;
                
                device = (struct device *)getObject(&registryPool, hRef);
                deviceClass = (struct device_class *)getObject(&registryPool, hrefClass);
                
                // loop over service classes
                for(i = 0; i < deviceClass->service_classes.nb; i++) {
                        phrefServiceClass = ((short *)deviceClass->service_classes.tab) + i;
                        hrefServiceEndpoint = create_hosted_service(device, *phrefServiceClass);
                        hrefServicePort = create_random_service_port();
                        bind_service(hrefServiceEndpoint, hrefServicePort);
                }
        }
        unlock(registry_lock_handle);

        return hRef;
}

int dpws_set_ptr_att(short handleRef, int att, const void * value)
{
        int ret = DPWS_OK;
        struct device *device;
        struct service_endpoint *service;
        struct service_port *servicePort;
        handle_s *handle;
        
        lock(registry_lock_handle);
        if (handleRef >= 0)
        {
                handle = getHandle(&registryPool, handleRef);
                if (!handle)
                        ret = DPWS_ERR_NO_HANDLE_FOUND;
                else
                        switch (handle->type)
                        {
                                case DEVICE_INSTANCE_TYPE:
                                        device = (struct device *)handle->pObject;
                                        if (tk_config.server_up && device->device_up)   // not a shared object // TODO: should disappear with hot configuration
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_ptr_att(device, att, value, TRUE);
                                        break;
                                case SERVICE_ENDPOINT_TYPE:
                                        service = (struct service_endpoint *)handle->pObject;
                                        if (tk_config.server_up && service->device->device_up && service->device->href != hidden_device_href)   // not a shared object (except for hidden device to allow plug hot plug of event handler)
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_endpoint_ptr_att(service, att, value, TRUE);
                                        break;
                                case SERVICE_PORT_TYPE:
                                        servicePort = (struct service_port *)handle->pObject;
                                        if (tk_config.server_up)        // TODO: be more permissive ?
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_port_ptr_att(servicePort, att, value, TRUE);
                                        break;
                                case DEVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_class_ptr_att((struct device_class *)handle->pObject, att, value, TRUE);
                                        break;
                                case SERVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_class_ptr_att((struct service_class *)handle->pObject, att, value, TRUE);
                                        break;
                                default:
                                        ret = DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
                        }
        }
        else
        {
                if (tk_config.server_up)        // used by more than the user
                        ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                else
                        ret = set_config_ptr_att(att, value, TRUE);
        }
        unlock(registry_lock_handle);
        
        return ret;
}

int dpws_set_int_att(short handleRef, int att, unsigned long value)
{
        int ret = DPWS_OK;
        struct device *device;
        struct service_endpoint *service;
        handle_s *handle;
        
        lock(registry_lock_handle);
        if (handleRef >= 0)
        {
                handle = getHandle(&registryPool, handleRef);
                if (!handle)
                        ret = DPWS_ERR_NO_HANDLE_FOUND;
                else
                        switch (handle->type)
                        {
                                case DEVICE_INSTANCE_TYPE:
                                        device = (struct device *)handle->pObject;
                                        if (tk_config.server_up && device->device_up)   // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_ptr_att(device, att, &value, TRUE);
                                        break;
                                case SERVICE_ENDPOINT_TYPE:
                                        service = (struct service_endpoint *)handle->pObject;
                                        if (tk_config.server_up && service->device->device_up)  // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_endpoint_ptr_att(service, att, &value, TRUE);
                                        break;
                                case DEVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_class_ptr_att((struct device_class *)handle->pObject, att, &value, TRUE);
                                        break;
                                case SERVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_class_ptr_att((struct service_class *)handle->pObject, att, &value, TRUE);
                                        break;
                                default:
                                        ret = DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
                        }
        }
        else
        {
                if (tk_config.server_up)        // used by more than the user
                        ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                else
                        ret = set_config_ptr_att(att, &value, TRUE);
        }
        unlock(registry_lock_handle);

        return ret;
}

int dpws_add_ptr_att(short handleRef, int att, const void * value)
{
        int ret = DPWS_OK;
        struct device *device;
        struct service_endpoint *service;
        handle_s *handle;
        
        lock(registry_lock_handle);
        if (handleRef >= 0)
        {
                handle = getHandle(&registryPool, handleRef);
                if (!handle)
                        ret = DPWS_ERR_NO_HANDLE_FOUND;
                else
                        switch (handle->type)
                        {
                                case DEVICE_INSTANCE_TYPE:
                                        device = (struct device *)handle->pObject;
                                        if (tk_config.server_up && device->device_up)   // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_ptr_att(device, att, value, FALSE);
                                        break;
                                case SERVICE_ENDPOINT_TYPE:
                                        service = (struct service_endpoint *)handle->pObject;
                                        if (tk_config.server_up && service->device->device_up && service->device->href != hidden_device_href)   // not a shared object (except for hidden device to allow plug hot plug of event handler)
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_endpoint_ptr_att(service, att, value, FALSE);
                                        break;
                                case DEVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_class_ptr_att((struct device_class *)handle->pObject, att, value, FALSE);
                                        break;
                                case SERVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_class_ptr_att((struct service_class *)handle->pObject, att, value, FALSE);
                                        break;
                                default:
                                        ret = DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
                        }
        }
        else
        {
                if (tk_config.server_up)        // used by more than the user
                        ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                else
                        ret = set_config_ptr_att(att, value, FALSE);
        }
        unlock(registry_lock_handle);
        return ret;
}

int dpws_add_int_att(short handleRef, int att, unsigned long value)
{
        int ret = DPWS_OK;
        struct device *device;
        struct service_endpoint *service;
        handle_s *handle;
        
        lock(registry_lock_handle);
        if (handleRef >= 0)
        {
                handle = getHandle(&registryPool, handleRef);
                if (!handle)
                        ret = DPWS_ERR_NO_HANDLE_FOUND;
                else
                        switch (handle->type)
                        {
                                case DEVICE_INSTANCE_TYPE:
                                        device = (struct device *)handle->pObject;
                                        if (tk_config.server_up && device->device_up)   // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_ptr_att(device, att, &value, FALSE);
                                        break;
                                case SERVICE_ENDPOINT_TYPE:
                                        service = (struct service_endpoint *)handle->pObject;
                                        if (tk_config.server_up && service->device->device_up)  // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_endpoint_ptr_att(service, att, &value, FALSE);
                                        break;
                                case DEVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_device_class_ptr_att((struct device_class *)handle->pObject, att, &value, FALSE);
                                        break;
                                case SERVICE_CLASS_TYPE:
                                        if (tk_config.server_up && handle->useCount > 1)        // used by more than the user
                                                ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                                        else
                                                ret = set_service_class_ptr_att((struct service_class *)handle->pObject, att, &value, FALSE);
                                        break;
                                default:
                                        ret = DPWS_ERR_NO_SUCH_ATT_ON_OBJECT;
                        }
        }
        else
        {
                if (tk_config.server_up)        // used by more than the user
                        ret = DPWS_ERR_IMMUTABLE_USED_OBJECT;
                else
                        ret = set_config_ptr_att(att, &value, FALSE);
        }
        unlock(registry_lock_handle);
        return ret;
}

void * dpws_get_ptr_att_item(short handleRef, int att, int index)
{
        void * ret;
        handle_s *handle;
        struct device * device;
        
        lock(registry_lock_handle);
        if (handleRef >= 0)
        {
                handle = getHandle(&registryPool, handleRef);
                if (!handle)
                        ret = NULL;
                else
                        switch (handle->type)
                        {
                                case DEVICE_INSTANCE_TYPE:
                                        device = (struct device *)handle->pObject;
                                        ret = get_device_ptr_att((struct device *)handle->pObject, att, index);
                                        break;
                                case SERVICE_ENDPOINT_TYPE:
                                        ret = get_service_endpoint_ptr_att((struct service_endpoint *)handle->pObject, att, index);
                                        break;
                                case SERVICE_PORT_TYPE:
                                        ret = get_service_port_ptr_att((struct service_port *)handle->pObject, att, index);
                                        break;
                                case DEVICE_CLASS_TYPE:
                                        ret = get_device_class_ptr_att((struct device_class *)handle->pObject, att, index);
                                        break;
                                case SERVICE_CLASS_TYPE:
                                        ret = get_service_class_ptr_att((struct service_class *)handle->pObject, att, index);
                                        break;
                                default:
                                        ret = NULL;
                        }
        }
        else
                ret = get_config_ptr_att(att, index);
        unlock(registry_lock_handle);
        return ret;
}

void * dpws_get_ptr_att(short handleRef, int att)
{
        return dpws_get_ptr_att_item(handleRef, att, 0);
}

unsigned long dpws_get_int_att(short handleRef, int att)
{
        unsigned long ret;
        handle_s *handle;
        
        lock(registry_lock_handle);
        if (handleRef >= 0)
        {
                handle = getHandle(&registryPool, handleRef);
                if (!handle)
                        ret = 0xFFFFFFFF;
                else
                        switch (handle->type)
                        {
                                case DEVICE_INSTANCE_TYPE:
                                        ret = *(unsigned long *)get_device_ptr_att((struct device *)handle->pObject, att, -1);
                                        break;
                                case SERVICE_ENDPOINT_TYPE:
                                        ret = *(unsigned long *)get_service_endpoint_ptr_att((struct service_endpoint *)handle->pObject, att, -1);
                                        break;
                                case DEVICE_CLASS_TYPE:
                                        ret = *(unsigned long *)get_device_class_ptr_att((struct device_class *)handle->pObject, att, -1);
                                        break;
                                case SERVICE_CLASS_TYPE:
                                        ret = *(unsigned long *)get_service_class_ptr_att((struct service_class *)handle->pObject, att, -1);
                                        break;
                                default:
                                        ret = 0xFFFFFFFF;
                        }
        }
        else
                ret = *(unsigned long *)get_config_ptr_att(att, -1);

        unlock(registry_lock_handle);
        return ret ? ret : 0xFFFFFFFF;
}

struct wsa_endpoint_ref * dpws_get_local_endpoint_ref(short service_port_href)
{
        struct wsa_endpoint_ref *local = malloc(sizeof(struct wsa_endpoint_ref));
        return dpws_build_local_endpoint_ref(local, service_port_href);
}

struct wsa_endpoint_ref * dpws_create_device_endpoint_ref(char * uuid)
{
        struct wsa_endpoint_ref *dev_endpoint = NULL;
        if (uuid) {
                dev_endpoint = malloc(sizeof(struct wsa_endpoint_ref));
                dpws_default_wsa_endpoint_ref(dev_endpoint);
                dev_endpoint->address = dpws_strdup(uuid);
        }
        return dev_endpoint;
}

short dpws_get_default_service_port(short href)
{
        dyn_array * result = find_service_endpoint_ports(href);
        return result == NULL ? -1 : *(short *) get_entry(result, 0);   // TODO: better
}

char* dpws_get_transport_address(short hrefServPort)
{
        return get_transport_address(NULL, hrefServPort);
}

int dpws_get_service_handles(short deviceHandleRef, short* hrefs, int* len)
{
        struct device *device;
        struct service_endpoint *wEndpoint;
        int count = 0, res = DPWS_OK;
                
        lock(registry_lock_handle);
        device = getObject(&registryPool, deviceHandleRef);
        if (!device) {
                res = -1;
                goto exit;
        }
                
        for (wEndpoint = device->hosted_services; wEndpoint; wEndpoint = wEndpoint->next) {
                if (wEndpoint->service_id) {
                        if (count < *len) 
                                hrefs[count] = wEndpoint->href;
                        else {
                                res = 1;
                                goto exit;
                        }
                        count++;
                }
        }
        *len = count;
exit:
        unlock(registry_lock_handle);
        return res;
}

int dpws_get_service_port_handles(short endpoint_href, short* hrefs, int* len)
{
        int res = DPWS_OK;
        dyn_array * result;
        
        lock(registry_lock_handle);
        result = find_service_endpoint_ports(endpoint_href);
        if (!result) {
                res = -1;
                goto exit;
        }
        if (result->nb > *len){
                res = 1;
                goto exit;
        }
        else
                *len = result->nb;
        memcpy(hrefs, result->tab, *len);

exit:
        unlock(registry_lock_handle);
        return res;
}


void * dpws_get_device_user_data(struct dpws* dpws)
{
        struct service_endpoint * endpoint = get_endpoint(dpws->href_endpoint);
        return endpoint ? (void *) endpoint->device->user_data: NULL;
}

void * dpws_get_endpoint_user_data(struct dpws* dpws)
{
        struct service_endpoint * endpoint = get_endpoint(dpws->href_endpoint);
        return endpoint ? (void *) endpoint->user_data: NULL;
}
