/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.99.2.3 $
|                                              $Date: 2007/03/23 16:41:34 $
\*============================================================================*/
 
/*******************************************************************************
*                              SERVER DATA HEADER                              *
*******************************************************************************/
#ifndef CONFIG_H
# define CONFIG_H

#include "tools.h"
#include "_dpws.h"
/*----------------------------------------------------------------------------*\
 *                                  INTERNAL                                  *
\*----------------------------------------------------------------------------*/
#ifdef USE_ONVIF /* Added by HG.W */
#define WAKEUP_SOCK_FNAME_CLI       "wakesock_cli"
#define WAKEUP_SOCK_FNAME_SVR_V4    "wakesock_svr_v4"
#ifdef WITH_IPV6
#define WAKEUP_SOCK_FNAME_SVR_V6    "wakesock_svr_v6"
#endif
#endif

// NOTE: fields that have default value in another object, are set to null when the default is not set (=> dynamic resolution)
struct dpws_config {
	unsigned long	boot_seq;
	void (*hello_hook) (struct dpws *, short);
	void (*bye_hook) (struct dpws *, short);
	BOOLEAN server_up;
	BOOLEAN tcp_listen;

	/* network interface (will be moved when multi-interface is supported) */
	struct na_info	network_adapter;
	
	/* default values for devices configuration */
	time_t			max_subsc_duration;
	char *			preferred_lang;

	/* default values for devices class */
	BOOLEAN			mdm_enabled;
	/* default device for dispatching request to physical address */
	short			default_device_handle;
#ifdef USE_ONVIF /* Added by HG.W */
    BOOLEAN         wakeup_socks_full_path_generated;
    char *          wakeup_socks_dir;
    char *          wakeup_sock_cli;
    char *          wakeup_sock_svr_v4;
#ifdef WITH_IPV6
    BOOLEAN         server_up_v6;
    char *          wakeup_sock_svr_v6;
#endif
#endif
};

struct listener {
	struct transport * transport;
	char *			url;	// TODO: remove http prefix
	BOOLEAN			url_generated;	// TODO: generalize
	int				port;
	int				tcp_backlog;
#ifdef USE_ONVIF /* Added by HG.W */
    char *          external_path;
#ifdef WITH_IPV6
    char *          url_v6_g;
    char *          url_v6_l;
#endif
#endif

	dyn_array		service_ports;	// ordered array of service ports pointer (alphabetic order of context_path) (slow for insertion, rather quick for retrieval with dichotomy)
	struct listener * next;
};

struct transport {
	char *			protocol_name;
	char *			scheme;
};

struct event_subscription {
	char id[SCHEME_UUID_STRING_SIZE];	// NOTE: prefix stored to make serialization more efficient
	struct wsa_endpoint_ref * notify_to;
	struct wsa_endpoint_ref * end_to;
	dyn_array actions;
	time_t expiration;	// seconds since 00:00:00 GMT, January 1, 1970
	struct event_subscription * next;
};

struct service_endpoint {	// NOTE: not named hosted service because the endpoint can be a device or a hosted service
	short href;	// because returned by an API
	char * service_id;	// NULL for the device endpoint
	short hrefServiceClass;
	time_t max_subsc_duration;
	struct device * device;	// TODO: why a direct pointer whereas the handle indirection has been introduced (for example for hot-configuration)
	struct service_endpoint * next;	// TODO: replace with a dynamic array of handles in device ? what about performance during invocation ?
	struct service_endpoint * previous;
	dyn_array service_ports;	// store handles

	// runtime immutable fields
	const void * user_data;
	
	// runtime mutable fields
	BOOLEAN subsc_manager;	// read: eventing, write: eventing
	struct event_subscription * subscriptions;
};

struct service_port {
	char * context_path;
	struct listener * listener;
	struct service_endpoint * service;	// TODO: pointer for perf reasons, replace by a href in case of full hot configuration
};


struct service_class {	// TODO: move handler fields to other structure to be able to attach it also to the endpoint ?
	char * id;
	dyn_array types;		// struct prefixed_qname
	dyn_array wsdl_files;	// char *
	dyn_array funcs;		// int (*fserve)(struct dpws*);	// TODO: doc. 1 .gsoap par port type => 1 fonction de dispatch
	int (*fault_func)(struct dpws *);
	void (*event_end_hook) (struct event_end_info *);
	dyn_array all_types;	// types as qName strings

	// runtime immutable fields
	const void * user_data;
};

struct device_class {
	char * id;
	char * manufacturer;
	char * manufacturer_url;
	char * model_name;
	char * model_number;
	char * model_url;
	char * presentation_url;

	short hrefBuiltinServiceClass;
	dyn_array service_classes;	// device class href
	BOOLEAN mdm_enabled;

	// runtime immutable fields
	const void * user_data;

	// compiled data
	dyn_array discovery_namespaces; // namespace table for WS-Discovery messages (no hosted services ns, but the device types ones)
	dyn_array services_namespaces;	// namespace table for WS-MetadataExchange (hosted services ns, but not the device types one)
	// NOTE: there should be one ns table per service class if we were able to merge it when mex ask for metadata of a device

	// runtime mutable fields
	unsigned short metadata_version;	// read: discovery, write: API (during config)
};

struct device {
	short			hrefEndpoint;	// a device is also a service endpoint
	short			href;	// because returned by an API
	unsigned short	local_id;
	short			hrefClass;
	char			uuid[SCHEME_UUID_STRING_SIZE];	// NOTE: prefix stored to make serialization more efficient
	char *			firmware_version;
	char *			serial_number;
	struct service_endpoint *hosted_services;
	short			service_count;
	struct device 	*next;
	struct device 	*previous;
	
	// runtime immutable fields
	const void * user_data;
	
	// runtime mutable fields
	short			lock;
	unsigned short	metadata_version;	// read: discovery, write ???
	dyn_array 		scopes;	// read: discovery, write: API (during config), MDM
	char *			friendly_name;	// read: MEX, write: API (during config), MDM
	BOOLEAN			device_up;	// read: message routing (fin_device() function), write: API
	BOOLEAN			bye_scheduled;	// read: dpws_stop_server() waiting to exit, write: send_eventing_end
};

int initRegistry();
void uninitRegistry();
int get_device_mode(struct device* pDevice);
int get_toolkit_mode();
int check_configuration();
int init_device_class(short hrefDeviceClass);
void free_configuration();
struct service_endpoint * find_endpoint(char * address);
struct service_endpoint * get_endpoint(short href);
struct event_subscription * find_subscription(struct service_endpoint * endpoint, char * id);
void remove_subscription(struct service_endpoint * endpoint, char * id);
dyn_array * get_device_types(struct device * device);
dyn_array * get_service_types(struct service_endpoint * service);
dyn_array * get_string_device_types(struct device * device);
dyn_array * get_string_service_types(struct service_endpoint * service);
unsigned long getDeviceMetadataVersion(struct device * device);
int count_tns_wsdl(struct service_endpoint *endpoint);
void free_subscription(struct event_subscription * wes);
int set_device_ptr_att(struct device * pDevice, int att, const void * value, BOOLEAN reset);
void release_device_classes(struct device *device);
int check_device(struct device *device);
void checkout_device_classes(struct device *device);
char* get_transport_address(struct dpws* dpws, short hrefServicePort);
char* get_default_transport_address(struct dpws* dpws, short hrefEndpoint);
// NOTE: hook: Return TRUE if the iteration must be stopped can browse a device or a service endpoint
typedef BOOLEAN (* service_port_cbk)(short *, void *);
int browse_service_ports(short href, service_port_cbk hook, void * param);


#define REGISTRY_POOL_NB_BUCKETS	32

#define RESERVED_DEVICE_LOCAL_ID	0x7FFF

#define DEVICE_INSTANCE_TYPE		1
#define SERVICE_ENDPOINT_TYPE		2
#define DEVICE_CLASS_TYPE			3
#define SERVICE_CLASS_TYPE 			4
#define SERVICE_PORT_TYPE 			5

#define DPWS_CLIENT_USE		0
#define DPWS_DISCOVERY_USE	1	// discovery monitor
#define DPWS_EVENTING_USE	2	// full device
#define DPWS_DEVICE_USE		3	// full device

#endif
