/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.94.2.5 $
|                                              $Date: 2007/04/20 14:59:20 $
\*============================================================================*/
 
#include "discovery.h"
#include "tools.h"
#include "cache.h"
#include "config.h"

extern struct Namespace default_namespaces[];
extern struct dpws_config tk_config;
extern struct listener * listeners;
extern short cache_lock_handle;
extern handlePool_t registryPool;
extern struct device *registry;

// TODO: fix header files
extern void dpws_set_udp_mode(struct dpws * dpws);
extern void dpws_set_anonymous_wsa_endpoint_ref(struct wsa_endpoint_ref *a);
extern int dpws_udp_close(struct soap *soap);

handlePool_t discoveryHandlePool;
short discovery_lock_handle = -1;


extern struct Namespace discovery_ns[];

static int dpws_handle_directed_probe(struct dpws* dpws, struct wsd__ProbeType * probe, struct wsd__ProbeMatchesType * probeMatches);
static int dpws_handle_probe_match(struct dpws* dpws, struct wsd__ProbeMatchesType* probe_matches);

/*----------------------------------------------------------------------------*\
 *                           APP sequence management                          *
\*----------------------------------------------------------------------------*/
static unsigned short app_seq_message_nb = 0;

unsigned short new_app_seq_msg_nb() {
    lock(discovery_lock_handle);
    app_seq_message_nb++;
    unlock(discovery_lock_handle);
    return app_seq_message_nb;
}

/*----------------------------------------------------------------------------*\
 *                          SOAP over UDP undoubling                          *
\*----------------------------------------------------------------------------*/

static char uuid_cache[UUID_CACHE_SIZE][UUID_STRING_SIZE];
static int next_index = 0;

int is_new_message(char * uuid)
{
    int ret = 1, i;

#ifdef USE_ONVIF /* Added by HG.W */
    /** Bugfix-2009.03.24:
     *  check the value of uuid, and
     *  reject (do nothing with) the bad UUID 
     *  (include null string, bad UUID length, non-UUID).
     */

    char *tmp_uuid_ptr;

    /* null string */
    if(uuid == NULL){
        ONVIF_LOG(ONVIF_LOG_ERROR,
                  "[ERROR]: Unsupported MessageId.(null string)\n");
        return -1;
    }

    /* urn:uuid: */
    if(strncasecmp(uuid, "urn:uuid:", 9) == 0){ /* strlen("urn:uuid:") = 9 */
        tmp_uuid_ptr = uuid + 9;
    }
    /* urn: */
    else if(strncasecmp(uuid, "uuid:", 5) == 0){ /* strlen("uuid:") = 5 */
        tmp_uuid_ptr = uuid + 5;
    }
    /* other (unknown) */
    else{
        ONVIF_LOG(ONVIF_LOG_ERROR,
                  "[ERROR]: Unsupported MessageId.(not uuid scheme)\n");
        return -1;
    }

    if(strlen(tmp_uuid_ptr) != UUID_SIZE){ /* check if uuid size is 36 bytes */
        ONVIF_LOG(ONVIF_LOG_ERROR,
                  "[ERROR]: Unsupported MessageId.(bad uuid)\n");
        return -1;
    }
#endif /* Added by HG.W */

    lock(discovery_lock_handle);
    if (next_index == 0)
        i = UUID_CACHE_SIZE - 1;
    else
        i = next_index - 1;
    while (i != next_index)
    {
        if (!dpws_strcasecmp(uuid + UUID_SCHEME_SIZE, uuid_cache[i])) {
            ret = 0;
            break;
        }
        if (i == 0) 
            i = UUID_CACHE_SIZE - 1;
        else
            i--;
    }
    
    // store new uuid
    strcpy(uuid_cache[next_index], uuid + UUID_SCHEME_SIZE);
    next_index++;
    if (next_index == UUID_CACHE_SIZE)
        next_index = 0;
    unlock(discovery_lock_handle);
    
    return ret;
}

/*******************************************************************************
*                                   RUNTIME                                    *
*******************************************************************************/
static int free_discovery_msg(handle_s *handleInfo)
{
    free(handleInfo->pObject);
    return DPWS_OK;
}

int initDiscovery()
{
    if ((discovery_lock_handle = lock_create()) < 0)
        return discovery_lock_handle;
        
    createHandlePool(
                    &discoveryHandlePool,
                    DISCOVERY_HANDLE_POOL_SIZE,
                    DISCOVERY_MAX_SCHEDULED_MSG,
                    discovery_lock_handle,
                    (handle_release_cbk)free_discovery_msg
                    );
    return DPWS_OK;
}

void uninitDiscovery()
{
    if (discovery_lock_handle >= 0)
    {
        deleteHandlePool(&discoveryHandlePool);
        lock_destroy(discovery_lock_handle);
        discovery_lock_handle = -1;
    }
}

int schedule_discovery_msg(scheduler_t *pScheduler, short msgType, char * req_id,
    struct device *device, int (*handler) (struct dpws*, void *), short hrefLock)
{
    short send_nb = 0, href;
    long delay = 0;
    discovery_msg_t *discMsg = malloc(sizeof(discovery_msg_t)); // TODO: recycle
    handle_s * handle = NULL;
    
    schemed_uuid_create(discMsg->uuid);
    if (msgType & (DPWS_MSG_PROBE_MATCH|DPWS_MSG_RESOLVE_MATCH))
        strcpy(discMsg->req_uuid, req_id);
    discMsg->msg_nb = new_app_seq_msg_nb();
    discMsg->param = device;
    
    while ((delay = get_soap_over_udp_schedule(msgType, &send_nb, delay)) != 0xFFFFFFFF)
    {
        if (handle == NULL)
        {
            href = createHandle(&discoveryHandlePool, msgType, discMsg);
            if (href >= 0)
                handle = getHandle(&discoveryHandlePool, href);
            else
                return href;
        }
        else
            checkoutHandle(&discoveryHandlePool, handle->handleRef);
        if (hrefLock >= 0)
            lock(hrefLock);
        add_schedule(pScheduler, delay, (schedule_cbk)handler, handle);
        if (hrefLock >= 0)
            unlock(hrefLock);
    }
    return DPWS_OK;
}

/*******************************************************************************
*                                 MARSHALLING                                  *
*******************************************************************************/
#define WITH_NOGLOBAL
#ifdef SOAP_FMAC3
# undef SOAP_FMAC3
#endif
#define SOAP_FMAC3 static
#include "gen/discoveryC.c"
#undef WITH_NOGLOBAL

/*******************************************************************************
*                                    STUB                                      *
*******************************************************************************/
// NOTE: in order to keep a standard generation (setting is performed by the caller)
#define discovery_namespaces (dpws->namespaces)
#define discovery_snd_namespaces (dpws->namespaces)
#include "gen/discoveryClient.c"

/*******************************************************************************
*                                  SKELETON                                    *
*******************************************************************************/
#undef discovery_snd_namespaces
#define discovery_snd_namespaces (default_namespaces)   // NOTE: same than for directed probe reception (no real reset). Set by the processing function
# define WITH_NOSERVEREQUEST
#  include "gen/discoveryServer.c"
# undef WITH_NOSERVEREQUEST
#undef discovery_namespaces
#undef discovery_snd_namespaces

// NOTE: the 3 next dispactch functions were written splitting the dispatch of discoveryServer.c
static int soap_mcast_recv_request(struct dpws *dpws)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    const char* action = dpws->action ? dpws->action : "";

    soap_peek_element(soap);
#ifdef USE_ONVIF /* Added by HG.W */
    if(DPWS_IS_MULTICAST(dpws->status)){
        ONVIF_LOG1(ONVIF_LOG_INFO,
                   "[INFO]: Received mcast request.(action=[%s])\n",action);
    }
    else{
        ONVIF_LOG1(ONVIF_LOG_INFO,
                   "[INFO]: Received ucast request.(action=[%s])\n",action);
    }
#endif

#ifdef USE_ONVIF /* Added by HG.W (to not serve Hello, Bye, Resolve at nvt_wsd_server */
    if(!strcmp(action,
               "http://schemas.xmlsoap.org/ws/2005/04/discovery/Hello")){
        ONVIF_LOG(ONVIF_LOG_INFO,
                  "[INFO]: Do nothing with received Hello msg.\n");
        return SOAP_OK;
    }
    if(!strcmp(action,
               "http://schemas.xmlsoap.org/ws/2005/04/discovery/Bye")){
        ONVIF_LOG(ONVIF_LOG_INFO,
                  "[INFO]: Do nothing with received Bye msg.\n");
        return SOAP_OK;
    }
    if (!strcmp(action,
                "http://schemas.xmlsoap.org/ws/2005/04/discovery/Resolve")){
        ONVIF_LOG(ONVIF_LOG_INFO,
                  "[INFO]: Do nothing with received Resolve msg.\n");
        return SOAP_OK;
    }
#else
    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/Hello"))
        return soap_serve___wsd__HelloOp(soap);
    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/Bye"))
        return soap_serve___wsd__ByeOp(soap);
    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/Resolve"))
        return soap_serve___wsd__ResolveOp(soap);
#endif

    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/Probe"))
        return soap_serve___wsd__ProbeOp(soap);

    return dpws_dpws2soap(dpws)->error = SOAP_NO_METHOD;
}

static int soap_ucast_recv_probematch(struct dpws *dpws)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    const char* action = dpws->action ? dpws->action : "";
    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/ProbeMatches"))
        return soap_serve___wsd__ProbeMatchOp(soap);
    return dpws_dpws2soap(dpws)->error = SOAP_NO_METHOD;
}

static int soap_ucast_recv_resolvematch(struct dpws *dpws)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    const char* action = dpws->action ? dpws->action : "";
    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/ResolveMatches"))
        return soap_serve___wsd__ResolveMatchOp(soap);
    return dpws_dpws2soap(dpws)->error = SOAP_NO_METHOD;
}

int wsd_serve_directed_probe(struct dpws *dpws)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    const char* action = dpws->action ? dpws->action : "";
    dpws_set_namespaces(dpws, default_namespaces);
    if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2005/04/discovery/Probe"))
        return soap_serve___wsd__DirectedProbeOp(soap);
    return dpws_dpws2soap(dpws)->error = SOAP_NO_METHOD;
}

static int wsd_recv(struct dpws *dpws, int (*pFunc)(struct dpws *))
{
    struct soap *soap = dpws_dpws2soap(dpws);
//  dpws_set_namespaces(dpws, default_namespaces);  // NOTE was discovery namespaces initially, will generate a bug
    dpws_set_namespaces(dpws, discovery_ns);
    soap_begin(soap);

    if (soap_begin_recv(soap))
    { 
        if (soap->error < SOAP_STOP)
        {
            if (soap->errnum == DPWS_ERR_UDP_TIMEOUT)   // dpws_send_fault overwrites errnum
                return soap->error;
#ifdef USE_ONVIF /* Added by HG.W */
            else if(dpws->err == WSD_ERR_NON_XML_DATA ||
                    dpws->err == WSD_ERR_UNKNOWN_UDP_DESTINATION_ADDRESS){
                return dpws->err;
            }
#endif
            else
                return dpws_send_fault(dpws);
        }
        soap_closesock(soap);
    }
    else if (soap_envelope_begin_in(soap)
     || soap_recv_header(soap)
     || soap_body_begin_in(soap)
     || pFunc(dpws)) {
        return soap->error;
    }

    return SOAP_OK;
}

int wsd_mcast_recv(struct dpws *dpws)
{
     int ret = wsd_recv(dpws, &soap_mcast_recv_request);
     dpws_dpws2soap(dpws)->socket = SOAP_INVALID_SOCKET;    // no to close discovery socket
     return ret;
}

int wsd_ucast_recv_probematch(struct dpws *dpws) { 
    return wsd_recv(dpws, &soap_ucast_recv_probematch);
}

int wsd_ucast_recv_resolvematch(struct dpws *dpws) { 
    return wsd_recv(dpws, &soap_ucast_recv_resolvematch);
}

/******************************************************************************\
 *                                                                            *
 * Service Operations (Server-side)                                           *
 *                                                                            *
\******************************************************************************/

/*----------------------------------------------------------------------------*\
 *                                Reception only                              *
\*----------------------------------------------------------------------------*/

SOAP_FMAC5 int SOAP_FMAC6 __wsd__HelloOp(struct dpws* dpws, struct wsd__HelloType *wsd__Hello)
{
    DPWS_SET_HELLO(dpws->status);
    dpws->discovery_msg = wsd__Hello;
    return SOAP_OK;
}

SOAP_FMAC5 int SOAP_FMAC6 __wsd__ByeOp(struct dpws* dpws, struct wsd__ByeType *wsd__Bye)
{
    DPWS_SET_BYE(dpws->status);
    dpws->discovery_msg = wsd__Bye;
    return SOAP_OK;
}

SOAP_FMAC5 int SOAP_FMAC6 __wsd__ProbeOp(struct dpws* dpws, struct wsd__ProbeType *wsd__Probe)
{
    DPWS_SET_PROBE(dpws->status);
    dpws->discovery_msg = wsd__Probe;
    return SOAP_OK;
}

SOAP_FMAC5 int SOAP_FMAC6 __wsd__ProbeMatchOp(struct dpws* dpws, struct wsd__ProbeMatchesType *wsd__ProbeMatches)
{
    DPWS_SET_PROBE_MATCH(dpws->status);
    dpws->discovery_msg = wsd__ProbeMatches;
    return SOAP_OK;
}

SOAP_FMAC5 int SOAP_FMAC6 __wsd__ResolveOp(struct dpws* dpws, struct wsd__ResolveType *wsd__Resolve)
{
    DPWS_SET_RESOLVE(dpws->status);
    dpws->discovery_msg = wsd__Resolve;
    return SOAP_OK;
}

SOAP_FMAC5 int SOAP_FMAC6 __wsd__ResolveMatchOp(struct dpws* dpws, struct wsd__ResolveMatchesType *wsd__ResolveMatches)
{
    DPWS_SET_RESOLVE_MATCH(dpws->status);
    dpws->discovery_msg = wsd__ResolveMatches;
    return SOAP_OK;
}

SOAP_FMAC5 int SOAP_FMAC6 __wsd__DirectedProbeOp(struct dpws* dpws, struct wsd__ProbeType *wsd__Probe, struct wsd__ProbeMatchesType *wsd__ProbeMatches)
{
    return dpws_handle_directed_probe(dpws, wsd__Probe, wsd__ProbeMatches);
}


/*----------------------------------------------------------------------------*\
 *                              delayed processing                            *
\*----------------------------------------------------------------------------*/
static struct wsd__HelloType * build_hello_body(
                                struct device *device,
                                struct wsd__HelloType *body,
                                struct wsa_endpoint_ref * local,
                                dyn_array * XAddrs,
                                struct wsd__ScopesType *scopes
                                )
{
    /* set message data */
    dpws_default_wsa_endpoint_ref(local);
    body->wsa__EndpointReference = local;
    local->address = device->uuid;
    body->XAddrs = XAddrs;
    body->Types = get_string_device_types(device);

    // No need to lock the registry since hello and mdm modification cannnot occur at the same time
    if (device->scopes.nb == 1 && !strcmp(((char **)device->scopes.tab)[0], WSD_DEFAULT_SCOPE))
        body->Scopes = NULL;    // do not send if default value (reduce Hello payload)
    else {
        scopes->__item = &device->scopes;
        scopes->MatchBy = NULL; // spec bug
        body->Scopes = scopes;
    }
        
    body->MetadataVersion = getDeviceMetadataVersion(device);
        
    return body;
}

static int dpws_register_device(struct dpws* dpws, struct wsd__HelloType *hello)
{
    struct device_proxy * cache_entry;
    lock(cache_lock_handle);
    cache_entry = find_device_proxy(hello->wsa__EndpointReference->address);
    if (cache_entry) {
        if (hello->MetadataVersion >= cache_entry->metadata_version || cache_entry->transport_addresses.nb == 0) { //TODO: A more complete test on metadata completeness should be used.
            if (!DPWS_IS_HELLO(dpws->status) || is_sequenced_msg(dpws, cache_entry))
                replace_cache_entry(cache_entry, dpws, hello);
        }
    }
    else
        add_cache_entry(dpws, hello);

    unlock(cache_lock_handle);
    return SOAP_OK;
}

struct XAddrInfo {
    struct dpws * dpws;
    dyn_array * XAddrs;
};
 
static BOOLEAN addXAddr(short *href, struct XAddrInfo * xaddrInfo)
{
    *(char **)new_entry(xaddrInfo->dpws, xaddrInfo->XAddrs) = get_transport_address(xaddrInfo->dpws, *href);
    return FALSE;
}

// TODO: create an API to send an HELLO when IP connectivity changes or metadata changes
// rather ensure that server_init() and shutdown() are reentrant
int send_hello(struct dpws* dpws, handle_s * handle)
{
    int ret = SOAP_OK;
    struct wsa_endpoint_ref hello_endpoint, discovery_endpoint;
    struct wsd__HelloType body;
    struct wsd__ScopesType scopes;
    struct device *device = ((discovery_msg_t *)handle->pObject)->param;
    dyn_array XAddrs;
    struct XAddrInfo xaddrInfo = {dpws, &XAddrs};
    
    init_dyn_array(&XAddrs, sizeof(char *), 1);
    browse_service_ports(device->href, (service_port_cbk)addXAddr, &xaddrInfo);
    
    /* set message data */
    dpws->message_id = ((discovery_msg_t *)handle->pObject)->uuid;
    dpws->instance_id = tk_config.boot_seq;
    dpws->msg_nb = ((discovery_msg_t *)handle->pObject)->msg_nb;

    DPWS_SET_UDP_MULTICAST(dpws->status);
    DPWS_SET_SEND(dpws->status);
    DPWS_SET_HELLO(dpws->status);
    dpws_set_udp_mode(dpws);
    dpws_default_wsa_endpoint_ref(&discovery_endpoint);
    discovery_endpoint.address = WSD_URN;
    // TODO: provide an API in config.c for objetc navigation
    dpws->namespaces = ((struct device_class *)getObject(&registryPool, device->hrefClass))->discovery_namespaces.tab;
    ret = dpws_send___wsd__HelloOp(
        dpws,
        &discovery_endpoint,
        build_hello_body(device, &body, &hello_endpoint, &XAddrs, &scopes)
        );
    soap_closesock(dpws_dpws2soap(dpws));   // one-way does not close the socket
    releaseHandle(&discoveryHandlePool, handle->handleRef); // to trigger free when not used any more
    return ret;
}

int send_bye(struct dpws* dpws, handle_s * handle)
{
    int ret = SOAP_OK;
    struct wsa_endpoint_ref bye_endpoint, discovery_endpoint;
    struct wsd__ByeType body;
    struct device *device = ((discovery_msg_t *)handle->pObject)->param;
    
    /* set message data */
    dpws->message_id = ((discovery_msg_t *)handle->pObject)->uuid;
    dpws->instance_id = tk_config.boot_seq;
    dpws->msg_nb = ((discovery_msg_t *)handle->pObject)->msg_nb;

    dpws_default_wsa_endpoint_ref(&bye_endpoint);
    body.wsa__EndpointReference = &bye_endpoint;
    bye_endpoint.address = device->uuid;
    
    DPWS_SET_UDP_MULTICAST(dpws->status);
    DPWS_SET_SEND(dpws->status);
    DPWS_SET_BYE(dpws->status);
    dpws_set_udp_mode(dpws);
    dpws_default_wsa_endpoint_ref(&discovery_endpoint);
    discovery_endpoint.address = WSD_URN;
    dpws->namespaces = discovery_ns;
    ret = dpws_send___wsd__ByeOp(dpws, &discovery_endpoint, &body);
    soap_closesock(dpws_dpws2soap(dpws));   // one-way does not close the socket
    releaseHandle(&discoveryHandlePool, handle->handleRef); // to trigger free when not used any more
    return ret;
}

int send_probe(struct dpws * dpws, struct wsd__ProbeType *body)
{
    struct wsa_endpoint_ref discovery_endpoint;
    
    dpws_default_wsa_endpoint_ref(&discovery_endpoint);
    discovery_endpoint.address = WSD_URN;
    DPWS_SET_UDP_MULTICAST(dpws->status);
    DPWS_SET_SEND(dpws->status);
    DPWS_SET_PROBE(dpws->status);
    dpws->namespaces = discovery_ns;
    return dpws_send___wsd__ProbeOp(dpws, &discovery_endpoint, body);
}

char * call_directed_probe(struct dpws * dpws, char * address, struct wsd__ProbeType *body)
{
    struct wsa_endpoint_ref endpoint;
    struct wsd__ProbeMatchesType probe_matches;
    char *result = NULL;
    
    dpws_default_wsa_endpoint_ref(&endpoint);
    soap_default_wsd__ProbeMatchesType(dpws_dpws2soap(dpws), &probe_matches);
    endpoint.address = address;
    dpws->namespaces = discovery_ns;
    if (dpws_call___wsd__DirectedProbeOp(dpws, &endpoint, NULL, body, &probe_matches)) {
        return NULL;
    }
    if (probe_matches.__sizeProbeMatch > 0) { // TODO: AM - What do we do when more than one device matches an address
        result = probe_matches.ProbeMatch[0]->wsa__EndpointReference->address;
        dpws_handle_probe_match(dpws, &probe_matches);      
    } else if (dpws_dpws2soap(dpws)->status == 202) {
        dpws_dpws2soap(dpws)->error = SOAP_OK;
    }
    return result;
}

int send_resolve(struct dpws * dpws, struct wsa_endpoint_ref *endpoint)
{
    struct wsa_endpoint_ref discovery_endpoint;
    struct wsd__ResolveType resolveBody;

    dpws_default_wsa_endpoint_ref(&discovery_endpoint);
    discovery_endpoint.address = WSD_URN;
    DPWS_SET_UDP_MULTICAST(dpws->status);
    DPWS_SET_SEND(dpws->status);
    DPWS_SET_RESOLVE(dpws->status);
    dpws->namespaces = discovery_ns;
    resolveBody.wsa__EndpointReference = endpoint;
    return dpws_send___wsd__ResolveOp(dpws, &discovery_endpoint, &resolveBody);
}

static int send_probe_match(struct dpws * dpws, struct device * device)
{
    int ret = DPWS_OK;
    struct wsa_endpoint_ref discovery_endpoint, this_endpoint;
    struct wsd__ProbeMatchesType probeMatchesBody;
    struct wsd__HelloType body, *pBody;
    struct wsd__ScopesType scopes;
    dyn_array XAddrs;
    struct XAddrInfo xaddrInfo = {dpws, &XAddrs};
    
    init_dyn_array(&XAddrs, sizeof(char *), 1);
    browse_service_ports(device->href, (service_port_cbk)addXAddr, &xaddrInfo);

    /* set message data */
    if (dpws_endpoint_ref_is_null(dpws->reply_to))
        dpws_set_anonymous_wsa_endpoint_ref(&discovery_endpoint);
    else
        memcpy(&discovery_endpoint, &dpws->reply_to, sizeof(struct wsa_endpoint_ref));

    probeMatchesBody.__sizeProbeMatch = 1;
    pBody = build_hello_body(device, &body, &this_endpoint, &XAddrs, &scopes);
    probeMatchesBody.ProbeMatch = &pBody;
        
    dpws->namespaces = ((struct device_class *)getObject(&registryPool, device->hrefClass))->discovery_namespaces.tab;
    ret = dpws_send___wsd__ProbeMatchOp(
        dpws,
        &discovery_endpoint,
        &probeMatchesBody
        );
    return ret;
}

static int send_scheduled_probe_match(struct dpws * dpws, handle_s * handle)
{
    int ret;
    struct device *device = ((discovery_msg_t *)handle->pObject)->param;
    dpws->message_id = ((discovery_msg_t *)handle->pObject)->uuid;
    dpws->relates_to = ((discovery_msg_t *)handle->pObject)->req_uuid;
    dpws->instance_id = tk_config.boot_seq;
    dpws->msg_nb = ((discovery_msg_t *)handle->pObject)->msg_nb;
    ret = send_probe_match(dpws, device);   
    releaseHandle(&discoveryHandlePool, handle->handleRef); // to trigger free when not used any more.
    return ret;
}

#ifdef USE_ONVIF /* Added by HG.W (support send discovery soap fault) */
static void set_discovery_fault(struct dpws *dpws)
{
    struct soap *soap = &dpws->soap;
    const char **c = soap_faultcode(soap);
    const char **sc = soap_faultsubcode(soap);
    const char **s = soap_faultstring(soap);
    const char **d = soap_faultdetail(soap);
      
    *c = "SOAP-ENV:Sender";
    *sc = "wsd:MatchingRuleNotSupported";
    *s = "the matching rule specified is not supported.";
    strcpy(soap->msgbuf, "<wsd:SupportedMatchingRules>");
    strcat(soap->msgbuf, WSD_MATCH_BY_RFC3986_URI);
    /**
     * if need to add other maching rules here, do as follow:
     * strcat(soap->msgbuf, " ");
     * strcat(soap->msgbuf, OTHER RULES URI);
     */
    strcat(soap->msgbuf, "</wsd:SupportedMatchingRules>");
    *d = soap->msgbuf;
        
    soap_set_fault(soap);
}

static int send_discovery_fault(struct dpws * dpws, struct device * device)
{
    int ret = DPWS_OK;
    struct soap *soap = &dpws->soap;

    dpws_dpws2soap(dpws)->error = WSD_ERR_MATCHING_RULE_NOT_SUPPORTED;

    /* wsa header */
    DPWS_SET_REQUEST(dpws->status);

    dpws->to = NULL;
    if (!dpws_endpoint_ref_is_null(dpws->fault_to)) {
        dpws->to = dpws->fault_to.address;
        dpws_default_wsa_endpoint_ref(&dpws->fault_to);
    }
    if (!dpws_endpoint_ref_is_null(dpws->reply_to)) {
        if (!dpws->to) {
            dpws->to = dpws->reply_to.address;
        }
        dpws_default_wsa_endpoint_ref(&dpws->reply_to);
    }
    if (!dpws_endpoint_ref_is_null(dpws->from)) {
        if (!dpws->to) {
            dpws->to = dpws->from.address;
            dpws->subscription_id = dpws->from.subscription_id;
        }
        dpws_default_wsa_endpoint_ref(&dpws->from);
    }
    if (!dpws->to) {
        dpws->to = WSA_ANONYMOUS_ENDPOINT_REF_URI;
        dpws->subscription_id = NULL;
    }
    
    dpws->action =  WSD_SOAP_FAULT_URI;

    dpws->namespaces = ((struct device_class *)getObject(&registryPool, device->hrefClass))->discovery_namespaces.tab;    

    /* body */
    set_discovery_fault(dpws);
    soap->encodingStyle = NULL;
    soap_begin(soap);
    soap_serializeheader(soap);
    soap_serializefault(soap);
    if(soap_begin_count(soap))
        return soap->error;
    if(soap->mode & SOAP_IO_LENGTH)
    {
        if(soap_envelope_begin_out(soap)
           || soap_putheader(soap)
           || soap_body_begin_out(soap)
           || soap_putfault(soap)
           || soap_body_end_out(soap)
           || soap_envelope_end_out(soap))
            return soap->error;
    }
    if(soap_end_count(soap))
        return soap->error;

    /* send */
    if(dpws_connect(dpws, dpws->to, WSD_SOAP_FAULT_URI)
       || soap_envelope_begin_out(soap)
       || soap_putheader(soap)
       || soap_body_begin_out(soap)
       || soap_putfault(soap)
       || soap_body_end_out(soap)
       || soap_envelope_end_out(soap)
       || soap_end_send(soap))
            return soap_closesock(soap);

    return SOAP_OK;
}

static int send_scheduled_discovery_fault(struct dpws *dpws, handle_s *handle)
{
    int ret;
    struct device *device = ((discovery_msg_t *)handle->pObject)->param;
    dpws->message_id = ((discovery_msg_t *)handle->pObject)->uuid;
    dpws->relates_to = ((discovery_msg_t *)handle->pObject)->req_uuid;
    dpws->instance_id = tk_config.boot_seq;
    dpws->msg_nb = ((discovery_msg_t *)handle->pObject)->msg_nb;
    ret = send_discovery_fault(dpws, device);   
    releaseHandle(&discoveryHandlePool, handle->handleRef); // to trigger free when not used any more.
    return ret;
}
#endif /* Added by HG.W (support send discovery soap fault) */

static int dpws_handle_hello(struct dpws* dpws, struct wsd__HelloType *hello)
{
    int ret;

    ret = dpws_register_device(dpws, hello);    // checks for doubles
    // call the callback after in order that the user can consult the cache
    if (tk_config.hello_hook) {
        struct device_proxy * device = find_device_proxy(hello->wsa__EndpointReference->address); 
        if (device) //  this section is not thread-safe
            tk_config.hello_hook(dpws, device->super.href);
    }
    return ret;
}

static int dpws_handle_bye(struct dpws* dpws, struct wsd__ByeType * bye)
{
    struct device_proxy * cache_entry;

    lock(cache_lock_handle);
    cache_entry = find_device_proxy(bye->wsa__EndpointReference->address);
    if (cache_entry && is_sequenced_msg(dpws, cache_entry)) {
        // call the callback before in order that the user can consult the cache
        if (tk_config.bye_hook)
            tk_config.bye_hook(dpws, cache_entry->super.href);
        invalidate_device_proxy(cache_entry);
    }
    unlock(cache_lock_handle);
    
    return SOAP_OK;
}

static int qname_struct_equals(char **type1, struct prefixed_qname *type2)
{
    return qname_equals(type2->qname.ns, type2->qname.lname, *type1);
}

static int dpws_handle_probe(struct dpws* dpws, struct wsd__ProbeType * probe)
{
    int ret = DPWS_OK;
    BOOLEAN match = FALSE;
    dyn_array *device_services;
    struct device * device;
    scheduler_t probeMatchScheduler = {NULL, NULL};
    unsigned32 delay = 0;
#ifdef USE_ONVIF /* Added by HG.W */
    int match_rule;
    scheduler_t discoveryFaultScheduler = {NULL, NULL};
#endif    
    
    // save msg data to be able
    for (device = registry; ret == DPWS_OK && device; device = device->next)
    {
        dyn_array* scopes = NULL;
        char * rule = NULL;
        /* check that the message must be processed */
        if (!device->device_up || get_device_mode(device) != DPWS_DEVICE_USE)
            continue;
            
        /* check that scope and types matches */
        lock(device->lock); // NOTE: because of mdm

        device_services = get_device_types(device);
        if (probe->Scopes) {
            scopes = probe->Scopes->__item;
            rule = probe->Scopes->MatchBy;
        }

#ifdef USE_ONVIF /* Added by HG.W */
        match_rule = parse_match_rule(rule);
        if(match_rule == WSD_MATCH_BY_UNSPECIFIED){
            /* if received by multicast do nothing */
            if(DPWS_IS_MULTICAST(dpws->status)){
                ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Unknown matching rule.(do nothing for multicast)\n");
                unlock(device->lock);
                continue;
            }
            else{ /* if received by unicast, then schedule a fault msg */
                ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Unknown matching rule.(send fault for unicast)\n");
                /**
                 * schedule the discovery fault msg.
                 * caution:
                 *   it should to define a "DPWS_MSG_DISCOVERY_FAULT" macro 
                 *   instead of DPWS_MSG_PROBE_MATCH used here.
                 *   But, to do this, there are too many changes have to be
                 *   done, because there are too many other relative macros.
                 *   For the behavior is same both the fault and ProbeMatch, 
                 *   so, sorry for using DPWS_MSG_PROBE_MATCH here temporary,
                 */
                ret = schedule_discovery_msg(&discoveryFaultScheduler,
                                             DPWS_MSG_PROBE_MATCH,
                                             dpws->message_id, device, 
                (int (*) (struct dpws*, void *))send_scheduled_discovery_fault,
                                             -1);

                unlock(device->lock);
                continue;
            }
        }
        else{
            match = probe_match(probe->Types, scopes, match_rule,
                                device_services, &device->scopes,
                                (type_compare_cbk)qname_struct_equals);
        }
#else
        match = probe_match(probe->Types, scopes, parse_match_rule(rule),
            device_services, &device->scopes, (type_compare_cbk)qname_struct_equals);
#endif 

        unlock(device->lock);

#ifdef USE_ONVIF /* Added by HG.W */
        if (match){
            ONVIF_LOG(ONVIF_LOG_INFO, "[INFO]: Probe matched.\n");
            ret = schedule_discovery_msg(&probeMatchScheduler, DPWS_MSG_PROBE_MATCH, dpws->message_id, device, (int (*) (struct dpws*, void *))send_scheduled_probe_match, -1);
        }
        else{
            ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Probe not matched.\n");
        }
#else
        if (match)
            ret = schedule_discovery_msg(&probeMatchScheduler, DPWS_MSG_PROBE_MATCH, dpws->message_id, device, (int (*) (struct dpws*, void *))send_scheduled_probe_match, -1);
#endif
        // TODO: using schedule_discovery_msg for "local" scheduling is not efficient since message is built several times (whereas dpws_end is not called)
    }

    // send messages in a raw to avoid delay accumulation
    /* send probe match messages */
    DPWS_SET_UNICAST(dpws->status);
    DPWS_SET_SEND(dpws->status);
    DPWS_SET_PROBE_MATCH(dpws->status);
    dpws_dpws2soap(dpws)->omode |= SOAP_IO_KEEPALIVE;   // NOTE: udp mode must have been set during probe receive
 
    dpws->relates_to = dpws->message_id;    // it is not a response so addressing won't do it automatically
    
    while (ret == DPWS_OK && (dpws->schedule = get_schedule(&probeMatchScheduler)))
    {
        delay = get_ms_schedule(dpws->schedule);
        if (delay > 0)
            sleep_ms(delay);
        ret = realize_schedule(&probeMatchScheduler, dpws, -1);
    }
    dpws_dpws2soap(dpws)->omode &= ~SOAP_IO_KEEPALIVE;
    dpws_dpws2soap(dpws)->keep_alive = 0;   // avoid dpws_serve to loop
    dpws_udp_close(dpws_dpws2soap(dpws));   // one-way does not close the socket
    free_scheduler(&probeMatchScheduler);
    dpws->schedule = NULL;

#ifdef USE_ONVIF /* Added by HG.W (to send discovery fault msg) */
    /* send soap fault messages */
    DPWS_SET_UNICAST(dpws->status);
    DPWS_SET_SEND(dpws->status);
    DPWS_SET_PROBE_MATCH(dpws->status);
    dpws_dpws2soap(dpws)->omode |= SOAP_IO_KEEPALIVE;   // NOTE: udp mode must have been set during probe receive
 
    dpws->relates_to = dpws->message_id;    // it is not a response so addressing won't do it automatically
    
    while (ret == DPWS_OK && (dpws->schedule = get_schedule(&discoveryFaultScheduler)))
    {
        delay = get_ms_schedule(dpws->schedule);
        if (delay > 0)
            sleep_ms(delay);
        ret = realize_schedule(&discoveryFaultScheduler, dpws, -1);
    }
    dpws_dpws2soap(dpws)->omode &= ~SOAP_IO_KEEPALIVE;
    dpws_dpws2soap(dpws)->keep_alive = 0;   // avoid dpws_serve to loop
    dpws_udp_close(dpws_dpws2soap(dpws));   // one-way does not close the socket
    free_scheduler(&discoveryFaultScheduler);
    dpws->schedule = NULL;
#endif

    return ret;
}

static int dpws_handle_directed_probe(struct dpws* dpws, struct wsd__ProbeType * probe, struct wsd__ProbeMatchesType * probeMatches)
{
    BOOLEAN match = FALSE;
    struct device * device = get_endpoint(dpws->href_endpoint)->device;

    /* check that the message must be processed */
    if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE) {
        dyn_array *device_services;
        dyn_array* scopes = NULL;
        char * rule = NULL;

        /* check that scope and types matches */
        lock(device->lock); // NOTE: because of mdm
        device_services = get_device_types(device);
        if (probe->Scopes) {
            scopes = probe->Scopes->__item;
            rule = probe->Scopes->MatchBy;
        }
        match = probe_match(probe->Types, scopes, parse_match_rule(rule),
                            device_services, &device->scopes, (type_compare_cbk)qname_struct_equals);
        unlock(device->lock);
        if (match) {
            struct wsd__HelloType *body, *pBody;
            struct wsd__ScopesType scopes;
            dyn_array* XAddrs;
            struct XAddrInfo xaddrInfo;
            struct wsa_endpoint_ref * this_endpoint;
    
            body = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct wsd__HelloType));
            XAddrs= soap_malloc(dpws_dpws2soap(dpws), sizeof(dyn_array));
            init_dyn_array(XAddrs, sizeof(char*), 1);
            xaddrInfo.dpws = dpws;
            xaddrInfo.XAddrs = XAddrs;
            browse_service_ports(dpws->href_endpoint, (service_port_cbk)addXAddr, &xaddrInfo);
            this_endpoint = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct wsa_endpoint_ref));
            probeMatches->__sizeProbeMatch = 1;
            pBody = build_hello_body(device, body, this_endpoint, XAddrs, &scopes);
            probeMatches->ProbeMatch = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct wsd_HelloType*));
            probeMatches->ProbeMatch[0] = pBody;
            dpws->namespaces = ((struct device_class *)getObject(&registryPool, device->hrefClass))->discovery_namespaces.tab;
            dpws_set_namespaces(dpws, dpws->namespaces);
//          dpws->instance_id = tk_config.boot_seq;
//          dpws->msg_nb = new_app_seq_msg_nb();
            return SOAP_OK;
        }
    }   
    return dpws_send_empty_response(dpws);
}

static int dpws_handle_probe_match(struct dpws* dpws, struct wsd__ProbeMatchesType* probe_matches)
{
    int i = 0, ret = DPWS_OK;

    for (; i < probe_matches->__sizeProbeMatch; i++) {
        if ((ret = dpws_register_device(dpws, probe_matches->ProbeMatch[i])))
            break;
    }
    return ret;
}


static int dpws_handle_resolve(struct dpws* dpws, struct wsd__ResolveType* resolve)
{
    int ret = DPWS_OK;
    struct service_endpoint * endpoint = find_endpoint(resolve->wsa__EndpointReference->address);

    if (endpoint) {
        // check a matching device logical id
        if (endpoint->device)
        {
            long delay = 0;
            short send_nb = 0;
            struct wsa_endpoint_ref this_endpoint, discovery_endpoint;
            struct wsd__ResolveMatchesType resolveMatches;
            struct wsd__HelloType body;
            struct wsd__ScopesType scopes;
            char uuid[SCHEME_UUID_STRING_SIZE];
            dyn_array* XAddrs;
            struct XAddrInfo xaddrInfo;

            DPWS_SET_UNICAST(dpws->status);
            DPWS_SET_SEND(dpws->status);
            DPWS_SET_RESOLVE_MATCH(dpws->status);
            dpws_dpws2soap(dpws)->omode |= SOAP_IO_KEEPALIVE;   // NOTE: udp mode must have been set during resolve receive
         
            dpws->relates_to = dpws->message_id;    // it is not a response so addressing won't do it automatically
            if (dpws_endpoint_ref_is_null(dpws->reply_to))
                dpws_set_anonymous_wsa_endpoint_ref(&discovery_endpoint);
            else
                memcpy(&discovery_endpoint, &dpws->reply_to, sizeof(struct wsa_endpoint_ref));
                
            schemed_uuid_create(uuid);
            dpws->message_id = uuid;
            dpws->instance_id = tk_config.boot_seq;
            dpws->msg_nb = new_app_seq_msg_nb();
    
            XAddrs= soap_malloc(dpws_dpws2soap(dpws), sizeof(dyn_array));
            init_dyn_array(XAddrs, sizeof(char*), 1);
            xaddrInfo.dpws = dpws;
            xaddrInfo.XAddrs = XAddrs;
            browse_service_ports(endpoint->href, (service_port_cbk)addXAddr, &xaddrInfo);
            resolveMatches.ResolveMatch = build_hello_body(endpoint->device, &body, &this_endpoint, XAddrs, &scopes);
            dpws->namespaces = ((struct device_class *)getObject(&registryPool, endpoint->device->hrefClass))->discovery_namespaces.tab;

            // TODO: Why not using a local scheduler like for PM if the message construction problem is solved ?
            while ((delay = get_soap_over_udp_schedule(DPWS_MSG_RESOLVE_MATCH, &send_nb, delay)) != 0xFFFFFFFF)
            {
                if (delay > 0)
                    sleep_ms(delay);
                
                ret = dpws_send___wsd__ResolveMatchOp(
                    dpws,
                    &discovery_endpoint,
                    &resolveMatches
                    );
            }
            dpws_dpws2soap(dpws)->omode &= ~SOAP_IO_KEEPALIVE;
            dpws_dpws2soap(dpws)->keep_alive = 0;   // avoid dpws_serve to loop
            dpws_udp_close(dpws_dpws2soap(dpws));   // one-way does not close the socket
        }
    }
    return ret;
}

int dpws_handle_resolve_match(struct dpws* dpws, struct wsd__ResolveMatchesType* resolve_matches)
{
    return dpws_register_device(dpws, resolve_matches->ResolveMatch);   // checks for doubles
}

#ifdef USE_ONVIF /* Added by HG.W */
/* bug fix */
int wsd_process(struct dpws *dpws)
{
    int err = 1;

    err = is_new_message(dpws->message_id);
    if(err == 1){
        switch (DPWS_GET_MSG_TYPE(dpws->status))
        {
            case DPWS_MSG_HELLO :
                return dpws_handle_hello(dpws, dpws->discovery_msg);
            case DPWS_MSG_BYE :
                return dpws_handle_bye(dpws, dpws->discovery_msg);
            case DPWS_MSG_PROBE :
                return dpws_handle_probe(dpws, dpws->discovery_msg);
            case DPWS_MSG_PROBE_MATCH :
                return dpws_handle_probe_match(dpws, dpws->discovery_msg);
            case DPWS_MSG_RESOLVE :
                return dpws_handle_resolve(dpws, dpws->discovery_msg);
            case DPWS_MSG_RESOLVE_MATCH :
                return dpws_handle_resolve_match(dpws, dpws->discovery_msg);
        }
    }
    else if(err == -1){
        return SOAP_ERR;
    }
    else{
        return SOAP_OK;
    }
}
#else
int wsd_process(struct dpws *dpws)
{
    if (is_new_message(dpws->message_id)) {
        switch (DPWS_GET_MSG_TYPE(dpws->status))
        {
            case DPWS_MSG_HELLO :
                return dpws_handle_hello(dpws, dpws->discovery_msg);
            case DPWS_MSG_BYE :
                return dpws_handle_bye(dpws, dpws->discovery_msg);
            case DPWS_MSG_PROBE :
                return dpws_handle_probe(dpws, dpws->discovery_msg);
            case DPWS_MSG_PROBE_MATCH :
                return dpws_handle_probe_match(dpws, dpws->discovery_msg);
            case DPWS_MSG_RESOLVE :
                return dpws_handle_resolve(dpws, dpws->discovery_msg);
            case DPWS_MSG_RESOLVE_MATCH :
                return dpws_handle_resolve_match(dpws, dpws->discovery_msg);
        }
    }
    return SOAP_OK;
}

#endif /* Added by HG.W */

