/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.273.2.13 $
|                                              $Date: 2007/06/28 17:10:18 $
\*============================================================================*/
 
/*******************************************************************************
*                               DPWS RUNTIME                                   *
*******************************************************************************/
#include "_dpws.h"
#include "discovery.h"
#include "metadata.h"
#include "tools.h"
#include "config.h"
#include "cache.h"
#include "event.h"
#include "subscmanager.h"
#include "mdm.h"
#include "my_stdsoap2.h"
#include <sys/un.h>

#ifdef VXWORKS
# include "inetLib.h"
#endif

// TODO: find a better solution, as new gSOAP modes could override this one
// New IO mode to specify that we are in DPWS mode. This allows pure gSOAP generated code
// to run with the DPWS library.
#define DPWS_HEADERS    0x10000000

scheduler_t scheduler = {NULL, NULL};
short scheduler_lock_handle = -1;

#ifdef WITH_IPV6
scheduler_t scheduler_v6 = {NULL, NULL};
short scheduler_lock_handle_v6 = -1;
#endif


extern struct Namespace default_namespaces[];
extern const char ** SOAP_FMAC4 soap_faultsubcode(struct soap *);
extern short cache_lock_handle;
extern struct device *registry;
extern handlePool_t registryPool;
extern struct dpws_config tk_config;
extern struct listener * listeners;

/*******************************************************************************
*                              PRIVATE STUFF                                   *
*******************************************************************************/
static struct dpws * dpws_in_wsd_AppSequence(struct dpws *dpws)
{
    struct soap *soap = &dpws->soap;
    if (soap_element_begin_in(soap, "wsd:AppSequence", 0))
        return NULL;
    if (soap_s2unsignedLong(soap, soap_attr_value(soap, "InstanceId", 1 /* REQUIRED */), &dpws->instance_id))
        return NULL;
    if (soap_s2unsignedShort(soap, soap_attr_value(soap, "MessageNumber", 1), &dpws->msg_nb))
        return NULL;
    if (soap->body && soap_element_end_in(soap, "wsd:AppSequence"))
            return NULL;
    return dpws;
}

int dpws_out_wsa_endpoint_ref(struct soap *soap, const char *tag, struct wsa_endpoint_ref *a) {
    return soap_out_endpoint_ref(soap, tag, -1, &a, NULL);
}

static struct wsa_endpoint_ref * dpws_in_wsa_endpoint_ref(struct soap *soap, const char *tag, struct wsa_endpoint_ref *a) {
    struct wsa_endpoint_ref ** ret = soap_in_endpoint_ref(soap, tag, &a, NULL);
    return ret ? *ret : NULL;
}

void dpws_set_anonymous_wsa_endpoint_ref(struct wsa_endpoint_ref *a) {
    a->address = WSA_ANONYMOUS_ENDPOINT_REF_URI;
    a->subscription_id = NULL;
}

static BOOLEAN generate_service_port(short * href, struct soap * soap)
{
    soap_send(soap, "<li>");
    soap_send(soap, (char*)get_transport_address(dpws_soap2dpws(soap), *href));
    soap_send(soap, "</li>");
    return FALSE;
}
#ifndef USE_ONVIF
static int dpws_http_get(struct soap* soap) {
    char uuid[SCHEME_UUID_STRING_SIZE];
    char* context_path = strchr(soap->path, '/');
    short href = -1;
    struct service_endpoint* endpoint;
    strcpy(uuid, UUID_SCHEME);
    strncpy(uuid + UUID_SCHEME_SIZE, context_path + 1, UUID_SIZE);
    uuid[SCHEME_UUID_STRING_SIZE-1] = 0;
    soap_response(soap, SOAP_HTML); // HTTP response header with text/html
    soap_send(soap, "<html><head><title>");
    soap_send(soap, "Device presentation page");
    soap_send(soap, "</title></head><body>");
    endpoint = find_endpoint(uuid);
    if (endpoint) {
        short hrefDevice = endpoint->device->href;
        short hrefDeviceClass = endpoint->device->hrefClass;
        char* friendlyName = (char*)dpws_get_ptr_att_item(hrefDevice, DPWS_STR_FRIENDLY_NAME, 0);
        int len = 10, i; 
        short hrefs[10];
        dyn_array * array = NULL;
        soap_send(soap, "<h1>");
        soap_send(soap, friendlyName);
        soap_send(soap, "</h1>");
        soap_send(soap, "<h2>Model information</h2>");
        soap_send(soap, "<p><b>Manufacturer: </b><a href='");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDeviceClass, DPWS_STR_MANUFACTURER_URL));
        soap_send(soap, "'>");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDeviceClass, DPWS_STR_MANUFACTURER));
        soap_send(soap, "</a></p>");
        soap_send(soap, "<p><b>Model name: </b><a href='");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDeviceClass, DPWS_STR_MODEL_URL));
        soap_send(soap, "'>");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDeviceClass, DPWS_STR_MODEL_NAME));
        soap_send(soap, "</a></p>");
        soap_send(soap, "<p><b>Model number: </b>");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDeviceClass, DPWS_STR_MODEL_NUMBER));
        soap_send(soap, "</p>");
        soap_send(soap, "<h2>Device discovery information</h2>");
        soap_send(soap, "<p><b>Logical address: </b>");
        soap_send(soap, uuid);
        soap_send(soap, "</p>");
        soap_send(soap, "<p><b>Supported types: </b></p><ul>");
        array = get_device_types(endpoint->device);
        for (i = 0; i < array->nb; i++) {
            struct prefixed_qname * qname = (struct prefixed_qname *)get_entry(array, i);
            soap_send(soap, "<li>{");
            soap_send(soap, qname->qname.ns);
            soap_send(soap, "}");
            soap_send(soap, qname->qname.lname);
            soap_send(soap, "</li>");
        }
        soap_send(soap, "</ul>");
        array = &endpoint->device->scopes;
        if (array->nb > 0) 
            soap_send(soap, "<p><b>Scopes: </b></p><ul>");
        for (i = 0; i < array->nb; i++) {
            char * scope = *(char **)get_entry(array, i);
            soap_send(soap, "<li>");
            soap_send(soap, scope);
            soap_send(soap, "</li>");
        }
        soap_send(soap, "</ul>");
        soap_send(soap, "<h2>Troubleshooting information</h2>");
        soap_send(soap, "<p><b>Serial number: </b>");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDevice, DPWS_STR_SERIAL_NUMBER));
        soap_send(soap, "</p>");
        soap_send(soap, "<p><b>Firmware version: </b>");
        soap_send(soap, (char*)dpws_get_ptr_att(hrefDevice, DPWS_STR_FIRMWARE_VERSION));
        soap_send(soap, "</p>");
        if (!dpws_get_service_handles(hrefDevice, hrefs, &len)) {
            soap_send(soap, "<h2>Hosted services</h2>");
            for (i = 0; i < len; i++) {
                int j, k;
                struct service_endpoint * service = get_endpoint(hrefs[i]);
                struct wsdl_info * info;
                array = get_service_types(service);
                soap_send(soap, "<h3>Service ");
                soap_send(soap, (char*)dpws_get_ptr_att(hrefs[i], DPWS_STR_SERVICE_ID));
                soap_send(soap, "</h3>");
                soap_send(soap, "<p><b>Addresses: </b></p><ul>");
                browse_service_ports(endpoint->href, (service_port_cbk)generate_service_port, soap);
                soap_send(soap, "</ul>");
                soap_send(soap, "<p><b>WSDL documents: </b><ul>");
                for (j = 0; (info = (struct wsdl_info *)dpws_get_ptr_att_item(service->hrefServiceClass, DPWS_PTR_WSDL, j))->target_ns; j++) {
                    soap_send(soap, "<li><b>Target namespace:  </b>");
                    soap_send(soap, info->target_ns);
                    soap_send(soap, "</li><ul><li>Location: <a href='");
                    soap_send(soap, info->location);
                    soap_send(soap, "'>");
                    soap_send(soap, info->location);
                    soap_send(soap, "</a></li><li>Supported port types: ");
                    for (k = 0; k < array->nb; k++) {
                        struct prefixed_qname * qname = (struct prefixed_qname *)get_entry(array, k);
                        if (!strcmp(info->target_ns, qname->qname.ns)) {
                            soap_send(soap, qname->qname.lname);
                            soap_send(soap, " ");
                        }
                    }
                    soap_send(soap, "</ul></li>");
                }
                soap_send(soap, "</ul>");
            }
        }
    } else {
        soap_send(soap, "<h1>Unknown device<h1>");
    }
    soap_send(soap, "</body></html>");
    soap_end_send(soap);
    return SOAP_OK; 
}
#endif
static void dpws_struct_init(struct dpws* dpws)
{
    memset(dpws, 0, sizeof(struct dpws));
    soap_init(&dpws->soap);
    dpws_init_wsa_headers(dpws);    // also reset status
    if (!dpws->soap.namespaces)
        dpws->soap.namespaces = default_namespaces;
    dpws->wsd_ls = SOAP_INVALID_SOCKET;
#ifdef USE_ONVIF /* Added by HG.W */
    dpws->ipproto_type = DPWS_IPPROTO_UNSPEC;
#ifdef WITH_IPV6
    dpws->xaddr_ipv6_scope = DPWS_XADDR_IPV6_SCOPE_LINKLOCAL;
#endif
    dpws->wsd_ws = SOAP_INVALID_SOCKET;
#endif
    dpws->err_detail_type = -1;
// Tell the SOAP runtime that we are in DPWS mode
    dpws->soap.imode |= DPWS_HEADERS;
    dpws->soap.omode |= DPWS_HEADERS;
// Add support for device page display
#ifndef USE_ONVIF
    dpws->soap.fget = dpws_http_get;
#endif
    init_dyn_array(&dpws->proxyHandles, sizeof(short), 10);
}

static int buffer_send(struct soap *soap, const char *s, size_t n)
{
    n = MIN(n, (size_t)(1024 - dpws_soap2dpws(soap)->tmp_buf_count));
    strncat(soap->msgbuf, s, n);
    dpws_soap2dpws(soap)->tmp_buf_count += n;
    return SOAP_OK;
}

static void buffer_marshalling_init(struct dpws *dpws)
{
    dpws->tmp_buf_count = 0;
    dpws->backup_send_f = dpws_dpws2soap(dpws)->fsend;
    dpws_dpws2soap(dpws)->fsend = &buffer_send; // QUESTION: is it correct ? Aren't flags modified ? A dedicated SOAP structure would be costly.
    strcpy(dpws_dpws2soap(dpws)->msgbuf, "<![CDATA[");
}

static char * buffer_marshalling_end(struct dpws *dpws)
{
    char *buf = dpws_dpws2soap(dpws)->msgbuf;
    buf += MIN(1018, dpws->tmp_buf_count);  // 1024 - 6
    strcpy(buf, "...]]>");
    dpws_dpws2soap(dpws)->fsend = dpws->backup_send_f;
    return dpws_dpws2soap(dpws)->msgbuf;
}

#ifdef USE_DISCOVERY_PORT_AS_SOURCE_PORT /* Added by HG.W */
static int my_dpws_udp_ucast_open(struct soap *soap, const char *endpoint, const char *host, int port)
{
    return dpws_soap2dpws(soap)->wsd_ls;
}

#endif

#ifdef USE_ONVIF /* Added by HG.W */
static int dpws_udp_ucast_open(struct soap *soap, const char *endpoint, const char *host, int port)
{
    int fd;
    struct addrinfo hints, *res, *res0;
    char *ip_addr = NULL;
    int err = 0;
    int sockopt_on = 1;
    struct dpws *dpws=dpws_soap2dpws(soap);

    soap->errmode = 0;

    memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_DGRAM;

    if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
        ip_addr = tk_config.network_adapter.ip_addr;
    }
#ifdef WITH_IPV6
    else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
        ip_addr = tk_config.network_adapter.ip_addr_v6_l_with_if;
    }
#endif
    else{
        return SOAP_INVALID_SOCKET;
    }

    err = getaddrinfo(ip_addr, NULL, &hints, &res0);
    if(err){
        ONVIF_LOG2(ONVIF_LOG_ERROR, "[ERROR]: Failed to get addrinfo(%s) in dpws_udp_ucast_open().(gai_strerror=%s)\n", ip_addr, gai_strerror(err));
        return SOAP_INVALID_SOCKET;
    }

    fd = socket(res0->ai_family, res0->ai_socktype, res0->ai_protocol);
    if(fd < 0){ 
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP socket creation failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to create socket in dpws_udp_ucast_open().(strerror=%s)\n", strerror(soap->errnum));
        return SOAP_INVALID_SOCKET;
    }    

    if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
        if(setsockopt(fd, IPPROTO_IP, IP_PKTINFO, &sockopt_on, sizeof(sockopt_on))< 0){
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ip pktinfo on UDP socket failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IP_PKTINFO on udp socket in dpws_udp_ucast_open().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
    }
#ifdef WITH_IPV6
    else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
#if defined IPV6_RECVPKTINFO
        if(setsockopt(fd, IPPROTO_IPV6, IPV6_RECVPKTINFO, &sockopt_on, sizeof(sockopt_on))< 0){
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ipv6 recvpktinfo on UDP socket failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_RECVPKTINFO on udp socket in dpws_udp_ucast_open().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
#elif defined IPV6_PKTINFO
        if(setsockopt(fd, IPPROTO_IPV6, IPV6_PKTINFO, &sockopt_on, sizeof(sockopt_on))< 0){
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ipv6 pktinfo on UDP socket failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_PKTINFO on udp socket in dpws_udp_ucast_open().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
#else
# error "can't determine socket option"
#endif
    }
#endif
    else{/* do nothing */}

    if(bind(fd, (struct sockaddr *) res0->ai_addr, res0->ai_addrlen) < 0){
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)fd);
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP bind failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to bind socket in dpws_udp_ucast_open().(strerror=%s)\n", strerror(soap->errnum));
        return SOAP_INVALID_SOCKET;
    }

    freeaddrinfo(res0);

    return fd;
}
#else /* !USE_ONVIF */
static int dpws_udp_ucast_open(struct soap *soap, const char *endpoint, const char *host, int port)
{
    register int fd;
    struct sockaddr_in cliAddr;
    
    /* create sending UDP socket */
    soap->errmode = 0;
    fd = (int)socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) { 
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP socket creation failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
        return SOAP_INVALID_SOCKET;
    }

    /* bind unicast UDP socket to any port number */
    cliAddr.sin_family = AF_INET;
    cliAddr.sin_addr.s_addr = dpws_soap2dpws(soap)->out_ip_addr ? inet_addr(dpws_soap2dpws(soap)->out_ip_addr) : htonl(INADDR_ANY);
    cliAddr.sin_port = htons(0);
    if (bind(fd, (struct sockaddr *) &cliAddr, sizeof(cliAddr)) < 0) {
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)fd);
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP bind failed in dpws_udp_ucast_open()", SOAP_UDP_ERROR);
        return SOAP_INVALID_SOCKET;
    }
    return fd;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W */
static int dpws_udp_mcast_open(struct soap *soap, const char *endpoint, const char *host, int port)
{
    int fd;
    unsigned char ttl = 1;
    int sockopt_on = 1;
#ifdef WITH_IPV6
    u_int ifindex, hops = 1;
#endif

    /* create sending UDP socket */
    fd = dpws_udp_ucast_open(soap, endpoint, host, port);
    if (fd < 0) {
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP socket creation failed in dpws_udp_mcast_open()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to open ucast socket in dpws_udp_mcast_open().(strerror=%s)\n", strerror(soap->errnum));
        return SOAP_INVALID_SOCKET;
    }

    if(dpws_soap2dpws(soap)->ipproto_type == DPWS_IPPROTO_IPV4){

        if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL,
                      &ttl, sizeof(ttl)) < 0) {
            soap->errnum = soap_socket_errno;
            closesocket((SOAP_SOCKET)fd);
            soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                  "UDP setsockopt IP_MULTICAST_TTL failed in dpws_udp_mcast_open()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IP_MULTICAST_TTL on udp socket in dpws_udp_mcast_open().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
    }
#ifdef WITH_IPV6
    else if(dpws_soap2dpws(soap)->ipproto_type == DPWS_IPPROTO_IPV6){

        ifindex = if_nametoindex(NA_IF);
        if(setsockopt(fd, IPPROTO_IPV6, IPV6_MULTICAST_IF,
                      &ifindex, sizeof(ifindex)) < 0) {
            soap->errnum = soap_socket_errno;
            closesocket((SOAP_SOCKET)fd);
            soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                  "UDP setsockopt IPV6_MULTICAST_IF failed in dpws_udp_mcast_open()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_MULTICAST_IF on udp socket in dpws_udp_mcast_open().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }

        if(setsockopt(fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
                      &hops, sizeof(hops)) < 0){
            soap->errnum = soap_socket_errno;
            closesocket((SOAP_SOCKET)fd);
            soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                  "UDP setsockopt IPV6_MULTICAST_HOPS failed in dpws_udp_mcast_open()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_MULTICAST_HOPS on udp socket in dpws_udp_mcast_open().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
    }
#endif
    else{
        return SOAP_INVALID_SOCKET;
    }

    return fd;
}
#else
static int dpws_udp_mcast_open(struct soap *soap, const char *endpoint, const char *host, int port)
{
    int fd;
#if (defined(TR_STACK_NEXGEN) || defined(UNDER_CE))
    unsigned int ttl = 1;
#else
    unsigned char ttl = 1;
#endif
    
    /* create sending UDP socket */
    fd = dpws_udp_ucast_open(soap, endpoint, host, port);
    if (fd < 0) { 
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP socket creation failed in dpws_udp_mcast_open()", SOAP_UDP_ERROR);
        return SOAP_INVALID_SOCKET;
    }
        
    if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0) {
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)fd);
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP setsockopt IP_MULTICAST_TTL failed in dpws_udp_mcast_open()", SOAP_UDP_ERROR);
        return SOAP_INVALID_SOCKET;
    }

    return fd;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W */
#ifdef WITH_IPV6
int send_wakeup_datagram_v6(void)
{ 
    struct sockaddr_un cliAddr, servAddr;
    register int s = SOAP_INVALID_SOCKET;
    int ret = DPWS_OK;

    /* create socket */
    s = (int)socket(AF_LOCAL, SOCK_DGRAM, 0);
    if (s < 0) {
        ret = DPWS_ERR_COULD_NOT_CREATE_SOCKET;
        goto exit;
    }
    
    /* prepare addresses */
    memset(&cliAddr, 0, sizeof(cliAddr));
    cliAddr.sun_family = AF_LOCAL;
    strcpy(cliAddr.sun_path, tk_config.wakeup_sock_cli);
    unlink(tk_config.wakeup_sock_cli);
    if(bind(s, (struct sockaddr*)&cliAddr, sizeof(cliAddr)) < 0){
        closesocket(s);
        ONVIF_LOG2(ONVIF_LOG_ERROR, "[ERROR]: Failed to bind unix domain socket.(sockpath=%s, strerror=%s)\n", tk_config.wakeup_sock_cli, strerror(errno));
        return DPWS_ERR_COULD_NOT_BIND_SOCKET;
    }
 
    /* send for v6 */
    memset(&servAddr, 0, sizeof(servAddr)); 
    servAddr.sun_family = AF_LOCAL;
    strcpy(servAddr.sun_path, tk_config.wakeup_sock_svr_v6);
    if(sendto(s, WAKE_UP_DATAGRAM, sizeof(WAKE_UP_DATAGRAM), 0,
              (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0) {
        ret = DPWS_ERR_COULD_NOT_SEND_DATA;
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to send wakeup datagram to wakeup_sock_v6.(strerror=%s)\n", strerror(errno));
    }
exit:
    closesocket((SOAP_SOCKET)s);

    return ret;
}
#endif /* WITH_IPV6 */

int send_wakeup_datagram(void)
{ 
    struct sockaddr_un cliAddr, servAddr;
    register int s = SOAP_INVALID_SOCKET;
    int ret = DPWS_OK;

    /* create socket */
    s = (int)socket(AF_LOCAL, SOCK_DGRAM, 0);
    if (s < 0) {
        ret = DPWS_ERR_COULD_NOT_CREATE_SOCKET;
        goto exit;
    }
    
    /* prepare addresses */
    memset(&cliAddr, 0, sizeof(cliAddr));
    cliAddr.sun_family = AF_LOCAL;
    strcpy(cliAddr.sun_path, tk_config.wakeup_sock_cli);
    unlink(tk_config.wakeup_sock_cli);
    if(bind(s, (struct sockaddr*)&cliAddr, sizeof(cliAddr)) < 0){
        closesocket(s);
        ONVIF_LOG2(ONVIF_LOG_ERROR, "[ERROR]: Failed to bind unix domain socket.(sockpath=%s, strerror=%s)\n", tk_config.wakeup_sock_cli, strerror(errno));
        return DPWS_ERR_COULD_NOT_BIND_SOCKET;
    }
 
    /* send for v4 */
    memset(&servAddr, 0, sizeof(servAddr)); 
    servAddr.sun_family = AF_LOCAL;
    strcpy(servAddr.sun_path, tk_config.wakeup_sock_svr_v4);
    if(sendto(s, WAKE_UP_DATAGRAM, sizeof(WAKE_UP_DATAGRAM), 0,
              (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0) {
        ret = DPWS_ERR_COULD_NOT_SEND_DATA;
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to send wakeup datagram to wakeup_sock_v4.(strerror=%s)\n", strerror(errno));
    }
exit:
    closesocket((SOAP_SOCKET)s);

    return ret;
}
#else /* !USE_ONVIF */
int send_wakeup_datagram()
{ 
    struct sockaddr_in sockaddr;
    register int s = SOAP_INVALID_SOCKET;
    int ret = DPWS_OK;
    
    /* create socket */
    s = (int)socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
        ret = DPWS_ERR_COULD_NOT_CREATE_SOCKET;
        goto exit;
    }
    
    /* prepare addresses */
    memset(&sockaddr, 0, sizeof(sockaddr));
    sockaddr.sin_family = AF_INET;
    dpws_gethost(NULL, "127.0.0.1", &sockaddr.sin_addr);    // soap struct should not be used and no error should occur since no DNS api should be called
    sockaddr.sin_port = htons((short)listeners->port);
    
    if (sendto(s, WAKE_UP_DATAGRAM, sizeof(WAKE_UP_DATAGRAM), 0, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0) {
        ret = DPWS_ERR_COULD_NOT_SEND_DATA;
        goto exit;
    }
    
exit:
    closesocket((SOAP_SOCKET)s);
    return ret;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W */
static int dpws_discovery_bind(struct dpws* dpws)
{
    int fd;
    struct soap *soap = dpws_dpws2soap(dpws);
    int sockopt_on = 1;
    struct addrinfo hints, *res, *res1;
    struct ip_mreq mreq;
#ifdef WITH_IPV6
    struct ipv6_mreq mreq_v6;
#endif
    dpws_ipproto_type ipproto = dpws->ipproto_type;
    char *mcast_addr = NULL;
    int err = 0;

    /* create multicast listening UDP socket */
    dpws->wsd_ls = SOAP_INVALID_SOCKET;
    soap->errmode = 0;

    memset(&hints, 0, sizeof(hints));
    if(ipproto == DPWS_IPPROTO_IPV4){
        hints.ai_family = AF_INET;
    }
#ifdef WITH_IPV6
    else if(ipproto == DPWS_IPPROTO_IPV6){
        hints.ai_family = AF_INET6;
    }
#endif
    else{
        return SOAP_INVALID_SOCKET;
    }
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_flags = AI_PASSIVE;

    err = getaddrinfo(NULL, DISCOVERY_PORT_STR, &hints, &res);
    if(err){
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to get addrinfo(NULL) in dpws_discovery_bind().(gai_strerror=%s)\n", gai_strerror(err));
        return SOAP_INVALID_SOCKET;
    }

    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if(fd < 0){ 
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                "UDP socket creation failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to create udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
        return SOAP_INVALID_SOCKET;
    }    
    if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &sockopt_on, sizeof(sockopt_on)) < 0 ) {
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                "Setting reuse address on UDP socket failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set REUSEADDR on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
        return dpws->err = DPWS_ERR_COULD_NOT_SET_REUSE_ADDR_OPTION;
    }

    if(ipproto == DPWS_IPPROTO_IPV4){
        if(setsockopt(fd, IPPROTO_IP, IP_PKTINFO, &sockopt_on, sizeof(sockopt_on))< 0){
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ip pktinfo on UDP socket failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IP_PKTINFO on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
    }
#ifdef WITH_IPV6
    else if(ipproto == DPWS_IPPROTO_IPV6){
#ifdef IPV6_V6ONLY
        if (setsockopt(fd, IPPROTO_IPV6,IPV6_V6ONLY, &sockopt_on, sizeof(sockopt_on)) < 0) {
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ipv6 v6only on UDP socket failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_V6ONLY on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
#endif

#if defined IPV6_RECVPKTINFO
        if(setsockopt(fd, IPPROTO_IPV6, IPV6_RECVPKTINFO, &sockopt_on, sizeof(sockopt_on))< 0){
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ipv6 recvpktinfo on UDP socket failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_RECVPKTINFO on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
#elif defined IPV6_PKTINFO
        if(setsockopt(fd, IPPROTO_IPV6, IPV6_PKTINFO, &sockopt_on, sizeof(sockopt_on))< 0){
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws),
                                    "Setting ipv6 pktinfo on UDP socket failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_PKTINFO on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_INVALID_SOCKET;
        }
#else
# error "can't determine socket option"
#endif
    }
#endif

    /* bind socket */
    if(bind(fd, (struct sockaddr *) res->ai_addr, res->ai_addrlen) < 0){
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)fd);
        soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                "UDP bind failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to bind udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
        return SOAP_INVALID_SOCKET;
    }

    /* set multicat options */
    memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_DGRAM;

    if(ipproto == DPWS_IPPROTO_IPV4){
        mcast_addr = DISCOVERY_MULTICAST_ADDR;
    }
#ifdef WITH_IPV6
    else if(ipproto == DPWS_IPPROTO_IPV6){
        mcast_addr = DISCOVERY_MULTICAST_ADDR_V6;
    }
#endif

    err = getaddrinfo(mcast_addr, DISCOVERY_PORT_STR, &hints, &res1);
    if(err){
        ONVIF_LOG2(ONVIF_LOG_ERROR, "[ERROR]: Failed to get addrinfo(%s) in dpws_discovery_bind().(gai_strerror=%s)\n", mcast_addr, gai_strerror(err));
        return SOAP_INVALID_SOCKET;
    }

    if(ipproto == DPWS_IPPROTO_IPV4){
        memcpy(&mreq.imr_multiaddr, &((struct sockaddr_in*)res1->ai_addr)->sin_addr, sizeof(struct in_addr));
        inet_aton(tk_config.network_adapter.ip_addr, &mreq.imr_interface);
        if(setsockopt(fd, IPPROTO_IP,IP_ADD_MEMBERSHIP, (void *) &mreq, sizeof(mreq)) < 0) {
            soap->errnum = soap_socket_errno;
            closesocket((SOAP_SOCKET)fd);
            soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                    "UDP join multicast group failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IP_ADD_MEMBERSHIP on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
            return dpws->err = DPWS_ERR_COULD_NOT_JOIN_MULTICAST_GROUP;
        }
    }
#ifdef WITH_IPV6
    else if(ipproto == DPWS_IPPROTO_IPV6){
        memcpy(&mreq_v6.ipv6mr_multiaddr, &((struct sockaddr_in6*)res1->ai_addr)->sin6_addr, sizeof(struct in6_addr));
        mreq_v6.ipv6mr_interface = if_nametoindex(NA_IF);
        if(setsockopt(fd, IPPROTO_IPV6,IPV6_ADD_MEMBERSHIP, (void *) &mreq_v6, sizeof(mreq_v6)) < 0) {
            soap->errnum = soap_socket_errno;
            closesocket((SOAP_SOCKET)fd);
            soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)),
                                    "UDP join ipv6 multicast group failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to set IPV6_ADD_MEMBERSHIP on udp socket in dpws_discovery_bind().(strerror=%s)\n", strerror(soap->errnum));
            return dpws->err = DPWS_ERR_COULD_NOT_JOIN_MULTICAST_GROUP;
        }
    }
#endif

    dpws->wsd_ls = soap->socket = fd;

    freeaddrinfo(res);
    freeaddrinfo(res1);

    ONVIF_LOG1(ONVIF_LOG_INFO,
               "[INFO]: Succeeded in binding discovery listen socket=%d.\n",
               dpws->wsd_ls);

    return SOAP_OK;
}
#else
static int dpws_discovery_bind(struct dpws* dpws) {
    register int fd;
    struct ip_mreq mreq;
    struct in_addr mcastAddr;
    struct sockaddr_in servAddr;
    struct soap *soap = dpws_dpws2soap(dpws);
    int reuse = 1;
    int size = sizeof(reuse);
    
    /* create multicast listening UDP socket */
    dpws->wsd_ls = SOAP_INVALID_SOCKET;
    soap->errmode = 0;
    fd = (int)socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) { 
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws), "UDP socket creation failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
        return dpws->err = DPWS_ERR_COULD_NOT_CREATE_SOCKET;
    }
    if( setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&reuse, size) < 0 ) {
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws), "Setting reuse address on UDP socket failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
        return dpws->err = DPWS_ERR_COULD_NOT_SET_REUSE_ADDR_OPTION;
    }
    
    /* bind multicast listening UDP socket */
    servAddr.sin_family=AF_INET;
    servAddr.sin_addr.s_addr=htonl(INADDR_ANY);
    servAddr.sin_port=htons(DISCOVERY_PORT);
    if (bind(fd, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0) {
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)fd);
        soap_set_receiver_error(soap, dpws_ip_error(dpws), "Could not bind UDP socket in dpws_discovery_bind()", SOAP_UDP_ERROR);
        return dpws->err = DPWS_ERR_COULD_NOT_BIND_SOCKET;
    }

#ifdef WIN32
    mcastAddr.s_addr = inet_addr(DISCOVERY_MULTICAST_ADDR);
#else
    inet_aton( DISCOVERY_MULTICAST_ADDR, &mcastAddr);
#endif
    mreq.imr_multiaddr.s_addr=mcastAddr.s_addr;
#ifdef TR_STACK_NEXGEN
    inet_aton(tk_config.ip_addr, &mreq.imr_interface.s_addr);
#else
    mreq.imr_interface.s_addr=htonl(INADDR_ANY);
#endif

    if (setsockopt(fd,IPPROTO_IP,IP_ADD_MEMBERSHIP, (void *) &mreq, sizeof(mreq)) < 0) {
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)fd);
        soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "UDP join multicast group failed in dpws_discovery_bind()", SOAP_UDP_ERROR);
        return dpws->err = DPWS_ERR_COULD_NOT_JOIN_MULTICAST_GROUP;
    }
    
    dpws->wsd_ls = soap->socket = fd;

    return SOAP_OK;
}
#endif

int dpws_udp_close(struct soap *soap) {
#ifdef USE_DISCOVERY_PORT_AS_SOURCE_PORT /* Added by HG.W */
    if(soap->socket == dpws_soap2dpws(soap)->wsd_ls){
        return SOAP_OK;
    }
#endif

    if (soap_valid_socket(soap->socket))
    { 
        closesocket((SOAP_SOCKET)soap->socket);
        soap->socket = SOAP_INVALID_SOCKET;
    }
    return SOAP_OK;
}

#ifdef USE_ONVIF /* Added by HG.W */
#define MAX_PORT_LEN (32)
static int dpws_udp_send(struct soap *soap, const char *s, size_t n)
{
    char *host = soap->host;
    char port_str[MAX_PORT_LEN];
    int fd = soap->socket;
    struct sockaddr_in sockaddr;
    struct addrinfo hints, *res;
    int ret = SOAP_OK;
    int err = 0;

    if (!soap_valid_socket(soap->socket)) {
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "invalid socket in udp_send()", SOAP_UDP_ERROR);
        closesocket((SOAP_SOCKET)fd);
        return SOAP_INVALID_SOCKET;
    }

    memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_DGRAM;

    snprintf(port_str, MAX_PORT_LEN, "%d", soap->port);

    err = getaddrinfo(host, port_str, &hints, &res);
    if(err){
        ONVIF_LOG3(ONVIF_LOG_ERROR, "[ERROR]: Failed to get addrinfo([%s]:%s) in dpws_udp_send().(gai_strerror=%s)\n",soap->host,port_str,gai_strerror(err));
        return SOAP_INVALID_SOCKET;
    }

    soap->errmode = 0;
    if (sendto(fd, (char *)s, n, 0, res->ai_addr, res->ai_addrlen) < 0) {
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "Send to failed in dpws_udp_send()", SOAP_UDP_ERROR);
        ONVIF_LOG3(ONVIF_LOG_ERROR, "[ERROR]: Failed to sendto [%s]:%s in dpws_udp_send().(strerror=%s)\n", host, port_str, strerror(soap->errnum));

#ifdef USE_DISCOVERY_PORT_AS_SOURCE_PORT
        if(fd == dpws_soap2dpws(soap)->wsd_ls){
            ret = SOAP_UDP_ERROR;
        }
        else{
            closesocket((SOAP_SOCKET)fd);
            ret = SOAP_INVALID_SOCKET;
        }
#else
        closesocket((SOAP_SOCKET)fd);
        ret = SOAP_INVALID_SOCKET;
#endif
    }

    freeaddrinfo(res);

    return ret;
}
#else
static int dpws_udp_send(struct soap *soap, const char *s, size_t n)
{
    const char *host = soap->host;
    int port = soap->port;
    int fd = soap->socket;
    struct sockaddr_in sockaddr;
    
    if (!soap_valid_socket(soap->socket)) {
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "invalid socket in udp_send()", SOAP_TCP_ERROR);
        closesocket((SOAP_SOCKET)fd);
        return SOAP_INVALID_SOCKET;
    }
    memset(&sockaddr, 0, sizeof(sockaddr));
    sockaddr.sin_family = AF_INET;
    DBGLOG(TEST,SOAP_MESSAGE(fdebug, "Open socket %d to host='%s'\n", fd, host));
    soap->errmode = 2;
    if (dpws_gethost(dpws_soap2dpws(soap), host, &sockaddr.sin_addr)) { 
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "get host by name failed in udp_send()", SOAP_TCP_ERROR);
        closesocket((SOAP_SOCKET)fd);
        return SOAP_INVALID_SOCKET;
    }
    sockaddr.sin_port = htons((short)port);
    soap->errmode = 0;
    if (sendto(fd, (char *)s, n, 0, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) < 0) {
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "send to failed in udp_send()", SOAP_TCP_ERROR);
        closesocket((SOAP_SOCKET)fd);
        return SOAP_INVALID_SOCKET;
    }
//  soap_set_endpoint(soap, NULL);  // QUESTION: was useful ?
    return SOAP_OK;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W (to get udp destination address) */
#ifdef WITH_IPV6
my_recvfrom_v6(int fd, void *buf, size_t len, int flags,
            struct sockaddr_in6 *srcaddr, socklen_t *srcaddrlen, 
            struct in6_addr *dstaddr)
{
#if defined IPV6_PKTINFO
# define DSTADDR_SOCKOPT_V6 IPV6_PKTINFO
# define DSTADDR_DATASIZE_V6 (CMSG_SPACE(sizeof(struct in6_pktinfo)))
# define _dstaddr_v6(x) (&(((struct in6_pktinfo *)(CMSG_DATA(x)))->ipi6_addr))
#else
# error "can't determine socket option"
#endif

    struct msghdr msg;
    struct iovec iov[1];
    ssize_t n;

    struct cmsghdr *cmptr;
    union{
        struct cmsghdr cm;
        u_char control[DSTADDR_DATASIZE_V6];
    }control_un;

    msg.msg_control = control_un.control;
    msg.msg_controllen = sizeof(control_un.control);
    msg.msg_flags = 0;

    msg.msg_name = srcaddr;
    msg.msg_namelen = *srcaddrlen;
    iov[0].iov_base = buf;
    iov[0].iov_len = len;
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;

    if((n = recvmsg(fd, &msg, flags)) < 0)
        return (n);

    *srcaddrlen = msg.msg_namelen;

    for(cmptr = CMSG_FIRSTHDR(&msg); cmptr != NULL;
        cmptr = CMSG_NXTHDR(&msg, cmptr)){
        if(cmptr->cmsg_level == IPPROTO_IPV6 &&
           cmptr->cmsg_type == DSTADDR_SOCKOPT_V6) {
            if(dstaddr)
                memcpy(dstaddr,_dstaddr_v6(cmptr), sizeof(struct in6_addr));
        }
    }

    return (n);
}
#endif

static ssize_t
my_recvfrom(int fd, void *buf, size_t len, int flags,
            struct sockaddr *srcaddr, socklen_t *srcaddrlen, 
            struct in_addr *dstaddr)
{
#if defined IP_PKTINFO
# define DSTADDR_SOCKOPT IP_PKTINFO
# define DSTADDR_DATASIZE (CMSG_SPACE(sizeof(struct in_pktinfo)))
# define _dstaddr(x) (&(((struct in_pktinfo *)(CMSG_DATA(x)))->ipi_addr))
#else
# error "can't determine socket option"
#endif

    struct msghdr msg;
    struct iovec iov[1];
    ssize_t n;

    struct cmsghdr *cmptr;
    union{
        struct cmsghdr cm;
        u_char control[DSTADDR_DATASIZE];
    }control_un;

    msg.msg_control = control_un.control;
    msg.msg_controllen = sizeof(control_un.control);
    msg.msg_flags = 0;

    msg.msg_name = srcaddr;
    msg.msg_namelen = *srcaddrlen;
    iov[0].iov_base = buf;
    iov[0].iov_len = len;
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;

    if((n = recvmsg(fd, &msg, flags)) < 0)
        return (n);

    *srcaddrlen = msg.msg_namelen;

    for(cmptr = CMSG_FIRSTHDR(&msg); cmptr != NULL;
        cmptr = CMSG_NXTHDR(&msg, cmptr)){
        if(cmptr->cmsg_level == IPPROTO_IP &&
           cmptr->cmsg_type == DSTADDR_SOCKOPT) {
            if(dstaddr)
                memcpy(dstaddr,_dstaddr(cmptr), sizeof(struct in_addr));
        }
    }

    return (n);
}
#endif

#ifdef USE_ONVIF /* Added by HG.W */
static size_t dpws_udp_recv(struct soap *soap, char *s, size_t n)
{
    struct sockaddr_in sockaddr;
    struct in_addr dstaddr;
    char dst_str[INET_ADDRSTRLEN+1], src_str[INET_ADDRSTRLEN+1];
#ifdef WITH_IPV6
    struct sockaddr_in6 sockaddr_v6;
    struct in6_addr dstaddr_v6;
    char dst_str_v6[INET6_ADDRSTRLEN+1], src_str_v6[INET6_ADDRSTRLEN+1];
#endif
    int len;
    socklen_t addrLen;
    char *host;
    int r;

    if (!soap_valid_socket(soap->socket)) {
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "invalid socket in udp_recv()", SOAP_TCP_ERROR);
        closesocket((SOAP_SOCKET)soap->socket);
        return 0;
    }
    
    if (dpws_soap2dpws(soap)->schedule)
    {
        struct timeval t_out;
        fd_set fd;
        
        get_timeval_schedule(dpws_soap2dpws(soap)->schedule, &t_out);

        FD_ZERO(&fd);
        FD_SET((SOAP_SOCKET)soap->socket, &fd);
        
        r = select((SOAP_SOCKET)(soap->socket + 1), &fd, NULL, NULL, &t_out);

        if (r == 0) {
            soap->errnum = DPWS_ERR_UDP_TIMEOUT;
            return 0;
        }
        else if (r < 0) {
            soap->errnum = soap_socket_errno;
            soap_closesock(soap);
            return 0;
        }
    }

    if(dpws_soap2dpws(soap)->ipproto_type == DPWS_IPPROTO_IPV4){
        memset(&dstaddr, 0, sizeof(struct in_addr));
        addrLen=sizeof(sockaddr);
        len = my_recvfrom(soap->socket, s, n, 0, (struct sockaddr *)&sockaddr,
                          &addrLen, &dstaddr);
        inet_ntop(AF_INET, &dstaddr, dst_str, INET_ADDRSTRLEN);
        inet_ntop(AF_INET, &((struct sockaddr_in*)&sockaddr)->sin_addr, src_str, INET_ADDRSTRLEN);

        if(len > 0){
            /* check s[0] == '<' */
            if(s[0] != '<'){
                dpws_soap2dpws(soap)->err = WSD_ERR_NON_XML_DATA;
                ONVIF_LOG3(ONVIF_LOG_ERROR, "[ERROR]: Got NON-XML udp data on socket=%d.(from %s to %s)\n", soap->socket, src_str, dst_str);
                return 0;
            }
            /* check multicast/unicast */
            if(!strcmp(dst_str, DISCOVERY_MULTICAST_ADDR)){
                ONVIF_LOG3(ONVIF_LOG_INFO, "[INFO]: Received udp multicast on socket=%d.(from %s to %s)\n", soap->socket, src_str, dst_str);
                DPWS_SET_UDP_MULTICAST(dpws_soap2dpws(soap)->status);
            }
            else if(!strcmp(dst_str, tk_config.network_adapter.ip_addr)){
                ONVIF_LOG3(ONVIF_LOG_INFO, "[INFO]: Received udp unicast on socket=%d.(from %s to %s)\n", soap->socket, src_str, dst_str);
                DPWS_SET_UDP(dpws_soap2dpws(soap)->status);
                DPWS_SET_UNICAST(dpws_soap2dpws(soap)->status);
            }
            else{
                dpws_soap2dpws(soap)->err = WSD_ERR_UNKNOWN_UDP_DESTINATION_ADDRESS;
                ONVIF_LOG3(ONVIF_LOG_ERROR, "[ERROR]: Received unexpected udp destination on socket=%d.(from %s to %s)\n", soap->socket, src_str, dst_str);
                return 0;
            }

            soap->port = ntohs(sockaddr.sin_port);
            strcpy(soap->host, src_str);
        }
        
        return len;
    }
#ifdef WITH_IPV6
    else if(dpws_soap2dpws(soap)->ipproto_type == DPWS_IPPROTO_IPV6){
        memset(&dstaddr_v6, 0, sizeof(struct in6_addr));
        addrLen=sizeof(sockaddr_v6);
        len = my_recvfrom_v6(soap->socket, s, n, 0, (struct sockaddr_in6 *)&sockaddr_v6,
                             &addrLen, &dstaddr_v6);

        inet_ntop(AF_INET6, &dstaddr_v6, dst_str_v6, INET6_ADDRSTRLEN);
        inet_ntop(AF_INET6, &((struct sockaddr_in6 *)&sockaddr_v6)->sin6_addr, src_str_v6, INET6_ADDRSTRLEN);

        if(len > 0){
            /* check s[0] == '<' */
            if(s[0] != '<'){
                dpws_soap2dpws(soap)->err = WSD_ERR_NON_XML_DATA;
                ONVIF_LOG3(ONVIF_LOG_ERROR, "[ERROR]: Got NON-XML udp data on socket=%d.(from %s to %s)\n", soap->socket, src_str_v6, dst_str_v6);
                return 0;
            }
            /* check multicast/unicast */
            if(!strcasecmp(dst_str_v6, DISCOVERY_MULTICAST_ADDR_V6)){
                ONVIF_LOG3(ONVIF_LOG_INFO, "[INFO]: Received udp multicast on socket=%d.(from %s to %s)\n", soap->socket, src_str_v6, dst_str_v6);
                DPWS_SET_UDP_MULTICAST(dpws_soap2dpws(soap)->status);
                /* set xaddr ipv6 scope (link-local) */
                dpws_soap2dpws(soap)->xaddr_ipv6_scope = DPWS_XADDR_IPV6_SCOPE_LINKLOCAL;
            }
            else if(tk_config.network_adapter.ip_addr_v6_l &&
                    !strcasecmp(dst_str_v6, tk_config.network_adapter.ip_addr_v6_l)){
                ONVIF_LOG3(ONVIF_LOG_INFO, "[INFO]: Received udp unicast on socket=%d.(from %s to %s)\n", soap->socket, src_str_v6, dst_str_v6);
                DPWS_SET_UDP(dpws_soap2dpws(soap)->status);
                DPWS_SET_UNICAST(dpws_soap2dpws(soap)->status);
                /* set xaddr ipv6 scope (link-local) */
                dpws_soap2dpws(soap)->xaddr_ipv6_scope = DPWS_XADDR_IPV6_SCOPE_LINKLOCAL;
            }
            else if(tk_config.network_adapter.ip_addr_v6_g &&
                    !strcasecmp(dst_str_v6, tk_config.network_adapter.ip_addr_v6_g)){
                ONVIF_LOG3(ONVIF_LOG_INFO, "[INFO]: Received udp unicast on socket=%d.(from %s to %s)\n", soap->socket, src_str_v6, dst_str_v6);
                DPWS_SET_UDP(dpws_soap2dpws(soap)->status);
                DPWS_SET_UNICAST(dpws_soap2dpws(soap)->status);
                /* set xaddr ipv6 scope (global) */
                dpws_soap2dpws(soap)->xaddr_ipv6_scope = DPWS_XADDR_IPV6_SCOPE_GLOBAL;
            }
            else{
                dpws_soap2dpws(soap)->err = WSD_ERR_UNKNOWN_UDP_DESTINATION_ADDRESS;
                ONVIF_LOG3(ONVIF_LOG_ERROR, "[ERROR]: Received unexpected udp destination on socket=%d.(from %s to %s)\n", soap->socket, src_str_v6, dst_str_v6);
                return 0;
            }

            soap->port = ntohs(sockaddr_v6.sin6_port);
            strcpy(soap->host, src_str_v6);
        }
        
        return len;
     }
#endif /* WITH_IPV6 */
    else{
        /* unknown ipproto */
        return 0;
    }

}
#else /* !USE_ONVIF */
static size_t dpws_udp_recv(struct soap *soap, char *s, size_t n)
{
    struct sockaddr_in sockaddr;

    int len;
    socklen_t addrLen;
    char *host;
    int r;

    if (!soap_valid_socket(soap->socket)) {
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws_soap2dpws(soap)), "invalid socket in udp_recv()", SOAP_TCP_ERROR);
        closesocket((SOAP_SOCKET)soap->socket);
        return 0;
    }
    
    if (dpws_soap2dpws(soap)->schedule)
    {
        struct timeval t_out;
        fd_set fd;
        
        get_timeval_schedule(dpws_soap2dpws(soap)->schedule, &t_out);

        FD_ZERO(&fd);
        FD_SET((SOAP_SOCKET)soap->socket, &fd);
        
#ifdef TR_STACK_NEXGEN
        r = selectsocket(soap->socket + 1, &fd, NULL, NULL, &t_out);
#else
        r = select((SOAP_SOCKET)(soap->socket + 1), &fd, NULL, NULL, &t_out);
#endif
        if (r == 0) {
            soap->errnum = DPWS_ERR_UDP_TIMEOUT;
            return 0;
        }
        else if (r < 0) {
            soap->errnum = soap_socket_errno;
            soap_closesock(soap);
            return 0;
        }
    }
    addrLen=sizeof(sockaddr);

    len = recvfrom(soap->socket, s, n, 0, (struct sockaddr *) &sockaddr, &addrLen);
    host = inet_ntoa(sockaddr.sin_addr);
    soap->port = ntohs(sockaddr.sin_port);
    strcpy(soap->host, host);

    // QUESTION: loop if msg oversize the buffer ?
    return len;
}
#endif /* USE_ONVIF */

void dpws_set_udp_mode(struct dpws * dpws)
{
    dpws_dpws2soap(dpws)->fopen = dpws_udp_mcast_open;
    dpws_dpws2soap(dpws)->fclose = dpws_udp_close;
    dpws_dpws2soap(dpws)->fsend = dpws_udp_send;
    dpws_dpws2soap(dpws)->frecv = dpws_udp_recv;
}

static struct device_proxy* lookup_by_id(struct dpws* dpws, char * id, long time_out)
{
    /* retrieve from cache 1st */
    struct device_proxy * cache_entry = find_device_proxy(id);
    
    /* send resolve */
    if (!cache_entry || cache_entry->transport_addresses.nb == 0)
    {
        short send_nb = 0;
        long delay = 0;
        scheduler_t resolveScheduler = {NULL, NULL};
        struct wsa_endpoint_ref discovery_endpoint;
        char uuid[SCHEME_UUID_STRING_SIZE];
        
        dpws_default_wsa_endpoint_ref(&discovery_endpoint);
        discovery_endpoint.address = id;
        
        dpws_dpws2soap(dpws)->omode |= SOAP_IO_KEEPALIVE;   // to avoid the socket be closed by the connect
        dpws_set_udp_mode(dpws);
        while ((delay = get_soap_over_udp_schedule(DPWS_MSG_RESOLVE, &send_nb, delay)) != 0xFFFFFFFF)
        {
            if (delay == 0) // normal for a resolve
            {
                dpws->message_id = schemed_uuid_create(uuid);
                if (send_resolve(dpws, &discovery_endpoint))
                    goto exit;
                dpws_end(dpws);
            }
            else
                add_schedule(&resolveScheduler, delay, (schedule_cbk)send_resolve, &discovery_endpoint);
        }
        
        add_schedule(&resolveScheduler, MIN(time_out, MATCH_TIMEOUT), NULL, NULL);

        /* receive the response until the endpoint is in the cache */
        dpws_dpws2soap(dpws)->imode |= SOAP_IO_KEEPALIVE;   // to avoid the socket be closed by the connect
        while (!cache_entry || cache_entry->transport_addresses.nb == 0)
        {
            dpws->schedule = get_schedule(&resolveScheduler);
            DPWS_SET_RCV(dpws->status);
            dpws_init_wsa_headers(dpws);
            if (!wsd_ucast_recv_resolvematch(dpws)) {
                if (dpws->schedule)
                    reschedule(&resolveScheduler, dpws->schedule);
                wsd_process(dpws);
            }
            else if (dpws_dpws2soap(dpws)->errnum == DPWS_ERR_UDP_TIMEOUT)
            {
                if (dpws->schedule->handler == NULL) {  // user timeout
                    dpws_dpws2soap(dpws)->errnum = 0;
                    dpws_dpws2soap(dpws)->error = WSA_ERR_DESTINATION_UNREACHABLE;
                    realize_schedule(&resolveScheduler, dpws, -1);  // recycle
                    goto exit;
                }
                dpws->message_id = uuid;
                realize_schedule(&resolveScheduler, dpws, -1);  // resolve send
            }
            dpws_end(dpws);
            dpws_dpws2soap(dpws)->keep_alive = 0;
            cache_entry = find_device_proxy(id);
        }
        // else look in the cache anyway
        dpws_end(dpws);
        dpws_udp_close(dpws_dpws2soap(dpws));   // NOTE: dpws_end() calls the dummy close
        free_scheduler(&resolveScheduler);
    }
    dpws_init_wsa_headers(dpws);
    
exit:
    dpws_dpws2soap(dpws)->omode &= ~SOAP_IO_KEEPALIVE;
    dpws_dpws2soap(dpws)->imode &= ~SOAP_IO_KEEPALIVE;
    return cache_entry;
}

static struct device_proxy * fill_device_services(struct dpws* dpws, short hrefDevice)
{
    struct device_proxy * device = NULL;
    BOOLEAN isDevice;
    
    // find the cache entry
    lock(cache_lock_handle);
        
    // perform a getMetadata if hosted services were not retrieved yet
    device = get_device_proxy(hrefDevice, &isDevice);

    if (!device) {
        dpws->err = DPWS_ERR_INVALID_PROXY;
        return NULL;
    }
    if (!device->hosted_retrieved) {
        unlock(cache_lock_handle);
        if (get_endpoint_metadata(dpws, hrefDevice, device, TRUE, NULL, NULL, NULL, NULL, NULL))
            return NULL;
        lock(cache_lock_handle);
    }
    // do not UNLOCK (performed by the caller)
    return device;
}

static struct dpws* dpws_copy_context(struct dpws* dest, struct dpws* src) {
    soap_copy_context(&dest->soap, &src->soap);
    memcpy((char *)dest + sizeof(dest->soap), (char *)src + sizeof(src->soap), sizeof(struct dpws) - sizeof(src->soap));
    return dest;
}

#ifdef USE_ONVIF /* Added by HG.W */
static int accept_msg(struct dpws* dpws, struct dpws* dpws_s)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    struct sockaddr_in sockaddr;
    struct timeval timeout, *ptimeout = NULL;
    fd_set fdset;
    int n = (int)sizeof(struct sockaddr), maxsock;
#ifndef WITH_LEAN
    int len = SOAP_BUFLEN;
    int set = 1;
#endif

    scheduler_t *tmp_scheduler = NULL;
    short tmp_scheduler_lock_handle = -1;
    BOOLEAN *tmp_server_up = NULL;

    if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
        tmp_scheduler = &scheduler;
        tmp_scheduler_lock_handle = scheduler_lock_handle;
        tmp_server_up = &tk_config.server_up;
    }
#ifdef WITH_IPV6
    else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
        tmp_scheduler = &scheduler_v6;
        tmp_scheduler_lock_handle = scheduler_lock_handle_v6;
        tmp_server_up = &tk_config.server_up_v6;
    }
#endif
    else{
        ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Unknown ipproto.\n");
        return DPWS_OK;
    }

    memset(&sockaddr, 0, sizeof(sockaddr));
    soap->socket = SOAP_INVALID_SOCKET;
    soap->errmode = 0;
    
    if (soap->accept_timeout)   // dummy event for user timeout
    {
        lock(tmp_scheduler_lock_handle);
        add_schedule(tmp_scheduler, soap->accept_timeout > 0
                    ? soap->accept_timeout*1000
                    : -soap->accept_timeout/1000,
                    NULL,
                    NULL
                    );
        unlock(tmp_scheduler_lock_handle);
    }
    
    /* poll for the listening sockets */
    FD_ZERO(&fdset);
    if (soap_valid_socket(soap->master))    // manager may not need TCP
        FD_SET((SOAP_SOCKET)soap->master, &fdset);
    FD_SET((SOAP_SOCKET)dpws->wsd_ls, &fdset);
    FD_SET((SOAP_SOCKET)dpws->wsd_ws, &fdset);
    maxsock = MAX(dpws->wsd_ws,MAX(soap->master,dpws->wsd_ls)) + 1;
    
    for (;;)
    {
        int r;

        // if shuting down start closing listening socket but do not exit to empty scheduler
        if (!*tmp_server_up)
        {
            /* close master TCP */
            soap_done(soap);
            
            /* close UDP sockets */
//          dpws_udp_close(soap);   // Needed ? Should have been closed
            if (soap_valid_socket(dpws->wsd_ls))
            { 
                closesocket((SOAP_SOCKET)dpws->wsd_ls);
                dpws->wsd_ls = SOAP_INVALID_SOCKET;
            }
            FD_ZERO(&fdset);
            FD_SET((SOAP_SOCKET)dpws->wsd_ws, &fdset);
            maxsock = dpws->wsd_ws + 1;
        }
        
        lock(tmp_scheduler_lock_handle);
        if ((dpws->schedule = get_schedule(tmp_scheduler))){
            ptimeout = get_timeval_schedule(dpws->schedule, &timeout);
        }
        else if (!*tmp_server_up)
        {
            /* Device is stopping and no message is scheduled */
            unlock(tmp_scheduler_lock_handle);

            /* close wake up UDP sockets */
            if (soap_valid_socket(dpws->wsd_ws))
            { 
                closesocket((SOAP_SOCKET)dpws->wsd_ws);
                dpws->wsd_ws = SOAP_INVALID_SOCKET;
            }
            /* empty scheduler */
            free_scheduler(tmp_scheduler);
            tmp_scheduler = NULL;
            
            *tmp_server_up = TRUE; // once the flag has been used to exit, reset it in order to be able to restart the main loop where stopped (in fact pause the server)
            return dpws->err = DPWS_ERR_SERVER_STOPPED;
        }
        unlock(tmp_scheduler_lock_handle);

        if (dpws->schedule && !ptimeout)
            r = 0;
        else
#ifdef TR_STACK_NEXGEN
            r = selectsocket(maxsock, &fdset, NULL, NULL, ptimeout);
#else
            r = select((SOAP_SOCKET)maxsock, &fdset, NULL, NULL, ptimeout);
#endif

        if (r > 0) {
            lock(tmp_scheduler_lock_handle);
            if (dpws->schedule) {
                reschedule(tmp_scheduler, dpws->schedule);
                dpws->schedule = NULL;
            }
            unlock(tmp_scheduler_lock_handle);

            DPWS_SET_RCV(dpws->status);
            DPWS_RESET_MSG_TYPE(dpws->status);
            
            /* UDP datagram received : schedule wake or discovery */
            if (FD_ISSET(dpws->wsd_ls, &fdset))
            {
                ONVIF_LOG1(ONVIF_LOG_INFO, "\n\n[INFO]: Received data on discovery listen socket=%d.\n", dpws->wsd_ls);

                int ret = SOAP_OK;
                if (dpws_s)
                    dpws_copy_context(dpws_s, dpws);
                else
                    dpws_s = dpws;
                dpws_dpws2soap(dpws_s)->socket = dpws_s->wsd_ls;
                dpws_dpws2soap(dpws_s)->imode |= SOAP_IO_KEEPALIVE; // prevent closing the socket
                dpws_set_udp_mode(dpws_s);
                if ((ret = wsd_mcast_recv(dpws_s))) {
                    if (ret == SOAP_NO_METHOD)
                    {
                        ONVIF_LOG(ONVIF_LOG_ERROR,
                                  "[ERROR]: Unsupported method.\n");
                        dpws_dpws2soap(dpws_s)->keep_alive = 0;
                        /**
                         * gSOAP parsing error so return DPWS_OK
                         * and cause dpws_serve() to end internal state.
                         */
                        return DPWS_OK;
                    }
                    else if(ret == WSA_ERR_REQUIRED_HEADER_INFO){
                        ONVIF_LOG(ONVIF_LOG_ERROR,
                                  "[ERROR]: Bad wsa headers.\n");
                        /**
                         * Non gSOAP parsing error so return error
                         * and cause dpws_end() to clean the env.
                         */
                        return ret;
                    }
                    /* Added by HG.W (to reject broadcast) */
                    else if(ret == WSD_ERR_UNKNOWN_UDP_DESTINATION_ADDRESS){
                        ONVIF_LOG(ONVIF_LOG_ERROR, 
                                  "[ERROR]: Bad udp destination.\n");
                        /**
                         * Non gSOAP parsing error so return error
                         * and cause dpws_end() to clean the env.
                         */
                        return ret;
                    }
                    /* Added by HG.W (to reject non-xml udp data) */
                    else if(ret == WSD_ERR_NON_XML_DATA){
                        ONVIF_LOG(ONVIF_LOG_ERROR, 
                                  "[ERROR]: Bad udp data.(non-xml)\n");
                        /**
                         * Non gSOAP parsing error so return error
                         * and cause dpws_end() to clean the env.
                         */
                        return ret;
                    }
                    else
                    {
                        ONVIF_LOG1(ONVIF_LOG_ERROR, 
                                   "[ERROR]: Unexpected error.(err=%d)\n",ret);
                        dpws_dpws2soap(dpws_s)->keep_alive = 0;
                        /**
                         * gSOAP parsing error so return DPWS_OK
                         * and cause dpws_serve() to end internal state
                         */
                        return DPWS_OK;
                    }
                }
                else
                {
                    dpws_dpws2soap(dpws_s)->keep_alive = 0;
                    return DPWS_OK;
                }
            }
            else if (FD_ISSET(dpws->wsd_ws, &fdset)) { // wakeup
                ONVIF_LOG1(ONVIF_LOG_INFO, "\n\n[INFO]: Received data on wakeup socket=%d.\n", dpws->wsd_ws);

                char buf[sizeof(WAKE_UP_DATAGRAM)];
                recv(dpws->wsd_ws, buf, sizeof(WAKE_UP_DATAGRAM), 0);  // if oversize (should not occur), select will exit again
                continue;
            }
            else
                break;  // should be a TCP connection
        }
        else if (r == 0)    // timeout
        {
            if (dpws->schedule->handler)
            {
                if (dpws_s)
                    dpws_copy_context(dpws_s, dpws);
                return DPWS_OK; // handler to be used in service thread
            }
            else
            {
                realize_schedule(tmp_scheduler,dpws,tmp_scheduler_lock_handle);
                dpws->schedule = NULL;
                soap->errnum = 0;
                soap_set_receiver_error(soap, "Timeout", "Select failed in dpws_accept()", SOAP_TCP_ERROR);
                ONVIF_LOG(ONVIF_LOG_ERROR,
                          "[ERROR]: Select timeout in dpws_accept().\n");
                return SOAP_TCP_ERROR;
            }
        }
        
        if (soap_socket_errno != SOAP_EINTR)
        {
            dpws->schedule = NULL;
            soap->errnum = soap_socket_errno;
            soap_closesock(soap);
            soap_set_sender_error(soap, dpws_ip_error(dpws), "Select failed in dpws_accept()", SOAP_TCP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Unknown error in dpws_accept().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_TCP_ERROR;
        }
    }

    return DPWS_OK;
}

#else /* !USE_ONVIF */
static int accept_msg(struct dpws* dpws, struct dpws* dpws_s)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    struct sockaddr_in sockaddr;
    struct timeval timeout, *ptimeout = NULL;
    fd_set fdset;
    int n = (int)sizeof(struct sockaddr), maxsock;
#ifndef WITH_LEAN
    int len = SOAP_BUFLEN;
    int set = 1;
#endif

    memset(&sockaddr, 0, sizeof(sockaddr));
    soap->socket = SOAP_INVALID_SOCKET;
    soap->errmode = 0;
    
    if (soap->accept_timeout)   // dummy event for user timeout
    {
        lock(scheduler_lock_handle);
        add_schedule(&scheduler, soap->accept_timeout > 0
                    ? soap->accept_timeout*1000
                    : -soap->accept_timeout/1000,
                    NULL,
                    NULL
                    );
        unlock(scheduler_lock_handle);
    }
    
    /* poll for the listening sockets */
    FD_ZERO(&fdset);
    if (soap_valid_socket(soap->master))    // manager may not need TCP
        FD_SET((SOAP_SOCKET)soap->master, &fdset);
    FD_SET((SOAP_SOCKET)dpws->wsd_ls, &fdset);
    FD_SET((SOAP_SOCKET)dpws->wsd_ws, &fdset);
    maxsock = MAX(dpws->wsd_ws,MAX(soap->master,dpws->wsd_ls)) + 1;
    
    for (;;)
    {
        int r;

        // if shuting down start closing listening socket but do not exit to empty scheduler
        if (!tk_config.server_up)
        {
            /* close master TCP */
            soap_done(soap);
            
            /* close UDP sockets */
//          dpws_udp_close(soap);   // Needed ? Should have been closed
            if (soap_valid_socket(dpws->wsd_ls))
            { 
                closesocket((SOAP_SOCKET)dpws->wsd_ls);
                dpws->wsd_ls = SOAP_INVALID_SOCKET;
            }
            FD_ZERO(&fdset);
            FD_SET((SOAP_SOCKET)dpws->wsd_ws, &fdset);
            maxsock = dpws->wsd_ws + 1;
        }
        
        lock(scheduler_lock_handle);
        if ((dpws->schedule = get_schedule(&scheduler))){
            ptimeout = get_timeval_schedule(dpws->schedule, &timeout);
        }
        else if (!tk_config.server_up)
        {
            /* Device is stopping and no message is scheduled */
            unlock(scheduler_lock_handle);

            /* close wake up UDP sockets */
            if (soap_valid_socket(dpws->wsd_ws))
            { 
                closesocket((SOAP_SOCKET)dpws->wsd_ws);
                dpws->wsd_ws = SOAP_INVALID_SOCKET;
            }
            /* empty scheduler */
            free_scheduler(&scheduler);
            
            tk_config.server_up = TRUE; // once the flag has been used to exit, reset it in order to be able to restart the main loop where stopped (in fact pause the server)
            return dpws->err = DPWS_ERR_SERVER_STOPPED;
        }
        unlock(scheduler_lock_handle);

        if (dpws->schedule && !ptimeout)
            r = 0;
        else
#ifdef TR_STACK_NEXGEN
            r = selectsocket(maxsock, &fdset, NULL, NULL, ptimeout);
#else
            r = select((SOAP_SOCKET)maxsock, &fdset, NULL, NULL, ptimeout);
#endif

        if (r > 0) {
            lock(scheduler_lock_handle);
            if (dpws->schedule) {
                reschedule(&scheduler, dpws->schedule);
                dpws->schedule = NULL;
            }
            unlock(scheduler_lock_handle);

            DPWS_SET_RCV(dpws->status);
            DPWS_RESET_MSG_TYPE(dpws->status);
            
            /* UDP datagram received : schedule wake or discovery */
            if (FD_ISSET(dpws->wsd_ls, &fdset))
            {
                int ret = SOAP_OK;
                DPWS_SET_UDP_MULTICAST(dpws->status);
                if (dpws_s)
                    dpws_copy_context(dpws_s, dpws);
                else
                    dpws_s = dpws;
                dpws_dpws2soap(dpws_s)->socket = dpws_s->wsd_ls;
                dpws_dpws2soap(dpws_s)->imode |= SOAP_IO_KEEPALIVE; // prevent closing the socket
                dpws_set_udp_mode(dpws_s);
                if ((ret = wsd_mcast_recv(dpws_s))) {
                    if (ret == SOAP_NO_METHOD)
                    // a bad datagram should not prevent: continue to serve
                        continue;
                    else
                    {
                        soap_set_receiver_error(soap, dpws_ip_error(dpws), "Select failed in dpws_accept() receiving a discovery datagram.", soap->error);
                        return soap->error;
                    }
                }
                else
                {
                    dpws_dpws2soap(dpws_s)->keep_alive = 0;
                    return DPWS_OK;
                }
            }
            else if (FD_ISSET(dpws->wsd_ws, &fdset)) { // wakeup
                char buf[sizeof(WAKE_UP_DATAGRAM)];
                recv(dpws->wsd_ws, buf, sizeof(WAKE_UP_DATAGRAM), 0);  // if oversize (should not occur), select will exit again
                continue;
            }
            else
                break;  // should be a TCP connection
        }
        else if (r == 0)    // timeout
        {
            if (dpws->schedule->handler)
            {
                if (dpws_s)
                    dpws_copy_context(dpws_s, dpws);
                return DPWS_OK; // handler to be used in service thread
            }
            else
            {
                realize_schedule(&scheduler, dpws, scheduler_lock_handle);
                dpws->schedule = NULL;
                soap->errnum = 0;
                soap_set_receiver_error(soap, "Timeout", "Select failed in dpws_accept()", SOAP_TCP_ERROR);
                ONVIF_LOG(ONVIF_LOG_ERROR,
                          "[ERROR]: Select timeout in dpws_accept().\n");
                return SOAP_TCP_ERROR;
            }
        }
        
        if (soap_socket_errno != SOAP_EINTR)
        {
            dpws->schedule = NULL;
            soap->errnum = soap_socket_errno;
            soap_closesock(soap);
            soap_set_sender_error(soap, dpws_ip_error(dpws), "Select failed in dpws_accept()", SOAP_TCP_ERROR);
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Unknown error in dpws_accept().(strerror=%s)\n", strerror(soap->errnum));
            return SOAP_TCP_ERROR;
        }
    }

    /* Receive TCP connections */
    if (soap_valid_socket(soap->master) && FD_ISSET(soap->master, &fdset)
            && (soap->socket = soap->faccept(soap, soap->master, (struct sockaddr*)&sockaddr, &n)) >= 0)
    {
        DPWS_SET_TCP(dpws->status);
        DPWS_SET_USER(dpws->status);
        dpws_set_tcp_mode(dpws);
        soap->ip = ntohl(sockaddr.sin_addr.s_addr);
        soap->port = (int)ntohs(sockaddr.sin_port);
        soap->keep_alive = ((soap->imode & SOAP_IO_KEEPALIVE) != 0);
#ifndef WITH_LEAN
        if (soap->accept_flags & SO_LINGER) {
            struct linger linger;
            memset(&linger, 0, sizeof(struct linger));
            linger.l_onoff = 1;
            linger.l_linger = 0;
            if (setsockopt((SOAP_SOCKET)soap->socket, SOL_SOCKET, SO_LINGER, (char*)&linger, sizeof(struct linger))) {
                soap->errnum = soap_socket_errno;
                soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP setsockopt SO_LINGER failed in dpws_accept()", SOAP_TCP_ERROR);
                return SOAP_TCP_ERROR;
            }
        }
        if ((soap->accept_flags & ~SO_LINGER) && setsockopt((SOAP_SOCKET)soap->socket, SOL_SOCKET, soap->accept_flags & ~SO_LINGER, (char*)&set, sizeof(int))) {
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP setsockopt failed in dpws_accept()", SOAP_TCP_ERROR);
            return SOAP_TCP_ERROR;
        }
        if (soap->keep_alive && setsockopt((SOAP_SOCKET)soap->socket, SOL_SOCKET, SO_KEEPALIVE, (char*)&set, sizeof(int))) {
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP setsockopt SO_KEEPALIVE failed in dpws_accept()", SOAP_TCP_ERROR);
            return SOAP_TCP_ERROR;
        }
        if (setsockopt((SOAP_SOCKET)soap->socket, SOL_SOCKET, SO_SNDBUF, (char*)&len, sizeof(int))) {
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP setsockopt SO_SNDBUF failed in dpws_accept()", SOAP_TCP_ERROR);
            return SOAP_TCP_ERROR;
        }
        if (setsockopt((SOAP_SOCKET)soap->socket, SOL_SOCKET, SO_RCVBUF, (char*)&len, sizeof(int))) {
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP setsockopt SO_RCVBUF failed in dpws_accept()", SOAP_TCP_ERROR);
            return SOAP_TCP_ERROR;
        }
# ifdef TCP_NODELAY
        if (setsockopt((SOAP_SOCKET)soap->socket, IPPROTO_TCP, TCP_NODELAY, (char*)&set, sizeof(int))) {
            soap->errnum = soap_socket_errno;
            soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP setsockopt TCP_NODELAY failed in dpws_accept()", SOAP_TCP_ERROR);
            return SOAP_TCP_ERROR;
        }
# endif
#endif
        if (soap->accept_timeout) {
#if defined(WIN32) || defined(TR_STACK_NEXGEN)
            {
                int blocking = 0;
                ioctlsocket((SOAP_SOCKET)soap->master, FIONBIO, &blocking);
                blocking = 0;
                ioctlsocket((SOAP_SOCKET)soap->socket, FIONBIO, &blocking);
            }
#elif defined(VXWORKS)
            {
                int blocking = 0;
                ioctl((SOAP_SOCKET)soap->master, FIONBIO, (int)(&blocking)); /* modified to use fd */
                blocking = 0;
                ioctl((SOAP_SOCKET)soap->socket, FIONBIO, (int)(&blocking)); /* modified to use fd */
            }
#else
            fcntl((SOAP_SOCKET)soap->master, F_SETFL, fcntl((SOAP_SOCKET)soap->master, F_GETFL)&~O_NONBLOCK);
            fcntl((SOAP_SOCKET)soap->socket, F_SETFL, fcntl((SOAP_SOCKET)soap->socket, F_GETFL)&~O_NONBLOCK);
#endif
        }
        if (dpws_s)
            dpws_copy_context(dpws_s, dpws);
    }
    else {
        soap->errnum = soap_socket_errno;
        soap_set_receiver_error(soap, dpws_ip_error(dpws), "TCP accept failed in dpws_accept()", SOAP_TCP_ERROR);
        return SOAP_TCP_ERROR;
    }
        
    return DPWS_OK;
}   
#endif

static char * spawn_resolve(struct dpws *dpws, const char * log_addr)
{
    char * phys_addr = NULL;
    struct dpws *spawn_dpws = malloc(sizeof(struct dpws));  // to avoid stack overflow
    // TODO: perform a minimal backup instead of the user DPWS structure instead of dynamic allocation
    struct device_proxy* device;
    
    dpws_struct_init(spawn_dpws);
    device = lookup_by_id(spawn_dpws, (char *)log_addr, WSD_MATCH_TIMEOUT);
    // TODO: add a better way to select the transport address
    if (device)
        phys_addr = soap_strdup(dpws_dpws2soap(dpws), *(char **)get_entry(&device->transport_addresses, 0));
    dpws_end(spawn_dpws);
    free(spawn_dpws);
    return phys_addr;
}

static int dpws_handle_fault(struct dpws *dpws, int (* ffault)(struct dpws*))
{
    struct soap *soap = &dpws->soap;
    const char* action = dpws->action ? dpws->action : "";
    if (!strcmp(action, WSA_SOAP_FAULT_URI))
    {
        dpws_set_namespaces(dpws, default_namespaces);
        soap_recv_fault(soap);
        return ffault(dpws);
    }
    else
        return soap->error = SOAP_NO_METHOD;
}

/*******************************************************************************
*                                    API                                       *
*******************************************************************************/
void dpws_default_wsa_endpoint_ref(struct wsa_endpoint_ref *a) {
    memset(a, 0, sizeof(struct wsa_endpoint_ref));
}

#ifdef USE_ONVIF /* Added by HG.W */
int dpws_init()
{
    int ret = DPWS_OK;

    initMutexPool();
    if ((ret = initRegistry()) != DPWS_OK)
        return ret;
        
    uuid_init();

    if(scheduler_lock_handle < 0)
    {
        if ((scheduler_lock_handle = lock_create()) < 0){
            ret = scheduler_lock_handle;
            return ret;
        }
    }
#ifdef WITH_IPV6
    if(scheduler_lock_handle_v6 < 0)
    {
        if ((scheduler_lock_handle_v6 = lock_create()) < 0){
            ret = scheduler_lock_handle_v6;
            goto exit_lock_destroy_v4;
        }
    }
#endif
    if ((ret = initDiscovery()) != DPWS_OK)
        goto exit_lock_destroy_v6;
            
    if ((ret = initCache()) != DPWS_OK)
        goto exit_lock_destroy_v6;
            
    if ((ret = retrieve_network_adapter_info(&tk_config.network_adapter)) != DPWS_OK)
        goto exit_lock_destroy_v6;
    else
        return ret;

exit_lock_destroy_v6:
#ifdef WITH_IPV6
    if(scheduler_lock_handle_v6 < 0){
        lock_destroy(scheduler_lock_handle_v6);
        scheduler_lock_handle_v6 = -1;
    }
#endif
exit_lock_destroy_v4:
    if(scheduler_lock_handle < 0){
        lock_destroy(scheduler_lock_handle);
        scheduler_lock_handle = -1;
    }

    return ret;
}
#else /* !USE_ONVIF */
int dpws_init()
{
    int ret = DPWS_OK;
    if (scheduler_lock_handle < 0)
    {
        initMutexPool();
        if ((ret = initRegistry()) != DPWS_OK)
            return ret;
        
        uuid_init();
        if ((scheduler_lock_handle = lock_create()) < 0)
            ret = scheduler_lock_handle;
            
        if ((ret = initDiscovery()) != DPWS_OK)
            return ret;
            
        if ((ret = initCache()) != DPWS_OK)
            return ret;
            
        if ((ret = retrieve_network_adapter_info(&tk_config.network_adapter)) != DPWS_OK)
            return ret;
    }
    return ret;
}
#endif

int dpws_client_init(struct dpws* dpws, struct wsa_endpoint_ref* source)
{
    dpws_struct_init(dpws);
    dpws->ipproto_type = DPWS_IPPROTO_IPV4;
    dpws->source = endpoint_ref_dup(dpws, source);
    return DPWS_OK;
}
#ifdef WITH_IPV6
int dpws_client_init_v6(struct dpws* dpws, struct wsa_endpoint_ref* source)
{
    dpws_struct_init(dpws);
    dpws->ipproto_type = DPWS_IPPROTO_IPV6;
    dpws->source = endpoint_ref_dup(dpws, source);
    return DPWS_OK;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W */
static int dpws_wakeup_bind(struct dpws* dpws)
{
    int fd;
    struct sockaddr_un servAddr;
    struct soap *soap = dpws_dpws2soap(dpws);
    
    soap->errmode = 0;
    fd = (int)socket(AF_LOCAL, SOCK_DGRAM, 0);

    if (fd < 0) { 
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws),
                              "Unix domain socket creation failed in dpws_wakeup_bind()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to create wakeup socket in dpws_wakeup_bind().(strerror=%s)\n", strerror(soap->errnum));
        return dpws->err = DPWS_ERR_COULD_NOT_CREATE_SOCKET;
    }

    memset(&servAddr, 0, sizeof(servAddr));
    servAddr.sun_family = AF_LOCAL;
    if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
        unlink(tk_config.wakeup_sock_svr_v4);
        strcpy(servAddr.sun_path, tk_config.wakeup_sock_svr_v4);
    }
#ifdef WITH_IPV6
    else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
        unlink(tk_config.wakeup_sock_svr_v6);
        strcpy(servAddr.sun_path, tk_config.wakeup_sock_svr_v6);
    }
#endif
    if(bind(fd, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0) {
        soap->errnum = soap_socket_errno;
        closesocket(fd);
        soap_set_sender_error(soap, dpws_ip_error(dpws),
                              "Could not bind unix domain socket in dpws_wakeup_bind()", SOAP_UDP_ERROR);
        ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to bind wakeup socket in dpws_wakeup_bind().(strerror=%s)\n", strerror(soap->errnum));
        return dpws->err = DPWS_ERR_COULD_NOT_BIND_SOCKET;
    }
    dpws->wsd_ws = fd;
    
    ONVIF_LOG1(ONVIF_LOG_INFO, "[INFO]: Succeeded in binding wakeup socket=%d\n", dpws->wsd_ws);

    return SOAP_OK;
}

#else
static int dpws_wakeup_bind(struct dpws* dpws)
{
    struct sockaddr_in servAddr;
    struct soap *soap = dpws_dpws2soap(dpws);
    
    /* create multicast listening UDP socket */
    wake_sock = SOAP_INVALID_SOCKET;
    soap->errmode = 0;
    wake_sock = (int)socket(AF_INET, SOCK_DGRAM, 0);
    if (wake_sock < 0) { 
        soap->errnum = soap_socket_errno;
        soap_set_sender_error(soap, dpws_ip_error(dpws), "UDP socket creation failed in dpws_wakeup_bind()", SOAP_UDP_ERROR);
        return dpws->err = DPWS_ERR_COULD_NOT_CREATE_SOCKET;
    }
    
    /* bind multicast listening UDP socket */
    servAddr.sin_family=AF_INET;
    servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servAddr.sin_port=htons(listeners->port);
    if (bind(wake_sock, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0) {
        soap->errnum = soap_socket_errno;
        closesocket((SOAP_SOCKET)wake_sock);
        soap_set_sender_error(soap, dpws_ip_error(dpws), "Could not bind UDP socket in dpws_wakeup_bind()", SOAP_UDP_ERROR);
        return dpws->err = DPWS_ERR_COULD_NOT_BIND_SOCKET;
    }
    
    return SOAP_OK;
}
#endif


#ifdef USE_ONVIF /* Added by HG.W */

#ifdef WITH_IPV6 
int dpws_server_init_v6(struct dpws* dpws, struct wsa_endpoint_ref* source)
{
    int ret = DPWS_OK;
    
    dpws_struct_init(dpws);
    dpws->ipproto_type = DPWS_IPPROTO_IPV6;

    if ((ret = check_configuration_v6()) == DPWS_OK)
    {
        dpws->source = endpoint_ref_dup(dpws, source);
    
        if (dpws_discovery_bind(dpws))
            ret = dpws_dpws2soap(dpws)->error;

        else if (dpws_wakeup_bind(dpws))
            ret = dpws_dpws2soap(dpws)->error;
        else {
            struct device * device;
    
            /* schedule hellos for enabled devices */
            for (device = registry; device; device = device->next) {
                lock(device->lock);
                if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE){
                    /* set to multicast */
                    short msgTypeHello = DPWS_MSG_HELLO;
                    ret = schedule_discovery_msg(&scheduler_v6, DPWS_SET_MULTICAST(msgTypeHello), NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle_v6);
                }
                unlock(device->lock);
            }

            tk_config.server_up_v6 = TRUE;
        }
    }
    return ret;
}
#endif

int dpws_server_init(struct dpws* dpws, struct wsa_endpoint_ref* source)
{
    int ret = DPWS_OK;
    
    dpws_struct_init(dpws);
    dpws->ipproto_type = DPWS_IPPROTO_IPV4;

    if ((ret = check_configuration()) == DPWS_OK)
    {
        dpws->source = endpoint_ref_dup(dpws, source);
    
        if (dpws_discovery_bind(dpws))
            ret = dpws_dpws2soap(dpws)->error;
        else if (dpws_wakeup_bind(dpws))
            ret = dpws_dpws2soap(dpws)->error;
        else {
            struct device * device;
    
            /* schedule hellos for enabled devices */
            for (device = registry; device; device = device->next) {
                lock(device->lock);
                if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE){
                    /* set to multicast */
                    short msgTypeHello = DPWS_MSG_HELLO;
                    ret = schedule_discovery_msg(&scheduler, DPWS_SET_MULTICAST(msgTypeHello), NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle);
                }
                unlock(device->lock);
            }

            tk_config.server_up = TRUE;
        }
    }
    return ret;
}
#else
int dpws_server_init(struct dpws* dpws, struct wsa_endpoint_ref* source)
{
    int ret = DPWS_OK;
    
    dpws_struct_init(dpws);

    if ((ret = check_configuration()) == DPWS_OK)
    {
        dpws->source = endpoint_ref_dup(dpws, source);
        if (dpws_discovery_bind(dpws))
            ret = dpws_dpws2soap(dpws)->error;
        else if (tk_config.tcp_listen && !soap_valid_socket(soap_bind(
                                                            &dpws->soap,
                                                            NULL,
                                                            listeners->port,
                                                            listeners->tcp_backlog
                                                            ))) // create and bind the TCP/HTTP socket
            ret = dpws_dpws2soap(dpws)->error;
        else if (dpws_wakeup_bind(dpws))
            ret = dpws_dpws2soap(dpws)->error;
        else {
            struct device * device;
    
            /* schedule hellos for enabled devices */
            for (device = registry; device; device = device->next) {
                lock(device->lock);
                if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE)
                    ret = schedule_discovery_msg(&scheduler, DPWS_MSG_HELLO, NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle);
                unlock(device->lock);
            }

            tk_config.server_up = TRUE;
        }
    }
    return ret;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W */
int dpws_enable_device(short deviceHandleRef)
{
    int ret = SOAP_OK;
    struct device *device;

    device = (struct device *)getObject(&registryPool, deviceHandleRef);
    if (!device) 
        return DPWS_ERR_NO_HANDLE_FOUND;
    
    lock(device->lock);
    if (!device->device_up) // initialized to FALSE
    {
        if (!(ret = init_device_class(device->hrefClass)) && !(ret = check_device(device)));
        {
            checkout_device_classes(device);
            device->device_up = TRUE;

            /* for ipv4 */
            if(tk_config.server_up && get_device_mode(device) == DPWS_DEVICE_USE) {    // else scheduling will be done by dpws_server_init()

                /* set to multicast */
                short msgTypeHello = DPWS_MSG_HELLO;
                ret = schedule_discovery_msg(&scheduler, DPWS_SET_MULTICAST(msgTypeHello), NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle);
                if (ret)
                    goto exit;
                /* Send a dummy connection on the TCP socket to wake up accept */
                ret = send_wakeup_datagram();
            }

#ifdef WITH_IPV6
            /* for ipv6 */
            if(tk_config.server_up_v6 && get_device_mode(device) == DPWS_DEVICE_USE) {    // else scheduling will be done by dpws_server_init()

                /* set to multicast */
                short msgTypeHello = DPWS_MSG_HELLO;
                ret = schedule_discovery_msg(&scheduler_v6, DPWS_SET_MULTICAST(msgTypeHello), NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle_v6);
                if (ret)
                    goto exit;
                /* Send a dummy connection on the TCP socket to wake up accept */
                ret = send_wakeup_datagram_v6();
            }
#endif 
        }
    }
exit:
    unlock(device->lock);
    return ret;
}

#else /* !USE_ONVIF */
int dpws_enable_device(short deviceHandleRef)
{
    int ret = SOAP_OK;
    struct device *device;

    device = (struct device *)getObject(&registryPool, deviceHandleRef);
    if (!device) 
        return DPWS_ERR_NO_HANDLE_FOUND;
    
    lock(device->lock);
    if (!device->device_up) // initialized to FALSE
    {
        if (!(ret = init_device_class(device->hrefClass)) && !(ret = check_device(device)));
        {
            checkout_device_classes(device);
            device->device_up = TRUE;
            if (tk_config.server_up && get_device_mode(device) == DPWS_DEVICE_USE) {    // else scheduling will be done by dpws_server_init()

                ret = schedule_discovery_msg(&scheduler, DPWS_MSG_HELLO, NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle);

                if (ret)
                    goto exit;
                /* Send a dummy connection on the TCP socket to wake up accept */
                ret = send_wakeup_datagram();
            }
        }
    }
exit:
    unlock(device->lock);
    return ret;
}
#endif

int dpws_shutdown()
{
    int ret = DPWS_OK;
    
    free_configuration();
    uninitRegistry();

    clean_cache();
    uninitCache();

    if (scheduler_lock_handle >= 0)
    {
        ret = lock_destroy(scheduler_lock_handle);
        scheduler_lock_handle = -1;
    }
#ifdef WITH_IPV6
    if (scheduler_lock_handle_v6 >= 0)
    {
        ret = lock_destroy(scheduler_lock_handle_v6);
        scheduler_lock_handle_v6 = -1;
    }
#endif

    uninitDiscovery();  // No need to clean pending discovery message because they all should have been sent
    destroyMutexPool();

    return ret;
}

static BOOLEAN all_bye_being_sent()
{
    struct device * device;
    BOOLEAN ret = TRUE;

    /* check all byes have been scheduled */
    for (device = registry; ret && device; device = device->next) {
        lock(device->lock);
        if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE && !device->bye_scheduled)
            ret = FALSE;
        unlock(device->lock);
    }
    /* check the scheduler is empty */
    if (ret) {
        lock(scheduler_lock_handle);
        if (scheduler.schedules != NULL)
            ret = FALSE;
        unlock(scheduler_lock_handle);
    }
    return ret;
}

#ifdef WITH_IPV6
static BOOLEAN all_bye_being_sent_v6()
{
    struct device * device;
    BOOLEAN ret = TRUE;

    /* check all byes have been scheduled */
    for (device = registry; ret && device; device = device->next) {
        lock(device->lock);
        if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE && !device->bye_scheduled)
            ret = FALSE;
        unlock(device->lock);
    }
    /* check the scheduler is empty */
    if (ret) {
        lock(scheduler_lock_handle_v6);
        if (scheduler_v6.schedules != NULL)
            ret = FALSE;
        unlock(scheduler_lock_handle_v6);
    }
    return ret;
}
#endif

static int stop_server(BOOLEAN blocking, long delay_ms)
{
    int ret = DPWS_OK;
    
    /* send byes */
    if (tk_config.server_up)
    {
        int i;
        struct device * device;
    
        if (get_toolkit_mode() == DPWS_DEVICE_USE)
        {
            /* schedule eventing ends and byes for enabled devices */
            for (device = registry; device; device = device->next) {
                lock(device->lock);
                if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE) {
#ifdef USE_ONVIF /* Added by HG.W */
                /* set to multicast */
                short msgTypeBye = DPWS_MSG_BYE;
                ret = schedule_discovery_msg(&scheduler, DPWS_SET_MULTICAST(msgTypeBye), NULL, device, (int (*) (struct dpws*, void *))send_bye, scheduler_lock_handle);
#else
                    lock(scheduler_lock_handle);
                    add_schedule(&scheduler, 0, (schedule_cbk)send_eventing_end, device);   // NOTE send_eventing_end will pipe the BYE
                    unlock(scheduler_lock_handle);
#endif
                }
                unlock(device->lock);
            }
    
            if (blocking) {
                /* Send a dummy connection on the UDP socket to wake up accept and perform schedule */
                ret = send_wakeup_datagram();
                if (ret)
                    return ret;
                
                /* loop until all schedules are processed */
                for (i = 0; i < SHUTDOWN_DELAY_MAX_REPEAT && all_bye_being_sent()== FALSE; i++) {
                    sleep_ms(SHUTDOWN_DELAY);   // NOTE : Should not be necessary (since exit should be performed when nothing is scheduled)
                }
                // Let the time for the byes to be sent
                sleep_ms(delay_ms);
            }
        }
        
        /* Set the shutdown flag to exit the loop */
        tk_config.server_up = FALSE;

        /* Send a dummy connection on the UDP socket to wake up accept and exit */
        ret = send_wakeup_datagram();
    }
    
    return ret;
}

#ifdef WITH_IPV6
static int stop_server_v6(BOOLEAN blocking, long delay_ms)
{
    int ret = DPWS_OK;
    
    /* send byes */
    if (tk_config.server_up_v6)
    {
        int i;
        struct device * device;
    
        if (get_toolkit_mode() == DPWS_DEVICE_USE)
        {
            /* schedule eventing ends and byes for enabled devices */
            for (device = registry; device; device = device->next) {
                lock(device->lock);
                if (device->device_up && get_device_mode(device) == DPWS_DEVICE_USE) {
                /* set to multicast */
                short msgTypeBye = DPWS_MSG_BYE;
                ret = schedule_discovery_msg(&scheduler_v6, DPWS_SET_MULTICAST(msgTypeBye), NULL, device, (int (*) (struct dpws*, void *))send_bye, scheduler_lock_handle_v6);
                }
                unlock(device->lock);
            }
    
            if (blocking) {
                /* Send a dummy connection on the UDP socket to wake up accept and perform schedule */
                ret = send_wakeup_datagram_v6();
                if (ret)
                    return ret;
                
                /* loop until all schedules are processed */
                for (i = 0; i < SHUTDOWN_DELAY_MAX_REPEAT && all_bye_being_sent_v6()== FALSE; i++) {
                    sleep_ms(SHUTDOWN_DELAY);   // NOTE : Should not be necessary (since exit should be performed when nothing is scheduled)
                }
                // Let the time for the byes to be sent
                sleep_ms(delay_ms);
            }
        }
        
        /* Set the shutdown flag to exit the loop */
        tk_config.server_up_v6 = FALSE;

        /* Send a dummy connection on the UDP socket to wake up accept and exit */
        ret = send_wakeup_datagram_v6();
    }
    
    return ret;
}
#endif

int dpws_stop_server(long delay_ms) {return stop_server(TRUE, delay_ms);}

#ifdef WITH_IPV6
int dpws_stop_server_v6(long delay_ms) {return stop_server_v6(TRUE, delay_ms);}
#endif

int dpws_stop_server_no_wait() {return stop_server(FALSE, 0);}

#ifdef USE_ONVIF /* Added by HG.W */
int dpws_disable_device(short deviceHandleRef)
{
    int ret = DPWS_OK;
    struct device *device;
    short msgTypeBye = DPWS_MSG_BYE;

    device = (struct device *)getObject(&registryPool, deviceHandleRef);
    if (!device) {
        return DPWS_ERR_NO_HANDLE_FOUND;
    }

    lock(device->lock);

    if (device->device_up)
    {
        release_device_classes(device);
        device->device_up = FALSE;

        /* for ipv4 */
        if (tk_config.server_up && get_device_mode(device) == DPWS_DEVICE_USE) {    // else scheduling will be done by dpws_stop_server()

            /* schedule BYE */
            ret = schedule_discovery_msg(&scheduler, DPWS_SET_MULTICAST(msgTypeBye), NULL, device, (int (*) (struct dpws*, void *))send_bye, scheduler_lock_handle);
            /* Send a dummy connection on the TCP socket to wake up accept */
            ret = send_wakeup_datagram();
        }

#ifdef WITH_IPV6
        /* for ipv6 */
        if (tk_config.server_up_v6 && get_device_mode(device) == DPWS_DEVICE_USE) {    // else scheduling will be done by dpws_stop_server()

            /* schedule BYE */
            ret = schedule_discovery_msg(&scheduler_v6, DPWS_SET_MULTICAST(msgTypeBye), NULL, device, (int (*) (struct dpws*, void *))send_bye, scheduler_lock_handle_v6);

            /* Send a dummy connection on the TCP socket to wake up accept */
            ret = send_wakeup_datagram_v6();
        }
#endif

    }

    unlock(device->lock);
    return ret;
}

#else /* !USE_ONVIF */
int dpws_disable_device(short deviceHandleRef)
{
    int ret = DPWS_OK;
    struct device *device;

    device = (struct device *)getObject(&registryPool, deviceHandleRef);
    if (!device) {
        ret = DPWS_ERR_NO_HANDLE_FOUND;
        goto exit;
    }

    lock(device->lock);

    if (device->device_up)
    {
        release_device_classes(device);
        device->device_up = FALSE;
        if (tk_config.server_up && get_device_mode(device) == DPWS_DEVICE_USE) {    // else scheduling will be done by dpws_stop_server()
            // NOTE send_eventing_end will pipe the BYE
            lock(scheduler_lock_handle);
            add_schedule(&scheduler, 0, (schedule_cbk)send_eventing_end, device);
            unlock(scheduler_lock_handle);
            /* Send a dummy connection on the TCP socket to wake up accept */
            ret = send_wakeup_datagram();
        }
    }

exit:
    unlock(device->lock);
    return ret;
}
#endif

int dpws_get_error(struct dpws* dpws) {
    return dpws->err ? dpws->err : dpws->soap.error;
}

const char * dpws_get_error_msg(struct dpws* dpws)
{
    struct soap *soap = dpws_dpws2soap(dpws);
    if (soap->error)
    {
        const char *c, *v = NULL, *s, **d;
        d = soap_faultcode(soap);
        if (!*d)
            soap_set_fault(soap);
        c = *d;
        if (soap->version == 2)
            v = *soap_faultsubcode(soap);
        s = *soap_faultstring(soap);
        d = soap_faultdetail(soap);
        sprintf(soap->tmpbuf, "%s%d fault: %s [%s]\n\"%s\"\nDetail: %s", soap->version ? "SOAP 1." : "Error ", soap->version ? (int)soap->version : soap->error, c, v ? v : "no subcode", s ? s : "[no reason]", d && *d ? *d : "[no detail]");
    }
    else
        return "No message available.";

    return soap->tmpbuf;
}

int dpws_accept(struct dpws* dpws)
{
    return accept_msg(dpws, NULL);
}

int dpws_accept_thr(struct dpws* dpws, struct dpws* dpws_thr)
{
    if (!dpws_thr)
        dpws_thr = malloc(sizeof(struct dpws));
    return accept_msg(dpws, dpws_thr);
}

int dpws_end(struct dpws* dpws) {
    dpws_init_wsa_headers(dpws);
    release_proxies(dpws);
    dpws->out_ip_addr = NULL;
    dpws->err = DPWS_OK;
    soap_end(&dpws->soap);
    return DPWS_OK;
}

int dpws_serve(struct dpws *dpws)
{
    unsigned int n = dpws_dpws2soap(dpws)->max_keep_alive, ret = SOAP_NO_METHOD;
    do {
        if (dpws->schedule) {
#ifdef USE_ONVIF
            if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
                return realize_schedule(&scheduler, dpws,
                                        scheduler_lock_handle);
            }
#ifdef WITH_IPV6
            else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
                return realize_schedule(&scheduler_v6, dpws,
                                        scheduler_lock_handle_v6);
            }
#endif
            else{/* do nothing */}
#else
            return realize_schedule(&scheduler, dpws, scheduler_lock_handle);   // calls handler
#endif
        }
        else if (DPWS_IS_DISCOVERY(dpws->status)) {
            ret = wsd_process(dpws);    // msg already retrieved
            dpws_dpws2soap(dpws)->imode &= ~SOAP_IO_KEEPALIVE;  // restore the flag set in accept_msg()
        }
        else {
            struct service_class *servClass;
            BOOLEAN smgr;
            int i;
            int (*fserv) (struct dpws *) = NULL;
            struct service_endpoint * service = NULL;
            
            dpws_set_namespaces(dpws, default_namespaces);
            soap_begin(dpws_dpws2soap(dpws));
            if (!--n)
                dpws_dpws2soap(dpws)->keep_alive = 0;
            if (soap_begin_recv(dpws_dpws2soap(dpws))) {
                if (dpws_dpws2soap(dpws)->error < SOAP_STOP)
                    return dpws_send_fault(dpws); // TODO: AM - Check what happens in one way case
                soap_closesock(dpws_dpws2soap(dpws));
                continue;
            }
            if (soap_envelope_begin_in(dpws_dpws2soap(dpws))
                || soap_recv_header(dpws_dpws2soap(dpws)))
                return dpws_send_fault(dpws_soap2dpws(dpws_dpws2soap(dpws)));

            // Added to complete the HTTP request/response in oneway cases
            // TODO: AM - Faults should not be sent back in oneway cases 
            if (dpws_is_asynchronous(dpws))
                if (dpws_send_empty_response(dpws))
                    return dpws_send_fault(dpws_soap2dpws(dpws_dpws2soap(dpws)));
                
            if (soap_body_begin_in(dpws_dpws2soap(dpws)))
                return dpws_send_fault(dpws_soap2dpws(dpws_dpws2soap(dpws)));
            
            service = find_endpoint(dpws->to);
            if (!service) {
                dpws_dpws2soap(dpws)->error = WSA_ERR_DESTINATION_UNREACHABLE;
                return dpws_send_fault(dpws);
            }
            dpws->href_endpoint = service->href;
            
            servClass = (struct service_class *)getObject(&registryPool, service->hrefServiceClass);
            
            for (i = 0; i < servClass->funcs.nb; i++) {
                fserv = *(int (**) (struct dpws *))get_entry(&servClass->funcs, i);
                ret = fserv(dpws);
                if (ret != SOAP_NO_METHOD) {                            // NOTE: this is optimized to return ASAP
                    if (ret)
                        return dpws_send_fault(dpws);
                    else 
                        return SOAP_OK;
                }
            }
                
            if (ret == SOAP_NO_METHOD)
                ret = metadata_serve_request(dpws); // NOTE: mex although attached to device can address a hosted service
            
            if (ret == SOAP_NO_METHOD)
            {
                lock(service->device->lock);
                smgr = service->subsc_manager;
                unlock(service->device->lock);
                if (smgr)
                    ret = subscmanager_serve_request(dpws);
            }

            if (ret == SOAP_NO_METHOD)
                ret = wsd_serve_directed_probe(dpws);

            // For asynchronous fault handling
            if (ret == SOAP_NO_METHOD)
            {
                if (servClass->fault_func)
                    ret = dpws_handle_fault(dpws, servClass->fault_func);
            }
        }
        if (ret == SOAP_NO_METHOD)
            dpws_dpws2soap(dpws)->error = WSA_ERR_ACTION_NOT_SUPPORTED;
            
        if (ret)
            return dpws_send_fault(dpws);
    }
    while (dpws_dpws2soap(dpws)->keep_alive);
    
    return DPWS_OK;
}

short * dpws_lookup(struct dpws* dpws, char * ns, char * type, char * scope, int *res_size)
{
    struct qname qn_type[2] = {{ns, type}, {NULL, NULL}};
    char * scopes_array[2] = {scope, NULL};
    if (!type) {
        qn_type[0].ns = NULL;
        qn_type[0].lname = NULL;
    }
    if (!scope)
        scopes_array[0] = NULL;
        
    return dpws_lookup_ex(dpws, qn_type, scopes_array, WSD_MATCH_BY_UNSPECIFIED, res_size, WSD_MATCH_TIMEOUT, WSD_LOOKUP_MODE_CACHE_FIRST);
}

static void init_probe_body(struct qname types[], char ** scopes, int match_rule,
                            struct wsd__ProbeType *body, struct wsd__ScopesType *scps, dyn_array *strTypes, dyn_array *scopesArray) {
    int i ;
    init_dyn_array(strTypes, sizeof(char *), 1);
    
    // TODO: make that qname_list be no more a string tab but a qname.
    // 1. Need to overload soap_QName2s and soap_s2QName
    // 2. Track other uses of qname_list (included metadata and the cache then !)
    
    if (types) {
        for (i = 0; types[i].lname; i++) {
            char **pstr = new_entry(NULL, strTypes);
            *pstr = malloc(strlen(types[i].ns) + strlen(types[i].lname) + 4);   // NOTE: will be freed with the dynamic array
            
            /* set message data */
            strcpy(*pstr, "\"");
            strcat(*pstr, types[i].ns);
            strcat(*pstr, "\":");
            strcat(*pstr, types[i].lname);
        }
    }
    
    if (strTypes->nb == 0)
        body->Types = NULL;
    else
        body->Types = strTypes;
    if (scopes && scopes[0] && *scopes[0] != 0) {
        body->Scopes = scps;
        scps->__item = scopesArray;
        scopesArray->tab = scopes;  // NOTE: custom uri list marshalling use the null boundary and not the nb field
        scps->MatchBy = scope_match_string(match_rule);
    }
    else
        body->Scopes = NULL;
}


short * dpws_lookup_ex(struct dpws* dpws, struct qname types[],
    char ** scopes, int match_rule, int * res_size, long time_out, int mode)
{
    int nb = 0;
    short * ret = NULL;
    
    /* retrieve endpoints from cache */
    if (mode == WSD_LOOKUP_MODE_CACHE_FIRST)
        nb = count_cache_entry(dpws, types, scopes, match_rule, *res_size);
    
    /* send probe */
    if (mode != WSD_LOOKUP_MODE_CACHE_ONLY && (*res_size <= 0 || nb < *res_size)) {
        short send_nb = 0;
        long delay = 0;
        scheduler_t probeScheduler = {NULL, NULL};
        struct wsd__ProbeType body;
        struct wsd__ScopesType scps;
        dyn_array strTypes, scopesArray;
        char uuid[SCHEME_UUID_STRING_SIZE];
        
        init_probe_body(types, scopes, match_rule, &body, &scps, &strTypes, &scopesArray);
        
        dpws_dpws2soap(dpws)->omode |= SOAP_IO_KEEPALIVE;   // to avoid the socket be closed by the connect
        dpws_set_udp_mode(dpws);

#ifdef USE_ONVIF /* Added by HG.W */
        /* set to multicast */
        short msgTypeProbe = DPWS_MSG_PROBE;
        while ((delay = get_soap_over_udp_schedule(DPWS_SET_MULTICAST(msgTypeProbe), &send_nb, delay)) != 0xFFFFFFFF)
#else
        while ((delay = get_soap_over_udp_schedule(DPWS_MSG_PROBE, &send_nb, delay)) != 0xFFFFFFFF)
#endif
        {
            if (delay == 0) // normal for a probe
            {
                dpws->message_id = schemed_uuid_create(uuid);
                if (send_probe(dpws, &body)) {
                    dpws_end(dpws);
                    goto exit;
                }
                dpws_end(dpws);
            }
            else
                add_schedule(&probeScheduler, delay, (schedule_cbk)send_probe, &body);
        }
        
        add_schedule(&probeScheduler, MIN(time_out, MATCH_TIMEOUT), NULL, NULL);

        /* receive until the nb of endpoint asked are retrieved (probe match) */
        dpws_dpws2soap(dpws)->imode |= SOAP_IO_KEEPALIVE;   // to avoid the socket be closed by the connect
        while (*res_size <= 0 || nb < *res_size)
        {
            dpws->schedule = get_schedule(&probeScheduler);
            DPWS_SET_RCV(dpws->status);
            dpws_init_wsa_headers(dpws);
            if (!wsd_ucast_recv_probematch(dpws)) {
                if (dpws->schedule)
                    reschedule(&probeScheduler, dpws->schedule);
#ifdef USE_ONVIF /* Added by HG.W */
                if(wsd_process(dpws)){
                    //fprintf(stderr, "WARNING: Failed to process cached ProbeMatches.\n");
                    break;
                }
#else
                wsd_process(dpws);
#endif /* Added by HG.W */
            }
            else if (dpws_dpws2soap(dpws)->errnum == DPWS_ERR_UDP_TIMEOUT)
            {
                if (dpws->schedule->handler == NULL) {  // user timeout
                    dpws_dpws2soap(dpws)->errnum = 0;
                    dpws_end(dpws);
                    realize_schedule(&probeScheduler, dpws, -1);    // recycle
                    break;
                }
                dpws->message_id = uuid;
                realize_schedule(&probeScheduler, dpws, -1);    // probe send
            }
            // else: ignore receive error that may be due to unepected packet
            // TODO: exit if socket is not operational.
            
            dpws_end(dpws);
            dpws_dpws2soap(dpws)->keep_alive = 0;
            nb =  count_cache_entry(dpws, types, scopes, match_rule, *res_size);
        }
        ret = cache_lookup(dpws, types, scopes, match_rule, res_size);  // NOTE: this function copy cache contents using dpws heap since dpws_end has been called to allow multiple receive

exit:
        dpws_udp_close(dpws_dpws2soap(dpws));   // NOTE: dpws_end() calls the dummy close
        free_scheduler(&probeScheduler);
        free_buffer_dyn_array(&strTypes);
        dpws_dpws2soap(dpws)->omode &= ~SOAP_IO_KEEPALIVE;
        dpws_dpws2soap(dpws)->imode &= ~SOAP_IO_KEEPALIVE;
        dpws_init_wsa_headers(dpws);
    }
    else
        ret = cache_lookup(dpws, types, scopes, match_rule, res_size);  // NOTE: this function copy cache contents using dpws heap since dpws_end has been called to allow multiple receive

    // TODO: in the WSD_LOOKUP_MODE_FORCE_NETWORK mode should be able to get results only from network ? Or take them first at least. Could simply ensure that most recently received are found first.
    lock(cache_lock_handle);
    checkout_proxies(dpws, ret, *res_size);
    unlock(cache_lock_handle);
    
    return ret;
}

short dpws_lookup_by_id(struct dpws* dpws, char * id)
{
    struct device_proxy* proxy;
    char uuid[SCHEME_UUID_STRING_SIZE];
    if (strncmp(UUID_SCHEME, id, UUID_SCHEME_SIZE)) {
        if (strncmp("uuid:", id, 5)) {
            strcpy(uuid, UUID_SCHEME);
            strncat(uuid, id, UUID_STRING_SIZE);
        }
        else {
            strcpy(uuid, "urn:");
            strncat(uuid, id, SCHEME_UUID_STRING_SIZE - 4);
        }
    }
    else
        strncpy(uuid, id, SCHEME_UUID_STRING_SIZE);
    proxy = lookup_by_id(dpws, uuid, WSD_MATCH_TIMEOUT);
    if (proxy) {
        lock(cache_lock_handle);
        checkout_proxy(dpws, proxy->super.href);
        unlock(cache_lock_handle);
        return proxy->super.href;
    }
    else
        return -1;
}

short dpws_probe_address_ex(struct dpws* dpws, char * addr, struct qname types[], char ** scopes, int match_rule) {
    struct wsd__ProbeType body;
    struct wsd__ScopesType scps;
    dyn_array strTypes, scopesArray;
    char * device_id;

    init_probe_body(types, scopes, match_rule, &body, &scps, &strTypes, &scopesArray);
    if ((device_id =  call_directed_probe(dpws, addr, &body))) {
        struct device_proxy * proxy = find_device_proxy(device_id);
        lock(cache_lock_handle);
        checkout_proxy(dpws, proxy->super.href);
        unlock(cache_lock_handle);
        return proxy->super.href;
    }
    else
        return -1;
}

short dpws_probe_address(struct dpws* dpws, char * addr, char* ns, char* type, char* scope) {
    struct qname qn_type[2] = {{ns, type}, {NULL, NULL}};
    char * scopes_array[2] = {scope, NULL};
    if (type) {
        qn_type[0].ns = ns;
        qn_type[0].lname = type;
        qn_type[1].ns = NULL;
        qn_type[1].lname = NULL;
    }
    else {
        qn_type[0].ns = NULL;
        qn_type[0].lname = NULL;
    }
    if (scope) {
        scopes_array[0] = scope;
        scopes_array[1] = NULL;
    }
    else
        scopes_array[0] = NULL;
        
    return dpws_probe_address_ex(dpws, addr, qn_type, scopes_array, WSD_MATCH_BY_UNSPECIFIED);
}

short * dpws_get_services(struct dpws* dpws,
    short href_device, char * ns, char * port_type, int * res_size)
{
    short * ret = NULL;
    struct device_proxy * device_proxy = fill_device_services(dpws, href_device);   // takes lock
    if (!device_proxy) {
        *res_size = dpws->err = DPWS_ERR_COULD_NOT_FIND_CACHE_ENTRY;
        goto exit;
    }
    // filter cache and copy to results for asked
    ret = filter_services_by_type(dpws, device_proxy, ns, port_type, res_size);
    if (ret)
        checkout_proxies(dpws, ret, *res_size);

exit:
    unlock(cache_lock_handle);
    dpws_init_wsa_headers(dpws);
    return ret;
}

short dpws_get_service_by_id(struct dpws* dpws, short href_device, char * id)
{
    short ret = -1;
    struct device_proxy * cache_entry = fill_device_services(dpws, href_device);    // takes lock
    if (!cache_entry)
        goto exit;
    
    // filter cache and copy to results for asked
    ret = filter_services_by_id(dpws, cache_entry, id);
    if (ret >= 0)
        checkout_proxy(dpws, ret);

exit:
    unlock(cache_lock_handle);
    dpws_init_wsa_headers(dpws);
    return ret;
}


struct device_metadata* dpws_get_device_metadata(struct dpws* dpws, short href_device)
{
    struct device_metadata* metadata = (struct device_metadata*)soap_malloc(dpws_dpws2soap(dpws), sizeof(struct device_metadata));
    dyn_array host_endpoints = DYN_ARRAY_INITIALIZER(sizeof(struct wsa_endpoint_ref *), 1);
    dyn_array wsdlList = DYN_ARRAY_INITIALIZER(sizeof(struct wsdl_info *), 4);
    dyn_array services = DYN_ARRAY_INITIALIZER(sizeof(struct service_info *), 4);
    struct device_proxy * device;
    BOOLEAN isDevice;
    
    lock(cache_lock_handle);            
    device = get_device_proxy(href_device, &isDevice);
    unlock(cache_lock_handle);
    
    if (!device) {
        dpws->err = DPWS_ERR_INVALID_PROXY;
        return NULL;
    }
    if (!isDevice) {
        dpws->err = DPWS_ERR_IS_A_HOSTED_SERVICE;
        return NULL;
    }
    if (get_endpoint_metadata(dpws, href_device, device, !device->hosted_retrieved, &metadata->model_info, &metadata->device_info, &host_endpoints, &services, &wsdlList))
        return NULL;

    metadata->host_endpoints = (struct wsa_endpoint_ref **)host_endpoints.tab;
    metadata->services = (struct service_info **)services.tab;
    metadata->wsdls = (struct wsdl_info **)wsdlList.tab;
    return metadata;    
}

struct service_metadata* dpws_get_service_metadata(struct dpws* dpws, short href_service) {
    struct service_metadata* metadata = (struct service_metadata*)soap_malloc(dpws_dpws2soap(dpws), sizeof(struct service_metadata));
    dyn_array host_endpoints = DYN_ARRAY_INITIALIZER(sizeof(struct wsa_endpoint_ref *), 1);
    dyn_array wsdlList = DYN_ARRAY_INITIALIZER(sizeof(struct wsdl_info *), 1);
    dyn_array services = DYN_ARRAY_INITIALIZER(sizeof(struct service_info *), 1);
    struct device_proxy * device;
    BOOLEAN isDevice;


    lock(cache_lock_handle);            
    device = get_device_proxy(href_service, &isDevice);
    unlock(cache_lock_handle);
    
    if (!device) {
        dpws->err = DPWS_ERR_INVALID_PROXY;
        return NULL;
    }
    if (isDevice) {
        dpws->err = DPWS_ERR_IS_A_DEVICE;
        return NULL;
    }
    if (get_endpoint_metadata(dpws, href_service, device, !device->hosted_retrieved, NULL, NULL, &host_endpoints, &services, &wsdlList))
        return NULL;

    metadata->host_endpoints = (struct wsa_endpoint_ref **)host_endpoints.tab;
    metadata->service_info = *(struct service_info **)get_entry(&services, 0);
    metadata->wsdls = (struct wsdl_info **)wsdlList.tab;
    return metadata;    
}

void update_metadata(struct device * device)
{
    device->metadata_version++; // no real need to protect it since hello is the only reader
    
    /* Schedule and send a datagram on the discovery socket */
    schedule_discovery_msg(&scheduler, DPWS_MSG_HELLO, NULL, device, (int (*) (struct dpws*, void *))send_hello, scheduler_lock_handle);
    send_wakeup_datagram();
}

int dpws_update_metadata(short hrefDevice)
{
    struct device * device = (struct device *)getObject(&registryPool, hrefDevice);
    
    if (!device)
        return DPWS_ERR_NO_HANDLE_FOUND; 
    
    if (tk_config.server_up && device->device_up)
    {
        lock(device->lock); // because of potential probe handling
        update_metadata(device);    
        unlock(device->lock);
    }
    
    return DPWS_OK;
}

char * dpws_event_renew(struct dpws* dpws, struct wsa_endpoint_ref* subscription_manager, char* expiration)
{
    struct _wse__Renew renew;
    struct _wse__RenewResponse response;
    char * ret = NULL;
    
    renew.wse__Expires = expiration;
    if (dpws_call___wse__RenewOp(dpws, subscription_manager, NULL, &renew, &response))
        goto exit;
        
    ret = response.wse__Expires;
    
exit:
    dpws_init_wsa_headers(dpws);
    return ret;
}

char * dpws_event_status(struct dpws* dpws, struct wsa_endpoint_ref* subscription_manager)
{
    struct _wse__GetStatus get_status;
    struct _wse__GetStatusResponse status;
    char * ret = NULL;
    
    if (dpws_call___wse__GetStatusOp(dpws, subscription_manager, NULL, &get_status, &status))
        goto exit;
        
    ret = status.wse__Expires;
    
exit:
    dpws_init_wsa_headers(dpws);
    return ret;
}

int dpws_event_unsubscribe(struct dpws* dpws, struct wsa_endpoint_ref* subscription_manager)
{
    int ret;
    struct _wse__Unsubscribe unsubscribe;
    struct __wse__UnsubscribeOpResponse response;

    ret = dpws_call___wse__UnsubscribeOp(dpws, subscription_manager, NULL, &unsubscribe, &response);
    dpws_init_wsa_headers(dpws);
    return ret;
}

char ** dpws_set_device_scopes(struct dpws* dpws, short href_device, char ** scopes)
{
    struct _mdm__ScopesMetadata mdm_scopes, dev_scopes;
    dyn_array da;
    
    if (*scopes == NULL)
    {
        dpws->err = DPWS_ERR_INVALID_PARAMETER;
        return NULL;
    }
        
    /* check first that device supports mdm */
    if (!dpws_check_device_type(dpws, MDM_URI, MDM_PORT_TYPE, -1, href_device))
    {
        dpws->err = DPWS_ERR_NO_AN_MDM_ENDPOINT;
        return NULL;
    }

    da.tab = scopes;    // NOTE: custom uri list marshalling use the null boundary and not the nb field
    mdm_scopes.Scopes = &da;

    if (dpws_call___mdm__SetScopesMetadataOp(dpws, dpws_get_default_endpoint_ref(dpws, href_device), NULL, &mdm_scopes, &dev_scopes))
        return NULL;
    
    return (char **)dev_scopes.Scopes->tab;
}

struct localized_string ** dpws_set_device_friendly_names(struct dpws* dpws, short href_device, struct localized_string **names)
{
    struct _mdm__SetFriendlyNameMetadata mdm_names;
    struct _mdm__SetFriendlyNameMetadataResponse dev_names;
    struct localized_string ** p = NULL;

    if (*names == NULL)
    {
        dpws->err = DPWS_ERR_INVALID_PARAMETER;
        return NULL;
    }
        
    /* check first that device supports mdm */
    if (!dpws_check_device_type(dpws, MDM_URI, MDM_PORT_TYPE, -1, href_device))
    {
        dpws->err = DPWS_ERR_NO_AN_MDM_ENDPOINT;
        return NULL;
    }
    /* count names */
    for(p = names; *p; p++);
    mdm_names.__sizefriendlyName = (int)(p - names);
    mdm_names.friendlyName = (struct _mdm__friendlyName **)names;   // TODO: structures are the same but definition made easier for API. Move API structure to gSOAP generated code ?
    
    if (dpws_call___mdm__SetFriendlyNameMetadataOp(dpws, dpws_get_default_endpoint_ref(dpws, href_device), NULL, &mdm_names, &dev_names))
        return NULL;
    
    /* allocate the result array (null-terminated) */
    p = soap_malloc(dpws_dpws2soap(dpws), (dev_names.__sizefriendlyName + 1) * sizeof(struct localized_string *));
    memcpy(p, dev_names.friendlyName, dev_names.__sizefriendlyName * sizeof(struct localized_string *));
    p[dev_names.__sizefriendlyName] = NULL;
    return p;
}

/*******************************************************************************
*                                     RUNTIME                                  *
*******************************************************************************/

/*----------------------------------------------------------------------------*\
 *                                  ADDRESSING                                *
\*----------------------------------------------------------------------------*/


void dpws_wsa_headers_out(struct dpws* dpws) {
    struct soap * soap = dpws_dpws2soap(dpws);
    // check that we are in DPWS context
    if (!(soap->omode & DPWS_HEADERS)) return;
    soap->part = SOAP_IN_ENVELOPE;  // NOTE: hack because gSoap 2.6.2 defines default namespaces locally for every element
    if (dpws->to)
        dpws_simple_element_out(soap, "wsa:To", dpws->to);
    if (dpws->action)
        dpws_simple_element_out(soap, "wsa:Action", dpws->action);
    if (dpws->message_id)
        dpws_simple_element_out(soap, "wsa:MessageID", dpws->message_id);
    if (dpws->relates_to)
        dpws_simple_element_out(soap, "wsa:RelatesTo", dpws->relates_to);

    if (!DPWS_IS_RESPONSE(dpws->status)) {
        if (!dpws_endpoint_ref_is_null(dpws->from))
            dpws_out_wsa_endpoint_ref(soap, "wsa:From", &dpws->from);
        if (!dpws_endpoint_ref_is_null(dpws->reply_to))
            dpws_out_wsa_endpoint_ref(soap, "wsa:ReplyTo", &dpws->reply_to);
        if (!dpws_endpoint_ref_is_null(dpws->fault_to))
            dpws_out_wsa_endpoint_ref(soap, "wsa:FaultTo", &dpws->fault_to);
    }
    if (DPWS_APP_SEQ_MSG(dpws->status)) {
        soap_set_attr(soap, "InstanceId", soap_unsignedLong2s(soap, dpws->instance_id));
#ifdef UNDER_CE
        soap_set_attr(soap, "MessageNumber", soap_short2s(soap, dpws->msg_nb));
#else
        soap_set_attr(soap, "MessageNumber", soap_unsignedShort2s(soap, dpws->msg_nb));
#endif
        dpws_empty_element_out(soap, "wsd:AppSequence");
    }
    if (dpws->subscription_id)
        dpws_simple_element_out(soap, "wse:Identifier", dpws->subscription_id);
    
    dpws->soap.part = SOAP_IN_HEADER;
}

int dpws_wsa_header_in(struct dpws* dpws) {
    struct soap * soap = dpws_dpws2soap(dpws);
    // check that we are in DPWS context
    if (!(soap->imode & DPWS_HEADERS)) return -1;
    if ((!dpws->to && dpws_simple_element_in(&dpws->soap, "wsa:To", &dpws->to))
        || (!dpws->action && dpws_simple_element_in(&dpws->soap, "wsa:Action", &dpws->action))
        || (!dpws->message_id && dpws_simple_element_in(&dpws->soap, "wsa:MessageID", &dpws->message_id))
        || (!dpws->relates_to && dpws_simple_element_in(&dpws->soap, "wsa:RelatesTo", &dpws->relates_to))
        || (!dpws->subscription_id && dpws_simple_element_in(&dpws->soap, "wse:Identifier", &dpws->subscription_id))
        || (DPWS_IS_UDP_RCV_MULTICAST(dpws->status) && dpws->msg_nb == 0 && dpws_in_wsd_AppSequence(dpws))
        || (dpws_endpoint_ref_is_null(dpws->from) && dpws_in_wsa_endpoint_ref(&dpws->soap, "wsa:From", &dpws->from))
        || (dpws_endpoint_ref_is_null(dpws->reply_to) && dpws_in_wsa_endpoint_ref(&dpws->soap, "wsa:ReplyTo", &dpws->reply_to))
        || (dpws_endpoint_ref_is_null(dpws->fault_to) && dpws_in_wsa_endpoint_ref(&dpws->soap, "wsa:FaultTo", &dpws->fault_to)))
        return 0;
    return -1;
}

void dpws_init_wsa_headers(struct dpws* dpws) {
    dpws->to = NULL;
    dpws->action = NULL;
    dpws->message_id = NULL;
    dpws->relates_to = NULL;
    dpws->instance_id = 0;
    dpws->msg_nb = 0;
    dpws->source = NULL;
    dpws_default_wsa_endpoint_ref(&dpws->from);
    dpws_default_wsa_endpoint_ref(&dpws->reply_to);
    dpws_default_wsa_endpoint_ref(&dpws->fault_to);
    DPWS_INIT_STATUS(dpws->status);
    dpws->subscription_id = NULL;
}

int dpws_check_wsa_headers(struct dpws *dpws)
{
    if (!dpws->to
        || (!dpws_endpoint_ref_is_null(dpws->reply_to) && !dpws->message_id)
        || !dpws->action)
        return dpws_dpws2soap(dpws)->error = WSA_ERR_REQUIRED_HEADER_INFO;
    // TODO: add the DPWS_WSA_ACTION_NOT_SUPPORTED test
    return SOAP_OK;
}


void dpws_set_call_wsa_headers(struct dpws* dpws,
        struct wsa_endpoint_ref *to, struct wsa_endpoint_ref *reply_to, char *wsa_action) {
    DPWS_SET_REQUEST(dpws->status);
    dpws->to = to->address;
    if (dpws->source)
        memcpy(&dpws->from, dpws->source, sizeof(struct wsa_endpoint_ref));
    dpws->action = wsa_action;
    if (reply_to)
        memcpy(&dpws->reply_to, reply_to, sizeof(struct wsa_endpoint_ref));
    else
        dpws_set_anonymous_wsa_endpoint_ref(&dpws->reply_to);
    // fault is optional
    dpws->message_id = schemed_uuid_create((char *)soap_malloc(&dpws->soap, SCHEME_UUID_STRING_SIZE));
    dpws->relates_to = NULL;
    dpws->subscription_id = to->subscription_id;
}

void dpws_set_oneway_wsa_headers(struct dpws* dpws, struct wsa_endpoint_ref *to, char *wsa_action)
{
    if (!DPWS_IS_DISCOVERY(dpws->status))   // NOTE: no need normally if the user call dpws_end between every call. Discovery APIs resets headers (no dpws_end after possibly them)
        dpws->message_id = NULL;
    DPWS_SET_REQUEST(dpws->status);
    dpws->to = to->address;
    if (dpws->source)
        memcpy(&dpws->from, dpws->source, sizeof(struct wsa_endpoint_ref));
    dpws->action = wsa_action;
    if (!(dpws->status & (DPWS_MSG_PROBE_MATCH|DPWS_MSG_RESOLVE_MATCH)))    // are responses even if formally not !
        dpws->relates_to = NULL;
    dpws->subscription_id = to->subscription_id;
}

void dpws_set_reply_wsa_headers(struct dpws* dpws, char *wsa_action) {
    DPWS_SET_RESPONSE(dpws->status);
    dpws->to = dpws->reply_to.address;
    if (dpws->source)
        memcpy(&dpws->from, dpws->source, sizeof(struct wsa_endpoint_ref));
    else
        dpws_default_wsa_endpoint_ref(&dpws->from);
    dpws_default_wsa_endpoint_ref(&dpws->reply_to);
    dpws_default_wsa_endpoint_ref(&dpws->fault_to);
    dpws->action = wsa_action;
    dpws->relates_to = dpws->message_id;
    dpws->message_id = schemed_uuid_create((char *)soap_malloc(&dpws->soap, SCHEME_UUID_STRING_SIZE));
    dpws->subscription_id = dpws->reply_to.subscription_id;
}

static int dpws_is_oneway(struct dpws* dpws) {
    return dpws_endpoint_ref_is_null(dpws->reply_to);
}

int dpws_is_asynchronous(struct dpws* dpws) {
    return dpws_endpoint_ref_is_null(dpws->reply_to) || strcmp(dpws->reply_to.address, WSA_ANONYMOUS_ENDPOINT_REF_URI);
}

int dpws_send_empty_response(struct dpws* dpws) {
    struct soap* soap = soap_copy(dpws_dpws2soap(dpws)); // TODO: AM - This is overkill for writing a few characters on the socket 
    // TODO: particularly that content-length should be explicitely (according to Microsoft) set to 0.
    if (!DPWS_IS_UDP(dpws->status)) { // TODO: AM - This won't work for plain SOAP over TCP. We need to detect that HTTP is used
        soap->count = 0;
        if (soap_response(soap, 202) || soap_end_send(soap)) {
            free(soap);
            return soap_closesock(soap);
        }
    }
    free(soap);
    return SOAP_OK;
}

int dpws_recv_empty_response(struct dpws* dpws) {
    struct soap* soap = dpws_dpws2soap(dpws);
    if (DPWS_IS_UDP(dpws->status)) // TODO: AM - This won't work for plain SOAP over TCP. We need to detect that HTTP is used
        return SOAP_OK;
    // The following mimick the recv part of soap_call...
    if (soap_recv_empty_response(dpws_dpws2soap(dpws)))
        return soap_closesock(soap);
    return soap_closesock(soap);
}

static void dpws_set_fault(struct dpws *dpws)
{
    struct soap *soap = &dpws->soap;
    const char **c = soap_faultcode(soap);
    const char **sc = soap_faultsubcode(soap);
    const char **s = soap_faultstring(soap);
    const char **d = soap_faultdetail(soap);
      
    /* Set the fault body */
    switch (soap->error)
    {
        case WSA_ERR_INVALID_HEADER_INFO_MESSAGE_ID: case WSA_ERR_INVALID_HEADER_INFO_RELATES_TO:
        case WSA_ERR_INVALID_HEADER_INFO_TO: case WSA_ERR_INVALID_HEADER_INFO_ACTION: // TODO: not set yet
        case WSA_ERR_INVALID_HEADER_INFO_FROM: case WSA_ERR_INVALID_HEADER_INFO_REPLY_TO:
        case WSA_ERR_INVALID_HEADER_INFO_FAULT_TO:
            *c = "SOAP-ENV:Sender";
            *sc = "wsa:InvalidMessageInformationHeader";
            *s = "A message information header is not valid and the message cannot be processed"; // spec message truncated
            buffer_marshalling_init(dpws);
            dpws_wsa_headers_out(dpws);
            *d = buffer_marshalling_end(dpws);
            break;
        case WSA_ERR_REQUIRED_HEADER_INFO:
            *c = "SOAP-ENV:Sender";
            *sc = "wsa:MessageInformationHeaderRequired";
            *s = "A required message information header, To, MessageID, or Action, is not present.";
            *d = soap->tag; // put it in when raising the error if not already done
            break;
        case WSA_ERR_DESTINATION_UNREACHABLE:
            *c = "SOAP-ENV:Sender";
            *sc = "wsa:DestinationUnreachable";
            *s = "No route can be determined to reach the destination role defined by the WS-Addressing To.";
            break;
        case WSA_ERR_ACTION_NOT_SUPPORTED:
            *c = "SOAP-ENV:Sender";
            *sc = "wsa:ActionNotSupported";
            *s = "The [action] cannot be processed at the receiver.";
            *d = dpws->action;
            break;
    //case WSA_ERR_ENDPOINT_UNAVAILABLE:
    //  *c = "SOAP-ENV:Receiver";
    //break;
        case WSE_ERR_DELIVERY_MODE_REQUESTED_UNAVAILABLE:
            *c = "SOAP-ENV:Sender";
            *sc = "wse:DeliveryModeRequestedUnavailable";
            *s = "The requested delivery mode is not supported.";
            strcpy(soap->msgbuf, "<wse:SupportedDeliveryMode>");
            strcat(soap->msgbuf, WSE_MODE_PUSH);
            strcat(soap->msgbuf, "</wse:SupportedDeliveryMode>");
            *d = soap->msgbuf;
            break;
        case WSE_ERR_INVALID_EXPIRATION_TIME:
            *c = "SOAP-ENV:Sender";
            *sc = "wse:InvalidExpirationTime";
            *s = "The expiration time requested is invalid.";
            break;
        case WSE_ERR_UNSUPPORTED_EXPIRATION_TYPE:
            *c = "SOAP-ENV:Sender";
            *sc = "wse:UnsupportedExpirationType";
            *s = "Only expiration durations are supported.";
            break;
        case WSE_ERR_FILTERING_NOT_SUPPORTED:
            *c = "SOAP-ENV:Sender";
            *sc = "wse:FilteringNotSupported";
            *s = "Filtering is not supported.";
            break;
        case WSE_ERR_FILTERING_REQUESTED_UNAVAILABLE:
            *c = "SOAP-ENV:Sender";
            *sc = "wse:FilteringRequestedUnavailable";
            *s = "The requested filter dialect is not supported.";
            strcpy(soap->msgbuf, "<wse:SupportedDialect>");
            strcat(soap->msgbuf, DPWS_WSE_FILTERING_DIALECT);
            strcat(soap->msgbuf, "</wse:SupportedDialect>");
            *d = soap->msgbuf;
            break;
        //case WSE_ERR_EVENT_SOURCE_UNABLE_TO_PROCESS:
        //  *c = "SOAP-ENV:Receiver";
        //break;
        case WSE_ERR_UNABLE_TO_RENEW:
            *c = "SOAP-ENV:Receiver";
            *sc = "wse:UnableToRenew";
            *s = "The referenced subscription does not exist";
            break;
        case WSE_ERR_INVALID_MESSAGE:
        case WSE_ERR_INVALID_SUBSCRIPTION:  // Not in the spec : invented
            *c = "SOAP-ENV:Sender";
            *sc = "wse:InvalidMessage";
            *s = "The message is not valid and cannot be processed.";
            buffer_marshalling_init(dpws);
            subscManager_putelement(soap, dpws->err_detail_data, dpws->err_detail_data_elt, -1, dpws->err_detail_type);
            *d = buffer_marshalling_end(dpws);
            break;
        case MDM_ERR_MISSING_SCOPES:
            *c = "SOAP-ENV:Sender";
            *sc = "mdm:MissingScopes";
            *s = "The message is not valid since no scope is specified.";
            break;
    }

    /* Set the fault WSA Headers */
    DPWS_SET_RESPONSE(dpws->status);
    
    dpws->to = NULL;
    if (!dpws_endpoint_ref_is_null(dpws->fault_to)) {
        dpws->to = dpws->fault_to.address;
        dpws->subscription_id = dpws->fault_to.subscription_id;
        dpws_default_wsa_endpoint_ref(&dpws->fault_to);
    }
    if (!dpws_endpoint_ref_is_null(dpws->reply_to)) {
        if (!dpws->to) {
            dpws->to = dpws->reply_to.address;
            dpws->subscription_id = dpws->reply_to.subscription_id;
        }
        dpws_default_wsa_endpoint_ref(&dpws->reply_to);
    }
    if (!dpws_endpoint_ref_is_null(dpws->from)) {
        if (!dpws->to) {
            dpws->to = dpws->from.address;
            dpws->subscription_id = dpws->from.subscription_id;
        }
        dpws_default_wsa_endpoint_ref(&dpws->from);
    }
    if (!dpws->to) {
        dpws->to = WSA_ANONYMOUS_ENDPOINT_REF_URI;
        dpws->subscription_id = NULL;
    }
    
        
    dpws->action = WSA_SOAP_FAULT_URI;
    dpws->relates_to = dpws->message_id;
    dpws->message_id = schemed_uuid_create((char *)soap_malloc(&dpws->soap, SCHEME_UUID_STRING_SIZE));
        
    soap_set_fault(soap);
}

int dpws_send_fault(struct dpws *dpws)
{
    struct soap *soap = &dpws->soap;
    register int status = soap->error;
    if (status == SOAP_STOP)
        return status;
    if (dpws_is_oneway(dpws))   // no need to send fault in one way (waiting for "robust oneway")
        return (soap->error = SOAP_OK);
    DBGLOG(TEST,SOAP_MESSAGE(fdebug, "Sending back fault struct for error code %d\n", soap->error));
    soap->keep_alive = 0; /* to terminate connection */
    dpws_set_fault(dpws);
//  if ((status != SOAP_EOF || (!soap->recv_timeout && !soap->send_timeout)) && soap_poll(soap) == SOAP_OK) //TODO: AM - check why soap_poll does not work
    if ((status != SOAP_EOF || (!soap->recv_timeout && !soap->send_timeout)))
    {
        soap->error = SOAP_OK;
        soap_serializeheader(soap);
        soap_serializefault(soap);
        soap_begin_count(soap);
        if (soap->mode & SOAP_IO_LENGTH)
        {
            soap_envelope_begin_out(soap);
            soap_putheader(soap);
            soap_body_begin_out(soap);
            soap_putfault(soap);
            soap_body_end_out(soap);
            soap_envelope_end_out(soap);
        }
        if (dpws_response(dpws, status)
            || soap_envelope_begin_out(soap)
            || soap_putheader(soap)
            || soap_body_begin_out(soap)
            || soap_putfault(soap)
            || soap_body_end_out(soap)
            || soap_envelope_end_out(soap))
            return soap_closesock(soap);
        soap_end_send(soap);
    }
    soap_closesock(soap);
    return soap->error = status;
}

static int dpws_use_http_response(struct dpws* dpws) {
    return !strcmp(dpws->to, WSA_ANONYMOUS_ENDPOINT_REF_URI);
}

static int dpws_udp_connect_command(struct soap *soap, const char *host, int port,
    const char* action, int (*fopen)(struct soap *, const char *, const char *, int))
{
    if (action)
        soap->action = soap_strdup(soap, action);
    if (soap_valid_socket(soap->socket) && !soap->keep_alive)
        dpws_udp_close(soap);   // one-way does not close the socket
    if (!soap_valid_socket(soap->socket))
        soap->socket = fopen(soap, NULL, host, port); 
    if (soap->socket < 0)
        return soap->error;
    if (soap_begin_send(soap))
        return soap->error;
    if ((soap->mode & SOAP_IO) != SOAP_IO_STORE && !(soap->mode & SOAP_ENC_XML)) { 
        unsigned int k = soap->mode;
        soap->mode &= ~(SOAP_IO | SOAP_ENC_ZLIB);
        if ((k & SOAP_IO) != SOAP_IO_FLUSH)
            soap->mode |= SOAP_IO_BUFFER;
        soap->mode = k;
    }
    return SOAP_OK;
}

#ifdef USE_ONVIF /* Added by HG.W */
int dpws_connect(struct dpws* dpws, const char* address, const char* action)
{
    char *endpoint = NULL;

    if(dpws->ipproto_type == DPWS_IPPROTO_IPV4){
        endpoint = DISCOVERY_ENDPOINT;
    }
#ifdef WITH_IPV6
    else if(dpws->ipproto_type == DPWS_IPPROTO_IPV6){
        endpoint = DISCOVERY_ENDPOINT_V6;
    }
#endif
    else{
        /* unknown ipproto type, set to ipv4 */
        endpoint = DISCOVERY_ENDPOINT;
    }

    if (!strcmp(address, WSD_URN))
    {
        soap_set_endpoint(dpws_dpws2soap(dpws), endpoint);
        return dpws_udp_connect_command(dpws_dpws2soap(dpws), dpws_dpws2soap(dpws)->host,
                                        dpws_dpws2soap(dpws)->port, action, &dpws_udp_mcast_open);
    }
    else if (!strcmp(WSA_ANONYMOUS_ENDPOINT_REF_URI, address))
    {
        if (DPWS_IS_UDP(dpws->status))
#ifdef USE_DISCOVERY_PORT_AS_SOURCE_PORT /* Added by HG.W */
            dpws_udp_connect_command(dpws_dpws2soap(dpws), dpws_dpws2soap(dpws)->host,
                                     dpws_dpws2soap(dpws)->port, action, &my_dpws_udp_ucast_open);
#else
            dpws_udp_connect_command(dpws_dpws2soap(dpws), dpws_dpws2soap(dpws)->host,
                                     dpws_dpws2soap(dpws)->port, action, &dpws_udp_ucast_open);
#endif
        return SOAP_OK;
    }
    else
    {
        if (!is_physical_address(address))
            address = spawn_resolve(dpws, address);
        if (address) {
            dpws_set_tcp_mode(dpws);
            return soap_connect(dpws_dpws2soap(dpws), address, action);
        }
        else
            return dpws_dpws2soap(dpws)->error = WSA_ERR_DESTINATION_UNREACHABLE;
    }
}
#else
int dpws_connect(struct dpws* dpws, const char* address, const char* action)
{
    if (!strcmp(address, WSD_URN))
    {
        soap_set_endpoint(dpws_dpws2soap(dpws), DISCOVERY_ENDPOINT);
        return dpws_udp_connect_command(dpws_dpws2soap(dpws), dpws_dpws2soap(dpws)->host, dpws_dpws2soap(dpws)->port, action, &dpws_udp_mcast_open);
    }
    else if (!strcmp(WSA_ANONYMOUS_ENDPOINT_REF_URI, address))
    {
        if (DPWS_IS_UDP(dpws->status))
            dpws_udp_connect_command(dpws_dpws2soap(dpws), dpws_dpws2soap(dpws)->host, dpws_dpws2soap(dpws)->port, action, &dpws_udp_ucast_open);
        return SOAP_OK;
    }
    else
    {
        if (!is_physical_address(address))
            address = spawn_resolve(dpws, address);
        if (address) {
            dpws_set_tcp_mode(dpws);
            return soap_connect(dpws_dpws2soap(dpws), address, action);
        }
        else
            return dpws_dpws2soap(dpws)->error = WSA_ERR_DESTINATION_UNREACHABLE;
    }
}
#endif

int dpws_response(struct dpws *dpws, int status)
{
    char * phys_addr = dpws->to;
    if (dpws_use_http_response(dpws))
        return soap_response(dpws_dpws2soap(dpws), status);
    if (!is_physical_address(phys_addr)) 
        phys_addr = spawn_resolve(dpws, phys_addr);
    soap_closesock(dpws_dpws2soap(dpws));
    return soap_connect(dpws_dpws2soap(dpws), phys_addr, dpws->action);
}

SOAP_FMAC1
void dpws_set_namespaces(struct dpws *dpws, struct Namespace *p)
{
    if (p != dpws->soap.namespaces)
        soap_set_namespaces(&dpws->soap, p);
}

/*----------------------------------------------------------------------------*\
 *                                   EVENTING                                 *
\*----------------------------------------------------------------------------*/
struct wsa_endpoint_ref** dpws_get_subscriptors(struct dpws* dpws, short event_source, char* action)
{
    struct service_endpoint * service;
    dyn_array sinks = DYN_ARRAY_INITIALIZER(sizeof(struct wsa_endpoint_ref *), EVENT_ALLOC_INCREMENT);
    service = get_endpoint(event_source);
    if (service) {
        char ** p_action;
        struct event_subscription * subscription = service->subscriptions, *previous = NULL;
        lock(service->device->lock);
        while (subscription)
        {
            // check that subscription must not be discarded (expired, lazy update)
            if ((unsigned32)subscription->expiration < system_time_s()) {
                struct event_subscription * r_sub;
                // shortcut expired subscription
                if (previous)
                    previous->next = subscription->next;
                else
                    service->subscriptions = subscription->next;
                r_sub = subscription;
                subscription = subscription->next;
                free_subscription(r_sub);
            }
            else {  // valid subscription
                if (subscription->actions.tab) {    // check action against filter
                    for (p_action = subscription->actions.tab; *p_action; p_action++) {
                        // collect endpoint
                        if (rfc2396_matches(*p_action, action)) {
                            * (struct wsa_endpoint_ref **)new_entry(dpws, &sinks) = endpoint_ref_dup(dpws, subscription->notify_to);
                            break;
                        }
                    }
                }
                else    // no filter
                    * (struct wsa_endpoint_ref **)new_entry(dpws, &sinks) = endpoint_ref_dup(dpws, subscription->notify_to);
                    
                previous = subscription;
                subscription = subscription->next;
            }
        }
        unlock(service->device->lock);
    }
    return (struct wsa_endpoint_ref **)sinks.tab;
}
