/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.22.2.2 $
|                                              $Date: 2007/03/23 16:41:34 $
\*============================================================================*/
 
/*******************************************************************************
*                                DEFINITIONS                                   *
*******************************************************************************/
#include "dpwsX.h"
#include "tools.h"
#include "config.h"

extern struct Namespace default_namespaces[];

/**
 * An entry that will be used by the generic handler to call a user callback.
 */
struct cb_entry {
	wscbk cbk; 	/**< callback */
	void * cbk_data;
};


/*******************************************************************************
*                          GENERIC INVOKATION UTILITIES                        *
*******************************************************************************/
static int buffer_write(struct soap *soap, const char *s, size_t n)
{
	size_t count = MIN(n, (size_t)(SOAP_BLKLEN - dpws_soap2dpws(soap)->tmp_buf_count));
	memcpy(dpws_soap2dpws(soap)->err_detail_data_elt + dpws_soap2dpws(soap)->tmp_buf_count, s, count);
	dpws_soap2dpws(soap)->tmp_buf_count += count;
	if (count < n)
	{
		dpws_soap2dpws(soap)->err_detail_data_elt = (char*)soap_push_block(soap, SOAP_BLKLEN);
		dpws_soap2dpws(soap)->tmp_buf_count = 0;
		buffer_write(soap, s + count, n - count);
	}
	return SOAP_OK;
}

static void add_missing_prefix_def_att(struct soap *soap, unsigned int root_level, const char *tag)
{
	struct soap_nlist *np;
	const char *s;
	
	np = soap->nlist;
	if (!(s = strchr(tag, ':')))
	{
		while (np && *np->id) /* find default namespace, if present */
			np = np->next;
	}
	else
	{
		while (np && (strncmp(np->id, tag, s - tag) || np->id[s - tag]))
			np = np->next;
	}
	if (np)
	{
		if (np->level < root_level)
		{
			const char *ns = np->index >= 0 ? soap->namespaces[np->index].ns : np->ns;
			if (np->id)
      			sprintf(soap->tmpbuf, "xmlns:%s", np->id);
			soap_set_attr(soap, np->id ? soap->tmpbuf : "xmlns", ns);	// replaces if already exists
			soap_push_namespace(soap, np->id, ns);	// to avoid multiple insertions
		}
	}
	else
		soap->error = SOAP_NAMESPACE;
}

static int browse_element(struct soap *soap, int root_level)
{
	int ret = SOAP_OK;
	register struct soap_attribute *tp;
	char * tag;
	
	if ((ret = soap_peek_element(soap)))
		return ret;

	/* ns processing (add soap->attributes with soap_set_attr() if necessary)*/
	tag = soap_strdup(soap, soap->tag);	// could try to unlink and free it...
	add_missing_prefix_def_att(soap, root_level, tag);

	for (tp = soap->attributes; tp; tp = tp->next)
	{
		if (tp->visible == 1)	// ignore xmlmns
			add_missing_prefix_def_att(soap, root_level, tp->name);
	}
	// NOTE: attribute for reading and writing are the same.
  
	/* Element processing */
	ret = soap_element_begin_in(soap, NULL, 1);
	ret = soap_element_begin_out(soap, tag, -1, NULL);
	soap->level--;
	
	if (soap->body)
	{
		if (!(ret = soap_peek_element(soap)))
		{	// recursive processing
			for (;;)
			{
				if ((ret = browse_element(soap, root_level)))
				{
					if (ret == SOAP_NO_TAG)
						soap->error = SOAP_OK;
					else
						return ret;
					break;
				}
			}
		}	// leaf processing
		else if (ret == SOAP_NO_TAG)
			soap_string_out(soap, soap_string_in(soap, 1, -1, -1), 1);
		else
			return ret;
			
		if ((ret = soap_element_end_out(soap, tag)))
			return ret;
		soap->level++;
		if ((ret = soap_element_end_in(soap, tag)))
			return ret;
	}
	return ret;
}

static int dpws_inliteral(struct dpws *dpws, char **out)
{
	int ret = SOAP_OK;
	struct soap *soap = dpws_dpws2soap(dpws);
	size_t bodySize;

	/* buffer init */
	dpws->tmp_buf_count = 0;
	dpws->backup_send_f = soap->fsend;
	soap->fsend = &buffer_write;
	soap_new_block(soap);
	dpws->err_detail_data_elt = (char*)soap_push_block(soap, SOAP_BLKLEN);	// used to store the current buffer

	/* browse XML tree */
	ret = browse_element(soap, soap->level);

	/* result allocation */
	soap->fsend = dpws->backup_send_f;
	bodySize = soap_size_block(soap, dpws->tmp_buf_count + 1);
	*out = soap_save_block(soap, NULL, 0);
	(*out)[bodySize - 1] = 0;
	dpws->err_detail_data_elt = NULL;
	return ret;
}

static int handle_body(struct dpws *dpws, struct cb_entry *entry)
{
	char *strBody = NULL;
	struct soap *soap = dpws_dpws2soap(dpws);
	
	soap->encodingStyle = NULL;
	if (dpws_inliteral(dpws, &strBody))
		return soap->error;
		
	if (soap_body_end_in(soap)
	 	|| soap_envelope_end_in(soap)
	 	|| soap_end_recv(soap)
	 	|| dpws_check_wsa_headers(dpws))
		return soap->error;
	soap->error = entry->cbk(dpws, entry->cbk_data, strBody);
	if (soap->error)
		return soap->error;
	return soap_closesock(soap);
}

static int handle_msg(struct dpws *dpws)
{
	struct soap *soap = &dpws->soap;
	struct cb_entry * entry = (struct cb_entry*)(get_endpoint(dpws->href_endpoint)->user_data);
	
	if (entry)
	{
//		dpws_set_namespaces(dpws, default_namespaces);
		soap->ns = 1;
		soap_peek_element(soap);
		// NOTE: To establish the source (if not explicitly given in from), 2 solutions:
		//	1. Use a reference parameter during subscription (not necessarily managed by DPWS servers, e.g. Ours does not),
		//	2. Use different endpoints (device endpoints) for every subscription.
		// The chosen one is the 2nd. This implies that action unicity is checked when subscriptions are added to an endpoint
		return handle_body(dpws, entry);
	}
	return soap->error = SOAP_NO_METHOD;
}


/*******************************************************************************
*                             GENERIC INVOKATION API                           *
*******************************************************************************/
int dpws_send_message(struct dpws *dpws, char * action, struct wsa_endpoint_ref *to, char *XML_msg_body)
{
	struct soap *soap = &dpws->soap;
	dpws_set_namespaces(dpws, default_namespaces);
	dpws_set_oneway_wsa_headers(dpws, to, action);
	soap->encodingStyle = NULL;
	soap_begin(soap);
	soap_serializeheader(soap);
	if (soap_begin_count(soap))
		return soap->error;
	if (soap->mode & SOAP_IO_LENGTH)
	{
		if (soap_envelope_begin_out(soap)
		 || soap_putheader(soap)
		 || soap_body_begin_out(soap)
		 ||	soap_outliteral(soap, NULL, (char **)&XML_msg_body)
		 || soap_body_end_out(soap)
		 || soap_envelope_end_out(soap))
			 return soap->error;
	}
	if (soap_end_count(soap))
		return soap->error;
	if (dpws_connect(dpws, to->address, action)
	 || soap_envelope_begin_out(soap)
	 || soap_putheader(soap)
	 || soap_body_begin_out(soap)
	 ||	soap_outliteral(soap, NULL, (char **)&XML_msg_body)
	 || soap_body_end_out(soap)
	 || soap_envelope_end_out(soap)
	 || soap_end_send(soap))
	{
		if (soap->error == SOAP_TCP_ERROR)
			dpws_invalidate(dpws, to);
		return soap_closesock(soap);
	}
	return SOAP_OK;
}

int dpws_call_service(
	struct dpws *dpws, char * action,
	struct wsa_endpoint_ref *to, struct wsa_endpoint_ref *reply_to,
	char *XML_request_body,	char **response
	)
{
	struct soap *soap = &dpws->soap;
	dpws_set_namespaces(dpws, default_namespaces);
	dpws_set_call_wsa_headers(dpws, to, reply_to, action);
	soap->encodingStyle = NULL;
	soap_begin(soap);
	soap_serializeheader(soap);
	if (soap_begin_count(soap))
		return soap->error;
	if (soap->mode & SOAP_IO_LENGTH)
	{
		if (soap_envelope_begin_out(soap)
			|| soap_putheader(soap)
			|| soap_body_begin_out(soap)
			||	soap_outliteral(soap, NULL, (char **)&XML_request_body)
			|| soap_body_end_out(soap)
			|| soap_envelope_end_out(soap))
			return soap->error;
	}
	if (soap_end_count(soap))
		return soap->error;
	if (dpws_connect(dpws, to->address, action)
		|| soap_envelope_begin_out(soap)
		|| soap_putheader(soap)
		|| soap_body_begin_out(soap)
		||	soap_outliteral(soap, NULL, (char **)&XML_request_body)
		|| soap_body_end_out(soap)
		|| soap_envelope_end_out(soap)
		|| soap_end_send(soap))
	{
		if (soap->error == SOAP_TCP_ERROR)
			dpws_invalidate(dpws, to);
		return soap_closesock(soap);
	}
	if (dpws_is_asynchronous(dpws)) {
		*response = dpws->message_id;
		return SOAP_OK;
	}
	else {
		dpws_init_wsa_headers(dpws);
		if (soap_begin_recv(soap)
			|| soap_envelope_begin_in(soap)
			|| soap_recv_header(soap)
			|| soap_body_begin_in(soap))
			return soap_closesock(soap);
		dpws_inliteral(dpws, (char **)response);	// TODO: dom-like parsing
		if (soap->error)
		{
			if (soap->error == SOAP_TAG_MISMATCH && soap->level == 2)
				return soap_recv_fault(soap);
			return soap_closesock(soap);
		}
		if (soap_body_end_in(soap)
			|| soap_envelope_end_in(soap)
			|| soap_end_recv(soap))
			return soap_closesock(soap);
		return soap_closesock(soap);
	}
}

/*******************************************************************************
*                              GENERIC HANDLING API                            *
*******************************************************************************/
int dpws_register_generic_handler(short href_endpoint, wscbk callback, void *cbk_data)
{
	struct service_endpoint * endpoint = get_endpoint(href_endpoint);
	struct cb_entry *pEntry;
	
	if (!endpoint)
		return DPWS_ERR_COULD_NOT_FIND_ENDPOINT;
	if (endpoint->user_data)
		return DPWS_ERR_CONFLICTING_OPERATION;
		
	// add the generic handler (doubles are checked)
	dpws_add_ptr_att(href_endpoint, DPWS_PTR_HANDLING_FUNCTION, handle_msg);
	
	// Create callback entry
	pEntry = (struct cb_entry *)malloc(sizeof(struct cb_entry));
	pEntry->cbk = callback;
	pEntry->cbk_data = cbk_data;
	endpoint->user_data = pEntry;

	return DPWS_OK;
}

int dpws_clean_endpoint(short href_endpoint, void (* clean_cbk) (void *, void *), void * cbk_param)
{
	struct service_endpoint * endpoint = get_endpoint(href_endpoint);

	if (!endpoint)
		return DPWS_ERR_COULD_NOT_FIND_ENDPOINT;

	if (endpoint->user_data) {
		struct cb_entry * pEntry = (struct cb_entry *)endpoint->user_data;
		if (clean_cbk)
			clean_cbk(pEntry->cbk_data, cbk_param);
		free(pEntry);
	}

	return DPWS_OK;
}
