/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.50.2.4 $
|                                              $Date: 2007/04/20 12:44:13 $
\*============================================================================*/
 
/*******************************************************************************
*                            WS-EVENTING STUB & SKELETON                       *
*******************************************************************************/
#include "event.h"
#include "tools.h"
#include "config.h"
#include "discovery.h"

extern struct Namespace default_namespaces[];
extern handlePool_t registryPool;
extern scheduler_t scheduler;
extern short scheduler_lock_handle;
extern struct dpws_config tk_config;

extern int schedule_discovery_msg(scheduler_t *pScheduler, short msgType, char * req_id,
	struct device *device, int (* handler) (struct dpws*, void *), short hrefLock);
extern int send_wakeup_datagram();

// wse: built-in compiler prefix (ns table not generated)
struct Namespace eventsource_snd_namespaces[] =
{
	{"SOAP-ENV", "http://www.w3.org/2003/05/soap-envelope", "http://www.w3.org/2003/05/soap-envelope", NULL},
	{"wsa", "http://schemas.xmlsoap.org/ws/2004/08/addressing", "http://schemas.xmlsoap.org/ws/*/addressing", NULL},
	{"wse", "http://schemas.xmlsoap.org/ws/2004/08/eventing", "http://schemas.xmlsoap.org/ws/*/eventing", NULL},
	{NULL, NULL, NULL, NULL}
};

/*******************************************************************************
*                                 MARSHALLING                                  *
*******************************************************************************/
#define WITH_NOGLOBAL
#ifdef SOAP_FMAC3
#undef SOAP_FMAC3
#endif
#define SOAP_FMAC3 static
#include "gen/eventsourceC.c"
#undef WITH_NOGLOBAL

/*******************************************************************************
*                                    STUB                                      *
*******************************************************************************/
#define eventsource_namespaces (default_namespaces)
#include "gen/eventsourceClient.c"

/*******************************************************************************
*                                  SKELETON                                    *
*******************************************************************************/
#define WITH_NOSERVEREQUEST
#include "gen/eventsourceServer.c"
#undef WITH_NOSERVEREQUEST
#undef eventsource_namespaces

int eventing_handle_event(struct dpws *dpws)
{
	struct soap *soap = &dpws->soap;
	const char* action = dpws->action ? dpws->action : "";
	dpws_set_namespaces(dpws, default_namespaces);
	soap_peek_element(soap);
	if (!strcmp(action, "http://schemas.xmlsoap.org/ws/2004/08/eventing/EventSource/SubscriptionEnd") || (!strlen(action) && !soap_match_tag(soap, soap->tag, "wse:SubscriptionEnd")))
		return soap_serve___wse__SubscriptionEnd(soap);
	return soap->error = SOAP_NO_METHOD;
}

/******************************************************************************\
 *                                                                            *
 * Service Operations (Server-side)                                           *
 *                                                                            *
\******************************************************************************/

// TODO: avoid doubles and discard subscriptions that fail + discard subscriptions on bye. Group susbcriptions by endpoints in order to cancel when one fails ?
int __wse__SubscribeOp(struct dpws * dpws, struct _wse__Subscribe * msg, struct _wse__SubscribeResponse *response)
{
	struct service_endpoint * service = NULL;
	struct event_subscription * subscription = NULL;
	time_t expiration;
	BOOLEAN hasFilter = FALSE;
	int ret = DPWS_OK;
	
	// checks
	if (!msg->wse__Delivery->Mode)
		msg->wse__Delivery->Mode = WSE_MODE_PUSH;
	else if (strcmp(msg->wse__Delivery->Mode, WSE_MODE_PUSH))
		return WSE_ERR_DELIVERY_MODE_REQUESTED_UNAVAILABLE;
		
	service = get_endpoint(dpws->href_endpoint);
	if (!msg->wse__Delivery->wse__NotifyTo) {
		ret = WSE_ERR_INVALID_MESSAGE;
		dpws->err_detail_type = SOAP_TYPE__wse__Subscribe;
		dpws->err_detail_data = msg;
		dpws->err_detail_data_elt = "wse:SubscribeOp";
		return ret;
	}
	if (msg->wse__Expires) {
		expiration = xmlduration2duration(msg->wse__Expires);
		if (expiration == 0)
			return WSE_ERR_UNSUPPORTED_EXPIRATION_TYPE;

		if (expiration < 0)
			return WSE_ERR_INVALID_EXPIRATION_TIME;
		lock(service->device->lock);
		if (expiration > service->max_subsc_duration)
			expiration = service->max_subsc_duration;
	} else {
		lock(service->device->lock);
		expiration = service->max_subsc_duration;
	}

	expiration += system_time_s();

	if (msg->wse__Filter) {
		if (
			msg->wse__Filter->Dialect
			&& !strcmp(msg->wse__Filter->Dialect, DPWS_WSE_FILTERING_DIALECT)
			&& msg->wse__Filter->__item
			)
			hasFilter = TRUE;
		else {
			ret = WSE_ERR_FILTERING_REQUESTED_UNAVAILABLE;
			goto exit;
		}
	}
	// else no filter

	// create or overwrite the subscription
	for(subscription = service->subscriptions; subscription; subscription = subscription->next) {
		if (!strcmp(subscription->notify_to->address, msg->wse__Delivery->wse__NotifyTo->address)) {
			// verify filters are equals
			if ((!hasFilter && subscription->actions.nb == 0) || (hasFilter
				&& is_rfc2396_included(msg->wse__Filter->__item, &subscription->actions)
				&& is_rfc2396_included(&subscription->actions, msg->wse__Filter->__item)))
				break;
		}
	}
	
	if (!subscription) {
		subscription = malloc(sizeof(struct event_subscription));
		subscription->next = service->subscriptions;
		service->subscriptions = subscription;
	}

	schemed_uuid_create(subscription->id);
	subscription->notify_to = dpws_endpoint_ref_dup(msg->wse__Delivery->wse__NotifyTo);
	if (hasFilter)
		string_array_dup(NULL, &subscription->actions, (dyn_array *)msg->wse__Filter->__item);
	else
		init_dyn_array(&subscription->actions, sizeof(char *), 1);

	subscription->end_to = dpws_endpoint_ref_dup((struct wsa_endpoint_ref *)msg->wse__EndTo);
	subscription->expiration = expiration;
	
	DPWSLOG1("<- Subscribing on %s\n", dpws->to);
	response->wse__SubscriptionManager = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct wsa_endpoint_ref));
	dpws_default_wsa_endpoint_ref(response->wse__SubscriptionManager);
	response->wse__SubscriptionManager->subscription_id = subscription->id;	// NOTE: memory risk if the registration is removed by a concurrent thread. Should not occur since registration is event sink private (except odd design)
	response->wse__SubscriptionManager->address = soap_malloc(dpws_dpws2soap(dpws), strlen(dpws->to) + 1);
	strcpy(response->wse__SubscriptionManager->address, dpws->to);	// Because the eventsource is also the subscription manager and the EPR protocol used for subscription is likely to be supported... 
	
	response->wse__Expires = soap_malloc(dpws_dpws2soap(dpws), DURATION_MAX_SIZE);
	time2xmlduration(response->wse__Expires, subscription->expiration); // DPWS authorizes to reply with a duration
	
	// hosted service will be the subscription manager : register server functions (once)
	// NOTE: won't be removed.
	service->subsc_manager = 1;

exit:

	unlock(service->device->lock);
	return ret;
}


int __wse__SubscriptionEnd(struct dpws* dpws, struct _wse__SubscriptionEnd *msg)
{
	struct event_end_info * event_end = NULL;
	int i;
	struct service_class * servClass = (struct service_class *)getObject(&registryPool, get_endpoint(dpws->href_endpoint)->hrefServiceClass);

	event_end = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct event_end_info));
	event_end->subscription_manager = msg->wse__SubscriptionManager;
	event_end->status = msg->wse__Status;
	
	// Status (if more than one take the one without @lang or in english (first of the 2)
	i = find_local_entry((void **)msg->wse__Reason, msg->__sizeReason, tk_config.preferred_lang);
	if (i >= 0)
		event_end->reason = msg->wse__Reason[i]->__item;
	else
		event_end->reason = NULL;
		
	servClass->event_end_hook(event_end);
	return SOAP_OK;
}


/*----------------------------------------------------------------------------*\
 *                                    API                                     *
\*----------------------------------------------------------------------------*/
struct wsa_endpoint_ref* dpws_event_subscribe(struct dpws* dpws,
	struct wsa_endpoint_ref* event_source, struct wsa_endpoint_ref* notify_to,
	struct wsa_endpoint_ref* end_to, char ** filter, char ** expiration)
{
	struct _wse__Subscribe subscription;
	struct wse__DeliveryType delivery;
	struct wse__FilterType s_filter;
	struct _wse__SubscribeResponse response;
	dyn_array actions;
	struct wsa_endpoint_ref* ret = NULL;
	
	soap_default__wse__Subscribe(dpws_dpws2soap(dpws), &subscription);
	subscription.wse__EndTo = end_to;
	subscription.wse__Delivery = &delivery;
	delivery.Mode = WSE_MODE_PUSH;
	delivery.wse__NotifyTo = notify_to;
	subscription.wse__Expires = *expiration;
	if (!filter || !*filter)
		subscription.wse__Filter = NULL;
	else {
		subscription.wse__Filter = &s_filter;
		s_filter.__item = &actions;
		s_filter.Dialect = DPWS_WSE_FILTERING_DIALECT;
		actions.tab = filter;	// NOTE: custom uri list marshalling use the null boundary and not the nb field
	}
	if (dpws_call___wse__SubscribeOp(dpws, event_source, NULL, &subscription, &response))
		goto exit;
	*expiration = response.wse__Expires;
	ret = response.wse__SubscriptionManager;
	
exit:
    dpws_init_wsa_headers(dpws);
	return ret;
}

/*----------------------------------------------------------------------------*\
 *                   RUNTIME (used by the generated code)                     *
\*----------------------------------------------------------------------------*/
int send_eventing_end(struct dpws* dpws, struct device *device)
{
	// TODO: handle it like an ordinary event ?
	int ret = DPWS_OK;
	struct wsa_endpoint_ref manager_endpoint;
	struct _wse__SubscriptionEnd body;
	struct service_endpoint * service = NULL;
	struct event_subscription * subscription = NULL;
	struct wse__LanguageSpecificStringType reason = {NULL, DPWS_EVENT_SUBSCRIPTION_END_REASON},
											*pReason = &reason;
	
	lock(device->lock);

	/* set message data */
	soap_default__wse__SubscriptionEnd(dpws_dpws2soap(dpws), &body);
	dpws_default_wsa_endpoint_ref(&manager_endpoint);
	body.wse__SubscriptionManager = &manager_endpoint;
	manager_endpoint.address = device->uuid;
	body.wse__Status = (char *)soap_wse__SubscriptionEndCodeType2s(dpws_dpws2soap(dpws),
		wse__SubscriptionEndCodeType__http___x002f_x002fschemas_x002exmlsoap_x002eorg_x002fws_x002f2004_x002f08_x002feventing_x002fSourceShuttingDown);	// NOTE: Failures are not implemented since altered working is not supported
	body.__sizeReason = 1;
	body.wse__Reason = &pReason;
	
	device->bye_scheduled = FALSE;
	
	for (service = device->hosted_services; service; service = service->next) {
		while (service->subscriptions) {
			subscription = service->subscriptions;
			if (subscription->end_to) {
				manager_endpoint.subscription_id = subscription->id;
				dpws_send___wse__SubscriptionEnd(dpws, subscription->end_to, &body);	// NOTE: continue even if an error encountered
				soap_closesock(dpws_dpws2soap(dpws));	// one-way does not close the socket
				// NOTE: no liberation since the whole configuration will be freed since we're shutting down
				dpws_end(dpws);
			}
			service->subscriptions = subscription->next;
			free_subscription(subscription);
		}
	}
	
	// now that event end  are sent, BYE can be scheduled
	schedule_discovery_msg(&scheduler, DPWS_MSG_BYE, NULL, device, (int (*) (struct dpws*, void *))send_bye, scheduler_lock_handle);
 	ret = send_wakeup_datagram();

	device->bye_scheduled = TRUE;
	unlock(device->lock);
 	return ret;
}

int dpws_serve_subscribe(struct dpws* dpws)
{
	return soap_serve___wse__SubscribeOp(dpws_dpws2soap(dpws));
}

// TODO: improve error processing when sending events (remove subscription or put a timeout for connection ?)
