/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.19.4.1 $
|                                              $Date: 2007/02/13 10:03:06 $
\*============================================================================*/
 
/*******************************************************************************
*                            WS-EVENTING STUB & SKELETON                       *
*******************************************************************************/
#include "mdm.h"
#include "tools.h"
#include "config.h"
#include "discovery.h"

extern handlePool_t registryPool;
extern struct dpws_config tk_config;

extern void update_metadata(struct device * device);

#include "gen/mdm.nsmap"

/*******************************************************************************
*                                 MARSHALLING                                  *
*******************************************************************************/
#define WITH_NOGLOBAL
#ifdef SOAP_FMAC3
#undef SOAP_FMAC3
#endif
#define SOAP_FMAC3 static
#include "gen/mdmC.c"
#undef WITH_NOGLOBAL

/*******************************************************************************
*                                    STUB                                      *
*******************************************************************************/
#include "gen/mdmClient.c"

/*******************************************************************************
*                                  SKELETON                                    *
*******************************************************************************/
#include "gen/mdmServer.c"

/******************************************************************************\
 *                                                                            *
 * Service Operations (Server-side)                                           *
 *                                                                            *
\******************************************************************************/

int __mdm__UpdateScopesMetadataOp(struct dpws* dpws, struct _mdm__UpdateScopesMetadata * cmd, struct _mdm__ScopesMetadata * response)
{
	return WSA_ERR_ACTION_NOT_SUPPORTED;
}

int __mdm__SetScopesMetadataOp(struct dpws* dpws, struct _mdm__ScopesMetadata *req, struct _mdm__ScopesMetadata *rsp)
{
	struct device * device;
	struct device_class * devClass;
	int i;

	if (!req->Scopes)
		return MDM_ERR_MISSING_SCOPES;

	device = get_endpoint(dpws->href_endpoint)->device;
	devClass = (struct device_class *)getObject(&registryPool, device->hrefClass);

	lock(device->lock);	// because of potential probe handling
	for (i = 0; i < req->Scopes->nb; i++) {
		if (set_device_ptr_att(device, DPWS_STR_SCOPE, *(char **)get_entry(req->Scopes,i), FALSE)) {
			unlock(device->lock);
			return soap_sender_fault(dpws_dpws2soap(dpws), "device scope size longer than MAX_URI_SIZE", NULL);	// TODO: create an error code ?
		}
	}
	update_metadata(device);
	unlock(device->lock);
	
	rsp->Scopes = req->Scopes;

	return DPWS_OK;
}

int __mdm__SetFriendlyNameMetadataOp(struct dpws* dpws, struct _mdm__SetFriendlyNameMetadata * req, struct _mdm__SetFriendlyNameMetadataResponse * rsp)
{
	int i;
	char * friendly_name = NULL;
	struct device * device;
	struct device_class * devClass;

	/* NOTE: DPWS 1.0; Device config can host only one friendlyName, and @lang
	 * not supported yet the following code partially takes into account such
	 * limit */
	device = get_endpoint(dpws->href_endpoint)->device;
	devClass = (struct device_class *)getObject(&registryPool, device->hrefClass);
	
	i = find_local_entry((void **)req->friendlyName, req->__sizefriendlyName, tk_config.preferred_lang);
	if (i >= 0) {
		if (tk_config.preferred_lang) {	// must be the preferred or default
			if (!req->friendlyName[i]->xml__lang_ || !strcmp(req->friendlyName[i]->xml__lang_, tk_config.preferred_lang))
				friendly_name = req->friendlyName[i]->__item;
			else
				return soap_sender_fault(dpws_dpws2soap(dpws),
					"No friendly name could match the device language", NULL);	// TODO: create an error code ?
		}
		else 
			friendly_name = req->friendlyName[i]->__item;
	}

	if (strlen(friendly_name) > MAX_FIELD_SIZE)
		return soap_sender_fault(dpws_dpws2soap(dpws), "device friendly name size longer than MAX_FIELD_SIZE", NULL);	// TODO: create an error code ?

	lock(device->lock);	// because of potential metadata exchange handling
	set_device_ptr_att(device, DPWS_STR_FRIENDLY_NAME, friendly_name, FALSE);
	update_metadata(device);
	unlock(device->lock);

	/* Build response */
	rsp->__sizefriendlyName = 1;
	rsp->friendlyName = req->friendlyName + i;

	return DPWS_OK;
}
