/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.11.4.1 $
|                                              $Date: 2007/02/13 10:03:05 $
\*============================================================================*/
 
#include "_dpws.h"
#include "stdsoap2.c"

void dpws_set_tcp_mode(struct dpws * dpws) {
	dpws_dpws2soap(dpws)->fopen = tcp_connect;
	dpws_dpws2soap(dpws)->fclose = tcp_disconnect;
	dpws_dpws2soap(dpws)->fsend = fsend;
	dpws_dpws2soap(dpws)->frecv = frecv;
}

const char* dpws_ip_error(struct dpws *dpws)
{
	register const char *msg;
#ifndef WITH_LEAN
	msg = soap_str_code(h_error_codes, dpws_dpws2soap(dpws)->errnum);
	if (!msg)
#endif
	{
		sprintf(dpws_dpws2soap(dpws)->msgbuf, "TCP/UDP IP error %d (%s)",
			dpws_dpws2soap(dpws)->errnum, soap_strerror(dpws_dpws2soap(dpws)));
		msg = dpws_dpws2soap(dpws)->msgbuf;
	}
	return msg;
}

int dpws_gethost(struct dpws *dpws, const char *addr, struct in_addr *inaddr)
{
#ifndef WITH_IPV6 /* Added by HG.W */
	return tcp_gethost(dpws_dpws2soap(dpws), addr, inaddr);
#else
    return DPWS_ERR_INVALID_PARAMETER;/* TO-DO define an error for this */
#endif
}

SOAP_FMAC1 const char* SOAP_FMAC2 soap_QNameStruct2s(struct soap *soap, const char *s)
{
  struct Namespace *p;
  char *t;
  int n;
  if (!s || *s != '"')
  {
#ifndef WITH_LEAN
    if (s && (soap->mode & SOAP_XML_CANONICAL))
    { t = strchr(s, ':');
      if (t)
        soap_utilize_ns(soap, s, t - s);
    }
#endif
    return s;
  }
  s++;
  if ((p = soap->local_namespaces))
  { for (; p->id; p++)
    { if (p->ns)
        if (!soap_tag_cmp(s, p->ns))
          break;
      if (p->in)
        if (!soap_tag_cmp(s, p->in))
          break;
    }
    if (p && p->id)
    { s = strchr(s, '"');
      if (s)
      { t = (char*)soap_malloc(soap, strlen(p->id) + strlen(s));
        strcpy(t, p->id);
	strcat(t, s + 1);
        return t;
      }
    }
  }
  t = (char*)strchr(s, '"');
  if (t)
    n = t - s;
  else
    n = 0;
  t = soap_strdup(soap, s);
  t[n] = '\0';
  sprintf(soap->tmpbuf, "xmlns:_%d", soap->idnum++);
  soap_set_attr(soap, soap->tmpbuf, t);
  s = strchr(s, '"');
  if (s)
  { t = (char*)soap_malloc(soap, strlen(soap->tmpbuf) + strlen(s) - 6);
    strcpy(t, soap->tmpbuf + 6);
    strcat(t, s + 1);
  }
  return t;
}

SOAP_FMAC1 int SOAP_FMAC2 soap_s2QNameStruct(struct soap *soap, const char *s, char **t)
{
  if (s)
  { struct soap_nlist *np;
    const char *p;
    if (!strncmp(s, "xml:", 4))
    { *t = soap_strdup(soap, s);
      return SOAP_OK;
    }
    np = soap->nlist;
    p = strchr(s, ':');
    if (p)
    { register int n = p - s;
      while (np && (strncmp(np->id, s, n) || np->id[n]))
        np = np->next;
      p++;
    }
    else
    { while (np && *np->id)
        np = np->next;
      p = s;
    }
    if (np)
    { if (np->index >= 0 && soap->local_namespaces)
      { register const char *q = soap->local_namespaces[np->index].id;
        if (q)
        { if ((*t = (char*)soap_malloc(soap, strlen(p) + strlen(q) + 2)))
            sprintf(*t, "%s:%s", q, p);
          return SOAP_OK;
        }
      }
      if (np->ns)
      { if ((*t = (char*)soap_malloc(soap, strlen(p) + strlen(np->ns) + 4)))
          sprintf(*t, "\"%s\":%s", np->ns, p);
        return SOAP_OK;
      }
      DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Namespace prefix of '%s' not defined (index=%d, URI=%s)\n", s, np->index, np->ns?np->ns:""));
      return soap->error = SOAP_NAMESPACE; 
    }
    DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Namespace prefix of '%s' not defined, assuming empty namespace\n", s));
    if ((*t = (char*)soap_malloc(soap, strlen(p) + 4)))
      sprintf(*t, "\"\":%s", p);
  }
  return soap->error;
}
