/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.14.2.2 $
|                                              $Date: 2007/03/15 17:26:19 $
\*============================================================================*/
 
/*******************************************************************************
*                            WS-EVENTING STUB & SKELETON                       *
*******************************************************************************/
#include "event.h"
#include "tools.h"
#include "config.h"

extern struct Namespace default_namespaces[];

// wse: built-in compiler prefix (ns table not generated)
struct Namespace subscmanager_snd_namespaces[] =
{
	{"SOAP-ENV", "http://www.w3.org/2003/05/soap-envelope", "http://www.w3.org/2003/05/soap-envelope", NULL},
	{"wsa", "http://schemas.xmlsoap.org/ws/2004/08/addressing", "http://schemas.xmlsoap.org/ws/*/addressing"},
	{"wse", "http://schemas.xmlsoap.org/ws/2004/08/eventing", "http://schemas.xmlsoap.org/ws/*/eventing"},
	{NULL, NULL, NULL, NULL}
};

/*******************************************************************************
*                                 MARSHALLING                                  *
*******************************************************************************/
#define WITH_NOGLOBAL
#ifdef SOAP_FMAC3
#undef SOAP_FMAC3
#endif
#define SOAP_FMAC3 static
#include "gen/subscmanagerC.c"
#undef WITH_NOGLOBAL

int subscManager_putelement(struct soap *soap, const void *ptr, const char *tag, int id, int type)
{
	return soap_putelement(soap, ptr, tag, id, type);
}

/*******************************************************************************
*                                    STUB                                      *
*******************************************************************************/
#define subscmanager_namespaces (default_namespaces)
#include "gen/subscmanagerClient.c"

/*******************************************************************************
*                                  SKELETON                                    *
*******************************************************************************/
#include "gen/subscmanagerServer.c"
#undef subscmanager_namespaces

/******************************************************************************\
 *                                                                            *
 * Service Operations (Server-side)                                           *
 *                                                                            *
\******************************************************************************/

int __wse__RenewOp(struct dpws * dpws, struct _wse__Renew * renew, struct _wse__RenewResponse * response)
{
	struct event_subscription * subscription = NULL;
	int ret = DPWS_OK;
	time_t expiration;
	struct service_endpoint * service = get_endpoint(dpws->href_endpoint);
	
	lock(service->device->lock);
	if (dpws->subscription_id)
		subscription = find_subscription(service, dpws->subscription_id);
	if (!subscription) {
		ret = WSE_ERR_UNABLE_TO_RENEW;	// TODO: Spec not correct (does not mean the subscription does not exist and WSA_ERR_ENDPOINT_UNAVAILABLE is temporarily unavailable)
		dpws->err_detail_type = SOAP_TYPE__wse__Renew;
		dpws->err_detail_data = NULL;
		dpws->err_detail_data_elt = NULL;
		goto exit;
	}

	if (renew->wse__Expires) {
/*
		expiration = xmlduration2time(renew->Expires);
		if (expiration == -1) {
			ret = WSE_ERR_UNSUPPORTED_EXPIRATION_TYPE;
			goto exit;
		}
		if (expiration < system_time_s()) {
			ret = WSE_ERR_INVALID_EXPIRATION_TIME;
			goto exit;
		}
*/
		expiration = xmlduration2duration(renew->wse__Expires);
		if (expiration == 0) {
			ret = WSE_ERR_UNSUPPORTED_EXPIRATION_TYPE;
			goto exit;
		}

		if (expiration < 0) {
			ret = WSE_ERR_INVALID_EXPIRATION_TIME;
			goto exit;
		}
		if (expiration > service->max_subsc_duration)
			expiration = service->max_subsc_duration;
	}
	else
		expiration = service->max_subsc_duration;

	expiration += system_time_s();
	subscription->expiration = expiration;
	
	response->wse__Expires = (char*)soap_malloc(dpws_dpws2soap(dpws), DURATION_MAX_SIZE);
	time2xmlduration(response->wse__Expires, subscription->expiration); // DPWS authorizes to reply with a duration

exit:
	unlock(service->device->lock);
	return ret;
}

int __wse__GetStatusOp(struct dpws * dpws, struct _wse__GetStatus * status, struct _wse__GetStatusResponse * response)
{
	struct event_subscription * subscription = NULL;
	int ret = DPWS_OK;
	struct service_endpoint * service = get_endpoint(dpws->href_endpoint);
	
	lock(service->device->lock);
	if (dpws->subscription_id)
		subscription = find_subscription(service, dpws->subscription_id);
		
	if (!subscription) {
		ret = WSE_ERR_INVALID_SUBSCRIPTION;
		dpws->err_detail_type = SOAP_TYPE__wse__GetStatus;
		dpws->err_detail_data = NULL;
		dpws->err_detail_data_elt = NULL;
		goto exit;
	}

	response->wse__Expires = soap_malloc(dpws_dpws2soap(dpws), DURATION_MAX_SIZE);
	time2xmlduration(response->wse__Expires, subscription->expiration); // DPWS authorizes to reply with a duration

exit:

	unlock(service->device->lock);
	return ret;
}

int __wse__UnsubscribeOp(struct dpws* dpws, struct _wse__Unsubscribe * in, struct __wse__UnsubscribeOpResponse *out)
{
	int ret = DPWS_OK;
	struct service_endpoint * service = get_endpoint(dpws->href_endpoint);
	
	DPWSLOG1("<- Unsubscribing on %s\n", dpws->to);
	lock(service->device->lock);
	if (dpws->subscription_id)
		remove_subscription(service, dpws->subscription_id);
	else {
		ret = WSE_ERR_INVALID_SUBSCRIPTION;
		dpws->err_detail_type = SOAP_TYPE__wse__Unsubscribe;
		dpws->err_detail_data = in;
		dpws->err_detail_data_elt = "wse:UnsubscribeOp";
		goto exit;
	}

exit:

	unlock(service->device->lock);
	return ret;
}

