/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.155.2.2 $
|                                              $Date: 2007/03/23 16:41:34 $
\*============================================================================*/
 

/******************************************************************************\
 *                                SHARED TOOLS                                *
\******************************************************************************/

/*----------------------------------------------------------------------------*\
 *                                   UUID                                     *
\*----------------------------------------------------------------------------*/
 /*
  ** Copyright (c) 1990- 1993, 1996 Open Software Foundation, Inc.
  ** Copyright (c) 1989 by Hewlett-Packard Company, Palo Alto, Ca. &
  ** Digital Equipment Corporation, Maynard, Mass.
  ** To anyone who acknowledges that this file is provided "AS IS"
  ** without any express or implied warranty: permission to use, copy,
  ** modify, and distribute this file for any purpose is hereby
  ** granted without fee, provided that the above copyright notices and
  ** this notice appears in all source code copies, and that none of
  ** the names of Open Software Foundation, Inc., Hewlett-Packard
  ** Company, or Digital Equipment Corporation be used in advertising
  ** or publicity pertaining to distribution of the software without
  ** specific, written prior permission.  Neither Open Software
  ** Foundation, Inc., Hewlett-Packard Company, nor Digital Equipment
  ** Corporation makes any representations about the suitability of
  ** this software for any purpose.
  */

#include "tools.h"
#include "config.h"
#include <ctype.h>

static int deleteHandle(handlePool_t *pool, short handleRef);
  
/*
**  Global variables.
*/
extern struct dpws_config tk_config;

static unsigned64_t     time_last;  // QUESTION: is it thread safe ?
static unsigned16       clock_seq;
static int              initialized = 0;

static void mult32(unsigned32 u, unsigned32 v, unsigned64_t *result)
{
    /* Following the notation in Knuth, Vol. 2. */
    unsigned32 uuid1, uuid2, v1, v2, temp;

    uuid1 = u >> 16;
    uuid2 = u & 0xFFFF;
    v1 = v >> 16;
    v2 = v & 0xFFFF;
    temp = uuid2 * v2;
    result->lo = temp & 0xFFFF;
    temp = uuid1 * v2 + (temp >> 16);
    result->hi = temp >> 16;
    temp = uuid2 * v1 + (temp & 0xFFFF);
    result->lo += (temp & 0xFFFF) << 16;
    result->hi += uuid1 * v1 + (temp >> 16);
}

static void get_uuid_system_time(unsigned64_t *uuid_time)
{
    unsigned64_t utc, usecs, os_basetime_diff;

    system_time(uuid_time);
    mult32(uuid_time->hi, 10000000, &utc);
    mult32(uuid_time->lo, 10, &usecs);
    ADD_64b_2_64b(&usecs, &utc, &utc);

    /* Offset between UUID formatted times and Unix formatted times.
    * UUID UTC base time is October 15, 1582.
    * Unix base time is January 1, 1970. */
    os_basetime_diff.lo = 0x13814000;
    os_basetime_diff.hi = 0x01B21DD2;
    ADD_64b_2_64b(&utc, &os_basetime_diff, uuid_time);
}

  /*
  ** See "The Multiple Prime Random Number Generator" by Alexander
  ** Hass pp. 368-381, ACM Transactions on Mathematical Software,
  ** 12/87.
  */
  static unsigned32 rand_m;
  static unsigned32 rand_ia;
  static unsigned32 rand_ib;
  static unsigned32 rand_irand;

static void true_random_init(void)
{
    unsigned64_t t;
    unsigned16 seed;
    
    /* Generating our 'seed' value Start with the current time, but,
    * since the resolution of clocks is system hardware dependent and
    * most likely coarser than our resolution (10 usec) we 'mixup' the
    * bits by xor'ing all the bits together.  This will have the effect
    * of involving all of the bits in the determination of the seed
    * value while remaining system independent.  Then for good measure
    * to ensure a unique seed when there are multiple processes
    * creating UUIDs on a system, we add in the PID.
    */
    rand_m = 971;
    rand_ia = 11113;
    rand_ib = 104322;
    rand_irand = 4181;
    get_uuid_system_time(&t);
    seed  =  (unsigned16)t.lo        & 0xFFFF;
    seed ^= (t.lo >> 16) & 0xFFFF;
    seed ^=  t.hi        & 0xFFFF;
    seed ^= (t.hi >> 16) & 0xFFFF;
    rand_irand += seed + GET_PID;
}

static unsigned16 true_random(void)
{
    if ((rand_m += 7) >= 9973)
        rand_m -= 9871;
    if ((rand_ia += 1907) >= 99991)
        rand_ia -= 89989;
    if ((rand_ib += 73939) >= 224729)
        rand_ib -= 96233;
    rand_irand = (rand_irand * rand_m) + rand_ia + rand_ib;
    return (unsigned16)((rand_irand >> 16) ^ (rand_irand & RAND_MASK));
}

/*
**  Startup initialization routine for the UUID module.
*/
void uuid_init(void)
{
    if (!initialized) {
        initialized = 1;
        true_random_init();
        get_uuid_system_time(&time_last);
        clock_seq = true_random();
    }
}

static int time_cmp(unsigned64_t *time1, unsigned64_t *time2)
{
    if (time1->hi < time2->hi) return -1;
    if (time1->hi > time2->hi) return 1;
    if (time1->lo < time2->lo) return -1;
    if (time1->lo > time2->lo) return 1;
    return 0;
}

static void new_clock_seq()
{
    clock_seq = (clock_seq + 1) % (CLOCK_SEQ_LAST + 1);
    if (clock_seq == 0) clock_seq = 1;
}

/* provide a 46 octet char buffer */
static char * uuid_generate(char *uuid, unsigned64_t time_now, unsigned16 clock_seq)
{
    /* Construct a uuid with the information we've gathered
    * plus a few constants. */
    sprintf(
        uuid, "%08lx-%04hx-%04hx-%02x%02x-%02x%02x%02x%02x%02x%02x",
            (unsigned32)time_now.lo,
            (unsigned16)time_now.hi & 0x0000FFFF,
            (unsigned16)((time_now.hi & 0x0FFF0000) >> 16) | (1 << 12),
            (unsigned8)((clock_seq & 0x3F00) >> 8) | 0x80,
            (unsigned8)clock_seq & 0xFF,
            (unsigned8)tk_config.network_adapter.mac_addr[0],
            (unsigned8)tk_config.network_adapter.mac_addr[1],
            (unsigned8)tk_config.network_adapter.mac_addr[2],
            (unsigned8)tk_config.network_adapter.mac_addr[3],
            (unsigned8)tk_config.network_adapter.mac_addr[4],
            (unsigned8)tk_config.network_adapter.mac_addr[5]
            );
    return uuid;
}

/* provide a UUID_STRING_SIZE octet char buffer */
char * uuid_create(char *uuid)  // QUESTION: should protect ?
{
    static unsigned64_t     time_now;
    static unsigned16       time_adjust;
    int                     got_no_time = 0;
    get_uuid_system_time(&time_now);

    do {
        switch (time_cmp(&time_now, &time_last)) {
            case -1: /* Time went backwards. */
                new_clock_seq();
                time_adjust = 0;
                break;
            case 1:
                time_adjust = 0;
                break;
            default:
                if (time_adjust == 0x7FFF)
                    /* We're going too fast for our clock; spin. */
                    got_no_time = 1;
                else
                    time_adjust++;
                break;
        }
    }
    while (got_no_time);

    time_last.lo = time_now.lo;
    time_last.hi = time_now.hi;

    if (time_adjust != 0) {
        ADD_16b_2_64b(&time_adjust, &time_now, &time_now);
    }

    return uuid_generate(uuid, time_now, clock_seq);
}

/* provide a SCHEME_UUID_STRING_SIZE octet char buffer */
char * schemed_uuid_create(char *uuid)  // QUESTION: should protect ?
{
    strcpy(uuid, UUID_SCHEME);
    uuid_create(uuid + UUID_SCHEME_SIZE);
    return uuid;
}

/* provide a SCHEME_UUID_STRING_SIZE octet char buffer */
char * stable_uuid_create(char *uuid, unsigned32 instance)
{
    unsigned64_t time = {0, 0};
    time.lo = instance; // NOTE: hope the time is small enough to be expired and avoid collision
    strcpy(uuid, UUID_SCHEME);
    uuid_generate(uuid + UUID_SCHEME_SIZE, time, 0);
    return uuid;
}

/*----------------------------------------------------------------------------*\
 *                                MEMORY STRUCTURES                           *
\*----------------------------------------------------------------------------*/
void init_dyn_array(dyn_array * array, size_t entry_size, int inc)
{
    array->tab = NULL;
    array->nb = 0;
    array->size = 0;
    array->f_size = entry_size;
    array->inc = inc;
}

void * get_entry(dyn_array * tab, int index) {
    return (char*)tab->tab + (index * tab->f_size);
}

void * new_entry(struct dpws* dpws, dyn_array * tab) {
    char * p = NULL;
    // test if table size is sufficient, if not allocate and copy
    if (tab->size == 0) {
        tab->size = tab->inc + 1;
        if (dpws_dpws2soap(dpws))
            tab->tab = soap_malloc(dpws_dpws2soap(dpws), tab->size * tab->f_size);
        else
            tab->tab = malloc(tab->size * tab->f_size);
    }
    else if (tab->nb == tab->size-1) {
        int s = tab->size;
        tab->size = tab->size + tab->inc;
        if (dpws) {
            void* t = tab->tab;
            tab->tab = soap_malloc(dpws_dpws2soap(dpws), tab->size * tab->f_size);
            memcpy(tab->tab, t, s * tab->f_size);
        } 
        else
            tab->tab = realloc(tab->tab, tab->size * tab->f_size);
    }
    // add null at end of table
    p = (char*)tab->tab + (tab->nb * tab->f_size);
    memset(p + tab->f_size, 0, tab->f_size);
    
    // increment count
    tab->nb++;
    
    // return entry
    return p;
}

void remove_entry(dyn_array * tab, int index, free_entry_cbk free_hook)
{
    char * p = get_entry(tab, index);
    int tail_size;
    if (free_hook)
        free_hook(p);
    tail_size = tab->nb - 1 - index;
    if (tail_size > 0)
        memmove(p, p + tab->f_size, tail_size * tab->f_size);
    tab->nb--;
}

void * insert_entry(struct dpws* dpws, dyn_array * tab, int index)
{
    char * p1 = NULL, * p2 = NULL;
    if (tab->nb > 0 && index < tab->nb) {
        int tail_size;
        p1 = get_entry(tab, index);
        tail_size = tab->nb - index;    // do it before !
        new_entry(dpws, tab);
        p1 = get_entry(tab, index); // do it after !
        for (p2 = p1 + (tail_size * tab->f_size) - 1; p2 >= p1; p2--)
            *(p2 + tab->f_size) = *p2;
    }
    else
        p1 = new_entry(dpws, tab);

    return p1;
}

int browse_dyn_array(dyn_array * da, browse_entry_cbk hook, void * param)
{
    int i; 
    for (i = 0; i < da->nb && !hook(get_entry(da, i), param); i++);
    return i;
}

static BOOLEAN free_dyn_array_hook(void * pEntry, free_entry_cbk free_hook) {
    free_hook(pEntry);
    return FALSE;
}

void free_dyn_array(dyn_array * da, free_entry_cbk free_hook)
{
    if (free_hook)
        browse_dyn_array(da, (browse_entry_cbk)free_dyn_array_hook, free_hook);
    free(da->tab);
    da->tab = NULL;
    da->nb = 0;
    da->size = 0;
}

static void free_buffer(void ** pBuf) { if (*pBuf) free(*pBuf);}

void free_buffer_dyn_array(dyn_array * da) { free_dyn_array(da, (void (*)(void *))free_buffer);}

//dyn_array * dyn_array_dup(dyn_array * da) {
//  size_t n = da->size * da->f_size;
//  dyn_array * copy = malloc(sizeof(dyn_array));
//  memcpy(copy, da, sizeof(dyn_array));
//  copy->tab = malloc(n);
//  memcpy(copy->tab, da->tab, n);
//  return copy;
//}

dyn_array * string_array_initialize(struct dpws* dpws, dyn_array * dest, char ** src, int inc) {
    init_dyn_array(dest, sizeof(char *), inc);
    for (; *src; src++)
        *(char **)new_entry(dpws, dest) = (dpws ? soap_strdup(dpws_dpws2soap(dpws), *src) : dpws_strdup(*src));
    return dest;
}
    
dyn_array * string_array_dup(struct dpws* dpws, dyn_array * dest, dyn_array * src) {
    if (src->nb > 0) 
        string_array_initialize(dpws, dest, src->tab, src->nb);
    else
        init_dyn_array(dest, sizeof(char *), src->inc);
    return dest;
}
void init_entries(dyn_array * tab, void* entries, int nentries) {
    tab->size = nentries+tab->inc;
    tab->tab = malloc(tab->size * tab->f_size);
    memcpy(tab->tab, entries, nentries * tab->f_size);
    tab->nb = nentries;
    memset(get_entry(tab, tab->nb), 0, tab->f_size);
}

struct wsa_endpoint_ref * endpoint_ref_dup(struct dpws* dpws, struct wsa_endpoint_ref * src) {
    struct wsa_endpoint_ref *dest = NULL;
    if (src) {
        dest = soap_malloc(dpws_dpws2soap(dpws), sizeof(struct wsa_endpoint_ref));
        endpoint_ref_copy(dpws, dest, src);
    }
    return dest;
}

struct wsa_endpoint_ref * endpoint_ref_copy(struct dpws* dpws, struct wsa_endpoint_ref * dest, struct wsa_endpoint_ref * src) {
    if (src && dest) {
        dest->address = soap_strdup(dpws_dpws2soap(dpws), src->address);
        dest->subscription_id = soap_strdup(dpws_dpws2soap(dpws), src->subscription_id);
    }
    return dest;
}

struct wsa_endpoint_ref * dpws_endpoint_ref_dup(struct wsa_endpoint_ref * src) {
    struct wsa_endpoint_ref *dest = NULL;
    if (src) {
        dest = malloc(sizeof(struct wsa_endpoint_ref));
        dpws_endpoint_ref_copy(dest, src);
    }
    return dest;
}

struct wsa_endpoint_ref * dpws_endpoint_ref_copy(struct wsa_endpoint_ref * dest, struct wsa_endpoint_ref * src) {
    if (src && dest) {
        dest->address = dpws_strdup(src->address);
        dest->subscription_id = dpws_strdup(src->subscription_id);
    }
    return dest;
}

void dpws_free(void * buffer) {free(buffer);}

void dpws_endpoint_ref_content_free(struct wsa_endpoint_ref * src) {
    if (src) {
        free(src->address);
        free(src->subscription_id);
        src->address = NULL;
        src->subscription_id = NULL;
    }
}

void dpws_endpoint_ref_free(struct wsa_endpoint_ref * src) {
    if (src) {
        dpws_endpoint_ref_content_free(src);
        free(src);
    }
}

void dpws_endpoint_ref_array_free(struct wsa_endpoint_ref ** endpoints, int nb_endpoints) {
    int i;
    for (i = 0; i < nb_endpoints; nb_endpoints++) {
        dpws_endpoint_ref_free(endpoints[i]);
    }
}

/* Look for a localized entry (the preferred lang one if found then the one
 * without @lang, then in english (first of the 3))
 * Assumes the array entries point on structures which first field is a string
 * containing an XML lang attribute.
 */
int find_local_entry(void ** entries, int entry_nb, char *preferred_lang)
{
// TODO: check that a lang attribute coming with multiple values (strange separator like ',' or ';') does not carry out a failure
    int i, p_index = -1, d_index = -1, e_index = -1, ret = -1;
    char * lang;

    // Status (if more than one take the one without @lang or in english (first of the 2)
    for (i = 0; i < entry_nb; i++) {
        lang = *(char **)entries[i];
        if (!lang)
            d_index = i;
        else if (preferred_lang && !strcmp(lang, preferred_lang))
            p_index = i;
        else if (!strncmp(lang, "en", 2))
            e_index = i;        
    }
    if (p_index >= 0)
        ret = p_index;
    else if (d_index >= 0)
        ret = d_index;
    else if (e_index >= 0)
        ret = e_index;
        
    return ret;
}

/*----------------------------------------------------------------------------*\
 *                                   CONVERSIONS                              *
\*----------------------------------------------------------------------------*/
static int parse_list(
                        struct dpws * dpws,
                        dyn_array * dest,
                        char * source,
                        int (*token_parser)(struct dpws * dpws, dyn_array * dest, char * source)
                        )
{
    char * p = source, * qname = NULL;
    int ret = SOAP_OK;
    for (;;p++) {
        switch (*p) {
            case '\0':
                if (qname)
                    ret = token_parser(dpws, dest, qname);
                return ret;
            case ' ': case '\t': case 0xA: case 0xD:
                if (qname) {
                    *p = '\0';
                    ret = token_parser(dpws, dest, qname);
                    if (ret)
                        return ret;
                    qname = NULL;
                }
                break;
            default:
                if (!qname)
                    qname = p;
        }
    }
}

static int parse_qname(struct dpws * dpws, dyn_array * dest, char * source) {
    void ** p;
    char * s;
    soap_s2QName(dpws_dpws2soap(dpws), source, &s); // allocates in soap context
    if (!s || !(p = (void **)new_entry(dpws, dest))
            || !(*p = soap_malloc(dpws_dpws2soap(dpws), strlen(s) + 1)))
        return  SOAP_EOM;
    strcpy(*p, s);
    return SOAP_OK;
}

static int parse_string(struct dpws * dpws, dyn_array * dest, char * source) {
    void ** p;
    if (!(p = (void **)new_entry(dpws, dest))
        || !(*p = soap_strdup(dpws_dpws2soap(dpws), source)))
        return  SOAP_EOM;
    return SOAP_OK;
}

int parse_qname_list(struct dpws * dpws, dyn_array * dest, char * source) {
#ifdef USE_ONVIF /* Added by HG.W */
    ONVIF_LOG1(ONVIF_LOG_INFO,
               "[INFO]: QNameList=[%s]\n", source?source:"null");
#endif
    return parse_list(dpws, dest,  source, &parse_qname);
}

int parse_token_list(struct dpws * dpws, dyn_array * dest, char * source) {
#ifdef USE_ONVIF /* Added by HG.W */
    ONVIF_LOG1(ONVIF_LOG_INFO,
               "[INFO]: UriList=[%s]\n", source?source:"null");
#endif
    return parse_list(dpws, dest,  source, &parse_string);
}

/*----------------------------------------------------------------------------*\
 *                       CUSTOM GSOAP TYPES MARSHALLING                       *
\*----------------------------------------------------------------------------*/
void soap_serialize_qname_list(struct soap* soap, struct dynamic_array *const* a) {
}

void soap_default_qname_list(struct soap* soap, struct dynamic_array ** a) {
    *a = NULL;
}

// NOTE: size is not used but null boundary
int soap_out_qname_list(struct soap* soap, const char * tag, int id, struct dynamic_array *const* a, const char * type)
{
    struct dynamic_array *p = *a;
    int i;
    const char ** qnames;
    if (!p)
        return soap_element_null(soap, tag, -1, "");

    qnames = (const char **)soap_malloc(soap, p->nb * sizeof(char*));
    // soap_QName2s has side-effects, hence the need for two loops
    // one before soap_element_begin_out, one after
    for (i = 0; i < p->nb; i++) {
        qnames[i] = soap_QName2s(soap, *(char **)get_entry(p, i));
    }
    soap_element_begin_out(soap, tag, -1, "");
    for (i = 0; i < p->nb; i++) {
        if (i > 0) soap_string_out(soap, " ", 0);
        soap_string_out(soap, qnames[i], 0);
    }
    soap_element_end_out(soap, tag);

    return soap->error;
}

// NOTE: size is not used but null boundary
struct dynamic_array ** soap_in_qname_list(struct soap* soap, const char* tag, struct dynamic_array **a, const char* type)
{
#ifdef USE_ONVIF /* Added by HG.W */
    const char *value = NULL;
#endif

    dyn_array *tab;
    if (*a)
        tab = *a;
    else
        *a = tab = soap_malloc(soap, sizeof(dyn_array));    // TODO: see if missing tag could not lead to a NULL (move after begin in)
    
    init_dyn_array(tab, sizeof(char *), DISCOVERY_ALLOC_INCREMENT);
    if (soap_element_begin_in(soap, tag, 0))
        return NULL;

#ifdef USE_ONVIF /* Added by HG.W */
    value = soap_value(soap);
    if(value){
        parse_qname_list(dpws_soap2dpws(soap), tab, soap_strdup(soap, value));
    }
    else{
        return NULL;
    }
#else
    parse_qname_list(dpws_soap2dpws(soap), tab, soap_string_in(soap, 1, -1, -1));
#endif

    if (soap->body && soap_element_end_in(soap, tag))
        return NULL;
    return a;
}

void soap_serialize_uri_list(struct soap* soap, struct dynamic_array *const* a) {
}

void soap_default_uri_list(struct soap* soap, struct dynamic_array ** a) {
    *a = NULL;
}

// does not use dyn array size but the null boundary (assumed by caller)
int soap_out_uri_list(struct soap* soap, const char * tag, int id, struct dynamic_array *const* a, const char * type)
{
    char ** s;
    if (!(*a))
        return soap_element_null(soap, tag, -1, "");

    s = (*a)->tab;

    soap_element_begin_out(soap, tag, -1, "");
    if (*s)
        soap_string_out(soap, *s, 0);
    s++;
    for (; *s; s++) {
        soap_string_out(soap, " ", 0);
        soap_string_out(soap, *s, 0);
    }
    soap_element_end_out(soap, tag);

    return soap->error;
}

struct dynamic_array ** soap_in_uri_list(struct soap* soap, const char* tag, struct dynamic_array **a, const char* type)
{
#ifdef USE_ONVIF /* Added by HG.W */
    const char *value = NULL;
#endif

    dyn_array *tab;
    if (*a)
        tab = *a;
    else
        *a = tab = soap_malloc(soap, sizeof(dyn_array));    // TODO: see if missing tag could not lead to a NULL (move after begin in)
    
    init_dyn_array(tab, sizeof(char *), DISCOVERY_ALLOC_INCREMENT);
    if (soap_element_begin_in(soap, tag, 0))
        return NULL;

#ifdef USE_ONVIF /* Added by HG.W */
    value = soap_value(soap);
    if(value){
        parse_token_list(dpws_soap2dpws(soap), tab, soap_strdup(soap, value));
    }
    else{
        return NULL;
    }
#else
    parse_token_list(dpws_soap2dpws(soap), tab, soap_string_in(soap, 1, -1, -1));
#endif

    if (soap->body && soap_element_end_in(soap, tag))
        return NULL;
    return a;
}

void dpws_simple_element_out(struct soap *soap, const char *tag, const char *value)
{
    soap_element_begin_out(soap, tag, -1, "");
    soap_string_out(soap, value, 0);
    soap_element_end_out(soap, tag);
} 

char** dpws_simple_element_in(struct soap *soap, const char *tag, char** value)
{
    if (soap_element_begin_in(soap, tag, 0) || !soap->body)
        return NULL;
    *value = soap_string_in(soap, 0, -1, -1);
    if (soap_element_end_in(soap, tag))
        return NULL;
    return value;
}

int dpws_empty_element_out(struct soap *soap, const char *tag)
{
    if (soap_element(soap, tag, -1, ""))
        return soap->error;
    return soap_element_start_end_out(soap, tag);
}


void soap_serialize_endpoint_ref(struct soap* soap, struct wsa_endpoint_ref *const*a) {
}

void soap_default_endpoint_ref(struct soap* soap, struct wsa_endpoint_ref **a) {
    *a = NULL;
}

int soap_out_endpoint_ref(struct soap *soap, const char *tag, int id, struct wsa_endpoint_ref *const*a, const char *type)
{
    struct wsa_endpoint_ref *p = *a;
    if (!p)
        return soap_element_null(soap, tag, -1, "");
    soap_element_begin_out(soap, tag, -1, "");
    dpws_simple_element_out(soap, "wsa:Address", p->address);
    if (p->subscription_id) {
        soap_element_begin_out(soap, "wsa:ReferenceParameters", -1, "");
        dpws_simple_element_out(soap, "wse:Identifier", p->subscription_id);
        soap_element_end_out(soap, "wsa:ReferenceParameters");
    }
    soap_element_end_out(soap, tag);
    return SOAP_OK;
}

static int dpws_ignore_element(struct soap *soap)
{
    if (!soap_peek_element(soap))
    {
        soap->peeked = 0;
        soap->level++;
        while (!dpws_ignore_element(soap));
        if (soap->error == SOAP_NO_TAG)
            soap->error = soap_element_end_in(soap, NULL);
    }
    return soap->error;
}

struct wsa_endpoint_ref ** soap_in_endpoint_ref(struct soap *soap, const char *tag, struct wsa_endpoint_ref **a, const char *type) {
    struct wsa_endpoint_ref *p = *a;
    if (soap_element_begin_in(soap, tag, 0))
        return NULL;
    if (soap->body) {
        if (!p) {
            *a= p = (struct wsa_endpoint_ref*)soap_malloc(soap, sizeof(struct wsa_endpoint_ref));
            if (!p)
                return NULL;
            dpws_default_wsa_endpoint_ref(p);
        }
        if (!dpws_simple_element_in(soap, "wsa:Address", &p->address))
            return NULL;
        
        // ignore next sequence element
        if (!soap_element_begin_in(soap, "wsa:ReferenceProperties", 1)) {
            if (soap_element_end_in(soap, "wsa:ReferenceProperties"))
                return NULL;
        }
        else if (soap->error != SOAP_TAG_MISMATCH && soap->error != SOAP_NO_TAG)
            return NULL;
            
        if (!soap_element_begin_in(soap, "wsa:ReferenceParameters", 1)) {
            if (!dpws_simple_element_in(soap, "wse:Identifier", &p->subscription_id))
                return NULL;
            if (soap_element_end_in(soap, "wsa:ReferenceParameters"))
                return NULL;
        }
        else if (soap->error != SOAP_TAG_MISMATCH && soap->error != SOAP_NO_TAG)
            return NULL;
            
        // ignore next sequence element
        if (!soap_element_begin_in(soap, "wsa:PortType", 1)) {
            if (soap_element_end_in(soap, "wsa:PortType"))
                return NULL;
        }
        else if (soap->error != SOAP_TAG_MISMATCH && soap->error != SOAP_NO_TAG)
            return NULL;

        // ignore next sequence element
        if (!soap_element_begin_in(soap, "wsa:ServiceName", 1)) {
            if (soap_element_end_in(soap, "wsa:ServiceName"))
                return NULL;
        }
        else if (soap->error != SOAP_TAG_MISMATCH && soap->error != SOAP_NO_TAG)
            return NULL;
        
        // loop over optional elements (interoperability)
        for (;;) {
            soap->error = SOAP_TAG_MISMATCH;
            if (soap->error == SOAP_TAG_MISMATCH)
                soap->error = dpws_ignore_element(soap);
            if (soap->error == SOAP_NO_TAG)
                break;
            if (soap->error)
                return NULL;
        }
    }
    else
        return NULL;
        
    if (soap_element_end_in(soap, tag))
        return NULL;
    return a;
}

/*----------------------------------------------------------------------------*\
 *                                   COMPARISONS                              *
\*----------------------------------------------------------------------------*/

BOOLEAN qname_equals(char * ns, char * local_name, char * qname)
{
    char *c1 = qname, *c2 = (ns ? ns : '\0');
    /* compare ns first */
    if (*c1 == '"') {   // ns is expanded
        c1++;
        while (*c1 != '"' && *c2 && *c2 == *c1) {
            c1++; c2++;
        }
        if (*c1 != '"' || *c2)
            return FALSE;
        c1++;   // jump over '"'
    }
    else
        return FALSE;

    /* compare local name */
    c1++;   // jump over ":"
    c2 = local_name;
    while (*c1 && *c1 == *c2) {
        c1++; c2++;
    }
    // check length are equal
    if (!*c1 && !*c2)
        return TRUE;
    else
        return FALSE;
}


#ifdef USE_ONVIF /* Added by HG.W (when onvif only support 3986) */
int parse_match_rule(char *uri) {
    if (!uri)
        return WSD_MATCH_BY_RFC3986;
    else{
        if (!strcmp(uri, WSD_MATCH_BY_RFC3986_URI))
            return WSD_MATCH_BY_RFC3986;
        else
            return WSD_MATCH_BY_UNSPECIFIED;
    }
}
#else
int parse_match_rule(char *uri) {
    if (!uri)
        return WSD_MATCH_BY_RFC2396;
    else if (!strcmp(uri, WSD_MATCH_BY_UUID_URI))
        return WSD_MATCH_BY_UUID;
    else if (!strcmp(uri, WSD_MATCH_BY_LDAP_URI))
        return WSD_MATCH_BY_LDAP;
    else if (!strcmp(uri, WSD_MATCH_BY_STRCMP_URI))
        return WSD_MATCH_BY_STRCMP;
    else    // included rfc2396 match rule
        return WSD_MATCH_BY_RFC2396;
}
#endif

#ifdef USE_ONVIF /* Added by HG.W (when onvif only support 3986) */
char * scope_match_string(int match_rule) {
    switch (match_rule) {
        case WSD_MATCH_BY_RFC3986: default:
            return NULL;
    }
}
#else
char * scope_match_string(int match_rule) {
    switch (match_rule) {
        case WSD_MATCH_BY_RFC2396: default:
            return NULL;
        case WSD_MATCH_BY_UUID:
            return WSD_MATCH_BY_UUID_URI;
        case WSD_MATCH_BY_LDAP:
            return WSD_MATCH_BY_LDAP_URI;
        case WSD_MATCH_BY_STRCMP :
            return WSD_MATCH_BY_STRCMP_URI;
    }
}
#endif

static BOOLEAN is_included(dyn_array *a, dyn_array *b, BOOLEAN (*compare)(void *, void *))
{
    int i, j;

    if (a == NULL) return TRUE;
    if (b == NULL) return FALSE;

    for (i = 0; i < a->nb; i++) {
        for (
            j = 0;
            j < b->nb && !compare(get_entry(a, i), get_entry(b, j));
            j++
            );
        if (j == b->nb)
            return FALSE;
    }
    return TRUE;
}


#ifdef USE_ONVIF /* Added by HG.W (when onvif only support 3986) */
static BOOLEAN rfc3986_match(char **pstr1, char **pstr2) {  // TODO: implement thoroughfully the spec
    char *s1, *s2;
    s1 = *pstr1;
    s2 = *pstr2;

    // compare scheme
    if (s1 == NULL || s2 == NULL) return s1 == s2;
    while (*s1 && *s2 && *s1 != ':') {
        if (tolower(*s1) != tolower(*s2))
            return FALSE;
        s1++;s2++;
    }
    // compare host name and paths
    do {    // TODO : canonize escapes
        s1++;s2++;
        while (*s1 && *s1 != '/') {
            if (*s1 == '?' || *s1 == ';' || *s1 == '#') {
                while (*s1 && *s1 != '/') s1++;
                continue;
            }
            
            if (*s2 == '?' || *s2 == ';' || *s2 == '#') {
                while (*s2 && *s2 != '/') s2++;
                continue;
            }
            if (*s2 == '\0' || tolower(*s1) != tolower(*s2)
                || (*s1 == '.' && (*(s1 + 1) == '.' || *(s1 + 1) == '/'))   // do not match with . or .. segments
                || (*s2 == '.' && (*(s2 + 1) == '.' || *(s2 + 1) == '/')))
                return FALSE;
            s1++;s2++;
        }
        if (*s2 == '?' || *s2 == ';' || *s2 == '#')
            while (*s2 && *s2 != '/') s2++;
        
        /* Added by HG.W (Fixed the /path/abc matches /path/abcd bug) */
        if(*s1 == '\0'){
            if(*s2 == '\0' || *s2 == '/'){
                return TRUE;
            }
            else{
                return FALSE;
            }
        }
        else{
            if(*s1 == '/' && *(s1+1) == '\0' && *s2 == '/') return TRUE;
        }
    }
    while (*s2);
    return FALSE;
}

BOOLEAN rfc3986_matches(char *base, char *uri) {return rfc3986_match(&base, &uri);}


BOOLEAN is_rfc3986_included(dyn_array * uri_set1, dyn_array * uri_set2) {
    return is_included(uri_set1, uri_set2, (BOOLEAN (*)(void*, void*))rfc3986_match);
}
#endif

static BOOLEAN rfc2396_match(char **pstr1, char **pstr2) {  // TODO: implement thoroughfully the spec
    char *s1, *s2;
    s1 = *pstr1;
    s2 = *pstr2;

    // compare scheme
    if (s1 == NULL || s2 == NULL) return s1 == s2;
    while (*s1 && *s2 && *s1 != ':') {
        if (tolower(*s1) != tolower(*s2))
            return FALSE;
        s1++;s2++;
    }
    // compare host name and paths
    do {    // TODO : canonize escapes
        s1++;s2++;
        while (*s1 && *s1 != '/') {
            if (*s1 == '?' || *s1 == ';' || *s1 == '#') {
                while (*s1 && *s1 != '/') s1++;
                continue;
            }
            
            if (*s2 == '?' || *s2 == ';' || *s2 == '#') {
                while (*s2 && *s2 != '/') s2++;
                continue;
            }
            if (*s2 == '\0' || tolower(*s1) != tolower(*s2)
                || (*s1 == '.' && (*(s1 + 1) == '.' || *(s1 + 1) == '/'))   // do not match with . or .. segments
                || (*s2 == '.' && (*(s2 + 1) == '.' || *(s2 + 1) == '/')))
                return FALSE;
            s1++;s2++;
        }
        if (*s2 == '?' || *s2 == ';' || *s2 == '#')
            while (*s2 && *s2 != '/') s2++;
        
        if (*s1 == '\0')    // s1 includes s2 is enough
            return TRUE;
    }
    while (*s2);

    return FALSE;
}

BOOLEAN rfc2396_matches(char *base, char *uri) {return rfc2396_match(&base, &uri);}

static BOOLEAN strcmp_match (char **a, char **b) {return strcmp(*a, *b) ? FALSE : TRUE;}

static BOOLEAN strcasecmp_match (char **a, char **b) {return dpws_strcasecmp(*a, *b) ? FALSE : TRUE;}

#ifdef USE_ONVIF /* Added by HG.W (when onvif only support 3986) */
static BOOLEAN (*get_scope_match_func(int match_rule))(void *, void *)
{
    switch (match_rule) {
        case WSD_MATCH_BY_RFC3986:
            return (BOOLEAN (*)(void*, void*))rfc3986_match;
        default:
            return NULL;
    }
}
#else
static BOOLEAN (*get_scope_match_func(int match_rule))(void *, void *)
{
    switch (match_rule) {
        case WSD_MATCH_BY_UUID:
            return (BOOLEAN (*)(void*, void*))strcasecmp_match;
        case WSD_MATCH_BY_LDAP:
            return NULL;
        case WSD_MATCH_BY_STRCMP :
            return (BOOLEAN (*)(void*, void*))strcmp_match;
        case WSD_MATCH_BY_RFC2396:
        default:
            return (BOOLEAN (*)(void*, void*))rfc2396_match;
    }
}
#endif /* Added by HG.W */

BOOLEAN is_rfc2396_included(dyn_array * uri_set1, dyn_array * uri_set2) {
    return is_included(uri_set1, uri_set2, (BOOLEAN (*)(void*, void*))rfc2396_match);
}

BOOLEAN probe_match(
        dyn_array * types, dyn_array * scopes, int match_rule,
        dyn_array *dev_types, dyn_array *dev_scopes,
        type_compare_cbk type_compare
        )
{
#ifdef USE_ONVIF /* Added by HG.W */
    if(match_rule == WSD_MATCH_BY_UNSPECIFIED) return FALSE;
#endif

    BOOLEAN (*match_func)(void *, void *) = get_scope_match_func(match_rule);
#ifdef USE_ONVIF /* Added by HG.W */
    BOOLEAN types_matched = FALSE, scopes_matched = FALSE;

    types_matched = is_included(types, dev_types, type_compare);
    if(!types_matched){
        ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Types not matched.\n");
        return FALSE;
    }

    scopes_matched = is_included(scopes, dev_scopes, match_func);
    if(!scopes_matched){
        ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Scopes not matched.\n");
        return FALSE;
    }
    return TRUE;

#else /* !USE_ONVIF */

    return is_included(types, dev_types, type_compare)  // will return true if types is empty
        && (dev_scopes->nb > 0
        ? is_included(scopes, dev_scopes, match_func)
        : scopes == NULL || scopes->nb == 0 || (scopes->nb == 1 && match_func((char **)get_entry(scopes, 0), &WSD_DEFAULT_SCOPE)));
#endif
}

/*----------------------------------------------------------------------------*\
 *                                     IPC                                    *
\*----------------------------------------------------------------------------*/
handlePool_t mutexPool;

void initMutexPool() { createHandlePool(&mutexPool, 2 * BUILTIN_MUTEX, BUILTIN_MUTEX + MAX_SERVER_OBJECTS, -1, NULL);}
void destroyMutexPool() { deleteHandlePool(&mutexPool);}

// ThreadX
#ifdef THREADX

# include <tx_api.h>

short lock_create()
{
    short handleRef = -1;
    handle_s * handle;
    
    handleRef = createHandle(&mutexPool, MUTEX_TYPE, NULL);
    if (handleRef >= 0) {
        handle = getHandle(&mutexPool, handleRef);
        if (!handle->pObject)
            handle->pObject = malloc(sizeof(TX_MUTEX)); // here mutex structure lifecycle is the handle one
    
        if (tx_mutex_create((TX_MUTEX *)handle->pObject, "dpws", TX_NO_INHERIT) != TX_SUCCESS) {
            deleteHandle(&mutexPool, handleRef);
            return DPWS_ERR_CREATING_MUTEX;
        }
    }
    return handleRef;
}

int lock(short handleRef)
{
    return tx_mutex_get((TX_MUTEX *)getObject(&mutexPool, handleRef), TX_WAIT_FOREVER);
}

int unlock(short handleRef)
{
    return tx_mutex_put((TX_MUTEX *)getObject(&mutexPool, handleRef));
}

int lock_destroy(short handleRef)
{
    TX_MUTEX * mutex = (TX_MUTEX *)getObject(&mutexPool, handleRef);
    deleteHandle(&mutexPool, handleRef);
    return tx_mutex_delete(mutex) ? DPWS_ERR_DELETING_MUTEX : DPWS_OK;
}

#elif defined(VXWORKS)

# include <semLib.h>

short lock_create()
{
    short handleRef = -1;
    handle_s * handle;
    
    handleRef = createHandle(&mutexPool, MUTEX_TYPE, NULL);
    if (handleRef >= 0) {
        handle = getHandle(&mutexPool, handleRef);
    
        if ((handle->pObject = semMCreate(SEM_Q_FIFO|SEM_Q_PRIORITY|SEM_INVERSION_SAFE)) == NULL) {
            deleteHandle(&mutexPool, handleRef);
            return DPWS_ERR_CREATING_MUTEX;
        }
    }
    return handleRef;
}

int lock(short handleRef)
{
    return semTake((SEM_ID)getObject(&mutexPool, handleRef), WAIT_FOREVER);
}

int unlock(short handleRef)
{
    return semGive((SEM_ID)getObject(&mutexPool, handleRef));
}

int lock_destroy(short handleRef)
{
    SEM_ID mutex = (SEM_ID)getObject(&mutexPool, handleRef);
    deleteHandle(&mutexPool, handleRef);
    return semDelete(mutex) ? DPWS_ERR_DELETING_MUTEX : DPWS_OK;
}

#elif defined(WIN32)

# include <Windows.h>

short lock_create()
{
    short handleRef = -1;
    handle_s * handle;
    
    handleRef = createHandle(&mutexPool, MUTEX_TYPE, NULL);
    if (handleRef >=0) {
        handle = getHandle(&mutexPool, handleRef);
        if (!handle->pObject)
            handle->pObject = malloc(sizeof(CRITICAL_SECTION)); // here mutex structure lifecycle is the handle one
    
        InitializeCriticalSection((CRITICAL_SECTION *)handle->pObject);
    }
    return handleRef;
}

//static int level = 0;
int lock(short handleRef)
{
//  int i;
    EnterCriticalSection((CRITICAL_SECTION *)getObject(&mutexPool, handleRef));
    //if (handleRef == 3) {
    //  char* p = NULL;
    //  if (level > 0) *p = 'a';
    //  for (i = 0; i < level; i++) fprintf(stderr, "  ");
    //  fprintf(stderr, "> %d\n", handleRef);
    //  level++;
    //}
    return DPWS_OK;
}

int unlock(short handleRef)
{
    //int i;
    //if (handleRef == 3) {
    //  level--;
    //  for (i = 0; i < level; i++) fprintf(stderr, "  ");
    //  fprintf(stderr, "< %d\n", handleRef);
    //}
    LeaveCriticalSection((CRITICAL_SECTION *)getObject(&mutexPool, handleRef));
    return DPWS_OK;
}

int lock_destroy(short handleRef)
{
    CRITICAL_SECTION * cs = (CRITICAL_SECTION *)getObject(&mutexPool, handleRef);
    deleteHandle(&mutexPool, handleRef);
    DeleteCriticalSection(cs);
    return DPWS_OK;
}

#else
// POSIX
# include <pthread.h>

short lock_create()
{
    short handleRef = -1;
    handle_s * handle;
    pthread_mutex_t mutex_init = PTHREAD_MUTEX_INITIALIZER;
    
    handleRef = createHandle(&mutexPool, MUTEX_TYPE, NULL);
    if (handleRef >=0) {
        handle = getHandle(&mutexPool, handleRef);
        if (!handle->pObject)
            handle->pObject = malloc(sizeof(pthread_mutex_t));  // here mutex structure lifecycle is the handle one
        
        *((pthread_mutex_t *)handle->pObject) = mutex_init;
        if (pthread_mutex_init((pthread_mutex_t *)handle->pObject, NULL)) {
            deleteHandle(&mutexPool, handleRef);
            return DPWS_ERR_CREATING_MUTEX;
        }
    }

    return handleRef;
}

int lock(short handleRef)
{
    return pthread_mutex_lock((pthread_mutex_t *)getObject(&mutexPool, handleRef));
}

int unlock(short handleRef)
{
    return pthread_mutex_unlock((pthread_mutex_t *)getObject(&mutexPool, handleRef));
}

int lock_destroy(short handleRef)
{
    pthread_mutex_t * mutex = getObject(&mutexPool, handleRef);
    deleteHandle(&mutexPool, handleRef);
#ifdef USE_ONVIF /* Added by HG.W (fixed memory leaks) */
    int ret;
    if(pthread_mutex_destroy(mutex))
        ret = DPWS_ERR_DELETING_MUTEX;
    else
        ret = DPWS_OK;
    free(mutex);
    return ret;
#else
    return pthread_mutex_destroy(mutex) ? DPWS_ERR_DELETING_MUTEX : DPWS_OK;
#endif
}

#endif

/*----------------------------------------------------------------------------*\
 *                                 NETWORKING                                 *
\*----------------------------------------------------------------------------*/

// NOTE: Here adapter info are not necessarily consistent (waiting for the multiple
// interface to be fully supported).
// - The getHostByName() method is preferred for IP (it's the simplest way to work
//   around the multiple interface pb).
// - Any MAC address will do the job for UUID unicity and stability.

#ifdef TR_STACK_NEXGEN

# include <ngeth.h>
extern NGethifnet ifnet0;

int retrieve_network_adapter_info(struct na_info * na)
{
    na->mac_addr = malloc(6);
    memcpy(na->mac_addr, ifnet0.eif_addr, 6);
    na->ip_addr = malloc(16);
    strcpy(na->ip_addr, inet_ntoa(ifnet0.eif_net.if_addr));
    return DPWS_OK;
}

# elif defined(VXWORKS)

# include <net/if.h>
# include <end.h> 

int retrieve_network_adapter_info(struct na_info * na)
{
    int s;
    struct ifreq buffer;
    struct sockaddr_in * addrin;

    s = socket(PF_INET, SOCK_DGRAM, 0);
    if (s < 0)
        return DPWS_ERR_RETRIEVING_NETWORK_INTERFACE_INFO;
        memset(&buffer, 0, sizeof(buffer));

#  ifdef FC
    na->mac_addr = malloc(6);
    memcpy(na->mac_addr, GetEthAddr(), 6);
#  else
#   if CPU==SIMNT
    strcpy(buffer.ifr_name, "nt0");
#   else
    strcpy(buffer.ifr_name, "fec0");
#   endif   

    if (ioctl(s, EIOCGADDR, (int)&buffer) < 0)
        return DPWS_ERR_RETRIEVING_MAC_ADDRESS;
    na->mac_addr = malloc(6);
    memcpy(na->mac_addr, buffer.ifr_data, 6);
#  endif


#  if CPU==SIMNT
    strcpy(buffer.ifr_name, "nt0");
#  else
    strcpy(buffer.ifr_name, "fec0");
#  endif    

    if (ioctl(s, SIOCGIFADDR, (int)&buffer) < 0)
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    addrin = (struct sockaddr_in *)&buffer.ifr_addr;
    na->ip_addr = dpws_strdup(inet_ntoa(addrin->sin_addr));

    close(s);
    return DPWS_OK;
}

#else

#ifdef USE_ONVIF /* Added by HG.W */

#ifdef WITH_IPV6 
#include <netdb.h>
#include <sys/socket.h>
#define PROC_IFINET6_PATH "/proc/net/if_inet6"
#define IPV6_ADDR_ANY           0x0000U
#define IPV6_ADDR_LOOPBACK      0x0010U
#define IPV6_ADDR_LINKLOCAL     0x0020U
#define IPV6_ADDR_SITELOCAL     0x0040U
#define IPV6_ADDR_COMPATv4      0x0080U

static int retrieve_ip_addrs_v6(struct na_info *na_info)
{
    FILE *fp;
    char addr[8][5];
    int index, plen, scope, flags;
    char ifname[8];
    char addr6[40];
    struct addrinfo hints, *res0;
    int err = 0;
    char host[NI_MAXHOST];
    BOOLEAN flag_ipv6_g = FALSE, flag_ipv6_l = FALSE;

    if(!na_info){
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }

    if((fp = fopen(PROC_IFINET6_PATH, "r")) == NULL) {
        ONVIF_LOG1(ONVIF_LOG_ERROR,
              "[ERROR]: Failed to open file(%s) to retrieve IPv6 addrs.\n",
                   PROC_IFINET6_PATH);
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }

    while (fscanf(fp, 
                  "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %8s\n",
                  addr[0],addr[1],addr[2],addr[3],
                  addr[4],addr[5],addr[6],addr[7],
                  &index, &plen, &scope, &flags, ifname) != EOF) {
        
        if(strcmp(ifname, NA_IF))
            continue;
        
        sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
                addr[0],addr[1],addr[2],addr[3],
                addr[4],addr[5],addr[6],addr[7]);

        memset(&hints, 0, sizeof(hints));
        hints.ai_family = AF_INET6;
        hints.ai_flags = NI_NUMERICHOST;
        err = getaddrinfo(addr6, NULL, &hints, &res0); 
        if(err){
            ONVIF_LOG2(ONVIF_LOG_ERROR, "[ERROR]: Failed to get addrinfo(%s) in retrieve_ip_addrs_v6().(gai_strerror=%s)\n", addr6, gai_strerror(err));
            goto go_out_err;
        }
        err = getnameinfo(res0->ai_addr, res0->ai_addrlen,
                          host, sizeof(host), NULL, 0, NI_NUMERICHOST);
        if(err){
            ONVIF_LOG1(ONVIF_LOG_ERROR, "[ERROR]: Failed to get nameinfo in retrieve_ip_addrs_v6().(gai_strerror=%s)\n", gai_strerror(err));
            goto go_out_err;
        }
        freeaddrinfo(res0);
        
        switch(scope){
        case IPV6_ADDR_ANY:
            ONVIF_LOG1(ONVIF_LOG_INFO,
                    "[INFO]: Got global IPv6 addr=[%s].\n", host);
            if(na_info->ip_addr_v6_g) free(na_info->ip_addr_v6_g);
            na_info->ip_addr_v6_g = strdup(host);
            flag_ipv6_g = TRUE;
            break;
        case IPV6_ADDR_LINKLOCAL:
            if(na_info->ip_addr_v6_l) free(na_info->ip_addr_v6_l);
            na_info->ip_addr_v6_l = strdup(host);
            /* linklocal with if */
            int len = strlen(host) + strlen(NA_IF) + 2;
            na_info->ip_addr_v6_l_with_if = malloc(len);
            snprintf(na_info->ip_addr_v6_l_with_if, len,
                     "%s%s%s", host, "%", NA_IF);
            flag_ipv6_l = TRUE;

            ONVIF_LOG1(ONVIF_LOG_INFO,
                       "[INFO]: Got link-local IPv6 addr=[%s].\n",
                    na_info->ip_addr_v6_l);
            ONVIF_LOG1(ONVIF_LOG_INFO,
                       "[INFO]: Got link-local IPv6 addr with if=[%s].\n",
                       na_info->ip_addr_v6_l_with_if);
            break;
        case IPV6_ADDR_LOOPBACK:break;
        case IPV6_ADDR_SITELOCAL:break;
        case IPV6_ADDR_COMPATv4:break;
        default:
            ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Unknown IPv6 addr scope.\n");
        }
    }

    if(flag_ipv6_l
       /* Always use ipv6 global address in <XAddrs>, so ipv6_g addr must be retrieved */
#ifdef USE_IPV6_GLOBAL_ADDR_IN_XADDRS_ALWAYS
       && flag_ipv6_g
#endif
    ){
        err = 0;
    }
    else{
        err = DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }

go_out_err:
    if(err){
        if(!flag_ipv6_l){
            ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Failed to retrieve IPv6 link-local address.\n");
        }
        if(!flag_ipv6_g){
            ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Failed to retrieve IPv6 global address.\n");
        }
    }

    fclose(fp);

    return err;
}

#endif

# include <net/if.h>
# include <sys/ioctl.h>
static char * retrieve_ip_addr()
{
    int fd, ret;
    char *ip_addr = NULL;
    struct ifreq ifr;

    bzero(&ifr, sizeof(struct ifreq));

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    ifr.ifr_addr.sa_family = AF_INET;
    strncpy(ifr.ifr_name, NA_IF, strlen(NA_IF));
  
    ret = ioctl(fd, SIOCGIFADDR, &ifr);
    close(fd);


    if(ret == 0){
        ip_addr = malloc(16);        
        strncpy(ip_addr,
                inet_ntoa(((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr),
                sizeof(char) * 16);
        ONVIF_LOG1(ONVIF_LOG_INFO, "[INFO]: Got IPv4 addr=[%s].\n", ip_addr);
    }
    else{
        ONVIF_LOG(ONVIF_LOG_ERROR, "[ERROR]: Failed to retrieve available IPv4 address.\n");
    }

    return ip_addr;
}

#else /* !USE_ONVIF */
static char * retrieve_ip_addr()
{
    char *ip_addr = NULL;
    struct in_addr addr;
    struct hostent * h = NULL;
    
    char hostname[256];
    if (gethostname(hostname, 256))
        return NULL;
    h = gethostbyname(hostname);
    if (!h)
        return NULL;
    memcpy(&addr, h->h_addr, h->h_length);
    ip_addr = malloc(16);

    return strcpy(ip_addr, inet_ntoa(addr));
}
#endif /* USE_ONVIF */

# ifdef WIN32

# include <IpHlpApi.h> 

int retrieve_network_adapter_info(struct na_info * na)
{
    IP_ADAPTER_INFO AdapterInfo;
    PIP_ADAPTER_INFO AdInfoBuf = &AdapterInfo, wAdapterInfo; 
    BOOLEAN found = FALSE;
    DWORD dwLen = sizeof(AdapterInfo);
    DWORD dwStatus = GetAdaptersInfo(
                        &AdapterInfo, 
                        &dwLen);
    if (dwStatus == ERROR_BUFFER_OVERFLOW)
    {
        AdInfoBuf = malloc(dwLen);
        dwStatus = GetAdaptersInfo(
                            AdInfoBuf, 
                            &dwLen);
        if (dwStatus != ERROR_SUCCESS)
            return DPWS_ERR_RETRIEVING_NETWORK_INTERFACE_INFO;
    }
    else if (dwStatus != ERROR_SUCCESS)
        return DPWS_ERR_RETRIEVING_NETWORK_INTERFACE_INFO;

    wAdapterInfo = AdInfoBuf;
    while (wAdapterInfo)
    {
        switch (wAdapterInfo->Type)
        {
            case MIB_IF_TYPE_ETHERNET:  // LAN
            case MIB_IF_TYPE_FDDI: case MIB_IF_TYPE_OTHER:
                found = TRUE;
                break;
            case MIB_IF_TYPE_PPP: case MIB_IF_TYPE_SLIP: // WAN
                break;  // skip

            case MIB_IF_TYPE_LOOPBACK:
                // take it if the only one
                if (wAdapterInfo == &AdapterInfo)
                    found = TRUE;
                break;
        }
        if (found)
            break;
        else
            wAdapterInfo = wAdapterInfo->Next;
    }
    if (found) {
        na->mac_addr = malloc(6);
        memcpy(na->mac_addr, wAdapterInfo->Address, 6);
        na->ip_addr = retrieve_ip_addr();
        if (na->ip_addr) {
            if (!strcmp(na->ip_addr, "127.0.0.1"))
                memcpy(na->ip_addr, &wAdapterInfo->IpAddressList.IpAddress.String, 16);
        }
        else
            return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }
    else
        return DPWS_ERR_RETRIEVING_NETWORK_INTERFACE_INFO;

    if (AdInfoBuf != &AdapterInfo)
        free(AdInfoBuf);
    return DPWS_OK;
}

# else  // Linux/POSIX

# include <net/if.h>
# include <sys/ioctl.h>

int retrieve_network_adapter_info(struct na_info * na)
{
    int s;
    struct ifreq buffer;
    struct sockaddr_in * addrin;
    s = socket(PF_INET, SOCK_DGRAM, 0);

    if (s < 0)
        return DPWS_ERR_RETRIEVING_NETWORK_INTERFACE_INFO;
//      memset(&buffer, 0, sizeof(buffer));

    strcpy(buffer.ifr_name, NA_IF);
    
    if (ioctl(s, SIOCGIFHWADDR, &buffer) < 0)
        return DPWS_ERR_RETRIEVING_MAC_ADDRESS;
    na->mac_addr = malloc(6);
    memcpy(na->mac_addr, buffer.ifr_hwaddr.sa_data, 6);

    na->ip_addr = retrieve_ip_addr();
    if (na->ip_addr) {
        if (!strcmp(na->ip_addr, "127.0.0.1")) {
            if (ioctl(s, SIOCGIFADDR, &buffer) < 0)
                return DPWS_ERR_RETRIEVING_IP_ADDRESS;
            addrin = (struct sockaddr_in *)&buffer.ifr_addr;
            strcpy(na->ip_addr, inet_ntoa(addrin->sin_addr));
        }       
    }
    else
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;

#ifdef WITH_IPV6 /* Added by HG.W */
    if(retrieve_ip_addrs_v6(na)){
        return DPWS_ERR_RETRIEVING_IP_ADDRESS;
    }
#endif

    close(s);
    return DPWS_OK;
}
# endif
#endif

byte* parse_mac_address(char * mac) {
    byte *mac_addr = malloc(6);
    int i = 0;
    do {
        mac_addr[i] = MAKE_BYTE(CHAR_HEXA2VALUE(*mac), CHAR_HEXA2VALUE(*(mac+1)));
        if (*(mac+2) == SEPARATOR_1 || *(mac+2) == SEPARATOR_2)
            mac += 3;
        else
            mac += 2;
        i++;
    }
    while (i < 6);
    return mac_addr;
}

/*----------------------------------------------------------------------------*\
 *                             TIMEOUT MANAGEMENT                             *
\*----------------------------------------------------------------------------*/
// an equal schedule is inserted last
void reschedule(scheduler_t *scheduler, schedule_t *schedule)
{
    schedule_t * wSchedule = scheduler->schedules;
    if (!wSchedule || IS_TIME_INF(schedule->expDate, wSchedule->expDate))
    {
        schedule->next = wSchedule;
        scheduler->schedules = schedule;    // change head
        return;
    }
    while (wSchedule && wSchedule->next &&
        (IS_TIME_EQUAL(schedule->expDate, wSchedule->next->expDate) || IS_TIME_SUP(schedule->expDate, wSchedule->next->expDate)))
        wSchedule = wSchedule->next;

    if (wSchedule)
    {
        if (wSchedule->next)
        {
            schedule->next = wSchedule->next;
            wSchedule->next = schedule;
        }
        else if (IS_TIME_INF(schedule->expDate, wSchedule->expDate))
        {
            schedule->next = wSchedule;
            scheduler->schedules = schedule;    // change head
        }
        else
        {
            schedule->next = NULL;
            wSchedule->next = schedule;
        }
    }
    else
    {
        schedule->next = NULL;
        scheduler->schedules = schedule;    // change head
    }
}

void add_schedule(scheduler_t *scheduler, unsigned32 duration, schedule_cbk handler, void *param)
{
    schedule_t *newSchedule;
    
    // recycle if possible
    if (scheduler->pool)
    {
        newSchedule = scheduler->pool;
        scheduler->pool = scheduler->pool->next;
    }
    else
        newSchedule = malloc(sizeof(schedule_t));
    
    system_time(&newSchedule->expDate);
    newSchedule->expDate.lo += duration % 1000 * 1000;
    if (newSchedule->expDate.lo > 1000000) {
        newSchedule->expDate.hi += newSchedule->expDate.lo / 1000000;
        newSchedule->expDate.lo %= 1000000;
    }
    newSchedule->expDate.hi += duration / 1000;
    
    newSchedule->handler = handler;
    newSchedule->param = param;
    reschedule(scheduler, newSchedule);
}

schedule_t * get_schedule(scheduler_t *scheduler)
{
    schedule_t *wSchedule = NULL;
    
    if (scheduler->schedules)
    {
        wSchedule = scheduler->schedules;
        scheduler->schedules = wSchedule->next;
    }
    return wSchedule;
}

unsigned32 get_ms_schedule(schedule_t *schedule)
{
    unsigned32 d = schedule->expDate.lo / 1000;
    unsigned64_t time;
    system_time(&time);
    if (schedule->expDate.lo < time.lo) {
        time.hi++;
        d += 1000;
    }
    d -= time.lo / 1000;
    if (schedule->expDate.hi < time.hi)
        d = 0;
    else
        d += (schedule->expDate.hi - time.hi) * 1000;

    return d;
}

struct timeval * get_timeval_schedule(schedule_t *schedule, struct timeval *out)
{
    unsigned64_t time;
    system_time(&time);
    if (schedule->expDate.lo < time.lo) {
        time.hi++;
        schedule->expDate.lo += 1000000;
    }
    out->tv_usec = schedule->expDate.lo - time.lo;
    out->tv_sec = schedule->expDate.hi - time.hi;
    return out->tv_sec >= 0 ? out : NULL;
}

static void recycle_schedule(scheduler_t *scheduler, schedule_t *schedule, short lock_handle)
{
    if (lock_handle >=0) {
        lock(lock_handle);
        schedule->next = scheduler->pool;
        scheduler->pool = schedule;
        unlock(lock_handle);
    } else {
        schedule->next = scheduler->pool;
        scheduler->pool = schedule;
    }
}

int realize_schedule(scheduler_t *scheduler, struct dpws *dpws, short lock_handle)
{
    int ret = DPWS_OK;
    if (dpws->schedule) {
        if (dpws->schedule->handler)
            ret = dpws->schedule->handler(dpws, dpws->schedule->param);
        recycle_schedule(scheduler, dpws->schedule, lock_handle);
        dpws->schedule = NULL;
    }
    return ret;
}

//int wait_and_call(struct dpws* dpws, scheduler_t *scheduler)
//{
//  int ret;
//  schedule_t *wSchedule = get_next_schedule(scheduler);
//  sleep_ms(wSchedule->delta);
//  recycle_schedule(scheduler, wSchedule);
//  return ret;
//}

void free_scheduler(scheduler_t *scheduler)
{
    schedule_t *wSchedule;
    while (scheduler->pool) {
        wSchedule = scheduler->pool;
        scheduler->pool = wSchedule->next;
        free(wSchedule);
    }
    while (scheduler->schedules) {
        wSchedule = scheduler->schedules;
        scheduler->schedules = wSchedule->next;
        free(wSchedule);
    }
}

/*----------------------------------------------------------------------------*\
 *                                    HANDLES                                 *
\*----------------------------------------------------------------------------*/
void createHandlePool(handlePool_t *pool, short bucketNumber, short maxHandle,
                        short lockHandle, handle_release_cbk releaseHook)
{
    pool->bucketNumber = bucketNumber;
    pool->lastHandle = 0;
    pool->lockHandle = lockHandle;
    pool->maxHandle = maxHandle;
    pool->buckets = malloc(bucketNumber * sizeof(handle_s *));
    memset(pool->buckets, 0, bucketNumber * sizeof(handle_s *));
    pool->unusedPool = NULL;
    pool->releaseHook = releaseHook;
}

void deleteHandlePool(handlePool_t *pool)
{
    handle_s * wHandle;
    
    // handles should be all in the recycling pool if the handle pool has been cleaned
    for (wHandle = pool->unusedPool; wHandle; wHandle = pool->unusedPool) {
        pool->unusedPool = wHandle->next;
        free(wHandle);
    }
    free(pool->buckets);
    pool->buckets = NULL;
}

short createHandle(handlePool_t *pool, short type, void * pObject)
{
    handle_s * newHandle;
    int bucketIndex;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    if (pool->lastHandle == pool->maxHandle)
        return DPWS_ERR_MAX_HANDLE_REACHED;
    
    /* try to recycle */
    if (pool->unusedPool) {
        newHandle = pool->unusedPool;
        pool->unusedPool = pool->unusedPool->next;
    }
    else {
        newHandle = malloc(sizeof(handle_s));
        newHandle->handleRef = pool->lastHandle++;
        newHandle->pObject = NULL;
    }
    
    /* fill handle */
    newHandle->useCount = 1;
    newHandle->type = type;
    if (pObject)    // to be able to recycle object too
        newHandle->pObject = pObject;
    
    /* insert in bucket */
    bucketIndex = newHandle->handleRef%pool->bucketNumber;
    newHandle->next = pool->buckets[bucketIndex];
    pool->buckets[bucketIndex] = newHandle;
    
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    return newHandle->handleRef;
}

static handle_s * _getHandle(handlePool_t *pool, short handleRef)
{
    handle_s * wHandle;

    for (
        wHandle = pool->buckets[handleRef%pool->bucketNumber];
        wHandle && wHandle->handleRef != handleRef;
        wHandle = wHandle->next
    );
        
    return wHandle;
}

handle_s * getHandle(handlePool_t *pool, short handleRef)
{
    handle_s * wHandle;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    wHandle = _getHandle(pool, handleRef);
        
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    return wHandle;
}

void * getObject(handlePool_t *pool, short handleRef)
{
    handle_s * wHandle;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    wHandle = _getHandle(pool, handleRef);
        
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    return wHandle ? wHandle->pObject : NULL;
}

void * getTypedObject(handlePool_t *pool, short handleRef, short type)
{
    handle_s * wHandle;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    wHandle = _getHandle(pool, handleRef);
        
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    return wHandle && wHandle->type == type ? wHandle->pObject : NULL;
}

handle_s * checkoutHandle(handlePool_t *pool, short handleRef)
{
    handle_s * wHandle;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    wHandle = _getHandle(pool, handleRef);
    if (wHandle)
        wHandle->useCount++;
    
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    return wHandle;
}

void * checkoutObject(handlePool_t *pool, short handleRef)
{
    handle_s * wHandle;
    void * ret = NULL;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    wHandle = _getHandle(pool, handleRef);
    if (wHandle) {
        wHandle->useCount++;
        ret = wHandle->pObject;
    }
    
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    return ret;
}

static int deleteHandle(handlePool_t *pool, short handleRef)
{
    int ret = DPWS_OK, bucketIndex = handleRef%pool->bucketNumber;
    handle_s * wHandle = pool->buckets[bucketIndex], *rHandle;
    
    if (wHandle->handleRef == handleRef)
    {
        pool->buckets[bucketIndex] = wHandle->next;
        wHandle->next = pool->unusedPool;
        pool->unusedPool = wHandle;
    }
    else
    {
        for (; wHandle; wHandle = wHandle->next)
        {
            if (wHandle->next && wHandle->next->handleRef == handleRef)
            {
                rHandle = wHandle->next;
                wHandle->next = wHandle->next->next;
                rHandle->next = pool->unusedPool;
                pool->unusedPool = rHandle;
                break;
            }
        }
    }
    
    if (pool->releaseHook)
        ret = pool->releaseHook(pool->unusedPool);

    return ret;
}


handle_s * handlePoolIterate(handlePool_t *pool, short typeFilter, handle_cbk callback, void * param)
{
    handle_s * wHandle;
    int i;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    for (i = 0; i < pool->bucketNumber; i++) {
        for (wHandle = pool->buckets[i]; wHandle;   wHandle = wHandle->next)
            if (wHandle && (typeFilter < 0 || wHandle->type == typeFilter)) {
                if (callback(wHandle->type, wHandle->pObject, param))
                    return wHandle;
            }
    }
    
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);
    return NULL;
}

void deleteHandles(handlePool_t *pool, short type)
{
    handle_s * wHandle, *rHandle;
    int i;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    for (i = 0; i < pool->bucketNumber; i++) {
        wHandle = pool->buckets[i];
        
        while (wHandle && (type < 0 || wHandle->type == type))
        {
            pool->buckets[i] = wHandle->next;
            wHandle->next = pool->unusedPool;
            pool->unusedPool = wHandle;
            if (pool->releaseHook)
                pool->releaseHook(pool->unusedPool);
            wHandle = pool->buckets[i];
        }
        for (; wHandle; wHandle = wHandle->next)
        {
            if (wHandle->next && (type < 0 || wHandle->next->type == type))
            {
                rHandle = wHandle->next;
                wHandle->next = wHandle->next->next;
                rHandle->next = pool->unusedPool;
                pool->unusedPool = rHandle;
                if (pool->releaseHook)
                    pool->releaseHook(pool->unusedPool);
            }
        }
    }
    
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);
}

int releaseHandle(handlePool_t *pool, short handleRef)
{
    handle_s * handle;
    int ret = 0;

    if (pool->lockHandle >= 0)
        lock(pool->lockHandle);
        
    handle = _getHandle(pool, handleRef);
    if (handle) {
        if (handle->useCount == 1)
            ret = deleteHandle(pool, handleRef);
        handle->useCount--;
    }
    else
        ret = DPWS_ERR_NO_HANDLE_FOUND;
    
    if (pool->lockHandle >= 0)
        unlock(pool->lockHandle);

    /**
     * Added by HG.W
     * According to freeXXXX()[eg.freeDevice()] in config.c,
     * a Non-Zero ret is expected
     */
#ifdef USE_ONVIF
    return 1;
#else
    return DPWS_OK;
#endif
}

/*----------------------------------------------------------------------------*\
 *                                     MISC                                   *
\*----------------------------------------------------------------------------*/
void sleep_ms(long delay)
{
#ifdef WIN32
    Sleep(delay);
#elif defined(TR_STACK_NEXGEN)
    ngSleep(delay);
#elif defined(VXWORKS)
    taskDelay(delay * sysClkRateGet()/1000L);
#else
    usleep(delay * 1000L);
#endif
}

void system_time(unsigned64_t *time)
{
#ifdef TR_STACK_NEXGEN
    NGtimeval tv;
    ngOSClockGetTimeOfDay((NGtimeval *)&time);  // QUESTION: is it satisfying to call NexGen ?
    time->hi = tv.tv_sec;
    time->lo = tv.tv_usec;
#elif defined(WIN32)
    FILETIME        ft;
    LARGE_INTEGER   li;
    __int64         t;
    static int      tzflag;
# ifdef UNDER_CE
    SYSTEMTIME systime;
    GetSystemTime(&systime);
    SystemTimeToFileTime(&systime,&ft);
# else
    GetSystemTimeAsFileTime(&ft);
# endif
    li.LowPart  = ft.dwLowDateTime;
    li.HighPart = ft.dwHighDateTime;
    t  = li.QuadPart;           /* In 100-nanosecond intervals */
    t -= 116444736000000000i64; /* Offset to the Epoch time */
    t /= 10;                   /* In microseconds */
    time->hi = (long)(t / 1000000);
    time->lo = (long)(t % 1000000);
#elif defined(VXWORKS)
    struct timespec ts;
    clock_gettime (CLOCK_REALTIME, &ts);
    time->hi = ts.tv_sec;
    time->lo = ts.tv_nsec / 1000L;  // timespec has nsec, but timeval has usec
#else
    struct timeval tv;
    gettimeofday(&tv, (struct timezone *)0);
    time->hi = tv.tv_sec;
    time->lo = tv.tv_usec;
#endif
}

unsigned32 system_time_s()
{
    unsigned64_t time;
    system_time(&time);
    return time.hi;
}

static int range_random(unsigned int min, unsigned int max)
{
    return (true_random() % (max - min)) + min;
}

unsigned32 get_soap_over_udp_schedule(unsigned short msgType, short * send_nb, unsigned32 last_delay)   // ms
{
    unsigned32 delay = 0;
    int max_msg = (DPWS_IS_MULTICAST(msgType) ? MULTICAST_UDP_REPEAT : UNICAST_UDP_REPEAT);
    
    if (++(*send_nb) > max_msg)
        return 0xFFFFFFFF;
        
    switch (*send_nb) {
        case 1:
            // use initial delay to avoid network storm
            if (DPWS_APP_DELAY_MSG(msgType))
                delay = range_random(0, APP_MAX_DELAY);
            break;
        case 2:
            delay = last_delay + range_random(UDP_MIN_DELAY, UDP_MAX_DELAY);
            break;
        default:
            delay = 3 * last_delay; // last_delay + 2*last_delay
            if (delay > UDP_UPPER_DELAY)
                delay = UDP_UPPER_DELAY;
    }
    return delay;
}

int is_physical_address(const char * addr) {
    return !strncmp(addr, "http:", 5) || !strncmp(addr, "https:", 6);
}

char * dpws_strdup(const char *s)
{
    char *t = NULL;
    if (s && (t = (char*)malloc(strlen(s) + 1)))
        strcpy(t, s);
    return t;
}

int dpws_strcasecmp(const char *s1, const char *s2)
{
    while ((*s1 != '\0') && (*s2 != '\0') 
        && (tolower(*(unsigned char *)s1) == tolower(*(unsigned char *)s2))) {
        s1++;
        s2++;
    }
    return tolower(*(unsigned char *) s1) - tolower(*(unsigned char *) s2);
}

#define DAY_IN_YEAR (365)   // NOTE: average value
#define DAY_IN_MONTH (30)   // NOTE: average value
#define HOUR_IN_DAY (24)
#define SEC_IN_HOUR (3600)
#define MIN_IN_HOUR (60)
#define SEC_IN_MIN (60)

// NOTE: alter the input string, does not control overflow
time_t xmlduration2duration(char *p)
{
    // TODO: control overflow ?
    time_t d = 0;
    char *t = NULL, *del = {"-PYMDTHMS."}, buf[10];
    int minus = 0, p_found = 0, t_found = 0;
    for (;*p;p++) {
        if (*p >= '0' && *p <= '9') {
            if (!p_found)   // datetime or invalid duration
                return 0;
            if (!t) t = p;
        }
        else {
            while (*del != *p) {
                del++;
                if (!del)
                    return 0;
            }
            // delimitor found
            switch (*p) {
                int f;
                case '-':
                    minus = 1;
                    break;
                case 'P':
                    p_found = 1;
                    break;
                case 'T':
                    t_found = 1;
                    break;
                case '.':
                    *(p + 1) = '\0';    // NOTE: ignore second decimals
                    break;
                case 'Y': case 'D': case 'H': case 'M': case 'S':
                    if (!t)
                        return 0;
                    strncpy(buf, t, p - t);
                    switch (*p) {
                        case 'Y':
                            f = DAY_IN_YEAR * HOUR_IN_DAY * SEC_IN_HOUR;    // NOTE: could avoid some multiplications
                            break;
                        case 'D':
                            f = HOUR_IN_DAY * SEC_IN_HOUR;
                            break;
                        case 'H':
                            f = SEC_IN_HOUR;
                            break;
                        case 'M':   // trick: month or minute
                            if (t_found)
                                f = SEC_IN_MIN;
                            else
                                f = DAY_IN_MONTH * HOUR_IN_DAY * SEC_IN_HOUR;
                            break;
                        case 'S':
                            f = 1;
                            break;
                    }
                    d +=  atoi(buf) * f;
                    break;
                default:
                    return 0;
            }
            t = NULL;
        }
    }
        
    return minus ? -d : d;
}

time_t get_system_time() {
    return (time_t)system_time_s();
}

time_t xmlduration2time(char *p)
{
    time_t ret = xmlduration2duration(p);
    return ret == 0 ? -1 : get_system_time() + xmlduration2duration(p);
}

char * dpws_duration2xmlduration(char *dest, time_t duration)
{
    long d;
    short h, m;
    char * minus = "";
    if (duration < 0) {
        minus = "-";
        duration = -duration;
    }
    d = (long)(duration / (HOUR_IN_DAY * SEC_IN_HOUR));
    duration = duration % (HOUR_IN_DAY * SEC_IN_HOUR);
    h = (short)(duration / SEC_IN_HOUR);
    duration = duration % SEC_IN_HOUR;
    m = (short)(duration / SEC_IN_MIN);
    duration = duration % SEC_IN_MIN;
    if (d > 0)
        sprintf(dest, "%sP%ldDT%dH%dM%dS", minus, d, h, m, (int)duration);
    else if (h > 0)
        sprintf(dest, "%sPT%dH%dM%dS", minus, h, m, (int)duration);
    else if (m > 0)
        sprintf(dest, "%sPT%dM%dS", minus, m, (int)duration);
    else
        sprintf(dest, "%sPT%dS", minus, (int)duration);
    return dest;
}

char * time2xmlduration(char *dest, time_t time)
{
    return dpws_duration2xmlduration(dest, time - system_time_s());
}
