/*============================================================================*\
|                                                                              |
|       =S= Device Profile for Web Services C Toolkit (DPWS toolkit)           |
|                                                                              |
|           ->>  Copyright 2004-2007 Schneider Electric SA <<-                 |
|                                                                              |
|                                                                              |
|                                + File info:                                  |
|                                              $Revision: 1.86.2.3 $
|                                              $Date: 2007/03/23 16:41:34 $
\*============================================================================*/
 
 /*
  ** Copyright (c) 1990- 1993, 1996 Open Software Foundation, Inc.
  ** Copyright (c) 1989 by Hewlett-Packard Company, Palo Alto, Ca. &
  ** Digital Equipment Corporation, Maynard, Mass.
  ** To anyone who acknowledges that this file is provided "AS IS"
  ** without any express or implied warranty: permission to use, copy,
  ** modify, and distribute this file for any purpose is hereby
  ** granted without fee, provided that the above copyright notices and
  ** this notice appears in all source code copies, and that none of
  ** the names of Open Software Foundation, Inc., Hewlett-Packard
  ** Company, or Digital Equipment Corporation be used in advertising
  ** or publicity pertaining to distribution of the software without
  ** specific, written prior permission.  Neither Open Software
  ** Foundation, Inc., Hewlett-Packard Company, nor Digital Equipment
  ** Corporation makes any representations about the suitability of
  ** this software for any purpose.
  */
#ifndef TOOLS_H
# define TOOLS_H

# include "_dpws.h"

/*----------------------------------------------------------------------------*\
 *                                   UUID                                     *
\*----------------------------------------------------------------------------*/
#ifdef TR_STACK_NEXGEN
# include <ngos.h>
#endif

#define CLOCK_SEQ_LAST              0x3FFF
#define RAND_MASK                   CLOCK_SEQ_LAST

void uuid_init(void);
char * uuid_create(char *);
char * schemed_uuid_create(char *uuid);
char * stable_uuid_create(char *, unsigned32 instance);

#ifdef NONVOLATILE_CLOCK
  void write_clock(unsigned16); // to implement ?
#endif

#if defined(THREADX) || defined(VXWORKS)
# define GET_PID 1
#elif defined(WIN32)
# define GET_PID GetCurrentProcessId()
#else
# define GET_PID getpid()
#endif
  
/*----------------------------------------------------------------------------*\
 *                                MEMORY STRUCTURES                           *
\*----------------------------------------------------------------------------*/

#define DYN_ARRAY_INITIALIZER(s, i) {NULL, 0, 0, s, i}

void init_dyn_array(dyn_array * array, size_t entry_size, int inc);
void * get_entry(dyn_array * tab, int index);
void * new_entry(struct dpws *dpws, dyn_array * tab);
void * insert_entry(struct dpws* dpws, dyn_array * tab, int index);
typedef void (*free_entry_cbk) (void *);
void remove_entry(dyn_array * tab, int index, free_entry_cbk free_hook);
// Stops when the hook returns TRUE
typedef BOOLEAN (*browse_entry_cbk) (void *, void *);
int browse_dyn_array(dyn_array * da, browse_entry_cbk hook, void *);
void free_dyn_array(dyn_array * da, free_entry_cbk free_hook);
void free_buffer_dyn_array(dyn_array * da);
//dyn_array * dyn_array_dup(dyn_array * da);
dyn_array * string_array_initialize(struct dpws* dpws, dyn_array * dest, char ** src, int inc);
dyn_array * string_array_dup(struct dpws* dpws, dyn_array * dest, dyn_array * src);
void init_entries(dyn_array * tab, void* entries, int nentries);
struct wsa_endpoint_ref * endpoint_ref_dup(struct dpws* dpws, struct wsa_endpoint_ref * src);
struct wsa_endpoint_ref * endpoint_ref_copy(struct dpws* dpws, struct wsa_endpoint_ref * dest, struct wsa_endpoint_ref * src);
int find_local_entry(void ** entries, int entry_nb, char *preferred_lang);

/*----------------------------------------------------------------------------*\
 *                                   CONVERSIONS                              *
\*----------------------------------------------------------------------------*/
/* Hexa/ASCII conversion */
static const char FIG_OFFSET = '0';
static const char FIG_MAX = '9';
static const char UPPERCASE_OFFSET = 'A';
static const char UPPERCASE_MAX = 'F';
static const char LOWERCASE_OFFSET = 'a';
static const char LOWERCASE_MAX = 'f';
static const char SEPARATOR_1 = '-';
static const char SEPARATOR_2 = ':';


#define MAKE_BYTE(hi, lo) ((hi<<4)|lo)
#define CHAR_HEXA2VALUE(c) (c >= FIG_OFFSET && c <= FIG_MAX ? c - FIG_OFFSET \
		:(c >= UPPERCASE_OFFSET && c <= UPPERCASE_MAX ? c - UPPERCASE_OFFSET + 10 \
		:(c >= LOWERCASE_OFFSET && c <= LOWERCASE_MAX ? c - LOWERCASE_OFFSET + 10 : -1)))
		
#define IS_WS(c) (c == ' ' || c == '\t' || c == 0xA || c == 0xD)		

byte* parse_mac_address(char * mac_addr);
int parse_qname_list(struct dpws * dpws, dyn_array * dest, char * source);
int parse_token_list(struct dpws * dpws, dyn_array * dest, char * source);

/*----------------------------------------------------------------------------*\
 *                       CUSTOM GSOAP TYPES MARSHALLING                       *
\*----------------------------------------------------------------------------*/

void dpws_simple_element_out(struct soap *soap, const char *tag, const char *value);
char** dpws_simple_element_in(struct soap *soap, const char *tag, char** value);
int dpws_empty_element_out(struct soap *soap, const char *tag);

/*----------------------------------------------------------------------------*\
 *                                   COMPARISONS                              *
\*----------------------------------------------------------------------------*/
BOOLEAN qname_equals(char * ns, char * local_name, char * qname);
int parse_match_rule(char *uri);
char * scope_match_string(int match_rule);
BOOLEAN rfc2396_matches(char *base, char *uri);
BOOLEAN is_rfc2396_included(dyn_array * uri_set1, dyn_array * uri_set2);
typedef BOOLEAN (*type_compare_cbk) (void *, void *);
BOOLEAN probe_match(dyn_array * types, dyn_array * scopes, int match_rule,
	dyn_array *dev_types, dyn_array *dev_scopes, type_compare_cbk type_compare);

/*----------------------------------------------------------------------------*\
 *                                     IPC                                    *
\*----------------------------------------------------------------------------*/
# define BUILTIN_MUTEX 4

void initMutexPool();
void destroyMutexPool();
short lock_create();
int lock(short handle);
int unlock(short handle);
int lock_destroy(short handle);

/*----------------------------------------------------------------------------*\
 *                                 NETWORKING                                 *
\*----------------------------------------------------------------------------*/
#ifdef WIN32
int winsock_init();
#endif

#ifdef USE_ONVIF /* Added by HG.W */
#define NA_IF    "eth0"
#endif

struct na_info {
	byte * mac_addr;
	char * ip_addr;
#ifdef WITH_IPV6 /* Added by HG.W (to support ipv6) */
    char * ip_addr_v6_g;/* global */
    char * ip_addr_v6_l;/* link-local */ 
    char * ip_addr_v6_l_with_if; /* link-local%index */
#endif
};
int retrieve_network_adapter_info(struct na_info * na);

/*----------------------------------------------------------------------------*\
 *                             TIMEOUT MANAGEMENT                             *
\*----------------------------------------------------------------------------*/
typedef struct scheduler {
	schedule_t* schedules;
	schedule_t* pool;
} scheduler_t;

// TODO: adding an earlier timeout should wake up the select()
typedef int (*schedule_cbk) (struct dpws*, void *);
void add_schedule(scheduler_t *scheduler, unsigned32 duration, schedule_cbk handler, void *param);	// ms
schedule_t * get_schedule(scheduler_t *scheduler);
unsigned32 get_ms_schedule(schedule_t *schedule);
struct timeval * get_timeval_schedule(schedule_t *schedule, struct timeval *out);
void reschedule(scheduler_t *scheduler, schedule_t *schedule);
int realize_schedule(scheduler_t *scheduler, struct dpws *dpws, short lock_handle);
//int wait_and_call(struct dpws* dpws, scheduler_t *scheduler);
void free_scheduler(scheduler_t *scheduler);

/*----------------------------------------------------------------------------*\
 *                                    HANDLES                                 *
\*----------------------------------------------------------------------------*/
#define ANY_TYPE		-1
#define MUTEX_TYPE		0	// NOTE: must be created first, in a mono thread context

typedef struct handleInfo {
	short handleRef;	// used as key
	short type;
	void * pObject;
	short useCount;
	struct handleInfo * next;
} handle_s;

typedef struct handlePoolInfo {
	short bucketNumber;
	short lastHandle;
	short maxHandle;
	handle_s **buckets;
	handle_s *unusedPool;
	short lockHandle;
	int (*releaseHook)(struct handleInfo *);
} handlePool_t;

typedef int (*handle_release_cbk) (handle_s *);
void createHandlePool(handlePool_t *pool, short bucketNumber, short maxHandle, short lockIndex, handle_release_cbk releaseHook);	// returns the handle
void deleteHandlePool(handlePool_t *pool);
short createHandle(handlePool_t *pool, short type, void * pObject);	// returns the handle and add a use count
handle_s * getHandle(handlePool_t *pool, short handleRef);
handle_s * checkoutHandle(handlePool_t *pool, short handleRef);	// NOTE: the "Handle" functions should be used when object need to be recycled (should be no use when object creation is moved to handle)
void * getObject(handlePool_t *pool, short handleRef);
void * getTypedObject(handlePool_t *pool, short handleRef, short type);
void * checkoutObject(handlePool_t *pool, short handleRef);
// the callback returns TRUE when the iterator must stop. If typeFilter is ANY_TYPE all handles are browsed
typedef BOOLEAN (*handle_cbk) (short, void *, void *);
handle_s * handlePoolIterate(handlePool_t *pool, short typeFilter, handle_cbk callback, void * param);
// TODO: add a copy and a replace function for dynamic update (object are immutable by defaut). Move object creation to handle.
int releaseHandle(handlePool_t *pool, short handleRef);
void deleteHandles(handlePool_t *pool, short type);

/*----------------------------------------------------------------------------*\
 *                                     MISC                                   *
\*----------------------------------------------------------------------------*/
void sleep_ms(long delay);
/* hi is seconds, lo is s */
void system_time(unsigned64_t *time);
unsigned32 system_time_s();
unsigned32 get_soap_over_udp_schedule(unsigned short msgType, short * send_nb, unsigned32 last_delay);	// ms
int is_physical_address(const char * addr);
char * dpws_strdup(const char *s);
time_t xmlduration2duration(char *p);
time_t xmlduration2time(char *p);
char * dpws_duration2xmlduration(char * dest, time_t duration);
char * time2xmlduration(char * dest, time_t time);

#endif
